Package: GitStats
Title: Standardized Git Repository Data
Version: 2.5.2
Authors@R: c(
    person(given = "Maciej", family = "Banas", email = "banasmaciek@gmail.com", role = c("aut", "cre")),
    person(given = "Kamil", family = "Koziej", email = "koziej.k@gmail.com", role = "aut"),
    person(given = "Karolina", family = "Marcinkowska", email = "karolina_marcinkowska@onet.pl", role = "aut"),
    person(given = "Matt", family = "Secrest", email = "secrestm@gene.com", role = "aut")
  )
Maintainer: Maciej Banas <banasmaciek@gmail.com>
Description: Obtain standardized data from multiple 'Git' services, including 'GitHub' and 'GitLab'. 
    Designed to be 'Git' service-agnostic, this package assists teams with activities spread across various 
    'Git' platforms by providing a unified way to access repository data.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
URL: https://r-world-devs.github.io/GitStats/,
        https://github.com/r-world-devs/GitStats
BugReports: https://github.com/r-world-devs/GitStats/issues
Imports: cli, dplyr, glue, httr2, lubridate (>= 1.8.0), mirai (>=
        2.0.0), rlang (>= 1.1.0), R6, purrr (>= 1.0.0), stringr
Suggests: DBI, jsonlite, RPostgres, RSQLite, spelling, knitr, mockery,
        rmarkdown, testthat (>= 3.0.0), withr
Config/testthat/edition: 3
VignetteBuilder: knitr
Language: en-US
Depends: R (>= 4.1.0)
NeedsCompilation: no
Packaged: 2026-05-05 10:39:52 UTC; banasm
Author: Maciej Banas [aut, cre],
  Kamil Koziej [aut],
  Karolina Marcinkowska [aut],
  Matt Secrest [aut]
Repository: CRAN
Date/Publication: 2026-05-05 13:14:10 UTC
Built: R 4.5.2; ; 2026-05-05 13:44:04 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,gitstats_repos_urls)
export(create_gitstats)
export(get_commits)
export(get_commits_stats)
export(get_files)
export(get_issues)
export(get_issues_stats)
export(get_orgs)
export(get_pull_requests)
export(get_pull_requests_stats)
export(get_release_logs)
export(get_repos)
export(get_repos_trees)
export(get_repos_urls)
export(get_storage)
export(get_storage_metadata)
export(get_users)
export(is_parallel)
export(remove_from_storage)
export(remove_postgres_storage)
export(remove_sqlite_storage)
export(set_github_host)
export(set_gitlab_host)
export(set_local_storage)
export(set_parallel)
export(set_postgres_storage)
export(set_sqlite_storage)
export(show_hosts)
export(show_orgs)
export(verbose_off)
export(verbose_on)
importFrom(R6,R6Class)
importFrom(cli,cli_abort)
importFrom(cli,cli_alert_danger)
importFrom(cli,cli_alert_info)
importFrom(cli,cli_alert_success)
importFrom(cli,cli_alert_warning)
importFrom(cli,col_green)
importFrom(cli,col_yellow)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,glimpse)
importFrom(dplyr,mutate)
importFrom(dplyr,relocate)
importFrom(glue,glue)
importFrom(httr2,req_headers)
importFrom(httr2,req_perform)
importFrom(httr2,request)
importFrom(httr2,resp_body_json)
importFrom(lubridate,floor_date)
importFrom(mirai,daemons)
importFrom(mirai,everywhere)
importFrom(mirai,mirai_map)
importFrom(purrr,keep)
importFrom(purrr,map)
importFrom(purrr,map_chr)
importFrom(rlang,"%||%")
importFrom(rlang,expr)
importFrom(stats,runif)
importFrom(stringr,str_length)
importFrom(stringr,str_replace)

gitstats is a statistics generator for git repositories.  It is mostly intended
for developers, as a way to check some development statistics for a project.

Currently it produces only HTML output with tables and graphs.

Requirements
============
- Python (>= 2.6.0)
- Git (>= 1.5.2.4)
- Gnuplot (>= 4.0.0)
- a git repository (bare clone will work as well)

The above versions are not absolute requirements; older versions may work also.

Recommended
===========
- Lots of memory and fast disk for large projects

Contributions
=============
Patches should be sent under "GPLv2 or later" license - this will allow
upgrading to newer GPL versions if they are sensible.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# GitStats <img src="man/figures/GitStats_logo.png" align="right" height="138" style="float:right; height:138px;"/>

<!-- badges: start -->

[![lifecycle](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![CRAN
status](https://www.r-pkg.org/badges/version/GitStats)](https://CRAN.R-project.org/package=GitStats)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/grand-total/GitStats)](https://CRAN.R-project.org/package=GitStats)
[![R-CMD-check](https://github.com/r-world-devs/GitStats/workflows/R-CMD-check/badge.svg)](https://github.com/r-world-devs/GitStats/actions)
[![Codecov test
coverage](https://codecov.io/gh/r-world-devs/GitStats/branch/devel/graph/badge.svg)](https://app.codecov.io/gh/r-world-devs/GitStats?branch=devel)
<!-- badges: end -->

With `GitStats` you can pull git data in a uniform way from GitHub and
GitLab. For the time-being you can get data on:

- 🏛️ organizations,
- 📘 repositories,
- 🕒 commits,
- 🐛 issues,
- 🔂 pull requests,
- 🙍 users,
- 🚀 release logs,
- 🌲 files tree (structure),
- 📄 text files content.

You can also prepare basic statistics with `get_*_stats()` functions for
commits and issues.

## Installation

From CRAN:

``` r
install.packages("GitStats")
```

From GitHub:

``` r
devtools::install_github("r-world-devs/GitStats")
```

## Examples:

Setup your `GitStats`:

``` r
library(GitStats)

git_stats <- create_gitstats() |>
  set_gitlab_host(
    repos = "mbtests/gitstatstesting"
  ) |>
  set_github_host(
    orgs = "r-world-devs",
    repos = "openpharma/DataFakeR"
  ) 
```

Optionally, you can run `GitStats` functions in parallel (with `mirai`
package underneath):

``` r
git_stats |>
  set_parallel()
```

Get commits:

``` r
commits <- git_stats |>
  get_commits(
    since = "2022-01-01"
  )

commits
#> # A tibble: 3,713 × 11
#>    repo_name id    committed_date      author author_login author_name additions
#>    <chr>     <chr> <dttm>              <chr>  <chr>        <chr>           <int>
#>  1 gitstats… 7f48… 2024-09-10 11:12:59 Macie… <NA>         Maciej Ban…         0
#>  2 gitstats… 9c66… 2024-09-10 10:35:37 Macie… <NA>         Maciej Ban…         0
#>  3 gitstats… fca2… 2024-09-10 10:31:24 Macie… <NA>         Maciej Ban…         0
#>  4 gitstats… e8f2… 2023-03-30 14:15:33 Macie… <NA>         Maciej Ban…         1
#>  5 gitstats… 7e87… 2023-02-10 09:48:55 Macie… <NA>         Maciej Ban…         1
#>  6 gitstats… 62c4… 2023-02-10 09:17:24 Macie… <NA>         Maciej Ban…         2
#>  7 gitstats… 55cf… 2023-02-10 09:07:54 Macie… <NA>         Maciej Ban…        92
#>  8 shinyGiz… C_kw… 2026-04-03 22:27:09 Kryst… krystian8207 Krystian I…      1638
#>  9 shinyGiz… C_kw… 2026-04-03 21:23:33 Kryst… krystian8207 Krystian I…        42
#> 10 shinyGiz… C_kw… 2026-04-03 20:29:59 Kryst… krystian8207 Krystian I…        18
#> # ℹ 3,703 more rows
#> # ℹ 4 more variables: deletions <int>, organization <chr>, repo_url <chr>,
#> #   api_url <glue>

commits |>
  get_commits_stats(
    time_aggregation = "month",
    group_var = author
  )
#> # A tibble: 382 × 4
#>    stats_date          githost author             stats
#>    <dttm>              <chr>   <chr>              <int>
#>  1 2022-01-01 00:00:00 github  Admin_mschuemi         1
#>  2 2022-01-01 00:00:00 github  Gowtham Rao            5
#>  3 2022-01-01 00:00:00 github  Krystian Igras         1
#>  4 2022-01-01 00:00:00 github  Martijn Schuemie       1
#>  5 2022-02-01 00:00:00 github  Hadley Wickham         3
#>  6 2022-02-01 00:00:00 github  Martijn Schuemie       2
#>  7 2022-02-01 00:00:00 github  Maximilian Girlich    13
#>  8 2022-02-01 00:00:00 github  Reijo Sund             1
#>  9 2022-02-01 00:00:00 github  eitsupi                1
#> 10 2022-03-01 00:00:00 github  Maximilian Girlich    14
#> # ℹ 372 more rows
```

Get repositories with specific code:

``` r
git_stats |>
  get_repos(
    with_code = "shiny",
    add_contributors = FALSE
  )
#> # A tibble: 9 × 17
#>   repo_id repo_name repo_fullpath default_branch stars forks created_at         
#>   <chr>   <chr>     <chr>         <chr>          <int> <int> <dttm>             
#> 1 R_kgDO… GitAI     r-world-devs… main               8     2 2024-11-07 11:51:03
#> 2 R_kgDO… hypothes… r-world-devs… master             2     0 2023-04-13 13:52:24
#> 3 R_kgDO… shinyGiz… r-world-devs… dev               21     0 2022-04-20 10:04:32
#> 4 R_kgDO… shinyTim… r-world-devs… master             3     0 2023-02-21 16:41:59
#> 5 R_kgDO… shinyCoh… r-world-devs… dev               10     0 2022-05-22 19:04:12
#> 6 R_kgDO… cohortBu… r-world-devs… dev                9     2 2022-05-22 18:31:55
#> 7 R_kgDO… GitStats  r-world-devs… master             9     2 2023-01-09 14:02:20
#> 8 R_kgDO… shinyQue… r-world-devs… master             3     0 2024-09-20 18:59:56
#> 9 R_kgDO… queryBui… r-world-devs… master             1     1 2024-09-20 14:54:12
#> # ℹ 10 more variables: last_activity_at <dttm>, languages <chr>,
#> #   issues_open <int>, issues_closed <int>, organization <chr>, repo_url <chr>,
#> #   commit_sha <chr>, api_url <chr>, githost <chr>, last_activity <drtn>
```

Get files:

``` r
git_stats |>
  get_files(
    pattern = "\\.md",
    depth = 2L
  )
#> # A tibble: 63 × 10
#>    repo_id      repo_name  organization file_path file_content file_size file_id
#>    <chr>        <chr>      <chr>        <chr>     <chr>            <int> <chr>  
#>  1 43398933     gitstatst… mbtests      README.md "# GitStats…       122 fe2407…
#>  2 R_kgDOHNMr2w shinyGizmo r-world-devs NEWS.md   "# shinyGiz…      2382 466c53…
#>  3 R_kgDOHNMr2w shinyGizmo r-world-devs README.md "\n# shinyG…      4271 388469…
#>  4 R_kgDOHNMr2w shinyGizmo r-world-devs cran-com… "## Test en…      1700 cfcaf4…
#>  5 R_kgDOHYNOFQ cohortBui… r-world-devs NEWS.md   "# cohortBu…      1369 8c5cf3…
#>  6 R_kgDOHYNOFQ cohortBui… r-world-devs README.md "\n# cohort…     13382 ebc919…
#>  7 R_kgDOHYNrJw shinyCoho… r-world-devs NEWS.md   "# shinyCoh…      2213 8b0e9a…
#>  8 R_kgDOHYNrJw shinyCoho… r-world-devs README.md "\n# shinyC…      3355 e49dfe…
#>  9 R_kgDOHYNxtw cohortBui… r-world-devs README.md "\n# cohort…      3472 d4687c…
#> 10 R_kgDOIvtxsg GitStats   r-world-devs LICENSE.… "# MIT Lice…      1075 141471…
#> # ℹ 53 more rows
#> # ℹ 3 more variables: repo_url <chr>, commit_sha <chr>, api_url <chr>
```

Get pull requests:

``` r
git_stats |>
  get_pull_requests(
    since = "2022-01-01"
  )
#> # A tibble: 444 × 10
#>    repo_name       number created_at          merged_at           state  author 
#>    <chr>           <chr>  <dttm>              <dttm>              <chr>  <chr>  
#>  1 gitstatstesting 2      2026-02-25 09:34:18 NA                  closed <NA>   
#>  2 gitstatstesting 1      2026-02-19 11:58:14 NA                  open   <NA>   
#>  3 shinyGizmo      1      2022-04-22 12:49:10 2022-04-22 12:58:32 merged krysti…
#>  4 shinyGizmo      2      2022-05-18 08:21:46 2022-06-02 11:16:13 merged galach…
#>  5 shinyGizmo      10     2022-06-13 09:13:59 2022-06-13 09:25:45 merged krysti…
#>  6 shinyGizmo      11     2022-06-13 10:22:16 2022-06-13 20:08:35 merged krysti…
#>  7 shinyGizmo      13     2022-06-13 21:34:05 2022-06-17 08:49:33 merged stla   
#>  8 shinyGizmo      16     2022-06-15 13:48:08 2022-06-15 14:12:52 merged krysti…
#>  9 shinyGizmo      17     2022-06-17 10:55:10 2022-06-17 10:56:09 merged krysti…
#> 10 shinyGizmo      19     2022-07-05 10:01:44 2022-07-05 11:05:30 merged krysti…
#> # ℹ 434 more rows
#> # ℹ 4 more variables: source_branch <chr>, target_branch <chr>,
#> #   organization <chr>, api_url <glue>
```

Print `GitStats` to see what it stores:

``` r
git_stats
#> A GitStats object for 2 hosts: 
#> Hosts: https://gitlab.com/api/v4, https://api.github.com
#> Scanning scope: 
#>  Organizations: [1] r-world-devs
#>  Repositories: [2] mbtests/gitstatstesting, openpharma/DataFakeR
#> Storage [local]:
#>  Commits: 3713 [date range: 2022-01-01 - 2026-04-21]
#>  Repositories: 9 
#>  Files: 63 [file pattern: \.md]
#>  Pull_requests: 444 [date range: 2022-01-01 - 2026-04-21]
```

## See also

`GitStats` is used to facilitate workflow of the `GitAI` R package, a
tool for gathering AI-based knowledge about git repositories:
<https://r-world-devs.github.io/GitAI/>

## Acknowledgement

Special thanks to [James Black](https://github.com/epijim), [Karolina
Marcinkowska](https://github.com/marcinkowskak), [Kamil
Koziej](https://github.com/Cotau), [Matt
Secrest](https://github.com/mattsecrest), [Krystian
Igras](https://github.com/krystian8207), [Kamil
Wais](https://github.com/kalimu), [Adam
Forys](https://github.com/galachad) - for the support in the package
development.

# to TEST create tokens with insufficient scopes

create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs",
    token = Sys.getenv("GITHUB_PAT_INSUFFICIENT")
  ) |>
  get_repos()

create_gitstats() |>
  set_gitlab_host(
    orgs = "mbtests"
  ) |>
  get_repos()

create_gitstats() |>
  set_gitlab_host(
    orgs = "mbtests",
    token = Sys.getenv("GITLAB_PAT_PUBLIC_INSUFFICIENT")
  ) |>
  get_repos()

create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs"
  )

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4
)

## ----eval = FALSE-------------------------------------------------------------
# library(GitStats)
# git_stats <- create_gitstats() |>
#   set_github_host(
#     orgs = c("r-world-devs", "openpharma"),
#     token = Sys.getenv("GITHUB_PAT")
#   ) |>
#   set_gitlab_host(
#     orgs = c("mbtests"),
#     token = Sys.getenv("GITLAB_PAT_PUBLIC")
#   )

## ----eval = FALSE-------------------------------------------------------------
# git_stats <- create_gitstats() |>
#   set_github_host(
#     host = "github.internal.com",
#     orgs = c("org_1", "org_2", "org_3"),
#     token = Sys.getenv("YOUR_GITHUB_PAT")
#   ) |>
#   set_gitlab_host(
#     host = "internal.host.com",
#     orgs = c("internal_org"),
#     token = Sys.getenv("YOUR_GITLAB_PAT")
#   )

## ----eval = FALSE-------------------------------------------------------------
# git_stats <- create_gitstats() |>
#   set_github_host(
#     orgs = c("r-world-devs", "openpharma"),
#     token = Sys.getenv("GITHUB_PAT")
#   ) |>
#   set_gitlab_host(
#     orgs = c("mbtests"),
#     token = Sys.getenv("GITLAB_PAT_PUBLIC")
#   )

## ----eval = FALSE-------------------------------------------------------------
# git_stats <- create_gitstats() |>
#   set_github_host(
#     repos = c("r-world-devs/GitStats", "r-world-devs/shinyCohortBuilder", "openpharma/DataFakeR"),
#     token = Sys.getenv("GITHUB_PAT")
#   ) |>
#   set_gitlab_host(
#     repos = "mbtests/gitstatstesting",
#     token = Sys.getenv("GITLAB_PAT_PUBLIC")
#   )

## ----eval = FALSE-------------------------------------------------------------
# git_stats <- create_gitstats() |>
#   set_github_host(
#     orgs = "openpharma",
#     repos = c("r-world-devs/GitStats", "r-world-devs/shinyCohortBuilder"),
#     token = Sys.getenv("GITHUB_PAT")
#   )

## ----eval = FALSE-------------------------------------------------------------
# git_stats <- create_gitstats() |>
#   set_github_host(
#     host = "github.internal.com",
#     token = Sys.getenv("YOUR_GITHUB_PAT")
#   ) |>
#   set_gitlab_host(
#     host = "internal.host.com",
#     token = Sys.getenv("YOUR_GITLAB_PAT")
#   )

## ----eval = FALSE-------------------------------------------------------------
# git_stats <- create_gitstats() |>
#   set_github_host(
#     orgs = c("r-world-devs", "openpharma")
#   ) |>
#   set_gitlab_host(
#     orgs = c("mbtests")
#   )

github_stats <- create_gitstats() |>
  set_github_host(
    host = "github.roche.com"
  )

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "https://code.roche.com"
  )

get_orgs(github_stats)

get_orgs(gitlab_stats)

git_stats <- create_gitstats() |>
  set_github_host(
    host = "github.roche.com"
  ) |>
  set_gitlab_host(
    host = "code.roche.com"
  )

get_orgs(git_stats)

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = c("datascience/rwd", "botany")
  )

get_orgs(gitlab_stats)

devtools::load_all(".")

git_stats <- create_gitstats() %>%
  set_github_host(
    orgs = c("r-world-devs", "openpharma"),
    token = Sys.getenv("GITHUB_PAT")
  ) %>%
  set_gitlab_host(
    orgs = c("mbtests"),
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  )

release_logs <- get_release_logs(
  gitstats = git_stats,
  since = "2024-01-01",
  verbose = FALSE
)

release_logs

# Check printing in storage
git_stats

get_commits(
  gitstats = git_stats,
  since = "2024-01-01"
)

# Check printing in storage
git_stats

commits_stats <- create_gitstats() %>%
  set_github_host(
    orgs = c("r-world-devs", "openpharma"),
    token = Sys.getenv("GITHUB_PAT")
  ) %>%
  set_gitlab_host(
    orgs = c("mbtests"),
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  ) |>
  get_commits(
    since = "2024-01-01"
  ) |>
  get_commits_stats(
    time_aggregation = "year",
    group_var = author_name
  )

test_gitstats <- create_gitstats() |>
  set_github_host(
    orgs = "openpharma"
  )

repos_urls <- get_repos_urls(test_gitstats, with_code = "pharma")
# should use cache
repos_urls <- get_repos_urls(test_gitstats, with_code = "pharma")
# should NOT use cache
repos_urls <- get_repos_urls(test_gitstats)
# should use cache
repos_urls <- get_repos_urls(test_gitstats)

# should NOT use cache
repos_urls <- get_repos_urls(test_gitstats, with_files = c("renv.lock", "DESCRIPTION"))
# should use cache
repos_urls <- get_repos_urls(test_gitstats, with_files = c("renv.lock", "DESCRIPTION"))
# should NOT use cache
repos_urls <- get_repos_urls(test_gitstats, with_files = "renv.lock")
# should use cache
repos_urls <- get_repos_urls(test_gitstats, with_files = "renv.lock")

repos <- get_repos(test_gitstats)
repos <- get_repos(test_gitstats)

repos <- get_repos(test_gitstats, with_code = "shiny", in_files = c("renv.lock", "DESCRIPTION"))
repos <- get_repos(test_gitstats, with_code = "shiny", in_files = c("renv.lock", "DESCRIPTION"))

repos <- get_repos(test_gitstats, with_code = "shiny")
repos <- get_repos(test_gitstats, with_code = "shiny")

# no cache
commits <- get_commits(test_gitstats, since = "2024-01-01")
# cache
commits <- get_commits(test_gitstats, since = "2024-01-01")

# no cache
commits <- get_commits(test_gitstats, since = "2024-01-02")
# cache
commits <- get_commits(test_gitstats, since = "2024-01-02")

# no cache
commits <- get_commits(test_gitstats, since = "2024-01-02", until = "2024-06-30")
# cache
commits <- get_commits(test_gitstats, since = "2024-01-02", until = "2024-06-30")

# no cache
users <- get_users(test_gitstats, c("kalimu", "maciekbanas"))
# cache
users <- get_users(test_gitstats, c("kalimu", "maciekbanas"))

# no cache
users <- get_users(test_gitstats, c("kalimu", "maciekbanas", "marcinkowskak"))
users <- get_users(test_gitstats, c("kalimu", "maciekbanas", "marcinkowskak"))

# no cache
files_data <- get_files(test_gitstats, file_path = "DESCRIPTION")
# cache
files_data <- get_files(test_gitstats, file_path = "DESCRIPTION")

# no cache
files_data <- get_files(test_gitstats, file_path = c("DESCRIPTION", "LICENSE"))
# cache
files_data <- get_files(test_gitstats, file_path = c("DESCRIPTION", "LICENSE"))

# no cache
files_data <- get_files(test_gitstats, file_path = "DESCRIPTION")
# cache
files_data <- get_files(test_gitstats, file_path = "DESCRIPTION")

# no cache
release_logs <- get_release_logs(test_gitstats, since = "2024-01-01")
# cache
release_logs <- get_release_logs(test_gitstats, since = "2024-01-01")

devtools::load_all(".")

git_stats <- create_gitstats() %>%
  set_github_host(
    orgs = c("r-world-devs", "openpharma"),
    token = Sys.getenv("GITHUB_PAT")
  )

release_logs <- get_release_logs(
  gitstats = git_stats,
  since = "2024-01-01"
)

repos_urls <- get_repos_urls(
  gitstats = git_stats,
  with_code = "shiny"
)

files_structure <- get_files_structure(
  gitstats = git_stats,
  pattern = "\\.md",
  depth = 1L
)

git_stats

get_storage(git_stats)

get_storage(git_stats, "commits")
get_storage(git_stats, "repos_urls")
get_storage(git_stats, "release_logs")
get_storage(git_stats, "files_structure")

devtools::load_all(".")

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    orgs = c("mbtests", "mbtestapps")
  )

get_repos_trees(
  gitstats = gitlab_stats
)

github_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs")
  )

get_repos_trees(
  gitstats = github_stats
)

github_stats <- create_gitstats() |>
  set_github_host(
    repos = c("r-world-devs/GitStats", "openpharma/DataFakeR")
  )

get_repos_trees(
  gitstats = github_stats,
  pattern = "\\.md"
)

git_stats <- create_gitstats() |>
  set_github_host(
    repos = c("openpharma/DataFakeR", "openpharma/visR")
  ) |>
  set_gitlab_host(
    orgs = c("mbtests")
  )

get_commits(git_stats, since = "2023-01-01", verbose = FALSE)

git_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs", "maciekbanas", "openpharma", "pharmaverse", "insightsengineering")
  )

get_orgs(git_stats)

git_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs"),
    repos = "openpharma/DataFakeR"
  )

get_orgs(git_stats)


git_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs"),
    repos = c("openpharma/DataFakeR", "maciekbanas/gitlabr")
  )

get_orgs(git_stats)

git_stats

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    orgs = "mbtests"
  )
get_orgs(gitlab_stats)

devtools::load_all(".")

test_gitstats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs", "openpharma")
  )

get_repos_urls(test_gitstats, with_code = "shiny")

# should return 2 repos URLs
create_gitstats() %>%
  set_github_host(
    repos = c("r-world-devs/GitStats", "r-world-devs/shinyCohortBuilder")
  ) %>%
  get_repos_urls()

# should return 2 repos URLs
create_gitstats() %>%
  set_github_host(
    repos = c("r-world-devs/GitStats", "r-world-devs/shinyCohortBuilder")
  ) %>%
  get_repos_urls(type = "api")

create_gitstats() %>%
  set_gitlab_host(
    orgs = "mbtests"
  ) %>%
  get_repos_urls()

# should return 1 repo URL
create_gitstats() %>%
  set_gitlab_host(
    repos = "mbtests/gitstatstesting"
  ) %>%
  get_repos_urls(type = "api")

devtools::load_all(".")

test_gitstats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs", "openpharma")
  ) |>
  set_gitlab_host(
    orgs = c("mbtests", "mbtestapps")
  )

get_files(
  gitstats = test_gitstats,
  file_path = c("LICENSE", "DESCRIPTION")
)

github_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs"),
    repos = "openpharma/DataFakeR"
  ) |>
  set_gitlab_host(
    repos = "mbtests/graphql_tests"
  )

github_stats

get_files(
  gitstats = github_stats,
  file_path = "DESCRIPTION"
)

datafaker_stats <- create_gitstats() |>
  set_github_host(
    repos = "openpharma/DataFakeR"
  )

get_files(
  gitstats = datafaker_stats,
  file_path = "DESCRIPTION"
)

md_files <- get_files(
  gitstats = test_gitstats,
  pattern = "\\.md|\\.Rmd",
  depth = 2L
)

get_files(
  gitstats = test_gitstats,
  pattern = "DESCRIPTION|\\.md|\\.qmd|\\.Rmd",
  depth = 2L,
  verbose = FALSE
)

devtools::load_all(".")

git_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs",
    repos = "openpharma/visR"
  ) |>
  set_gitlab_host(
    host = "code.roche.com",
    repos = c("datascience/rwd/543_eylea_hd", "RWDInsightsEngineering/RocheMeta")
  )

md_files <- get_files(
  gitstats = git_stats,
  pattern  = "\\.md|\\.Rmd",
  depth    = 2L,
  verbose = TRUE
)

rwd_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs"
  )

get_files(rwd_stats,
          "DESCRIPTION")

rwd_stats <- create_gitstats() |>
  set_github_host(
    repos = c("r-world-devs/GitStats", "openpharma/DataFakeR")
  )
get_files(rwd_stats, "DESCRIPTION")

rwd_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs",
    repos = "openpharma/DataFakeR"
  )
get_files(rwd_stats)

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "RWDInsightsEngineering",
    repos = "datascience/rwd/543_eylea_hd"
  )
get_repos_trees(gitlab_stats)
get_files(gitlab_stats, file_path = "DESCRIPTION")

banasm_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "banasm"
  )
get_files(banasm_stats, "DESCRIPTION")

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    repos = "datascience/rwd/543_eylea_hd"
  )
get_files(gitlab_stats)
get_files(gitlab_stats)

# Issue #605

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    repos = "datascience/rwd/546_copd_marketscan"
  )
get_files(gitlab_stats)
get_files(gitlab_stats, pattern = "\\.md")
get_files(gitlab_stats, pattern = "\\.R$")
get_files(gitlab_stats, pattern = "*")

gitai_stats <- create_gitstats() |>
  set_github_host(
    repos = "r-world-devs/GitAI"
  )
get_files(gitai_stats)
get_files(gitai_stats, pattern = "\\.md")
get_files(gitai_stats, pattern = "\\.R$")
get_files(gitai_stats, pattern = "*")

rwd_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs"
  )

get_files(rwd_stats) # fails

pharma_stats <- create_gitstats() |>
  set_github_host(
    orgs = "pharmaverse"
  )

pharma_repos <- pharma_stats |>
  get_repos(
    add_contributors = FALSE
  )

pharma_stats <- create_gitstats() |>
  set_github_host(
    repos = pharma_repos$fullname
  )

pharma_stats

pharma_stats |>
  get_release_logs(
    since = "2020-01-01",
    cache = FALSE
  )

pharma_stats |>
  get_commits(
    since = "2020-01-01",
    cache = FALSE
  )

pharma_stats |>
  get_repos(
    cache = FALSE
  )

# very slow, better to run it when whole orgs are set
pharma_stats |>
  get_repos(
    with_code = "shiny",
    cache = FALSE
  )

pharma_stats |>
  get_repos_urls(
    cache = FALSE
  )

pharma_stats |>
  get_files(
    pattern = "\\.md",
    depth = 1L
  )

devtools::load_all()

git_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs", "openparma"),
    repos = c("openpharma/DataFakR", "r-world-devs/GitAI"),
    token = Sys.getenv("GITHUB_PAT"),
    .error = FALSE
  ) |>
  set_gitlab_host(
    orgs = c("mbtests", "makbest"),
    repos = c("makbest/something", "mbtests/gitstatstesting", "krystianigras/gitlab-test"),
    token = Sys.getenv("GITLAB_PAT_PUBLIC"),
    .error = FALSE
  )

git_stats

# Setting users instead of orgs

git_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("maciekbanas", "ddsjoberg", "mattsecrest"),
    token = Sys.getenv("GITHUB_PAT")
  )
get_repos(git_stats)
get_repos_urls(git_stats)

git_stats <- create_gitstats() |>
  set_gitlab_host(
    orgs = c("krystianigras"),
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  )

get_repos(git_stats, add_contributors = FALSE)
get_repos_urls(git_stats)

devtools::load_all(".")

test_gitstats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs",
    repos = "openpharma/visR"
  ) |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "datascience/rwd",
    repos = "RWDInsightsEngineering/RocheMeta"
  )

get_release_logs(test_gitstats, since = "2024-01-01")

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "RWDInsightsEngineering"
  )

get_release_logs(gitlab_stats, since = "2024-01-01")


gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "datascience/rwd",
    repos = "RWDInsightsEngineering/RocheMeta"
  )

get_release_logs(gitlab_stats, since = "2024-01-01")

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    repos = "RWDInsightsEngineering/RocheMeta"
  )

get_release_logs(gitlab_stats, since = "2024-01-01")

github_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs",
    repos = "openpharma/visR"
  )

get_release_logs(github_stats, since = "2024-01-01")

github_stats <- create_gitstats() |>
  set_github_host(
    repos = "openpharma/visR"
  )

get_release_logs(github_stats, since = "2024-01-01")

github_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs"
  )

get_release_logs(github_stats, since = "2024-01-01")

github_stats <- create_gitstats() |>
  set_github_host(
    repos = "dgkf/ggpackets"
  )

get_release_logs(github_stats, since = "2022-01-01")

github_stats <- create_gitstats() |>
  set_github_host(
    repos = "dgkf/shinyDataFilter"
  )

get_release_logs(github_stats, since = "2022-01-01")

git_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = c("banasm", "waisk"),
    token = Sys.getenv("GITLAB_PAT")
  )

get_repos(git_stats, add_contributors = FALSE)
get_repos_urls(git_stats)

git_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "RWDInsightsEngineering",
    token = Sys.getenv("GITLAB_PAT")
  )

get_repos(git_stats, add_contributors = FALSE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4,
  message = FALSE
)

## ----eval = FALSE-------------------------------------------------------------
# library(GitStats)
# 
# github_stats <- create_gitstats() %>%
#   set_github_host(
#     orgs = c("r-world-devs", "openpharma"),
#     token = Sys.getenv("GITHUB_PAT")
#   ) %>%
#   verbose_off()
# 
# repos_urls <- get_repos_urls(
#   gitstats = github_stats,
#   with_code = "shiny"
# )

## ----eval = FALSE-------------------------------------------------------------
# repos_urls <- get_repos_urls(
#   gitstats = github_stats,
#   with_code =  c("purrr", "shiny"),
#   in_files = c("DESCRIPTION", "NAMESPACE", "renv.lock")
# )

## ----eval = FALSE-------------------------------------------------------------
# repos_urls <- get_repos_urls(
#   gitstats = github_stats,
#   with_files = c("renv.lock", "DESCRIPTION")
# )

roche_meta_api <- RocheMeta::RocheMetaAPI$new()
roche_projects <- roche_meta_api$get_projects(outputs = "table")
roche_projects$web_url

# remove duplicated data
roche_projects <- roche_projects |>
  dplyr::group_by(name) |>
  dplyr::arrange(desc(lubridate::as_datetime(repo_date)), .by_group = TRUE) |>
  dplyr::slice(1) |>
  dplyr::ungroup()

github_orgs <- roche_projects |> dplyr::filter(host == "github.com") |> dplyr::select(org) |> unlist() |> unique()
github_roche_orgs <- roche_projects |> dplyr::filter(host == "github.roche.com") |> dplyr::select(org) |> unlist() |> unique()
gitlab_roche_orgs <- roche_projects |> dplyr::filter(host == "code.roche.com") |> dplyr::select(org) |> unlist() |> unique()

git_stats <- create_gitstats() |>
  set_github_host(
    orgs = github_orgs,
    token = Sys.getenv("GITHUB_PAT"),
    .error = FALSE
  ) |>
  set_github_host(
    host = "github.roche.com",
    orgs = github_roche_orgs,
    token = Sys.getenv("GITHUB_PAT_ROCHE"),
    .error = FALSE
  ) |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = gitlab_roche_orgs,
    token = Sys.getenv("GITLAB_PAT"),
    .error = FALSE
  )

commits_table <- git_stats |>
  GitStats::get_commits(
    since = "2022-01-01"
  )

releases_table <- git_stats |>
  GitStats::get_release_logs(
    since = "2022-01-01"
  )

issues_table <- git_stats |>
  GitStats::get_issues(
    since = "2022-01-01"
  )

devtools::load_all(".")

roche_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs"
  ) |>
  set_gitlab_host(
    host = "code.roche.com",
    repos = c("RWDInsightsEngineering/RocheData", "RWDInsightsEngineering/GitProjectsExplorer")
  )
roche_stats
roche_stats |>
  get_repos(add_contributors = FALSE)

roche_stats |>
  get_repos(with_code = "Shiny")

roche_stats |>
  get_repos_with_R_packages("dplyr")

commits_stats <- roche_stats |>
  get_commits(since = "2025-01-01") |>
  get_commits_stats(time_aggregation = "month")

roche_stats

commits_stats

git_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs",
    repos = "openpharma/DataFakeR"
  )

get_repos(git_stats)

get_repos_urls(git_stats,
               with_files = "project_metadata.yaml",
               progress = FALSE)

get_repos_urls(git_stats,
               with_files = "project_metadata.yaml",
               cache = FALSE,
               verbose = FALSE,
               progress = TRUE)

get_repos_urls(git_stats,
               with_code = "Shiny",
               in_files = "DESCRIPTION",
               cache = FALSE,
               verbose = FALSE)

get_repos(git_stats)

get_repos(git_stats,
          cache = FALSE,
          verbose = FALSE,
          progress = TRUE)

get_repos(git_stats,
          cache = FALSE,
          verbose = FALSE)

get_repos(git_stats,
          with_code = "Shiny")

get_repos(git_stats,
          with_code = "Shiny",
          cache = FALSE,
          verbose = FALSE)

get_repos(git_stats,
          with_code = "Shiny",
          cache = FALSE,
          verbose = FALSE,
          progress = TRUE)

get_repos(git_stats,
          with_code = "Shiny",
          in_files = "DESCRIPTION",
          cache = FALSE)

get_repos(git_stats,
          with_code = c("shiny", "purrr"),
          in_files = c("DESCRIPTION", "NAMESPACE"),
          verbose = FALSE)

get_commits(git_stats, since = "2024-06-01")

get_commits(git_stats,
            since = "2024-06-02",
            verbose = FALSE,
            progress = TRUE)

get_release_logs(
  gitstats = git_stats,
  since = "2024-06-02",
  verbose = FALSE
)

get_release_logs(
  gitstats = git_stats,
  since = "2024-06-01",
  verbose = TRUE
)

devtools::load_all(".")

git_stats <- create_gitstats() %>%
  set_github_host(
    orgs = c("r-world-devs"),
    repos = c("openpharma/DataFakeR"),
    token = Sys.getenv("GITHUB_PAT")
  ) %>%
  set_gitlab_host(
    orgs = c("mbtests"),
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  )

issues <- get_issues(
  gitstats = git_stats,
  since = "2024-01-01",
  verbose = TRUE
)

issues

open_issues <- get_issues(
  gitstats = git_stats,
  since = "2024-01-01",
  state = "open",
  verbose = TRUE
)

closed_issues <- get_issues(
  gitstats = git_stats,
  since = "2023-01-01",
  state = "closed",
  verbose = TRUE
)

# Check printing in storage
git_stats

issues_stats <- create_gitstats() %>%
  set_github_host(
    orgs = c("r-world-devs", "openpharma"),
    token = Sys.getenv("GITHUB_PAT")
  ) %>%
  set_gitlab_host(
    orgs = c("mbtests"),
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  ) |>
  get_issues(
    since = "2024-01-01"
  ) |>
  get_issues_stats(
    time_aggregation = "month",
    group_var = author
  )

devtools::load_all(".")

create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs"
  ) |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "RWDInsightsEngineering"
  ) |>
  get_commits(
    since = "2022-06-01",
    verbose = TRUE
  )

create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    repos = c("datascience/rwd/543_eylea_hd", "datascience/rwd/2016_eylea_vestrum_analysis")
  ) |>
  get_commits(since = "2020-01-01")

create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    repos = c("nest/insights-engineering/rtables")
  ) |>
  get_commits(since = "2020-01-01")

create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "nest/insights-engineering"
  ) |>
  get_commits(since = "2022-01-01")

create_gitstats() |>
  set_github_host(
    host = "github.roche.com",
    orgs = c("E4ATI", "RWDScodeshare", "PHC")
  ) |>
  get_commits(since = "2022-01-01")

create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs"
  ) |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "RWDInsightsEngineering"
  ) |>
  set_github_host(
    host = "github.roche.com",
    orgs = "RWDScodeshare"
  )

get_commits(
  gitstats = rwdie_stats,
  since = "2022-06-01",
  verbose = TRUE
)

get_commits_stats(
  gitstats = rwdie_stats,
  group_var = author,
  time_aggregation = "year"
)

git_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs",
    repos = "openpharma/visR"
  ) |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "RWDInsightsEngineering",
    repos = c("datascience/rwd/543_eylea_hd", "datascience/rwd/2016_eylea_vestrum_analysis")
  )

get_commits(git_stats, since = "2024-01-01")

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "RWDInsightsEngineering"
  )

get_commits(gitlab_stats, since = "2024-01-01")

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "RWDInsightsEngineering",
    repos = c("datascience/rwd/543_eylea_hd", "datascience/rwd/2016_eylea_vestrum_analysis")
  )

get_commits(gitlab_stats, since = "2024-01-01")

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    repos = "RWDInsightsEngineering/RocheMeta"
  )

get_commits(gitlab_stats, since = "2024-01-01")

github_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs",
    repos = "openpharma/visR"
  )

get_commits(github_stats, since = "2024-01-01")

github_stats <- create_gitstats() |>
  set_github_host(
    repos = "openpharma/visR"
  )

get_commits(github_stats, since = "2024-01-01")

github_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs"
  )

get_commits(github_stats, since = "2024-01-01")

github_stats <- create_gitstats() |>
  set_github_host(
    repos = "dgkf/shinyDataFilter"
  )

get_commits(github_stats, since = "2022-01-01") |>
  get_commits_stats(group_var = author_name)

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    repos = "datascience/rwd/543_eylea_hd"
  )

get_repos(gitlab_stats)

get_repos_trees(gitlab_stats)

get_repos_urls(gitlab_stats)

get_repos_urls(gitlab_stats, with_files = "README.md")

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    repos = "RWDInsightsEngineering/EEADashboard"
  )

get_repos(gitlab_stats, with_code = "shiny")

get_repos_urls(gitlab_stats, with_files = "DESCRIPTION")

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    repos = c("datascience/rwd/543_eylea_hd", "datascience/rwd/4308_es-sclc-atezo-vs-durva", "RWDInsightsEngineering/RocheMetrics")
  )

get_repos(gitlab_stats)

get_repos(gitlab_stats, with_code = "tests")
get_repos(gitlab_stats, with_code = "quarto")
get_repos(gitlab_stats, with_files = "README.md")

get_repos_urls(gitlab_stats)

get_repos_urls(gitlab_stats, with_code = "tests")
get_repos_urls(gitlab_stats, with_code = "quarto")
get_repos_urls(gitlab_stats, with_files = "README.md")


gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "RWDInsightsEngineering"
  )

get_repos(gitlab_stats)

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "RWDInsightsEngineering",
    repos = c("datascience/rwd/543_eylea_hd", "datascience/rwd/4308_es-sclc-atezo-vs-durva")
  )

get_repos(gitlab_stats, add_contributors = FALSE)

get_repos_urls(gitlab_stats)

get_repos_urls(gitlab_stats, with_code = "quarto")
get_repos_urls(gitlab_stats, with_files = "DESCRIPTION")
get_repos_urls(gitlab_stats, with_files = "project_metadata.yaml")
get_repos_urls(gitlab_stats, with_files = c("project_metadata.yaml", "meta_data.yaml"))

github_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs",
    repos = c("dgkf/shinyDataFilter", "ddsjoberg/gtsummary", "mattsecrest/survivalCCW", "mattsecrest/ggTreatmentSequence")
  )

get_repos(github_stats)

get_repos(github_stats, with_code = "plot", add_contributors = FALSE)

get_repos_urls(github_stats)

get_repos_urls(github_stats, with_code = "shiny")
get_repos_urls(github_stats, type = "web", with_code = "shiny")

git_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    repos = c("datascience/rwd/543_eylea_hd", "datascience/rwd/4308_es-sclc-atezo-vs-durva", "RWDInsightsEngineering/RocheMetrics")
  ) |>
  set_github_host(
    orgs = "r-world-devs",
    repos = c("dgkf/shinyDataFilter", "ddsjoberg/gtsummary", "mattsecrest/survivalCCW", "mattsecrest/ggTreatmentSequence")
  )

get_repos(git_stats)
get_repos(git_stats, with_files = "DESCRIPTION")
get_repos_urls(git_stats, with_files = c("project_metadata.yaml", "meta_data.yaml"))

get_repos_urls(git_stats, with_code = c("shiny", "purrr"))

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com"
  )

get_repos_urls(gitlab_stats, type = "api", with_code = "RocheData")
get_repos_urls(gitlab_stats, type = "web", with_code = "RocheTemplates")
get_repos_urls(gitlab_stats, type = "api", with_files = c("meta_data.yaml", "project_metadata.yaml"))

github_stats <- create_gitstats() |>
  set_github_host(
    host = "github.roche.com"
  )

get_repos_urls(github_stats, type = "api", with_code = "RocheData")
get_repos_urls(github_stats, type = "web", with_code = "RocheTemplates")
get_repos_urls(github_stats, type = "api", with_files = c("meta_data.yaml", "project_metadata.yaml"))

get_repos(gitlab_stats)

get_repos(github_stats)

git_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com"
  ) |>
  set_github_host(
    host = "github.roche.com"
  )

repos_with_desc <- get_repos(git_stats, with_files = "DESCRIPTION", cache = FALSE)

get_repos_urls(git_stats, type = "api", with_files = c("meta_data.yaml", "project_metadata.yaml"))

git_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs", "openpharma")
  ) |>
  set_gitlab_host(
    host = "code.roche.com"
  ) |>
  set_github_host(
    host = "github.roche.com"
  )

get_repos_urls(git_stats, type = "api", with_files = c("meta_data.yaml", "project_metadata.yaml"))

create_gitstats() |>
  set_github_host(
    orgs = c("openpharma")
  ) |> get_repos_urls(type = "api", with_files = c("meta_data.yaml", "project_metadata.yaml"))

gitlab_stats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "pred-clinops"
  )

get_repos(gitlab_stats)


test_gitstats <- create_gitstats() |>
  set_gitlab_host(
    host = "code.roche.com",
    orgs = "rapp/aosp/platform/packages/modules"
  )

get_repos_trees(test_gitstats, depth = 0)

EngineGraphQLGitHub <- R6::R6Class(
  classname = "EngineGraphQLGitHub",
  inherit = EngineGraphQL,
  public = list(
    initialize = function(gql_api_url,
                          token,
                          scan_all = FALSE) {
      super$initialize(
        gql_api_url = gql_api_url,
        token = token,
        scan_all = scan_all
      )
      self$gql_query <- GQLQueryGitHub$new()
    },

    set_owner_type = function(owners, verbose) {
      user_or_org_query <- self$gql_query$user_or_org_query
      login_types <- gitstats_map(owners, function(owner) {
        cached <- private[["owner_types_cache"]][[owner]]
        if (!is.null(cached)) {
          return(cached)
        }
        response <- self$gql_response(
          gql_query = user_or_org_query,
          vars = list(
            "login" = owner
          ),
          verbose = verbose
        )
        if (length(response$errors) < 2) {
          type <- purrr::discard(response$data, is.null) |>
            names()
          attr(owner, "type") <- type
        } else {
          attr(owner, "type") <- "not found"
        }
        private[["owner_types_cache"]][[owner]] <- owner
        return(owner)
      })
      return(login_types)
    },

    get_orgs = function(output = c("only_names", "full_table"), verbose) {
      if (verbose) {
        cli::cli_alert("[Host:GitHub][Engine:{cli::col_yellow('GraphQL')}] Pulling organizations {cli_icons$org}...")
      }
      end_cursor <- ""
      has_next_page <- TRUE
      full_orgs_list <- list()
      while (has_next_page) {
        response <- self$gql_response(
          gql_query = self$gql_query$orgs(
            end_cursor = end_cursor
          ),
          verbose = verbose
        )
        if (!inherits(response, "graphql_error")) {
          if (output == "only_names") {
            orgs_list <- purrr::map(response$data$search$edges, ~ .$node$login)
          } else {
            orgs_list <- purrr::map(response$data$search$edges, ~ .$node)
          }
          full_orgs_list <- append(full_orgs_list, orgs_list)
          has_next_page <- response$data$search$pageInfo$hasNextPage
          end_cursor <- response$data$search$pageInfo$endCursor
        } else {
          full_orgs_list <- response
          break
        }
      }
      all_orgs <- if (output == "only_names") {
        unlist(full_orgs_list)
      } else {
        full_orgs_list
      }
      return(all_orgs)
    },

    prepare_orgs_table = function(full_orgs_list) {
      orgs_table <- purrr::map(full_orgs_list, function(orgs_node) {
        orgs_node$description <- orgs_node$description %||% ""
        orgs_node$repos_count <- orgs_node$repositories$totalCount
        orgs_node$members_count <- orgs_node$membersWithRole$totalCount
        orgs_node$membersWithRole <- NULL
        orgs_node$repositories <- NULL
        data.frame(orgs_node)
      }) |>
        purrr::list_rbind() |>
        dplyr::rename(path = login,
                      avatar_url = avatarUrl) |>
        dplyr::relocate(description, .after = name) |>
        tibble::as_tibble()
      return(orgs_table)
    },

    get_org = function(org, verbose = TRUE) {
      response <- self$gql_response(
        gql_query = self$gql_query$org(),
        vars = list("org" = org),
        verbose = verbose
      )
      return(response$data$organization)
    },

    get_repos = function(repos_ids, verbose) {
      repos_query <- self$gql_query$repos_by_ids()
      if (length(repos_ids) > 100) {
        iterations_number <- ceiling(length(repos_ids) / 100)
        x <- 1
        repos_nodes <- purrr::map(c(1:iterations_number), function(i) {
          if (i == iterations_number) {
            repos_ids_length <- length(repos_ids)
          } else {
            repos_ids_length <- i * 100
          }
          repos_response <- self$gql_response(
            gql_query = repos_query,
            vars = list(
              "ids" = repos_ids[x:repos_ids_length]
            ),
            verbose = verbose
          )
          x <<- x + 100
          return(repos_response$data$nodes)
        }) |>
          purrr::list_flatten()
      } else {
        repos_response <- self$gql_response(
          gql_query = repos_query,
          vars = list(
            "ids" = repos_ids
          ),
          verbose = verbose
        )
        repos_nodes <- repos_response$data$nodes
      }
      return(repos_nodes)
    },

    get_repos_from_org = function(org = NULL,
                                  owner_type = c("organization", "user"),
                                  verbose = TRUE) {
      full_repos_list <- list()
      next_page <- TRUE
      repo_cursor <- ""
      while (next_page) {
        repos_response <- private$get_repos_page(
          login = org,
          type = owner_type,
          repo_cursor = repo_cursor,
          verbose = verbose
        )
        repositories <- if (owner_type == "organization") {
          repos_response$data$repositoryOwner$repositories
        } else {
          repos_response$data$user$repositories
        }
        repos_list <- repositories$nodes
        next_page <- repositories$pageInfo$hasNextPage
        if (is.null(next_page)) next_page <- FALSE
        if (is.null(repos_list)) repos_list <- list()
        if (next_page) {
          repo_cursor <- repositories$pageInfo$endCursor
        } else {
          repo_cursor <- ""
        }
        full_repos_list <- append(full_repos_list, repos_list)
      }
      return(full_repos_list)
    },

    prepare_repos_table = function(repos_list, org) {
      if (length(repos_list) > 0) {
        repos_table <- purrr::map(repos_list, function(repo) {
          default_branch <- if (!is.null(repo$default_branch)) {
            repo$default_branch$name
          } else {
            ""
          }
          last_activity_at <- as.POSIXct(repo$last_activity_at)
          if (length(last_activity_at) == 0) {
            last_activity_at <- gts_to_posixt(repo$created_at)
          }
          data.frame(
            repo_id = repo$repo_id,
            repo_name = repo$repo_path,
            repo_fullpath = repo$repo_fullpath,
            default_branch = default_branch,
            stars = repo$stars,
            forks = repo$forks,
            created_at = gts_to_posixt(repo$created_at),
            last_activity_at = last_activity_at,
            languages = purrr::map_chr(repo$languages$nodes, ~ .$name) |>
              paste0(collapse = ", "),
            issues_open = repo$issues_open$totalCount,
            issues_closed = repo$issues_closed$totalCount,
            organization = repo$organization$login,
            repo_url = repo$repo_url,
            commit_sha = repo$defaultBranchRef$target$oid %||% NA_character_
          )
        }) |>
          purrr::list_rbind()
      } else {
        repos_table <- NULL
      }
      return(repos_table)
    },

    get_commits_from_repos = function(org,
                                      repos_names,
                                      since,
                                      until,
                                      verbose) {
      repos_list_with_commits <- gitstats_map(repos_names, function(repo) {
        private$get_commits_from_one_repo(
          org   = org,
          repo  = repo,
          since = since,
          until = until,
          verbose = verbose
        )
      })
      names(repos_list_with_commits) <- repos_names
      repos_list_with_commits <- repos_list_with_commits |>
        purrr::discard(~ length(.) == 0)
      return(repos_list_with_commits)
    },

    prepare_commits_table = function(repos_list_with_commits,
                                     org) {
      commits_table <- purrr::imap(repos_list_with_commits, function(repo, repo_name) {
        commits_row <- purrr::map(repo, function(commit) {
          if (!is.null(commit$node)) {
            commit_author <- commit$node$author
            data.frame(
              id = commit$node$id,
              committed_date = gts_to_posixt(commit$node$committed_date),
              author = commit_author$name,
              author_login = commit_author$user$login %||% NA_character_,
              author_name = commit_author$user$name %||% NA_character_,
              additions = commit$node$additions,
              deletions = commit$node$deletions,
              repo_url = commit$node$repository$url
            )
          } else {
            NULL
          }
        }) |>
          purrr::list_rbind()
        commits_row$repo_name <- repo_name
        commits_row
      }) |>
        purrr::discard(~ length(.) == 1) |>
        purrr::list_rbind()
      if (nrow(commits_table) > 0) {
        commits_table <- commits_table |>
          dplyr::mutate(
            organization = org,
            api_url = self$gql_api_url
          ) |>
          dplyr::relocate(
            repo_name,
            .before = id
          ) |>
          dplyr::relocate(
            any_of(c("author_login", "author_name")),
            .after = author
          ) |>
          dplyr::relocate(
            repo_url,
            .before = api_url
          )
      }
      commits_table <- private$fill_empty_authors(commits_table)
      return(commits_table)
    },

    get_files_from_org = function(org,
                                  owner_type,
                                  repos_data,
                                  file_paths = NULL,
                                  host_files_structure = NULL,
                                  verbose = TRUE) {
      repositories <- repos_data[["paths"]]
      def_branches <- repos_data[["def_branches"]]
      org_files_list <- private$get_repositories_with_files(
        repositories = repositories,
        def_branches = def_branches,
        org = org,
        file_paths = file_paths,
        host_files_structure = host_files_structure,
        verbose = verbose
      )
      names(org_files_list) <- repositories
      org_files_list <- purrr::map(org_files_list, function(repo_data) {
        purrr::discard(repo_data, function(file_data) {
          length(file_data$file) == 0
        })
      })
      return(org_files_list)
    },

    prepare_files_table = function(files_response, org) {
      if (!is.null(files_response)) {
        files_table <- purrr::map(files_response, function(repository) {
          purrr::discard(repository, ~ length(.$file) == 0) |>
            purrr::imap(function(file_data, file_name) {
              data.frame(
                "repo_id" = file_data$repo_id,
                "repo_name" = file_data$repo_name,
                "organization" = org,
                "file_path" = file_name,
                "file_content" = file_data$file$text %||% NA,
                "file_size" = file_data$file$byteSize,
                "file_id" = file_data$file$oid,
                "repo_url" = file_data$repo_url,
                "commit_sha" = file_data$defaultBranchRef$target$oid
              )
            }) |>
            purrr::list_rbind()
        }) |>
          purrr::list_rbind()
      } else {
        files_table <- NULL
      }
      return(files_table)
    },

    prepare_user_table = function(user_response) {
      if (!is.null(user_response$data$user)) {
        user_data <- user_response$data$user
        user_data[["name"]] <- user_data$name %||% ""
        user_data[["starred_repos"]] <- user_data$starred_repos$totalCount
        user_data[["commits"]] <- user_data$contributions$totalCommitContributions
        user_data[["issues"]] <- user_data$contributions$totalIssueContributions
        user_data[["pull_requests"]] <- user_data$contributions$totalPullRequestContributions
        user_data[["reviews"]] <- user_data$contributions$totalPullRequestReviewContributions
        user_data[["contributions"]] <- NULL
        user_data[["email"]] <- user_data$email %||% ""
        user_data[["location"]] <- user_data$location %||% ""
        user_data[["web_url"]] <- user_data$web_url %||% ""
        user_table <- tibble::as_tibble(user_data) |>
          dplyr::relocate(c(commits, issues, pull_requests, reviews),
            .after = starred_repos
          )
      } else {
        user_table <- NULL
      }
      return(user_table)
    },

    get_release_logs_from_org = function(repos_names, org, verbose = TRUE) {
      release_responses <- gitstats_map(repos_names, function(repository) {
        releases_from_repo_query <- self$gql_query$releases_from_repo()
        response <- self$gql_response(
          gql_query = releases_from_repo_query,
          vars = list(
            "org" = org,
            "repo" = repository
          ),
          verbose = verbose
        )
        return(response)
      }) |>
        purrr::discard(~ length(.$data$repository$releases$nodes) == 0)
      return(release_responses)
    },

    prepare_releases_table = function(releases_response, org) {
      if (length(releases_response) > 0) {
        releases_table <-
          purrr::map(releases_response, function(release) {
            release_table <- purrr::map(release$data$repository$releases$nodes, function(node) {
              data.frame(
                release_name = node$name,
                release_tag = node$tagName,
                published_at = gts_to_posixt(node$publishedAt),
                release_url = node$url,
                release_log = node$description
              )
            }) |>
              purrr::list_rbind() |>
              dplyr::mutate(
                repo_name = release$data$repository$name,
                repo_url = release$data$repository$url
              ) |>
              dplyr::relocate(
                repo_name, repo_url,
                .before = release_name
              )
            return(release_table)
          }) |>
          purrr::list_rbind()
      } else {
        releases_table <- NULL
      }
      return(releases_table)
    }
  ),
  private = list(
    get_repos_page = function(login = NULL,
                              type = c("organization"),
                              repo_cursor = "",
                              verbose = TRUE) {
      if (type == "organization") {
        repos_query <- self$gql_query$repos_by_org(
          repo_cursor = repo_cursor
        )
      } else if (type == "user") {
        repos_query <- self$gql_query$repos_by_user(
          repo_cursor = repo_cursor
        )
      }
      response <- self$gql_response(
        gql_query = repos_query,
        vars = list(
          "login" = login
        ),
        verbose = verbose
      )
      private$handle_graphql_error(
        responses_list = response,
        verbose = verbose
      )
      return(response)
    },

    get_commits_from_one_repo = function(org,
                                         repo,
                                         since,
                                         until,
                                         verbose = TRUE) {
      next_page <- TRUE
      full_commits_list <- list()
      commits_cursor <- ""
      while (next_page) {
        commits_response <- private$get_commits_page_from_repo(
          org = org,
          repo = repo,
          since = since,
          until = parse_until_param(until),
          commits_cursor = commits_cursor,
          verbose = verbose
        )
        if (!inherits(commits_response, "graphql_error")) {
          commits_list <- commits_response$data$repository$defaultBranchRef$target$history$edges
          next_page <- commits_response$data$repository$defaultBranchRef$target$history$pageInfo$hasNextPage
          if (is.null(next_page)) next_page <- FALSE
          if (is.null(commits_list)) commits_list <- list()
          if (next_page) {
            commits_cursor <- commits_response$data$repository$defaultBranchRef$target$history$pageInfo$endCursor
          } else {
            commits_cursor <- ""
          }
        } else {
          commits_list <- NULL
          next_page <- FALSE
        }
        full_commits_list <- append(full_commits_list, commits_list)
      }
      return(full_commits_list)
    },

    get_commits_page_from_repo = function(org,
                                          repo,
                                          since,
                                          until,
                                          commits_cursor = "",
                                          verbose = TRUE) {
      commits_from_repo_query <- self$gql_query$commits_from_repo(
        commits_cursor = commits_cursor
      )
      response <- self$gql_response(
        gql_query = commits_from_repo_query,
        vars = list(
          "org" = org,
          "repo" = repo,
          "since" = date_to_gts(since),
          "until" = date_to_gts(until)
        ),
        verbose = verbose
      )
      if (inherits(response, "graphql_error")) {
        response <- self$gql_response(
          gql_query = commits_from_repo_query,
          vars = list(
            "org" = org,
            "repo" = repo,
            "since" = date_to_gts(since),
            "until" = date_to_gts(until)
          ),
          verbose = verbose
        )
      }
      return(response)
    },

    get_pr_page_from_repo = function(org,
                                     repo,
                                     pr_cursor = "",
                                     verbose = TRUE) {
      pr_from_repo_query <- self$gql_query$pull_requests_from_repo(
        pr_cursor = pr_cursor
      )
      response <- self$gql_response(
        gql_query = pr_from_repo_query,
        vars = list(
          "org" = org,
          "repo" = repo
        ),
        verbose = verbose
      )
      if (inherits(response, "graphql_error")) {
        response <- self$gql_response(
          gql_query = pr_from_repo_query,
          vars = list(
            "org" = org,
            "repo" = repo
          ),
          verbose = verbose
        )
      }
      return(response)
    },

    get_pr_from_one_repo = function(org,
                                    repo,
                                    verbose = TRUE) {
      next_page <- TRUE
      full_pr_list <- list()
      pr_cursor <- ""
      while (next_page) {
        pr_response <- private$get_pr_page_from_repo(
          org = org,
          repo = repo,
          pr_cursor = pr_cursor,
          verbose = verbose
        )
        pr_list <- pr_response$data$repository$pullRequests$edges
        next_page <- pr_response$data$repository$pullRequests$pageInfo$hasNextPage
        if (is.null(next_page)) next_page <- FALSE
        if (is.null(pr_list)) pr_list <- list()
        if (next_page) {
          pr_cursor <- pr_response$data$repository$pullRequests$pageInfo$endCursor
        } else {
          pr_cursor <- ""
        }
        full_pr_list <- append(full_pr_list, pr_list)
      }
      return(full_pr_list)
    },

    get_issues_from_one_repo = function(org,
                                        repo,
                                        verbose = TRUE) {
      next_page <- TRUE
      full_issues_list <- list()
      issues_cursor <- ""
      while (next_page) {
        issues_response <- private$get_issues_page_from_repo(
          org = org,
          repo = repo,
          issues_cursor = issues_cursor,
          verbose = verbose
        )
        issues_list <- issues_response$data$repository$issues$edges
        next_page <- issues_response$data$repository$issues$pageInfo$hasNextPage
        if (is.null(next_page)) next_page <- FALSE
        if (is.null(issues_list)) issues_list <- list()
        if (next_page) {
          issues_cursor <- issues_response$data$repository$issues$pageInfo$endCursor
        } else {
          issues_cursor <- ""
        }
        full_issues_list <- append(full_issues_list, issues_list)
      }
      return(full_issues_list)
    },

    get_issues_page_from_repo = function(org,
                                         repo,
                                         issues_cursor = "",
                                         verbose = TRUE) {
      issues_from_repo_query <- self$gql_query$issues_from_repo(
        issues_cursor = issues_cursor
      )
      response <- self$gql_response(
        gql_query = issues_from_repo_query,
        vars = list(
          "org" = org,
          "repo" = repo
        ),
        verbose = verbose
      )
      if (inherits(response, "graphql_error")) {
        response <- self$gql_response(
          gql_query = issues_from_repo_query,
          vars = list(
            "org" = org,
            "repo" = repo
          ),
          verbose = verbose
        )
      }
      return(response)
    },

    get_repositories_with_files = function(repositories,
                                           def_branches,
                                           org,
                                           host_files_structure,
                                           file_paths,
                                           verbose) {
      gitstats_map2(repositories, def_branches, function(repo, def_branch) {
        if (!is.null(host_files_structure)) {
          file_paths <- private$get_path_from_files_structure(
            host_files_structure = host_files_structure,
            org = org,
            repo = repo
          )
        } else if (is.null(host_files_structure)) {
          file_paths <- file_paths[grepl(text_files_pattern, file_paths)]
        }
        repo_files_list <- purrr::map(file_paths, function(file_path) {
          private$get_file_response(
            org = org,
            repo = repo,
            def_branch = def_branch,
            file_path = file_path,
            files_query = self$gql_query$file_blob_from_repo(),
            verbose = verbose
          )
        }) |>
          purrr::map(~ .$data$repository)
        names(repo_files_list) <- file_paths
        return(repo_files_list)
      })
    },

    get_file_response = function(org, repo, def_branch, file_path, files_query, verbose = TRUE) {
      expression <- paste0(def_branch, ":", file_path)
      files_response <- self$gql_response(
        gql_query = files_query,
        vars = list(
          "org" = org,
          "repo" = repo,
          "expression" = expression
        ),
        verbose = verbose
      )
      return(files_response)
    },

    fill_empty_authors = function(commits_table) {
      if (length(commits_table) > 0) {
        commits_table <- commits_table |>
          dplyr::rowwise() |>
          dplyr::mutate(
            author_name = ifelse(is.na(author_name) & is_name(author), author, author_name),
            author_login = ifelse(is.na(author_login) & is_login(author), author, author_login)
          )
      }
      return(commits_table)
    }
  )
)

test_that("file queries for GitHub are built properly", {
  gh_file_blobs_from_repo_query <-
    test_gqlquery_gh$file_blob_from_repo()
  expect_snapshot(
    gh_file_blobs_from_repo_query
  )
  test_mocker$cache(gh_file_blobs_from_repo_query)
})

test_that("GitHub GraphQL Engine pulls file response", {
  mockery::stub(
    test_graphql_github_priv$get_file_response,
    "self$gql_response",
    test_fixtures$github_file_response
  )
  github_file_response <- test_graphql_github_priv$get_file_response(
    org         = "r-world-devs",
    repo        = "GitStats",
    def_branch  = "master",
    file_path   = "LICENSE",
    files_query = test_mocker$use("gh_file_blobs_from_repo_query")
  )
  expect_files_github_response(github_file_response)
  test_mocker$cache(github_file_response)
})

test_that("get_repositories_with_files works", {
  mockery::stub(
    test_graphql_github_priv$get_repositories_with_files,
    "private$get_file_response",
    test_mocker$use("github_file_response")
  )
  gh_repositories_with_files <- test_graphql_github_priv$get_repositories_with_files(
    repositories = c("test_repo_1", "test_repo_2", "test_repo_3", "test_repo_4", "test_repo_5"),
    def_branches = c("test_branch_1", "test_branch_2", "test_branch_3", "test_branch_4", "test_branch_5"),
    org = "test_org",
    file_paths = "test_files.txt",
    host_files_structure = NULL
  )
  expect_type(gh_repositories_with_files, "list")
  test_mocker$cache(gh_repositories_with_files)
})

test_that("get_repositories_with_files works for two files even if some of them do not exist in one of repositories", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_github_priv$get_repositories_with_files,
      "private$get_file_response",
      test_mocker$use("github_file_response")
    )
  }  
  gh_repositories_with_two_files <- test_graphql_github_priv$get_repositories_with_files(
    repositories = c("GitStats", "GitAI", "cohortBuilder"),
    def_branches = c("master", "main", "dev"),
    org = "r-world-devs",
    file_paths = c("DESCRIPTION", "project_metadata.yaml"),
    host_files_structure = NULL
  )
  expect_type(gh_repositories_with_two_files, "list")
  test_mocker$cache(gh_repositories_with_two_files)
})

test_that("GitHub GraphQL Engine pulls files from organization", {
  mockery::stub(
    test_graphql_github$get_files_from_org,
    "private$get_repos_data",
    test_mocker$use("gh_repos_data")
  )
  mockery::stub(
    test_graphql_github$get_files_from_org,
    "private$get_repositories_with_files",
    test_mocker$use("gh_repositories_with_files")
  )
  github_files_response <- test_graphql_github$get_files_from_org(
    org = "test_org",
    owner_type = "organization",
    repos = NULL,
    file_paths = "test_files.txt",
    host_files_structure = NULL,
    verbose = FALSE
  )
  expect_files_github_response(github_files_response)
  test_mocker$cache(github_files_response)
})

test_that("GitHub GraphQL Engine pulls files from organization", {
  mockery::stub(
    test_graphql_github$get_files_from_org,
    "private$get_repos_data",
    test_mocker$use("gh_repos_data")
  )
  mockery::stub(
    test_graphql_github$get_files_from_org,
    "private$get_repositories_with_files",
    test_mocker$use("gh_repositories_with_two_files")
  )
  github_two_files_response <- test_graphql_github$get_files_from_org(
    org = "r-world-devs",
    owner_type = "organization",
    repos = NULL,
    file_paths = c("DESCRIPTION", "project_metadata.yaml"),
    host_files_structure = NULL,
    verbose = FALSE
  )
  expect_files_github_response(github_two_files_response)
  test_mocker$cache(github_two_files_response)
})

test_that("GitHubHost prepares table from files response", {
  gh_files_table <- test_graphql_github$prepare_files_table(
    files_response = test_mocker$use("github_files_response"),
    org = "r-world-devs"
  )
  expect_files_table(gh_files_table)
  test_mocker$cache(gh_files_table)
})

test_that("GitHubHost prepares table from files response", {
  gh_two_files_table <- test_graphql_github$prepare_files_table(
    files_response = test_mocker$use("github_two_files_response"),
    org = "r-world-devs"
  )
  expect_files_table(gh_two_files_table)
  if (!integration_tests_skipped) {
    expect_true(all(unique(gh_two_files_table$repo_name) == c("GitStats", "GitAI", "cohortBuilder")))
  }  
  test_mocker$cache(gh_two_files_table)
})

test_that("get_files_content_from_orgs for GitHub works", {
  mockery::stub(
    github_testhost_priv$get_files_content_from_orgs,
    "graphql_engine$prepare_files_table",
    test_mocker$use("gh_files_table")
  )
  github_testhost_priv$searching_scope <- "org"
  gh_files_from_orgs <- github_testhost_priv$get_files_content_from_orgs(
    file_path = "DESCRIPTION",
    verbose = FALSE
  )
  expect_files_table(
    gh_files_from_orgs,
    with_cols = "api_url"
  )
  test_mocker$cache(gh_files_from_orgs)
})

test_that("get_files_content_from_orgs for GitHub works with two files", {
  mockery::stub(
    github_testhost_priv$get_files_content_from_orgs,
    "graphql_engine$prepare_files_table",
    test_mocker$use("gh_two_files_table")
  )
  github_testhost_priv$searching_scope <- "org"
  gh_two_files_from_orgs <- github_testhost_priv$get_files_content_from_orgs(
    file_path = c("DESCRIPTION", "project_metadata.yaml"),
    verbose = FALSE
  )
  expect_files_table(
    gh_two_files_from_orgs,
    with_cols = "api_url"
  )
  test_mocker$cache(gh_two_files_from_orgs)
})

test_that("get_files_content_from_repos for GitHub works", {
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    github_testhost_priv$get_files_content_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  mockery::stub(
    github_testhost_priv$get_files_content_from_repos,
    "private$get_repos_data",
    test_mocker$use("gh_repos_data")
  )
  mockery::stub(
    github_testhost_priv$get_files_content_from_repos,
    "graphql_engine$prepare_files_table",
    test_mocker$use("gh_files_table")
  )
  github_testhost_priv$searching_scope <- "repo"
  gh_files_from_repos <- github_testhost_priv$get_files_content_from_repos(
    file_path = "DESCRIPTION",
    verbose = FALSE
  )
  expect_files_table(
    gh_files_from_repos,
    with_cols = "api_url"
  )
  test_mocker$cache(gh_files_from_repos)
})

test_that("get_files_content_from_orgs for GitHub prints messages", {
  mockery::stub(
    github_testhost_priv$get_files_content_from_orgs,
    "graphql_engine$prepare_files_table",
    test_mocker$use("gh_files_table")
  )
  mockery::stub(
    github_testhost_priv$get_files_content_from_orgs,
    "private$get_repos_data",
    test_mocker$use("gh_repos_data")
  )
  github_testhost_priv$searching_scope <- "org"
  expect_snapshot(
    gh_files_from_orgs <- github_testhost_priv$get_files_content_from_orgs(
      file_path = "DESCRIPTION",
      verbose = TRUE
    )
  )
})

test_that("get_files_content_from_repos for GitHub prints messages", {
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    github_testhost_priv$get_files_content_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  mockery::stub(
    github_testhost_priv$get_files_content_from_repos,
    "private$get_repos_data",
    test_mocker$use("gh_repos_data")
  )
  mockery::stub(
    github_testhost_priv$get_files_content_from_repos,
    "graphql_engine$prepare_files_table",
    test_mocker$use("gh_files_table")
  )
  github_testhost_priv$searching_scope <- "repo"
  expect_snapshot(
    gh_files_from_repos <- github_testhost_priv$get_files_content_from_repos(
      file_path = "DESCRIPTION",
      verbose = TRUE
    )
  )
})

test_that("get_files_content makes use of files_structure", {
  mockery::stub(
    github_testhost_priv$get_files_content_from_files_structure,
    "private$get_repos_data",
    test_mocker$use("gh_repos_data")
  )
  mockery::stub(
    github_testhost_priv$get_files_content_from_files_structure,
    "graphql_engine$prepare_files_table",
    test_mocker$use("gh_files_table")
  )
  expect_snapshot(
    files_content <- github_testhost_priv$get_files_content_from_files_structure(
      files_structure = test_mocker$use("gh_files_structure_from_orgs"),
      verbose = TRUE,
      progress = FALSE
    )
  )
  expect_files_table(
    files_content,
    with_cols = "api_url"
  )
})

test_that("get_files_content skips if no files found in files_structure", {
  expect_snapshot(
    files_content <- github_testhost_priv$get_files_content_from_files_structure(
      files_structure = test_mocker$use("gh_empty_files_structure"),
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("`get_files_content()` pulls files in the table format", {
  github_testhost <- create_github_testhost(orgs = "test_org")
  mockery::stub(
    github_testhost$get_files_content,
    "private$get_files_content_from_orgs",
    test_mocker$use("gh_files_from_orgs")
  )
  mockery::stub(
    github_testhost$get_files_content,
    "private$get_files_content_from_repos",
    test_mocker$use("gh_files_from_repos")
  )
  gh_files_table <- github_testhost$get_files_content(
    file_path = "DESCRIPTION"
  )
  expect_files_table(gh_files_table, with_cols = "api_url")
  test_mocker$cache(gh_files_table)
})

test_that("`get_files_content()` pulls two files in the table format", {
  github_testhost <- create_github_testhost(orgs = "r-world-devs")
  mockery::stub(
    github_testhost$get_files_content,
    "private$get_files_content_from_orgs",
    test_mocker$use("gh_two_files_from_orgs")
  )
  gh_files_table <- github_testhost$get_files_content(
    file_path = c("DESCRIPTION", "project_metadata.yaml")
  )
  expect_files_table(gh_files_table, with_cols = "api_url")
})

test_that("set_owner_types sets attributes to owners list", {
  mockery::stub(
    test_graphql_github$set_owner_type,
    "self$gql_response",
    test_fixtures$github_user_login
  )
  owner <- test_graphql_github$set_owner_type(
    owners = c("test_user")
  )
  expect_equal(attr(owner[[1]], "type"), "user")
  expect_equal(owner[[1]], "test_user", ignore_attr = TRUE)

  mockery::stub(
    test_graphql_github$set_owner_type,
    "self$gql_response",
    test_fixtures$github_org_login
  )
  owner <- test_graphql_github$set_owner_type(
    owners = c("test_org")
  )
  expect_equal(attr(owner[[1]], "type"), "organization")
  expect_equal(owner[[1]], "test_org", ignore_attr = TRUE)
})

test_that("set_owner_type returns cached result on second call for GitHub", {
  test_engine <- EngineGraphQLGitHub$new(
    gql_api_url = "https://api.github.com/graphql",
    token = "test_token"
  )
  mockery::stub(
    test_engine$set_owner_type,
    "self$gql_response",
    test_fixtures$github_user_login
  )
  owner_first <- test_engine$set_owner_type(owners = c("cached_user"))
  owner_second <- test_engine$set_owner_type(owners = c("cached_user"))
  expect_equal(owner_first, owner_second)
  expect_equal(attr(owner_second[[1]], "type"), "user")
})

test_that("set_owner_type returns cached result on second call for GitLab", {
  test_engine <- EngineGraphQLGitLab$new(
    gql_api_url = "https://gitlab.com/api/graphql",
    token = "test_token"
  )
  mockery::stub(
    test_engine$set_owner_type,
    "self$gql_response",
    list(
      "data" = list(
        "user" = list("username" = "cached_user"),
        "group" = NULL
      )
    )
  )
  owner_first <- test_engine$set_owner_type(owners = c("cached_user"))
  owner_second <- test_engine$set_owner_type(owners = c("cached_user"))
  expect_equal(owner_first, owner_second)
  expect_equal(attr(owner_second[[1]], "type"), "user")
})

test_that("set_api_url works", {
  expect_equal({
    github_testhost_priv$set_api_url(
      host = "github.com"
    )
  }, "https://api.github.com")
  expect_equal({
    github_testhost_priv$set_api_url(
      host = "https://github.com"
    )
  }, "https://api.github.com")
  expect_equal({
    github_testhost_priv$set_api_url(
      host = "http://github.com"
    )
  }, "https://api.github.com")
  expect_equal({
    github_testhost_priv$set_api_url(
      host = "https://github.company.com"
    )
  }, "https://github.company.com/api/v3")
  expect_equal({
    gitlab_testhost_priv$set_api_url(
      host = "https://gitlab.com"
    )
  }, "https://gitlab.com/api/v4")
  expect_equal({
    gitlab_testhost_priv$set_api_url(
      host = "gitlab.com"
    )
  }, "https://gitlab.com/api/v4")
})

test_that("set_custom_api_url works", {
  expect_equal({
    gitlab_testhost_priv$set_custom_api_url(
      host = "http://gitlab.com"
    )
  }, "https://gitlab.com/api/v4")
  expect_equal({
    github_testhost_priv$set_custom_api_url(
      host = "http://github.company.com"
    )
  }, "https://github.company.com/api/v3")
})

test_that("orgs GitHub query is built properly", {
  gh_orgs_query <-
    test_gqlquery_gh$orgs(
      end_cursor = ""
    )
  expect_snapshot(
    gh_orgs_query
  )
})

test_that("org GitHub query is built properly", {
  gh_org_query <-
    test_gqlquery_gh$org()
  expect_snapshot(
    gh_org_query
  )
})

test_that("get_orgs pulls responses from GraphQL", {
  mockery::stub(
    test_graphql_github$get_orgs,
    "self$gql_response",
    test_fixtures$graphql_gh_orgs_response
  )
  gh_orgs_raw_response <- test_graphql_github$get_orgs(
    output = "only_names",
    verbose = FALSE
  )
  expect_type(
    gh_orgs_raw_response,
    "character"
  )
  test_mocker$cache(gh_orgs_raw_response)
  gh_orgs_full_response <- test_graphql_github$get_orgs(
    output = "full_table",
    verbose = FALSE
  )
  expect_type(
    gh_orgs_full_response,
    "list"
  )
  expect_orgs_github_full_list(gh_orgs_full_response)
  test_mocker$cache(gh_orgs_full_response)
})


test_that("get_orgs breaks when meet GraphQL error", {
  graphql_error <- ""
  class(graphql_error) <- "graphql_error"
  mockery::stub(
    test_graphql_github$get_orgs,
    "self$gql_response",
    graphql_error
  )
  gh_orgs_raw_response <- test_graphql_github$get_orgs(
    output = "only_names",
    verbose = FALSE
  )
  expect_s3_class(
    gh_orgs_raw_response,
    "graphql_error"
  )
})

test_that("get_orgs prints message", {
  mockery::stub(
    test_graphql_github$get_orgs,
    "self$gql_response",
    test_fixtures$graphql_gh_orgs_response
  )
  expect_snapshot(
    gh_orgs_full_response <- test_graphql_github$get_orgs(
      output = "full_table",
      verbose = TRUE
    )
  )
})

test_that("prepare_orgs_table works", {
  github_orgs_table <- test_graphql_github$prepare_orgs_table(
    full_orgs_list = test_mocker$use("gh_orgs_full_response")
  )
  expect_orgs_table(
    github_orgs_table
  )
  test_mocker$cache(github_orgs_table)
})

test_that("get_org pulls response for one org from GraphQL", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_github$get_org,
      "self$gql_response",
      test_fixtures$graphql_gh_org_response
    )
    org <- "test_org"
  } else {
    org <- "r-world-devs"
  }
  gh_org_response <- test_graphql_github$get_org(
    org = org
  )
  expect_type(gh_org_response, "list")
  test_mocker$cache(gh_org_response)
})

test_that("get_orgs_from_host returns error when GitHost is public", {
  github_testhost_priv$is_public <- TRUE
  expect_snapshot_error(
    github_testhost_priv$get_orgs_from_host(
      output = "full_table",
      verbose = FALSE
    )
  )
})

test_that("get_orgs_from_host works on GitHost level", {
  github_testhost_priv$is_public <- FALSE
  mockery::stub(
    github_testhost_priv$get_orgs_from_host,
    "graphql_engine$get_orgs",
    test_mocker$use("gh_orgs_full_response")
  )
  github_orgs_table <- github_testhost_priv$get_orgs_from_host(
    output = "full_table",
    verbose = FALSE
  )
  expect_orgs_table(
    github_orgs_table
  )
  test_mocker$cache(github_orgs_table)
})

test_that("get_orgs_from_orgs_and_repos works on GitHost level", {
  mockery::stub(
    github_testhost_priv$get_orgs_from_orgs_and_repos,
    "graphql_engine$get_org",
    test_mocker$use("gh_org_response")
  )
  github_orgs_from_orgs_table <- github_testhost_priv$get_orgs_from_orgs_and_repos(
    verbose = FALSE
  )
  expect_orgs_table(
    github_orgs_from_orgs_table
  )
  test_mocker$cache(github_orgs_from_orgs_table)
})

test_that("get_orgs works on GitHost level", {
  mockery::stub(
    github_testhost$get_orgs,
    "private$get_orgs_from_hosts",
    test_mocker$use("github_orgs_table")
  )
  mockery::stub(
    github_testhost$get_orgs,
    "private$get_orgs_from_orgs_and_repos",
    test_mocker$use("github_orgs_from_orgs_table")
  )
  github_orgs_table <- github_testhost$get_orgs(
    output = "full_table",
    verbose = FALSE
  )
  expect_orgs_table(
    github_orgs_table,
    add_cols = c("host_url", "host_name")
  )
  test_mocker$cache(github_orgs_table)
})

test_that("commits_by_repo GitHub query is built properly", {
  gh_commits_from_repo_query <-
    test_gqlquery_gh$commits_from_repo()
  expect_snapshot(
    gh_commits_from_repo_query
  )
})

test_that("`get_commits_page_from_repo()` pulls commits page from repository", {
  mockery::stub(
    test_graphql_github_priv$get_commits_page_from_repo,
    "self$gql_response",
    test_fixtures$github_commits_response
  )
  commits_page <- test_graphql_github_priv$get_commits_page_from_repo(
    org = "r-world-devs",
    repo = "GitStats",
    since = "2023-01-01",
    until = "2023-02-28"
  )
  expect_commit_github_gql_response(
    commits_page$data$repository$defaultBranchRef$target$history$edges[[1]]
  )
  test_mocker$cache(commits_page)
})

test_that("`get_commits_page_from_repo()` runs for the second time when fails with error", {
  error_response <- ""
  class(error_response) <- "graphql_error"
  mockery::stub(
    test_graphql_github_priv$get_commits_page_from_repo,
    "self$gql_response",
    error_response
  )
  expect_no_error(
    commits_page <- test_graphql_github_priv$get_commits_page_from_repo(
      org = "r-world-devs",
      repo = "GitStats",
      since = "2023-01-01",
      until = "2023-02-28"
    )
  )
})

test_that("`get_commits_from_one_repo()` prepares formatted list", {
  mockery::stub(
    test_graphql_github_priv$get_commits_from_one_repo,
    "private$get_commits_page_from_repo",
    test_mocker$use("commits_page")
  )
  commits_from_repo <- test_graphql_github_priv$get_commits_from_one_repo(
    org = "r-world-devs",
    repo = "GitStats",
    since = "2023-01-01",
    until = "2023-02-28"
  )
  expect_commit_github_gql_response(
    commits_from_repo[[1]]
  )
  test_mocker$cache(commits_from_repo)
})

test_that("`get_commits_from_one_repo()` handles 502 error and returns empty list", {
  bad_gateway_error <- "502 Bad Gateway Error"
  class(bad_gateway_error) <- c("graphql_error", class(bad_gateway_error))
  mockery::stub(
    test_graphql_github_priv$get_commits_from_one_repo,
    "private$get_commits_page_from_repo",
    bad_gateway_error
  )
  commits_from_repo <- test_graphql_github_priv$get_commits_from_one_repo(
    org = "r-world-devs",
    repo = "GitStats",
    since = "2023-01-01",
    until = "2023-02-28"
  )
  expect_equal(
    commits_from_repo,
    list()
  )
})

test_that("`get_commits_from_repos()` pulls commits from repos", {
  mockery::stub(
    test_graphql_github$get_commits_from_repos,
    "private$get_commits_from_one_repo",
    test_mocker$use("commits_from_repo")
  )
  commits_from_repos <- test_graphql_github$get_commits_from_repos(
    org = "r-world-devs",
    repo = "GitStats",
    since = "2023-01-01",
    until = "2023-02-28"
  )
  expect_commit_github_gql_response(
    commits_from_repos[[1]][[1]]
  )
  test_mocker$cache(commits_from_repos)
})

test_that("`prepare_commits_table()` prepares commits table", {
  gh_commits_table <- test_graphql_github$prepare_commits_table(
    repos_list_with_commits = test_mocker$use("commits_from_repos"),
    org = "r-world-devs"
  )
  expect_commits_table(
    gh_commits_table
  )
  test_mocker$cache(gh_commits_table)
})

test_that("fill_empty_authors() works as expected", {
  commits_table <- test_mocker$use("gh_commits_table")
  commits_table$author_name <- NA
  commits_table <- test_graphql_github_priv$fill_empty_authors(
    commits_table = commits_table
  )
  expect_true(
    all(
      c("Bob Test", "Barbara Check", "John Test") %in% commits_table$author_name
    )
  )
})

test_that("get_commits_from_orgs for GitHub works", {
  mockery::stub(
    github_testhost_priv$get_commits_from_orgs,
    "graphql_engine$prepare_commits_table",
    test_mocker$use("gh_commits_table")
  )
  mockery::stub(
    github_testhost_priv$get_commits_from_orgs,
    "private$get_repos_data",
    test_mocker$use("gh_repos_data")
  )
  github_testhost_priv$searching_scope <- "org"
  gh_commits_from_orgs <- github_testhost_priv$get_commits_from_orgs(
    since = "2023-03-01",
    until = "2023-04-01",
    verbose = FALSE,
    progress = FALSE
  )
  expect_commits_table(
    gh_commits_from_orgs
  )
  test_mocker$cache(gh_commits_from_orgs)
})


test_that("get_commits_from_orgs for GitHub prints messages", {
  mockery::stub(
    github_testhost_priv$get_commits_from_orgs,
    "graphql_engine$prepare_commits_table",
    test_mocker$use("gh_commits_table")
  )
  mockery::stub(
    github_testhost_priv$get_commits_from_orgs,
    "private$get_repos_data",
    test_mocker$use("gh_repos_data")
  )
  github_testhost_priv$searching_scope <- "org"
  expect_snapshot(
    gh_commits_from_orgs <- github_testhost_priv$get_commits_from_orgs(
      since = "2023-03-01",
      until = "2023-04-01",
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("get_commits_from_repos for GitHub works", {
  mockery::stub(
    github_testhost_priv$get_commits_from_repos,
    "graphql_engine$prepare_commits_table",
    test_mocker$use("gh_commits_table")
  )
  github_testhost_priv$searching_scope <- "repo"
  github_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    github_testhost_priv$get_commits_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  gh_commits_from_repos <- github_testhost_priv$get_commits_from_repos(
    since    = "2023-03-01",
    until    = "2023-04-01",
    verbose  = FALSE,
    progress = FALSE
  )
  expect_commits_table(
    gh_commits_from_repos
  )
  test_mocker$cache(gh_commits_from_repos)
})

test_that("get_commits_from_repos for GitHub prints messages", {
  mockery::stub(
    github_testhost_priv$get_commits_from_repos,
    "graphql_engine$prepare_commits_table",
    test_mocker$use("gh_commits_table")
  )
  github_testhost_priv$searching_scope <- "repo"
  github_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    github_testhost_priv$get_commits_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  expect_snapshot(
    gh_commits_from_repos <- github_testhost_priv$get_commits_from_repos(
      since    = "2023-03-01",
      until    = "2023-04-01",
      verbose  = TRUE,
      progress = FALSE
    )
  )
})

test_that("`get_commits()` retrieves commits in the table format", {
  mockery::stub(
    github_testhost$get_commits,
    "private$get_commits_from_orgs",
    test_mocker$use("gh_commits_from_orgs")
  )
  mockery::stub(
    github_testhost$get_commits,
    "private$get_commits_from_repos",
    test_mocker$use("gh_commits_from_repos")
  )
  gh_commits_table <- github_testhost$get_commits(
    since    = "2023-01-01",
    until    = "2023-02-28",
    verbose  = FALSE,
    progress = FALSE
  )
  expect_commits_table(
    gh_commits_table
  )
  test_mocker$cache(gh_commits_table)
})

test_that("`get_commits()` is set to scan whole git host", {
  github_testhost_all <- create_github_testhost_all(orgs = "test_org")
  mockery::stub(
    github_testhost_all$get_commits,
    "private$get_orgs_from_host",
    "test_org"
  )
  mockery::stub(
    github_testhost_all$get_commits,
    "private$get_commits_from_orgs",
    test_mocker$use("gh_commits_from_orgs")
  )
  mockery::stub(
    github_testhost_all$get_commits,
    "private$get_commits_from_repos",
    test_mocker$use("gh_commits_from_repos")
  )
  expect_commits_table(
    github_testhost_all$get_commits(
      since    = "2023-01-01",
      until    = "2023-02-28",
      verbose  = TRUE,
      progress = FALSE
    )
  )
})

test_fixtures$github_releases_from_repo <- list(
  "data" = list(
    "repository" = list(
      "name" = "TestProject1",
      "url"  = "test_url",
      "releases" = list(
        "nodes" = list(
          list(
            "name"    = "2.1.0",
            "tagName" = "v2.1.0",
            "publishedAt" = "2024-01-01T00:00:00Z",
            "url"         = "https://test_url/tag/v2.1.0",
            "description" = "Great features come with this release."
          ),
          list(
            "name"    = "1.1.1",
            "tagName" = "v1.1.1",
            "publishedAt" = "2023-09-28T00:00:00Z",
            "url"         = "https://test_url/tag/v1.1.0",
            "description" = "Great features come with this release."
          ),
          list(
            "name"    = "0.1.0",
            "tagName" = "v0.1.0",
            "publishedAt" = "2023-04-01T00:00:00Z",
            "url"         = "https://test_url/tag/v0.1.0",
            "description" = "Great features come with this release."
          ),
          list(
            "name"    = "0.1.1",
            "tagName" = "v0.1.1",
            "publishedAt" = "2023-05-02T00:00:00Z",
            "url"         = "https://test_url/tag/v0.1.0",
            "description" = "Great features come with this release."
          )
        )
      )
    )
  )
)

test_fixtures$gitlab_releases_from_repo <- list(
  "data" = list(
    "project" = list(
      "name" = "TestProject",
      "webUrl" = "test_url",
      "releases" = list(
        "nodes" = list(
          list(
            "name" = "TestProject 1.0.0",
            "tagName" = "1.0.0",
            "releasedAt" = "2024-06-08T00:00:00+02:00",
            "links" = list(
              "selfUrl" = "test_url/releases/1.0.0"
            ),
            "description" = "This release comes with awesome features."
          ),
          list(
            "name" = "TestProject 2.0.0",
            "tagName" = "2.0.0",
            "releasedAt" = "2024-08-08T00:00:00+02:00",
            "links" = list(
              "selfUrl" = "test_url/releases/2.0.0"
            ),
            "description" = "This release comes with awesome features."
          ),
          list(
            "name" = "TestProject 0.1.0",
            "tagName" = "0.1.0",
            "releasedAt" = "2023-06-08T00:00:00+02:00",
            "links" = list(
              "selfUrl" = "test_url/releases/0.1.0"
            ),
            "description" = "This release comes with awesome features."
          ),
          list(
            "name" = "TestProject 3.0.0",
            "tagName" = "3.0.0",
            "releasedAt" = "2023-09-08T00:00:00+02:00",
            "links" = list(
              "selfUrl" = "test_url/releases/3.0.0"
            ),
            "description" = "This release comes with awesome features."
          )
        )
      )
    )
  )
)

test_fixtures_pr <- list()

github_open_pr_edge <- function(created_at) {
  list(
    "node" = list(
      "number" = 1,
      "created_at" = created_at,
      "merged_at" = "",
      "state" = "open",
      "author" = list(
        "login" = "test"
      ),
      "source_branch" = "dev",
      "target_branch" = "main"
    )
  )
}

github_closed_pr_edge <- function(created_at) {
  list(
    "node" = list(
      "number" = 1,
      "created_at" = created_at,
      "merged_at" = "2027-01-01T00:00:00Z",
      "state" = "closed",
      "author" = list(
        "login" = "test"
      ),
      "source_branch" = "dev",
      "target_branch" = "main"
    )
  )
}

open_pr_timestamps <- generate_random_timestamps(25, 2024, 2026)
set.seed(234)
closed_pr_timestamps <- generate_random_timestamps(25, 2024, 2026)

open_prs <- purrr::map(open_pr_timestamps, github_open_pr_edge)
closed_prs <- purrr::map(closed_pr_timestamps, github_closed_pr_edge)

test_fixtures_pr$github_graphql_pr_response <- list(
  "data" = list(
    "repository" = list(
      "pullRequests" = list(
        "pageInfo" = list(
          "hasNextPage" = FALSE,
          "endCursor" = ""
        ),
        "edges" = append(open_prs, closed_prs)
      )
    )
  )
)

test_fixtures_pr$gitlab_graphql_pr_response <- list(
  "data" = list(
    "project" = list(
      "mergeRequests" = list(
        "pageInfo" = list(
          "hasNextPage" = FALSE,
          "endCursor" = ""
        ),
        "edges" = append(open_prs, closed_prs)
      )
    )
  )
)

#' @title Create a `GitStats` object
#' @name create_gitstats
#' @examples
#' my_gitstats <- create_gitstats()
#' @return A `GitStats` object.
#' @export
create_gitstats <- function() {
  GitStats$new()
}

#' @title Show hosts set in `GitStats`
#' @name show_hosts
#' @description Retrieves hosts set by `GitStats` with `set_*_host()` functions.
#' @param gitstats A GitStats object.
#' @return A list of hosts.
#' @export
show_hosts <- function(gitstats) {
  gitstats$show_hosts()
}

#' @title Show organizations set in `GitStats`
#' @name show_orgs
#' @description Retrieves organizations set or pulled by `GitStats`. Especially
#'   helpful when user is scanning whole git platform and wants to have a
#'   glimpse at organizations.
#' @param gitstats A GitStats object.
#' @return A vector of organizations.
#' @export
show_orgs <- function(gitstats) {
  gitstats$show_orgs()
}

#' @title Switch on verbose mode
#' @name verbose_on
#' @description Print all messages and output.
#' @param gitstats A GitStats object.
#' @return A GitStats object.
#' @export
verbose_on <- function(gitstats) {
  gitstats$verbose_on()
  return(invisible(gitstats))
}

#' @title Switch off verbose mode
#' @name verbose_off
#' @description Stop printing messages and output.
#' @param gitstats A GitStats object.
#' @return A GitStats object.
#' @export
verbose_off <- function(gitstats) {
  gitstats$verbose_off()
  return(invisible(gitstats))
}

#' @title Is verbose mode switched on
#' @name is_verbose
#' @param gitstats A GitStats object.
is_verbose <- function(gitstats) {
  gitstats$is_verbose()
}

GitStats <- R6::R6Class(
  classname = "GitStats",
  public = list(

    initialize = function() {
      private$storage_backend <- StorageLocal$new()
    },

    set_github_host = function(host,
                               token = NULL,
                               orgs = NULL,
                               repos = NULL,
                               verbose = TRUE,
                               .error = TRUE) {
      new_host <- NULL
      new_host <- GitHostGitHub$new(
        orgs = orgs,
        repos = repos,
        token = token,
        host = host,
        verbose = verbose,
        .error = .error
      )
      private$add_new_host(new_host)
    },

    set_gitlab_host = function(host,
                               token = NULL,
                               orgs = NULL,
                               repos = NULL,
                               verbose = TRUE,
                               .error = TRUE) {
      new_host <- NULL
      new_host <- GitHostGitLab$new(
        orgs = orgs,
        repos = repos,
        token = token,
        host = host,
        verbose = verbose,
        .error = .error
      )
      private$add_new_host(new_host)
    },

    get_orgs = function(cache = TRUE, output = "full_table", verbose = TRUE) {
      private$check_for_host()
      trigger <- private$trigger_pulling(
        cache = cache,
        storage = "organizations",
        verbose = verbose
      )
      if (trigger) {
        cli::cli_alert("Pulling organizations {cli_icons$org} data...")
        private$inform_parallel(verbose)
        organizations <- private$get_orgs_from_hosts(
          output = output,
          verbose = verbose
        ) |>
          private$set_object_class(
            class = "gitstats_orgs"
          )
        private$save_to_storage(
          table = organizations
        )
      } else {
        organizations <- private$get_from_storage(
          table = "organizations"
        )
      }
      return(organizations)
    },

    get_repos = function(add_contributors = FALSE,
                         add_languages = TRUE,
                         with_code = NULL,
                         in_files = NULL,
                         with_files = NULL,
                         language = NULL,
                         cache = TRUE,
                         verbose = TRUE,
                         progress = TRUE,
                         fill_empty_sha = FALSE) {
      private$check_for_host()
      private$check_params_conflict(
        with_code = with_code,
        in_files = in_files,
        with_files = with_files
      )
      args_list <- list("with_code" = with_code,
                        "in_files" = in_files,
                        "with_files" = with_files,
                        "language" = language)
      trigger <- private$trigger_pulling(
        cache = cache,
        storage = "repositories",
        args_list = args_list,
        verbose = verbose
      )
      if (trigger) {
        cli::cli_alert("Pulling repositories {cli_icons$repo} data...")
        private$inform_parallel(verbose)
        repositories <- private$get_repos_from_hosts(
          add_contributors = add_contributors,
          add_languages = add_languages,
          with_code = with_code,
          in_files = in_files,
          with_files = with_files,
          language = language,
          verbose = verbose,
          progress = progress,
          fill_empty_sha = fill_empty_sha
        )
        if (nrow(repositories) > 0) {
          repositories <- private$set_object_class(
            object = repositories,
            class = "gitstats_repos",
            attr_list = args_list
          )
          private$save_to_storage(
            table = repositories
          )
        } else {
          if (verbose) {
            cli::cli_alert_warning("No repositories {cli_icons$repo} found.")
          }
        }
      } else {
        repositories <- private$get_from_storage(
          table = "repositories"
        )
      }
      return(repositories)
    },

    get_repos_urls = function(type = "web",
                              with_code = NULL,
                              in_files = NULL,
                              with_files = NULL,
                              cache = TRUE,
                              verbose = TRUE,
                              progress = TRUE) {
      private$check_for_host()
      private$check_params_conflict(
        with_code = with_code,
        in_files = in_files,
        with_files = with_files
      )
      args_list <- list("type" = type,
                        "with_code" = with_code,
                        "in_files" = in_files,
                        "with_files" = with_files)
      trigger <- private$trigger_pulling(
        cache = cache,
        storage = "repos_urls",
        args_list = args_list,
        verbose = verbose
      )
      if (trigger) {
        cli::cli_alert("Pulling repositories {cli_icons$repo} URLs...")
        private$inform_parallel(verbose)
        repos_urls <- private$get_repos_urls_from_hosts(
          type = type,
          with_code = with_code,
          in_files = in_files,
          with_files = with_files,
          verbose = verbose,
          progress = progress
        )
        if (!is.null(repos_urls)) {
          repos_urls <- private$set_object_class(
            object = repos_urls,
            class = "gitstats_repos_urls",
            attr_list = args_list
          )
          private$save_to_storage(
            table = repos_urls
          )
        }
      } else {
        repos_urls <- private$get_from_storage(
          table = "repos_urls"
        )
      }
      return(repos_urls)
    },

    get_commits = function(since,
                           until,
                           cache = TRUE,
                           verbose = TRUE,
                           progress = TRUE) {
      private$check_for_host()
      args_list <- list(
        "date_range" = c(since, as.character(until)),
        "scope" = private$get_scope_fingerprint()
      )
      trigger <- private$trigger_pulling(
        cache = cache,
        storage = "commits",
        args_list = args_list,
        verbose = verbose
      )
      if (trigger) {
        private$inform_parallel(verbose)
        commits <- private$pull_incrementally(
          storage_name = "commits",
          fetch_fn = function(since, until) {
            private$get_commits_from_hosts(
              since = since, until = until,
              verbose = verbose, progress = progress
            )
          },
          args_list = args_list,
          dedup_columns = "id",
          cache = cache,
          verbose = verbose,
          icon = cli_icons$commit,
          pull_message = "Pulling commits {cli_icons$commit}..."
        )
        if (nrow(commits) > 0) {
          commits <- private$set_object_class(
            object = commits,
            class = "gitstats_commits",
            attr_list = args_list
          )
          private$save_to_storage(
            table = commits
          )
        } else {
          if (verbose) {
            cli::cli_alert_warning("No commits {cli_icons$commit} found.")
          }
        }
      } else {
        commits <- private$get_from_storage(
          table = "commits"
        )
      }
      return(commits)
    },

    get_issues = function(since,
                          until,
                          state,
                          cache    = TRUE,
                          verbose  = TRUE,
                          progress = TRUE) {
      private$check_for_host()
      args_list <- list(
        "state" = state,
        "date_range" = c(since, as.character(until)),
        "scope" = private$get_scope_fingerprint()
      )
      trigger <- private$trigger_pulling(
        cache = cache,
        storage = "issues",
        args_list = args_list,
        verbose = verbose
      )
      if (trigger) {
        private$inform_parallel(verbose)
        issues <- private$pull_incrementally(
          storage_name = "issues",
          fetch_fn = function(since, until) {
            private$get_issues_from_hosts(
              since = since, until = until, state = state,
              verbose = verbose, progress = progress
            )
          },
          args_list = args_list,
          dedup_columns = c("number", "repo_name"),
          cache = cache,
          verbose = verbose,
          icon = cli_icons$issue,
          pull_message = "Getting issues {cli_icons$issue}..."
        )
        if (nrow(issues) > 0) {
          issues <- private$set_object_class(
            object = issues,
            class = "gitstats_issues",
            attr_list = args_list
          )
          private$save_to_storage(
            table = issues
          )
        } else {
          if (verbose) {
            cli::cli_alert_warning("No issues {cli_icons$issue} found.")
          }
        }
      } else {
        issues <- private$get_from_storage(
          table = "issues"
        )
      }
      return(issues)
    },

    get_pull_requests = function(since,
                                 until,
                                 state,
                                 cache    = TRUE,
                                 verbose  = TRUE,
                                 progress = TRUE) {
      private$check_for_host()
      args_list <- list(
        "state" = state,
        "date_range" = c(since, as.character(until)),
        "scope" = private$get_scope_fingerprint()
      )
      trigger <- private$trigger_pulling(
        cache = cache,
        storage = "pull_requests",
        args_list = args_list,
        verbose = verbose
      )
      if (trigger) {
        private$inform_parallel(verbose)
        pull_requests <- private$pull_incrementally(
          storage_name = "pull_requests",
          fetch_fn = function(since, until) {
            private$get_pull_requests_from_hosts(
              since = since, until = until, state = state,
              verbose = verbose, progress = progress
            )
          },
          args_list = args_list,
          dedup_columns = c("number", "repo_name"),
          cache = cache,
          verbose = verbose,
          icon = cli_icons$pull_request,
          pull_message = "Getting pull requests {cli_icons$pull_request}..."
        )
        if (nrow(pull_requests) > 0) {
          pull_requests <- private$set_object_class(
            object = pull_requests,
            class = "gitstats_pull_requests",
            attr_list = args_list
          )
          private$save_to_storage(
            table = pull_requests
          )
        } else {
          if (verbose) {
            cli::cli_alert_warning("No pull requests {cli_icons$pull_request} found.")
          }
        }
      } else {
        pull_requests <- private$get_from_storage(
          table = "pull_requests"
        )
      }
      return(pull_requests)
    },

    get_users = function(logins, cache = TRUE, verbose = TRUE) {
      private$check_for_host()
      args_list <- list("logins" = logins)
      trigger <- private$trigger_pulling(
        cache = cache,
        storage = "users",
        args_list = args_list,
        verbose = verbose
      )
      if (trigger) {
        cli::cli_alert("Pulling users {cli_icons$user} data...")
        private$inform_parallel(verbose)
        users <- private$get_users_from_hosts(logins) |>
          private$set_object_class(
            class = "gitstats_users",
            attr_list = args_list
          )
        private$save_to_storage(users)
      } else {
        users <- private$get_from_storage(
          table = "users"
        )
      }
      return(users)
    },

    get_files = function(pattern,
                         depth,
                         file_path = NULL,
                         cache = TRUE,
                         verbose = TRUE,
                         progress = verbose) {
      private$check_for_host()
      args_list <- list(
        "file_pattern" = paste(file_path, pattern),
        "depth" = depth
      )
      trigger <- private$trigger_pulling(
        cache = cache,
        storage = "files",
        args_list = args_list,
        verbose = verbose
      )
      if (trigger) {
        cli::cli_alert("Pulling files {cli_icons$file} content...")
        private$inform_parallel(verbose)
        files <- private$get_files_from_hosts(
          pattern = pattern,
          depth = depth,
          file_path = file_path,
          verbose = verbose,
          progress = progress
        )
        if (nrow(files) > 0) {
          files <- private$set_object_class(
            object = files,
            class = "gitstats_files",
            attr_list = args_list
          )
          private$save_to_storage(files)
        } else {
          if (verbose) {
            cli::cli_alert_warning("No files {cli_icons$file} found.")
          }
        }
      } else {
        files <- private$get_from_storage(
          table = "files"
        )
      }
      return(files)
    },

    get_repos_trees = function(pattern,
                               depth,
                               cache,
                               verbose,
                               progress) {
      private$check_for_host()
      args_list <- list(
        "file_pattern" = pattern %||% "",
        "depth" = depth
      )
      trigger <- private$trigger_pulling(
        cache = cache,
        storage = "repos_trees",
        args_list = args_list,
        verbose = verbose
      )
      if (trigger) {
        cli::cli_alert("Pulling repositories {cli_icons$tree} structure...")
        private$inform_parallel(verbose)
        repos_trees <- private$get_repos_trees_from_hosts(
          pattern = pattern,
          depth = depth,
          verbose = verbose,
          progress = progress
        )
        if (nrow(repos_trees) > 0) {
          repos_trees <- private$set_object_class(
            object = repos_trees,
            class = "gitstats_repos_trees",
            attr_list = args_list
          )
          private$save_to_storage(repos_trees)
        } else {
          if (verbose) {
            cli::cli_alert_warning("No repos {cli_icons$tree} trees found.")
          }
        }
      } else {
        repos_trees <- private$get_from_storage(
          table = "repos_trees"
        )
      }
      return(repos_trees)
    },

    get_release_logs = function(since,
                                until = Sys.Date(),
                                cache = TRUE,
                                verbose = TRUE,
                                progress = TRUE) {
      private$check_for_host()
      args_list <- list(
        "date_range" = c(since, as.character(until)),
        "scope" = private$get_scope_fingerprint()
      )
      trigger <- private$trigger_pulling(
        storage = "release_logs",
        cache = cache,
        args_list = args_list,
        verbose = verbose
      )
      if (trigger) {
        private$inform_parallel(verbose)
        release_logs <- private$pull_incrementally(
          storage_name = "release_logs",
          fetch_fn = function(since, until) {
            private$get_release_logs_from_hosts(
              since = since, until = until,
              verbose = verbose, progress = progress
            )
          },
          args_list = args_list,
          dedup_columns = c("release_tag", "repo_name"),
          cache = cache,
          verbose = verbose,
          icon = cli_icons$release,
          pull_message = "Pulling release logs {cli_icons$release}.."
        )
        if (nrow(release_logs) > 0) {
          release_logs <- private$set_object_class(
            object = release_logs,
            class = "gitstats_releases",
            attr_list = args_list
          )
          private$save_to_storage(release_logs)
        } else {
          if (verbose) {
            cli::cli_alert_warning("No release logs {cli_icons$release} found.")
          }
        }
      } else {
        release_logs <- private$get_from_storage(
          table = "release_logs"
        )
      }
      return(release_logs)
    },

    show_hosts = function() {
      purrr::map(private$hosts, function(host) {
        host_priv <- host$.__enclos_env__$private
        list(
          host = host_priv$host_name,
          web_url = host_priv$web_url,
          api_url = host_priv$api_url
        )
      })
    },

    show_orgs = function() {
      purrr::map(private$hosts, function(host) {
        orgs <- host$.__enclos_env__$private$orgs
        purrr::map_vec(orgs, ~ url_decode(.))
      }) |>
        unlist()
    },

    verbose_on = function() {
      private$settings$verbose <- TRUE
    },

    verbose_off = function() {
      private$settings$verbose <- FALSE
    },

    is_verbose = function() {
      private$settings$verbose
    },

    set_parallel = function(workers = 10L) {
      if (isFALSE(workers) || identical(workers, 0L) || identical(workers, 0)) {
        status <- mirai::daemons(0)
        private$settings$parallel <- FALSE
        cli::cli_alert_info("Parallel processing disabled.")
      } else {
        if (isTRUE(workers)) {
          workers <- parallel::detectCores(logical = FALSE)
          if (is.na(workers) || workers < 2L) workers <- 2L
        }
        status <- mirai::daemons(workers)
        ns <- asNamespace("GitStats")
        ns_objects <- as.list(ns, all.names = TRUE)
        do.call(mirai::everywhere, c(list(quote({})), ns_objects))
        private$settings$parallel <- TRUE
        cli::cli_alert_success(
          "Parallel processing enabled with {workers} workers."
        )
      }
      return(invisible(status))
    },

    is_parallel = function() {
      private$settings$parallel && mirai_active()
    },

    set_local_storage = function() {
      private$storage_backend <- StorageLocal$new()
      private$propagate_storage_to_hosts()
      invisible(self)
    },

    set_postgres_storage = function(host = NULL,
                                    port = NULL,
                                    dbname = NULL,
                                    user = NULL,
                                    password = NULL,
                                    schema = "git_stats",
                                    ...) {
      args <- list(...)
      if (!is.null(host)) args$host <- host
      if (!is.null(port)) args$port <- port
      if (!is.null(dbname)) args$dbname <- dbname
      if (!is.null(user)) args$user <- user
      if (!is.null(password)) args$password <- password
      private$storage_backend <- do.call(
        StoragePostgres$new,
        c(list(schema = schema), args)
      )
      private$propagate_storage_to_hosts()
      cli::cli_alert_success("Storage set to {.val PostgreSQL}.")
      private$report_storage_contents()
      invisible(self)
    },

    set_sqlite_storage = function(dbname = ":memory:") {
      private$storage_backend <- StorageSQLite$new(dbname = dbname)
      private$propagate_storage_to_hosts()
      cli::cli_alert_success("Storage set to {.val SQLite}.")
      private$report_storage_contents()
      invisible(self)
    },

    get_storage = function(storage) {
      if (is.null(storage)) {
        stored <- private$storage_backend$list()
        result <- purrr::map(stored, ~ private$storage_backend$load(.))
        names(result) <- stored
        return(result)
      } else {
        private$storage_backend$load(storage)
      }
    },

    remove_from_storage = function(storage) {
      if (!private$storage_backend$exists(storage)) {
        cli::cli_abort("Table {.val {storage}} does not exist in storage.")
      }
      private$storage_backend$remove(storage)
      cli::cli_alert_success("Removed {.val {storage}} from storage.")
      invisible(self)
    },

    remove_postgres_storage = function() {
      if (!inherits(private$storage_backend, "StoragePostgres")) {
        cli::cli_abort("No PostgreSQL storage is set.")
      }
      private$storage_backend$drop_storage()
      cli::cli_alert_success(
        "PostgreSQL schema dropped and connection closed."
      )
      private$storage_backend <- StorageLocal$new()
      private$propagate_storage_to_hosts()
      cli::cli_alert_info("Storage set back to {.val local}.")
      invisible(self)
    },

    remove_sqlite_storage = function() {
      if (!inherits(private$storage_backend, "StorageSQLite")) {
        cli::cli_abort("No SQLite storage is set.")
      }
      dbname <- private$storage_backend$.__enclos_env__$private$dbname
      private$storage_backend$drop_storage()
      if (!is.null(dbname) && dbname != ":memory:") {
        cli::cli_alert_success(
          "SQLite database file {.file {dbname}} removed."
        )
      } else {
        cli::cli_alert_success(
          "In-memory SQLite database removed."
        )
      }
      private$storage_backend <- StorageLocal$new()
      private$propagate_storage_to_hosts()
      cli::cli_alert_info("Storage set back to {.val local}.")
      invisible(self)
    },

    get_storage_metadata = function(storage = NULL) {
      if (!is.null(storage) && !private$storage_backend$exists(storage)) {
        cli::cli_abort("Table {.val {storage}} does not exist in storage.")
      }
      private$storage_backend$get_metadata(storage)
    },

    print = function() {
      cat(paste0("A ", cli::col_blue('GitStats'), " object for ", length(private$hosts), " hosts: \n"))
      private$print_hosts()
      cat(cli::col_blue("Scanning scope: \n"))
      private$print_orgs_and_repos()
      private$print_storage()
    }
  ),
  private = list(

    hosts = list(),

    settings = list(
      verbose  = TRUE,
      cache    = TRUE,
      parallel = FALSE
    ),

    storage_backend = NULL,

    add_new_host = function(new_host) {
      if (!is.null(new_host)) {
        new_host <- new_host |>
          private$check_for_duplicate_hosts()
        if (!is.null(private$storage_backend)) {
          new_host$set_storage_backend(private$storage_backend)
        }
        private$hosts <- append(private$hosts, new_host)
      }
    },

    check_for_host = function() {
      if (length(private$hosts) == 0) {
        cli::cli_abort("Add first your hosts with `set_github_host()` or `set_gitlab_host()`.", call = NULL)
      }
    },

    inform_parallel = function(verbose) {
      if (verbose && self$is_parallel()) {
        cli::cli_alert_info("Running in parallel mode.")
      }
    },

    check_params_conflict = function(with_code, in_files, with_files) {
      if (!is.null(with_code) && !is.null(with_files)) {
        cli::cli_abort(c(
          "x" = "Both `with_code` and `with_files` parameters are defined.",
          "!" = "Use either `with_code` of `with_files` parameter.",
          "i" = "If you want to search for [{with_code}] code in given files
          - use `in_files` parameter together with `with_code` instead."
        ), call = NULL)
      }
      if (!is.null(in_files) && is.null(with_code)) {
        cli::cli_abort(c(
          "!" = "Passing files to `in_files` parameter works only when you
          search code with `with_code` parameter.",
          "i" = "If you want to search for repositories with [{in_files}] files
          you should instead use `with_files` parameter."
        ), call = NULL)
      }
    },

    set_verbose_param = function(verbose) {
      if (!is.null(verbose)) {
        if (!is.logical(verbose)) {
          cli::cli_abort("verbose parameter accepts only TRUE/FALSE values")
        }
        private$settings$verbose <- verbose
      }
    },

    storage_is_empty = function(table) {
      !private$storage_backend$exists(table)
    },

    save_to_storage = function(table) {
      table_name <- deparse(substitute(table))
      private$storage_backend$save(table_name, table)
    },

    propagate_storage_to_hosts = function() {
      purrr::walk(private$hosts, function(host) {
        host$set_storage_backend(private$storage_backend)
      })
    },

    report_storage_contents = function() {
      stored_names <- private$storage_backend$list()
      if (length(stored_names) == 0) {
        cli::cli_alert_info("Database is empty.")
      } else {
        cli::cli_alert_info("Database contains data:")
        purrr::walk(stored_names, function(name) {
          data <- private$storage_backend$load(name)
          n <- if (inherits(data, "data.frame")) nrow(data) else length(data)
          cli::cli_bullets(c(" " = "{name}: {n} records"))
        })
      }
    },

    get_from_storage = function(table) {
      cli::cli_alert_warning(cli::col_yellow(
        glue::glue("Getting cached {table} data.")
      ))
      cli::cli_alert_info(cli::col_cyan(
        "If you wish to pull the data from API once more, set `cache` parameter to `FALSE`."
      ))
      private$storage_backend$load(table)
    },

    trigger_pulling = function(cache, storage, args_list = NULL, verbose) {
      trigger <- FALSE
      if (private$storage_is_empty(storage)) {
        trigger <- TRUE
      } else {
        if (!is.null(args_list)) {
          repos_parameters_changed <- private$check_if_args_changed(
            storage = storage,
            args_list = args_list
          )
          if (repos_parameters_changed) {
            if (verbose) {
              cli::cli_alert_info(cli::col_blue(
                "Parameters changed, I will pull data from API."
              ))
            }
            trigger <- TRUE
          } else if (!repos_parameters_changed && !cache) {
            if (verbose) {
              cli::cli_alert_info(cli::col_blue(
                "Cache set to FALSE, I will pull data from API."
              ))
            }
            trigger <- TRUE
          }
        }
      }
      return(trigger)
    },

    check_if_args_changed = function(storage, args_list) {
      storage_data <- private$storage_backend$load(storage)
      stored_params <- purrr::map(names(args_list), ~ attr(storage_data, .) %||% "")
      new_params <- purrr::map(args_list, ~ . %||% "")
      !all(purrr::map2_lgl(new_params, stored_params, ~ identical(.x, .y)))
    },

    get_date_gaps = function(stored_range, requested_range) {
      stored_since <- as.Date(stored_range[1])
      stored_until <- as.Date(stored_range[2])
      req_since <- as.Date(requested_range[1])
      req_until <- as.Date(requested_range[2])
      gaps <- list()
      if (req_since < stored_since) {
        gaps <- c(gaps, list(list(
          since = as.character(req_since),
          until = as.character(stored_since - 1)
        )))
      }
      if (req_until > stored_until) {
        gaps <- c(gaps, list(list(
          since = as.character(stored_until + 1),
          until = as.character(req_until)
        )))
      }
      gaps
    },

    pull_incrementally = function(storage_name,
                                  fetch_fn,
                                  args_list,
                                  dedup_columns,
                                  cache,
                                  verbose,
                                  icon,
                                  pull_message = NULL) {
      stored_data <- private$storage_backend$load(storage_name)
      stored_range <- attr(stored_data, "date_range")
      requested_range <- args_list[["date_range"]]
      non_date_args_changed <- private$non_date_args_changed(
        stored_data, args_list
      )
      if (!is.null(stored_data) && !is.null(stored_range) &&
            cache && !non_date_args_changed) {
        gaps <- private$get_date_gaps(stored_range, requested_range)
        if (length(gaps) > 0) {
          if (verbose) {
            cli::cli_alert_info(
              "Using cached {storage_name} {icon} from {.val {stored_range[1]}} to {.val {stored_range[2]}}."
            )
            gap_ranges <- purrr::map_chr(
              gaps, ~ paste(.x$since, "to", .x$until)
            )
            cli::cli_alert(
              "Pulling {storage_name} {icon} from API for: {.val {gap_ranges}}."
            )
          }
          new_data <- purrr::map(gaps, function(gap) {
            fetch_fn(since = gap$since, until = gap$until)
          }) |>
            purrr::list_rbind()
          if (nrow(new_data) > 0) {
            class(stored_data) <- class(stored_data)[
              !grepl("^gitstats_", class(stored_data))
            ]
            result <- dplyr::bind_rows(stored_data, new_data) |>
              dplyr::distinct(
                dplyr::across(dplyr::all_of(dedup_columns)),
                .keep_all = TRUE
              )
          } else {
            result <- stored_data
            class(result) <- class(result)[
              !grepl("^gitstats_", class(result))
            ]
          }
        } else {
          result <- stored_data
        }
      } else {
        if (verbose && !is.null(pull_message)) {
          cli::cli_alert(pull_message)
        }
        result <- fetch_fn(
          since = requested_range[1],
          until = requested_range[2]
        )
      }
      result
    },

    get_scope_fingerprint = function() {
      orgs <- purrr::map(private$hosts, function(host) {
        host$.__enclos_env__$private$orgs
      }) |> unlist()
      repos <- purrr::map(private$hosts, function(host) {
        host_priv <- environment(host$initialize)$private
        if ("repo" %in% host_priv$searching_scope) {
          host_priv$repos_fullnames
        }
      }) |> unlist()
      scope <- sort(unique(c(orgs, repos)))
      paste(scope, collapse = ",")
    },

    non_date_args_changed = function(stored_data, args_list) {
      non_date_args <- args_list[names(args_list) != "date_range"]
      if (length(non_date_args) == 0) return(FALSE)
      stored_params <- purrr::map(
        names(non_date_args), ~ attr(stored_data, .) %||% ""
      )
      new_params <- purrr::map(non_date_args, ~ . %||% "")
      !all(purrr::map2_lgl(new_params, stored_params, ~ identical(.x, .y)))
    },

    set_object_class = function(object, class, attr_list = NULL) {
      class(object) <- append(class, class(object))
      if (!is.null(attr_list)) {
        purrr::iwalk(attr_list, function(attrib, attrib_name)  {
          attr(object, attrib_name) <<- attrib
        })
      }
      return(object)
    },

    get_orgs_from_hosts = function(output, verbose) {
      orgs_from_hosts <- purrr::map(private$hosts, function(host) {
        host$get_orgs(
          output = output,
          verbose = verbose
        )
      })
      if (output == "full_table") {
        orgs_from_hosts |>
          purrr::list_rbind()
      } else {
        orgs_from_hosts |>
          purrr::list_flatten()
      }
    },

    get_repos_from_hosts = function(add_contributors = FALSE,
                                    add_languages = TRUE,
                                    with_code,
                                    in_files = NULL,
                                    with_files,
                                    language = NULL,
                                    output = "table",
                                    verbose = TRUE,
                                    progress = TRUE,
                                    fill_empty_sha = FALSE) {
      repos_table <- purrr::map(private$hosts, function(host) {
        if (!is.null(with_code)) {
          private$get_repos_from_host_with_code(
            host = host,
            add_contributors = add_contributors,
            add_languages = add_languages,
            with_code = with_code,
            in_files = in_files,
            language = language,
            output = output,
            verbose = verbose,
            progress = progress
          )
        } else if (!is.null(with_files)) {
          private$get_repos_from_host_with_files(
            host = host,
            add_contributors = add_contributors,
            add_languages = add_languages,
            with_files = with_files,
            language = language,
            output = output,
            verbose = verbose,
            progress = progress
          )
        } else {
          host$get_repos(
            add_contributors = add_contributors,
            add_languages = add_languages,
            verbose = verbose,
            progress = progress,
            fill_empty_sha = fill_empty_sha
          )
        }
      }) |>
        purrr::list_rbind() |>
        dplyr::as_tibble() |>
        private$add_stats_to_repos() |>
        dplyr::as_tibble()
      return(repos_table)
    },

    get_repos_from_host_with_code = function(host,
                                             add_contributors,
                                             add_languages,
                                             with_code,
                                             in_files,
                                             language,
                                             output,
                                             verbose,
                                             progress) {
      purrr::map(with_code, function(with_code) {
        host$get_repos(
          add_contributors = add_contributors,
          add_languages = add_languages,
          with_code = with_code,
          in_files = in_files,
          language = language,
          output = output,
          verbose = verbose,
          progress = progress
        )
      }) |>
        purrr::list_rbind()
    },

    get_repos_from_host_with_files = function(host,
                                              add_contributors,
                                              add_languages,
                                              with_files,
                                              language,
                                              output,
                                              verbose,
                                              progress) {
      purrr::map(with_files, function(with_file) {
        host$get_repos(
          add_contributors = add_contributors,
          add_languages = add_languages,
          with_file = with_file,
          language = language,
          output = output,
          verbose = verbose,
          progress = progress
        )
      }) |>
        purrr::list_rbind()
    },

    get_repos_urls_from_hosts = function(type,
                                         with_code,
                                         in_files,
                                         with_files,
                                         verbose,
                                         progress) {
      purrr::map(private$hosts, function(host) {
        if (!is.null(with_code)) {
          private$get_repos_urls_from_host_with_code(
            host = host,
            type = type,
            with_code = with_code,
            in_files = in_files,
            verbose = verbose,
            progress = progress
          )
        } else if (!is.null(with_files)) {
          private$get_repos_urls_from_host_with_files(
            host = host,
            type = type,
            with_files = with_files,
            verbose = verbose,
            progress = progress
          )
        } else {
          host$get_repos_urls(
            type = type,
            verbose = verbose,
            progress = progress
          )
        }
      }) |>
        unlist() |>
        unique()
    },

    get_repos_urls_from_host_with_code = function(host,
                                                  type,
                                                  with_code,
                                                  in_files,
                                                  verbose,
                                                  progress) {
      purrr::map(with_code, function(code) {
        host$get_repos_urls(
          type      = type,
          with_code = code,
          in_files  = in_files,
          verbose   = verbose,
          progress  = progress
        )
      }) |>
        unlist()
    },

    get_repos_urls_from_host_with_files = function(host,
                                                   type,
                                                   with_files,
                                                   verbose,
                                                   progress) {
      purrr::map(with_files, function(file) {
        host$get_repos_urls(
          type      = type,
          with_file = file,
          verbose   = verbose,
          progress  = progress
        )
      }) |>
        unlist()
    },

    get_commits_from_hosts = function(since, until, verbose, progress) {
      commits_table <- purrr::map(private$hosts, function(host) {
        host$get_commits(
          since    = since,
          until    = until,
          verbose  = verbose,
          progress = progress
        )
      }) |>
        purrr::list_rbind() |>
        dplyr::as_tibble()
      return(commits_table)
    },

    get_issues_from_hosts = function(since, until, state, verbose, progress) {
      issues_table <- purrr::map(private$hosts, function(host) {
        host$get_issues(
          since    = since,
          until    = until,
          state    = state,
          verbose  = verbose,
          progress = progress
        )
      }) |>
        purrr::list_rbind() |>
        dplyr::as_tibble()
      return(issues_table)
    },

    get_pull_requests_from_hosts = function(since, until, state, verbose, progress) {
      pr_table <- purrr::map(private$hosts, function(host) {
        host$get_pull_requests(
          since    = since,
          until    = until,
          state    = state,
          verbose  = verbose,
          progress = progress
        )
      }) |>
        purrr::list_rbind() |>
        dplyr::as_tibble()
      return(pr_table)
    },

    get_users_from_hosts = function(logins) {
      purrr::map(private$hosts, function(host) {
        host$get_users(logins)
      }) |>
        unique() |>
        purrr::list_rbind() |>
        dplyr::as_tibble()
    },

    get_repos_trees_from_hosts = function(pattern,
                                          depth,
                                          verbose,
                                          progress) {
      purrr::map(private$hosts, function(host) {
        files_tree_table <- host$get_files_structure(
          pattern = pattern,
          depth = depth,
          verbose = verbose,
          progress = progress
        ) |>
          purrr::discard(~ length(.) == 0) |>
          purrr::imap(function(org_tree, org_name) {
            org_files_tree <- purrr::imap(org_tree, function(files_tree, repo_name) {
              dplyr::tibble(
                "repo_id" = attr(files_tree, "repo_id"),
                "repo_name" = repo_name,
                "files_tree" = list(files_tree)
              )
            }) |>
              purrr::list_rbind() |>
              dplyr::mutate(
                organization = org_name
              ) |>
              dplyr::relocate(
                organization, .before = files_tree
              )
          }) |>
          purrr::list_rbind() |>
          dplyr::mutate(
            githost = retrieve_githost(host$.__enclos_env__$private$api_url)
          )
      }) |>
        purrr::list_rbind()
    },

    get_files_from_hosts = function(pattern,
                                    depth,
                                    file_path,
                                    verbose,
                                    progress) {
      purrr::map(private$hosts, function(host) {
        if (is.null(file_path)) {
          files_structure <- host$get_files_structure(
            pattern = pattern,
            depth = depth,
            verbose = verbose,
            progress = progress
          ) |>
            purrr::discard(~ length(.) == 0)
        } else {
          files_structure <- NULL
        }
        host$get_files_content(
          file_path = file_path,
          files_structure = files_structure,
          verbose = verbose,
          progress = progress
        )
      }) |>
        purrr::list_rbind() |>
        dplyr::as_tibble()
    },

    get_release_logs_from_hosts = function(since, until, verbose, progress) {
      purrr::map(private$hosts, function(host) {
        host$get_release_logs(
          since = since,
          until = until,
          verbose = verbose,
          progress = progress
        )
      }) |>
        purrr::list_rbind() |>
        dplyr::as_tibble()
    },

    add_stats_to_repos = function(repos_table) {
      if (nrow(repos_table > 0)) {
        repos_table <- repos_table |>
          dplyr::mutate(
            last_activity = difftime(
              Sys.time(),
              last_activity_at,
              units = "days"
            ) |> round(2)
          )
        if ("contributors" %in% colnames(repos_table)) {
          repos_table <- dplyr::mutate(
            repos_table,
            contributors_n = purrr::map_vec(contributors, function(contributors_string) {
              length(stringr::str_split(contributors_string[1], ", ")[[1]])
            })
          ) |>
            dplyr::relocate(
              contributors,
              .after = last_activity
            )
        }
      } else {
        repos_table <- NULL
      }
      return(repos_table)
    },

    check_for_duplicate_hosts = function(host) {
      if (length(private$hosts) > 0) {
        hosts_to_check <- append(host, private$hosts)
        urls <- purrr::map_chr(hosts_to_check, function(host) {
          host_priv <- environment(host$initialize)$private
          host_priv$engines$rest$rest_api_url
        })

        if (length(urls) != length(unique(urls))) {
          cli::cli_abort(c(
            "x" = "You can not provide two hosts of the same API urls."
          ),
          call = NULL
          )
        }
      }

      host
    },

    print_item = function(item_name,
                          item_to_check,
                          item_to_print = item_to_check) {
      if (item_name %in% c(" Organizations", " Repositories")) {
        item_to_print <- unlist(item_to_print)
        item_to_print <- purrr::map_vec(item_to_print, function(element) {
          url_decode(element)
        })
        list_items <- cut_item_to_print(item_to_print)
        item_to_print <- paste0("[", cli::col_green(length(item_to_print)), "] ", list_items)
      }
      cat(paste0(
        cli::col_blue(paste0(item_name, ": ")),
        ifelse(
          is.null(item_to_check),
          cli::col_grey("<not defined>"),
          item_to_print
        ), "\n"
      ))
    },

    print_hosts = function() {
      hosts <- purrr::map_chr(private$hosts, function(host) {
        host_priv <- environment(host$initialize)$private
        host_priv$engines$rest$rest_api_url
      })
      private$print_item("Hosts", hosts, paste0(hosts, collapse = ", "))
    },

    print_orgs_and_repos = function() {
      orgs <- purrr::map(private$hosts, function(host) {
        host_priv <- environment(host$initialize)$private
        if ("org" %in% host_priv$searching_scope) {
          orgs <- host_priv$orgs
        }
      })
      repos <- purrr::map(private$hosts, function(host) {
        host_priv <- environment(host$initialize)$private
        if ("repo" %in% host_priv$searching_scope) {
          repos <- host_priv$repos_fullnames
        }
      })
      private$print_item(" Organizations", orgs)
      private$print_item(" Repositories", repos)
    },

    print_storage = function() {
      backend_type <- if (inherits(private$storage_backend, "StoragePostgres")) {
        "PostgreSQL"
      } else if (inherits(private$storage_backend, "StorageSQLite")) {
        "SQLite"
      } else {
        "local"
      }
      stored_names <- private$storage_backend$list()
      gitstats_storage <- purrr::map(stored_names, function(storage_name) {
        storage_object <- private$storage_backend$load(storage_name)
        if (!is.null(storage_object)) {
          storage_size <- if (inherits(storage_object, "data.frame")) {
            nrow(storage_object)
          } else {
            length(storage_object)
          }
          glue::glue(
            "{stringr::str_to_title(storage_name)}: {storage_size} {private$print_storage_attribute(storage_object, storage_name)}"
          )
        }
      }) |>
        purrr::discard(~is.null(.))
      if (length(gitstats_storage) == 0) {
        private$print_item(
          glue::glue("Storage [{backend_type}]"),
          cli::col_grey("<no data in storage>")
        )
      } else {
        cat(cli::col_blue(glue::glue("Storage [{backend_type}]:")))
        cat("\n")
        cat(paste0(" ", gitstats_storage, collapse = "\n"))
      }
    },

    print_storage_attribute = function(storage_data, storage_name) {
      if (!storage_name %in% c("repositories", "organizations")) {
        storage_attr <- switch(storage_name,
                               "repos_urls" = "type",
                               "repos_trees" = "file_pattern",
                               "files" = "file_pattern",
                               "commits" = "date_range",
                               "issues" = "date_range",
                               "pull_requests" = "date_range",
                               "release_logs" = "date_range",
                               "users" = "logins")
        attr_data <- attr(storage_data, storage_attr)
        attr_name <- switch(storage_attr,
                            "type" = "type",
                            "file_pattern" = "file pattern",
                            "date_range" = "date range",
                            "logins" = "logins")
        if (length(attr_data) > 1) {
          separator <- if (storage_attr == "date_range") {
            " - "
          } else {
            ", "
          }
          attr_data <- attr_data |> paste0(collapse = separator)
        }
        return(cli::col_grey(glue::glue("[{attr_name}: {trimws(attr_data)}]")))
      } else {
        return("")
      }
    }
  )
)

#' @title Enable parallel processing
#' @name set_parallel
#' @description Set up parallel processing for API calls using mirai daemons.
#'   When enabled, GitStats fetches data from multiple repositories
#'   concurrently. Call `set_parallel(FALSE)` or `set_parallel(0)` to revert
#'   to sequential execution.
#' @param gitstats A GitStats object.
#' @param workers Number of parallel workers. Set to `TRUE` for automatic
#'   detection, a positive integer for a specific count, or `FALSE`/`0` to
#'   disable parallelism.
#' @return A `GitStats` object.
#' @examples
#' \dontrun{
#'   my_gitstats <- create_gitstats() |>
#'     set_github_host(
#'       token = Sys.getenv("GITHUB_PAT"),
#'       orgs = c("r-world-devs", "openpharma")
#'     ) |>
#'     set_parallel(4)
#'   get_commits(my_gitstats, since = "2024-01-01")
#'   my_gitstats |> set_parallel(FALSE) # revert to sequential
#' }
#' @export
set_parallel <- function(gitstats, workers = 10L) {
  if (missing(gitstats) || !inherits(gitstats, "GitStats")) {
    cli::cli_abort(c(
      "{.fun set_parallel} requires a {.cls GitStats} object as its first argument.",
      "i" = "Use {.code my_gitstats |> set_parallel(workers)} or
             {.code set_parallel(my_gitstats, workers)}."
    ), call = NULL)
  }
  gitstats$set_parallel(workers = workers)
  return(invisible(gitstats))
}

#' @title Check if parallel processing is active
#' @name is_parallel
#' @description Returns `TRUE` when mirai daemons are running on the given
#'   `GitStats` object (i.e. `set_parallel()` has been called), `FALSE`
#'   otherwise.
#' @param gitstats A GitStats object.
#' @return A logical scalar.
#' @examples
#' \dontrun{
#'   my_gitstats <- create_gitstats()
#'   is_parallel(my_gitstats)
#' }
#' @export
is_parallel <- function(gitstats) {
  if (missing(gitstats) || !inherits(gitstats, "GitStats")) {
    cli::cli_abort(c(
      "{.fun is_parallel} requires a {.cls GitStats} object as its first argument.",
      "i" = "Use {.code is_parallel(my_gitstats)}."
    ), call = NULL)
  }
  gitstats$is_parallel()
}

test_that("get_users_from_hosts work", {
  mockery::stub(
    test_gitstats_priv$get_users_from_hosts,
    "host$get_users",
    test_mocker$use("github_users")
  )
  users_from_hosts <- test_gitstats_priv$get_users_from_hosts()
  expect_users_table(users_from_hosts)
  test_mocker$cache(users_from_hosts)
})

test_that("GitStats get users info", {
  test_gitstats <- create_test_gitstats(hosts = 2)
  mockery::stub(
    test_gitstats$get_users,
    "private$get_users_from_hosts",
    test_mocker$use("users_from_hosts")
  )
  users_result <- test_gitstats$get_users(
    c("test_user1", "test_user2"),
    verbose = FALSE
  )
  expect_users_table(
    users_result
  )
  test_mocker$cache(users_result)
})

test_that("GitStats get_users prints messages", {
  test_gitstats <- create_test_gitstats(hosts = 2)
  mockery::stub(
    test_gitstats$get_users,
    "private$get_users_from_hosts",
    test_mocker$use("users_from_hosts")
  )
  expect_snapshot(
    users_result <- test_gitstats$get_users(
      c("test_user1", "test_user2"),
      verbose = TRUE
    )
  )
})

test_that("get users works correctly", {
  test_gitstats <- create_test_gitstats(hosts = 2)
  mockery::stub(
    get_users,
    "gitstats$get_users",
    test_mocker$use("users_result")
  )
  users_result <- get_users(
    gitstats = test_gitstats,
    c("test_user1", "test_user2"),
    verbose = FALSE
  )
  expect_users_table(
    users_result
  )
})

test_that("get_users prints info on data used to pull data", {
  test_gitstats <- create_test_gitstats(hosts = 2)
  mockery::stub(
    get_users,
    "gitstats$get_users",
    test_mocker$use("users_result")
  )
  expect_snapshot(
    users_result <- get_users(
      gitstats = test_gitstats,
      c("test_user1", "test_user2"),
      verbose = TRUE
    )
  )
})

test_that("get_repos_urls_from_hosts gets data from the hosts", {
  mockery::stub(
    test_gitstats_priv$get_repos_urls_from_hosts,
    "host$get_repos_urls",
    c(test_mocker$use("gh_api_repos_urls"), test_mocker$use("gl_api_repos_urls"))
  )
  repos_urls_from_hosts <- test_gitstats_priv$get_repos_urls_from_hosts(
    type = "api",
    with_code = NULL,
    in_files = NULL,
    with_files = NULL,
    verbose = FALSE
  )
  expect_type(repos_urls_from_hosts, "character")
  expect_gt(length(repos_urls_from_hosts), 0)
  expect_true(any(grepl("gitlab.com/api", repos_urls_from_hosts)))
  expect_true(any(grepl("testhost.com/api", repos_urls_from_hosts)))
  test_mocker$cache(repos_urls_from_hosts)
})

test_that("get_repos_urls_from_hosts gets data with_code in_files from the hosts", {
  mockery::stub(
    test_gitstats_priv$get_repos_urls_from_hosts,
    "private$get_repos_urls_from_host_with_code",
    c(test_mocker$use("gh_repos_urls_with_code_in_files"), test_mocker$use("gl_repos_urls_with_code_in_files"))
  )
  repos_urls_from_hosts_with_code_in_files <- test_gitstats_priv$get_repos_urls_from_hosts(
    type = "api",
    with_code = "shiny",
    in_files = "DESCRIPTION",
    with_files = NULL,
    verbose = FALSE
  )
  expect_type(repos_urls_from_hosts_with_code_in_files, "character")
  expect_gt(length(repos_urls_from_hosts_with_code_in_files), 0)
  expect_true(any(grepl("gitlab.com", repos_urls_from_hosts_with_code_in_files)))
  expect_true(any(grepl("github", repos_urls_from_hosts_with_code_in_files)))
  test_mocker$cache(repos_urls_from_hosts_with_code_in_files)
})

test_that("set_object_class works correctly", {
  repos_urls <- test_gitstats_priv$set_object_class(
    object = test_mocker$use("repos_urls_from_hosts_with_code_in_files"),
    class = "repos_urls",
    attr_list = list(
      "type" = "api",
      "with_code" = "shiny",
      "in_files" = c("NAMESPACE", "DESCRIPTION"),
      "with_files" = NULL
    )
  )
  expect_s3_class(repos_urls, "repos_urls")
  expect_equal(attr(repos_urls, "type"), "api")
  expect_equal(attr(repos_urls, "with_code"), "shiny")
  expect_equal(attr(repos_urls, "in_files"), c("NAMESPACE", "DESCRIPTION"))
})

test_that("get_repos_urls gets vector of repository URLS", {
  test_gitstats <- create_test_gitstats(hosts = 2)
  mockery::stub(
    test_gitstats$get_repos_urls,
    "private$get_repos_urls_from_hosts",
    test_mocker$use("repos_urls_from_hosts")
  )
  repo_urls <- test_gitstats$get_repos_urls(
    verbose = FALSE
  )
  expect_type(
    repo_urls,
    "character"
  )
  expect_gt(
    length(repo_urls),
    1
  )
})

test_that("get_repos_urls prints messages", {
  test_gitstats <- create_test_gitstats(hosts = 2)
  mockery::stub(
    test_gitstats$get_repos_urls,
    "private$get_repos_urls_from_hosts",
    test_mocker$use("repos_urls_from_hosts")
  )
  expect_snapshot(
    repo_urls <- test_gitstats$get_repos_urls(
      verbose = TRUE
    )
  )
})

test_that("get_repos_urls gets vector of repository URLS", {
  test_gitstats <- create_test_gitstats(hosts = 2)
  mockery::stub(
    test_gitstats$get_repos_urls,
    "private$get_repos_urls_from_hosts",
    test_mocker$use("repos_urls_from_hosts_with_code_in_files")
  )
  repos_urls <- test_gitstats$get_repos_urls(
    with_code = "shiny",
    in_files = "DESCRIPTION",
    verbose = FALSE
  )
  expect_type(
    repos_urls,
    "character"
  )
  expect_gt(
    length(repos_urls),
    1
  )
  test_mocker$cache(repos_urls)
})

test_that("get_repos_urls gets vector of repository URLS", {
  test_gitstats <- create_test_gitstats(hosts = 2)
  mockery::stub(
    get_repos_urls,
    "gitstats$get_repos_urls",
    test_mocker$use("repos_urls")
  )
  repos_urls <- get_repos_urls(
    gitstats = test_gitstats,
    with_code = "shiny",
    in_files = "DESCRIPTION",
    verbose = FALSE
  )
  expect_type(
    repos_urls,
    "character"
  )
  expect_gt(
    length(repos_urls),
    1
  )
})

test_that("get_repos_urls prints time used to pull data", {
  test_gitstats <- create_test_gitstats(hosts = 2)
  mockery::stub(
    get_repos_urls,
    "gitstats$get_repos_urls",
    test_mocker$use("repos_urls")
  )
  expect_snapshot(
    get_repos_urls(
      gitstats = test_gitstats,
      with_code = "shiny",
      in_files = "DESCRIPTION",
      verbose = TRUE
    )
  )
})

test_that("get_repos_urls prints when data is longer than 5", {
  test_gitstats <- create_test_gitstats(hosts = 2)
  mocked_repos_urls <- rep(test_mocker$use("repos_urls"), 5)
  class(mocked_repos_urls) <- c(class(mocked_repos_urls), "gitstats_repos_urls")
  mockery::stub(
    get_repos_urls,
    "gitstats$get_repos_urls",
    mocked_repos_urls
  )
  expect_snapshot(
    get_repos_urls(
      gitstats = test_gitstats,
      with_code = "shiny",
      in_files = "DESCRIPTION",
      verbose = FALSE
    )
  )
})

test_that("print.gitstats_repos_urls prints message for empty URLs", {
  empty_urls <- character(0)
  class(empty_urls) <- c("gitstats_repos_urls", class(empty_urls))
  expect_output(print(empty_urls), "No repository URLs found")
})

test_that("print.gitstats_repos_urls prints all URLs when 5 or fewer", {
  urls <- c(
    "https://github.com/org/repo1",
    "https://github.com/org/repo2",
    "https://gitlab.com/group/project"
  )
  class(urls) <- c("gitstats_repos_urls", class(urls))
  output <- capture.output(print(urls))
  expect_true(any(grepl("Repository URLs:", output)))
  expect_true(any(grepl("repo1", output)))
  expect_true(any(grepl("repo2", output)))
  expect_true(any(grepl("project", output)))
  expect_true(any(grepl("Host Summary:", output)))
})

test_that("show_orgs() returns orgs", {
  expect_equal(
    show_orgs(test_gitstats),
    c("github_test_org", "gitlab_test_group")
  )
})

test_that("show_hosts() returns some details on hosts", {
  expect_equal(
    show_hosts(test_gitstats),
    list(
      list(
        "host" = "GitHub", 
        "web_url" = "https://github.com",
        "api_url" = "https://api.github.com"
      ),
      list(
        "host" = "GitLab", 
        "web_url" = "https://gitlab.com",
        "api_url" = "https://gitlab.com/api/v4"
      )
    )
  )
})

# get_date_gaps ----------------------------------------------------------------

test_that("get_date_gaps returns gap when extending forward", {
  test_gitstats <- create_test_gitstats(hosts = 1, priv_mode = TRUE)
  gaps <- test_gitstats$get_date_gaps(
    stored_range = c("2024-01-01", "2024-06-30"),
    requested_range = c("2024-01-01", "2024-12-31")
  )
  expect_length(gaps, 1)
  expect_equal(gaps[[1]]$since, "2024-07-01")
  expect_equal(gaps[[1]]$until, "2024-12-31")
})

test_that("get_date_gaps returns gap when extending backward", {
  test_gitstats <- create_test_gitstats(hosts = 1, priv_mode = TRUE)
  gaps <- test_gitstats$get_date_gaps(
    stored_range = c("2024-06-01", "2024-12-31"),
    requested_range = c("2024-01-01", "2024-12-31")
  )
  expect_length(gaps, 1)
  expect_equal(gaps[[1]]$since, "2024-01-01")
  expect_equal(gaps[[1]]$until, "2024-05-31")
})

test_that("get_date_gaps returns two gaps when extending both directions", {
  test_gitstats <- create_test_gitstats(hosts = 1, priv_mode = TRUE)
  gaps <- test_gitstats$get_date_gaps(
    stored_range = c("2024-04-01", "2024-09-30"),
    requested_range = c("2024-01-01", "2024-12-31")
  )
  expect_length(gaps, 2)
  expect_equal(gaps[[1]]$since, "2024-01-01")
  expect_equal(gaps[[1]]$until, "2024-03-31")
  expect_equal(gaps[[2]]$since, "2024-10-01")
  expect_equal(gaps[[2]]$until, "2024-12-31")
})

test_that("get_date_gaps returns empty when requested is within stored", {
  test_gitstats <- create_test_gitstats(hosts = 1, priv_mode = TRUE)
  gaps <- test_gitstats$get_date_gaps(
    stored_range = c("2024-01-01", "2024-12-31"),
    requested_range = c("2024-03-01", "2024-09-30")
  )
  expect_length(gaps, 0)
})

test_that("get_date_gaps returns empty when ranges are identical", {
  test_gitstats <- create_test_gitstats(hosts = 1, priv_mode = TRUE)
  gaps <- test_gitstats$get_date_gaps(
    stored_range = c("2024-01-01", "2024-12-31"),
    requested_range = c("2024-01-01", "2024-12-31")
  )
  expect_length(gaps, 0)
})

# Incremental get_commits ------------------------------------------------------

test_that("get_commits fetches only missing date range", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_commits <- dplyr::tibble(
    repo_name = "test/repo",
    id = "abc123",
    committed_date = as.POSIXct("2024-03-15"),
    author = "author1",
    author_login = "author1",
    author_name = "Author One",
    additions = 10L,
    deletions = 5L,
    organization = "test_org",
    repo_url = "https://github.com/test/repo",
    api_url = "https://api.github.com"
  )
  class(stored_commits) <- c("gitstats_commits", class(stored_commits))
  attr(stored_commits, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(stored_commits, "scope") <- "test_org"
  priv$storage_backend$save("commits", stored_commits)

  new_commits <- dplyr::tibble(
    repo_name = "test/repo",
    id = "def456",
    committed_date = as.POSIXct("2024-09-15"),
    author = "author2",
    author_login = "author2",
    author_name = "Author Two",
    additions = 20L,
    deletions = 10L,
    organization = "test_org",
    repo_url = "https://github.com/test/repo",
    api_url = "https://api.github.com"
  )

  mockery::stub(
    test_gitstats$get_commits,
    "private$get_commits_from_hosts",
    new_commits
  )

  suppressMessages(
    result <- test_gitstats$get_commits(
      since = "2024-01-01",
      until = "2024-12-31",
      verbose = FALSE
    )
  )

  expect_equal(nrow(result), 2)
  expect_in("abc123", result$id)
  expect_in("def456", result$id)
  expect_equal(attr(result, "date_range"), c("2024-01-01", "2024-12-31"))
})

test_that("get_commits deduplicates by commit id", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_commits <- dplyr::tibble(
    repo_name = "test/repo",
    id = c("abc123", "shared"),
    committed_date = as.POSIXct(c("2024-03-15", "2024-06-30")),
    author = "author1",
    author_login = "author1",
    author_name = "Author One",
    additions = 10L,
    deletions = 5L,
    organization = "test_org",
    repo_url = "https://github.com/test/repo",
    api_url = "https://api.github.com"
  )
  class(stored_commits) <- c("gitstats_commits", class(stored_commits))
  attr(stored_commits, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(stored_commits, "scope") <- "test_org"
  priv$storage_backend$save("commits", stored_commits)

  new_commits <- dplyr::tibble(
    repo_name = "test/repo",
    id = c("shared", "def456"),
    committed_date = as.POSIXct(c("2024-06-30", "2024-09-15")),
    author = "author1",
    author_login = "author1",
    author_name = "Author One",
    additions = 10L,
    deletions = 5L,
    organization = "test_org",
    repo_url = "https://github.com/test/repo",
    api_url = "https://api.github.com"
  )

  mockery::stub(
    test_gitstats$get_commits,
    "private$get_commits_from_hosts",
    new_commits
  )

  suppressMessages(
    result <- test_gitstats$get_commits(
      since = "2024-01-01",
      until = "2024-12-31",
      verbose = FALSE
    )
  )

  expect_equal(nrow(result), 3)
  expect_equal(sum(result$id == "shared"), 1)
})

test_that("get_commits does full fetch when storage is empty", {
  test_gitstats <- create_test_gitstats(hosts = 1)

  fresh_commits <- dplyr::tibble(
    repo_name = "test/repo",
    id = "abc123",
    committed_date = as.POSIXct("2024-03-15"),
    author = "author1",
    author_login = "author1",
    author_name = "Author One",
    additions = 10L,
    deletions = 5L,
    organization = "test_org",
    repo_url = "https://github.com/test/repo",
    api_url = "https://api.github.com"
  )

  mockery::stub(
    test_gitstats$get_commits,
    "private$get_commits_from_hosts",
    fresh_commits
  )

  suppressMessages(
    result <- test_gitstats$get_commits(
      since = "2024-01-01",
      until = "2024-12-31",
      verbose = FALSE
    )
  )

  expect_equal(nrow(result), 1)
  expect_s3_class(result, "gitstats_commits")
})

test_that("get_commits does full fetch when cache is FALSE", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_commits <- dplyr::tibble(
    repo_name = "test/repo",
    id = "old_commit",
    committed_date = as.POSIXct("2024-03-15"),
    author = "author1",
    author_login = "author1",
    author_name = "Author One",
    additions = 10L,
    deletions = 5L,
    organization = "test_org",
    repo_url = "https://github.com/test/repo",
    api_url = "https://api.github.com"
  )
  class(stored_commits) <- c("gitstats_commits", class(stored_commits))
  attr(stored_commits, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(stored_commits, "scope") <- "test_org"
  priv$storage_backend$save("commits", stored_commits)

  fresh_commits <- dplyr::tibble(
    repo_name = "test/repo",
    id = "new_commit",
    committed_date = as.POSIXct("2024-09-15"),
    author = "author2",
    author_login = "author2",
    author_name = "Author Two",
    additions = 20L,
    deletions = 10L,
    organization = "test_org",
    repo_url = "https://github.com/test/repo",
    api_url = "https://api.github.com"
  )

  mockery::stub(
    test_gitstats$get_commits,
    "private$get_commits_from_hosts",
    fresh_commits
  )

  suppressMessages(
    result <- test_gitstats$get_commits(
      since = "2024-01-01",
      until = "2024-12-31",
      cache = FALSE,
      verbose = FALSE
    )
  )

  # Full re-fetch, not incremental
  expect_equal(nrow(result), 1)
  expect_equal(result$id, "new_commit")
})

test_that("get_commits returns stored data when no gaps exist", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_commits <- dplyr::tibble(
    repo_name = "test/repo",
    id = "abc123",
    committed_date = as.POSIXct("2024-03-15"),
    author = "author1",
    author_login = "author1",
    author_name = "Author One",
    additions = 10L,
    deletions = 5L,
    organization = "test_org",
    repo_url = "https://github.com/test/repo",
    api_url = "https://api.github.com"
  )
  class(stored_commits) <- c("gitstats_commits", class(stored_commits))
  attr(stored_commits, "date_range") <- c("2024-01-01", "2024-12-31")
  attr(stored_commits, "scope") <- "test_org"
  priv$storage_backend$save("commits", stored_commits)

  result <- test_gitstats$get_commits(
    since = "2024-01-01",
    until = "2024-12-31",
    verbose = FALSE
  )

  expect_equal(nrow(result), 1)
  expect_s3_class(result, "gitstats_commits")
})

test_that("get_commits uses stored data when requested range is a subset", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_commits <- dplyr::tibble(
    repo_name = "test/repo",
    id = c("abc123", "def456"),
    committed_date = as.POSIXct(c("2024-03-15", "2024-09-15")),
    author = "author1",
    author_login = "author1",
    author_name = "Author One",
    additions = 10L,
    deletions = 5L,
    organization = "test_org",
    repo_url = "https://github.com/test/repo",
    api_url = "https://api.github.com"
  )
  class(stored_commits) <- c("gitstats_commits", class(stored_commits))
  attr(stored_commits, "date_range") <- c("2024-01-01", "2024-12-31")
  attr(stored_commits, "scope") <- "test_org"
  priv$storage_backend$save("commits", stored_commits)

  suppressMessages(
    result <- test_gitstats$get_commits(
      since = "2024-03-01",
      until = "2024-09-30",
      verbose = FALSE
    )
  )

  expect_equal(nrow(result), 2)
  expect_s3_class(result, "gitstats_commits")
  expect_equal(attr(result, "date_range"), c("2024-03-01", "2024-09-30"))
})

test_that("get_commits handles gaps with no new commits from API", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_commits <- dplyr::tibble(
    repo_name = "test/repo",
    id = "abc123",
    committed_date = as.POSIXct("2024-03-15"),
    author = "author1",
    author_login = "author1",
    author_name = "Author One",
    additions = 10L,
    deletions = 5L,
    organization = "test_org",
    repo_url = "https://github.com/test/repo",
    api_url = "https://api.github.com"
  )
  class(stored_commits) <- c("gitstats_commits", class(stored_commits))
  attr(stored_commits, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(stored_commits, "scope") <- "test_org"
  priv$storage_backend$save("commits", stored_commits)

  mockery::stub(
    test_gitstats$get_commits,
    "private$get_commits_from_hosts",
    dplyr::tibble(
      repo_name = character(),
      id = character(),
      committed_date = as.POSIXct(character()),
      author = character(),
      author_login = character(),
      author_name = character(),
      additions = integer(),
      deletions = integer(),
      organization = character(),
      repo_url = character(),
      api_url = character()
    )
  )

  suppressMessages(
    result <- test_gitstats$get_commits(
      since = "2024-01-01",
      until = "2024-12-31",
      verbose = FALSE
    )
  )

  expect_equal(nrow(result), 1)
  expect_equal(result$id, "abc123")
  expect_s3_class(result, "gitstats_commits")
})

test_that("get_commits prints cli messages when verbose is TRUE", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_commits <- dplyr::tibble(
    repo_name = "test/repo",
    id = "abc123",
    committed_date = as.POSIXct("2024-03-15"),
    author = "author1",
    author_login = "author1",
    author_name = "Author One",
    additions = 10L,
    deletions = 5L,
    organization = "test_org",
    repo_url = "https://github.com/test/repo",
    api_url = "https://api.github.com"
  )
  class(stored_commits) <- c("gitstats_commits", class(stored_commits))
  attr(stored_commits, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(stored_commits, "scope") <- "test_org"
  priv$storage_backend$save("commits", stored_commits)

  new_commits <- dplyr::tibble(
    repo_name = "test/repo",
    id = "def456",
    committed_date = as.POSIXct("2024-09-15"),
    author = "author2",
    author_login = "author2",
    author_name = "Author Two",
    additions = 20L,
    deletions = 10L,
    organization = "test_org",
    repo_url = "https://github.com/test/repo",
    api_url = "https://api.github.com"
  )

  mockery::stub(
    test_gitstats$get_commits,
    "private$get_commits_from_hosts",
    new_commits
  )

  output <- capture.output(
    result <- test_gitstats$get_commits(
      since = "2024-01-01",
      until = "2024-12-31",
      verbose = TRUE
    ),
    type = "message"
  )

  expect_true(any(grepl("cached commits", output)))
  expect_true(any(grepl("2024-01-01", output)))
  expect_true(any(grepl("Pulling commits", output)))
})

# non_date_args_changed --------------------------------------------------------

test_that("non_date_args_changed returns FALSE when no non-date args", {
  test_gitstats <- create_test_gitstats(hosts = 1, priv_mode = TRUE)
  data <- dplyr::tibble(x = 1)
  attr(data, "date_range") <- c("2024-01-01", "2024-06-30")
  result <- test_gitstats$non_date_args_changed(
    data, list("date_range" = c("2024-01-01", "2024-12-31"))
  )
  expect_false(result)
})

test_that("non_date_args_changed returns TRUE when state changed", {
  test_gitstats <- create_test_gitstats(hosts = 1, priv_mode = TRUE)
  data <- dplyr::tibble(x = 1)
  attr(data, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(data, "state") <- "open"
  result <- test_gitstats$non_date_args_changed(
    data, list("date_range" = c("2024-01-01", "2024-12-31"), "state" = "all")
  )
  expect_true(result)
})

test_that("non_date_args_changed returns FALSE when state is same", {
  test_gitstats <- create_test_gitstats(hosts = 1, priv_mode = TRUE)
  data <- dplyr::tibble(x = 1)
  attr(data, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(data, "state") <- "open"
  result <- test_gitstats$non_date_args_changed(
    data, list("date_range" = c("2024-01-01", "2024-12-31"), "state" = "open")
  )
  expect_false(result)
})

test_that("non_date_args_changed returns TRUE when scope changed", {
  test_gitstats <- create_test_gitstats(hosts = 1, priv_mode = TRUE)
  data <- dplyr::tibble(x = 1)
  attr(data, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(data, "scope") <- "old_org"
  result <- test_gitstats$non_date_args_changed(
    data, list("date_range" = c("2024-01-01", "2024-12-31"), "scope" = "new_org")
  )
  expect_true(result)
})

test_that("non_date_args_changed returns FALSE when scope is same", {
  test_gitstats <- create_test_gitstats(hosts = 1, priv_mode = TRUE)
  data <- dplyr::tibble(x = 1)
  attr(data, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(data, "scope") <- "test_org"
  result <- test_gitstats$non_date_args_changed(
    data, list("date_range" = c("2024-01-01", "2024-12-31"), "scope" = "test_org")
  )
  expect_false(result)
})

# Scope change triggers full re-fetch -----------------------------------------

test_that("get_commits does full re-fetch when scanning scope changes", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_commits <- dplyr::tibble(
    repo_name = "old_org/repo",
    id = "old_commit",
    committed_date = as.POSIXct("2024-03-15"),
    author = "author1",
    author_login = "author1",
    author_name = "Author One",
    additions = 10L,
    deletions = 5L,
    organization = "old_org",
    repo_url = "https://github.com/old_org/repo",
    api_url = "https://api.github.com"
  )
  class(stored_commits) <- c("gitstats_commits", class(stored_commits))
  attr(stored_commits, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(stored_commits, "scope") <- "old_org"
  priv$storage_backend$save("commits", stored_commits)

  fresh_commits <- dplyr::tibble(
    repo_name = "test_org/repo",
    id = "new_commit",
    committed_date = as.POSIXct("2024-03-15"),
    author = "author1",
    author_login = "author1",
    author_name = "Author One",
    additions = 10L,
    deletions = 5L,
    organization = "test_org",
    repo_url = "https://github.com/test_org/repo",
    api_url = "https://api.github.com"
  )

  mockery::stub(
    test_gitstats$get_commits,
    "private$get_commits_from_hosts",
    fresh_commits
  )

  suppressMessages(
    result <- test_gitstats$get_commits(
      since = "2024-01-01",
      until = "2024-06-30",
      verbose = FALSE
    )
  )

  # Full re-fetch: only new data, old data discarded
  expect_equal(nrow(result), 1)
  expect_equal(result$id, "new_commit")
})

test_that("get_issues does full re-fetch when scanning scope changes", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_issues <- dplyr::tibble(
    repo_name = "old_org/repo",
    number = 1L,
    title = "Old issue",
    description = "desc",
    created_at = as.POSIXct("2024-03-15"),
    closed_at = as.POSIXct(NA),
    state = "open",
    url = "https://github.com/old_org/repo/issues/1",
    author = "author1",
    organization = "old_org",
    api_url = "https://api.github.com"
  )
  class(stored_issues) <- c("gitstats_issues", class(stored_issues))
  attr(stored_issues, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(stored_issues, "state") <- "open"
  attr(stored_issues, "scope") <- "old_org"
  priv$storage_backend$save("issues", stored_issues)

  fresh_issues <- dplyr::tibble(
    repo_name = "test_org/repo",
    number = 5L,
    title = "New issue",
    description = "desc",
    created_at = as.POSIXct("2024-03-15"),
    closed_at = as.POSIXct(NA),
    state = "open",
    url = "https://github.com/test_org/repo/issues/5",
    author = "author1",
    organization = "test_org",
    api_url = "https://api.github.com"
  )

  mockery::stub(
    test_gitstats$get_issues,
    "private$get_issues_from_hosts",
    fresh_issues
  )

  suppressMessages(
    result <- test_gitstats$get_issues(
      since = "2024-01-01",
      until = "2024-06-30",
      state = "open",
      verbose = FALSE
    )
  )

  expect_equal(nrow(result), 1)
  expect_equal(result$number, 5L)
})

# Incremental get_issues -------------------------------------------------------

test_that("get_issues fetches only missing date range", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_issues <- dplyr::tibble(
    repo_name = "test/repo",
    number = 1L,
    title = "Old issue",
    description = "desc",
    created_at = as.POSIXct("2024-03-15"),
    closed_at = as.POSIXct(NA),
    state = "open",
    url = "https://github.com/test/repo/issues/1",
    author = "author1",
    organization = "test_org",
    api_url = "https://api.github.com"
  )
  class(stored_issues) <- c("gitstats_issues", class(stored_issues))
  attr(stored_issues, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(stored_issues, "state") <- "open"
  attr(stored_issues, "scope") <- "test_org"
  priv$storage_backend$save("issues", stored_issues)

  new_issues <- dplyr::tibble(
    repo_name = "test/repo",
    number = 2L,
    title = "New issue",
    description = "desc",
    created_at = as.POSIXct("2024-09-15"),
    closed_at = as.POSIXct(NA),
    state = "open",
    url = "https://github.com/test/repo/issues/2",
    author = "author2",
    organization = "test_org",
    api_url = "https://api.github.com"
  )

  mockery::stub(
    test_gitstats$get_issues,
    "private$get_issues_from_hosts",
    new_issues
  )

  suppressMessages(
    result <- test_gitstats$get_issues(
      since = "2024-01-01",
      until = "2024-12-31",
      state = "open",
      verbose = FALSE
    )
  )

  expect_equal(nrow(result), 2)
  expect_in(1L, result$number)
  expect_in(2L, result$number)
  expect_equal(attr(result, "date_range"), c("2024-01-01", "2024-12-31"))
})

test_that("get_issues does full re-fetch when state changes", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_issues <- dplyr::tibble(
    repo_name = "test/repo",
    number = 1L,
    title = "Open issue",
    description = "desc",
    created_at = as.POSIXct("2024-03-15"),
    closed_at = as.POSIXct(NA),
    state = "open",
    url = "https://github.com/test/repo/issues/1",
    author = "author1",
    organization = "test_org",
    api_url = "https://api.github.com"
  )
  class(stored_issues) <- c("gitstats_issues", class(stored_issues))
  attr(stored_issues, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(stored_issues, "state") <- "open"
  attr(stored_issues, "scope") <- "test_org"
  priv$storage_backend$save("issues", stored_issues)

  all_issues <- dplyr::tibble(
    repo_name = "test/repo",
    number = c(1L, 3L),
    title = c("Open issue", "Closed issue"),
    description = "desc",
    created_at = as.POSIXct(c("2024-03-15", "2024-04-01")),
    closed_at = as.POSIXct(c(NA, "2024-05-01")),
    state = c("open", "closed"),
    url = paste0("https://github.com/test/repo/issues/", c(1, 3)),
    author = "author1",
    organization = "test_org",
    api_url = "https://api.github.com"
  )

  mockery::stub(
    test_gitstats$get_issues,
    "private$get_issues_from_hosts",
    all_issues
  )

  suppressMessages(
    result <- test_gitstats$get_issues(
      since = "2024-01-01",
      until = "2024-06-30",
      state = "all",
      verbose = FALSE
    )
  )

  expect_equal(nrow(result), 2)
  expect_in("closed", result$state)
})

# Incremental get_pull_requests ------------------------------------------------

test_that("get_pull_requests fetches only missing date range", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_prs <- dplyr::tibble(
    repo_name = "test/repo",
    number = 10L,
    created_at = as.POSIXct("2024-03-15"),
    merged_at = as.POSIXct(NA),
    state = "open",
    author = "author1",
    source_branch = "feature",
    target_branch = "main",
    organization = "test_org",
    api_url = "https://api.github.com"
  )
  class(stored_prs) <- c("gitstats_pull_requests", class(stored_prs))
  attr(stored_prs, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(stored_prs, "state") <- "all"
  attr(stored_prs, "scope") <- "test_org"
  priv$storage_backend$save("pull_requests", stored_prs)

  new_prs <- dplyr::tibble(
    repo_name = "test/repo",
    number = 20L,
    created_at = as.POSIXct("2024-09-15"),
    merged_at = as.POSIXct("2024-09-20"),
    state = "merged",
    author = "author2",
    source_branch = "hotfix",
    target_branch = "main",
    organization = "test_org",
    api_url = "https://api.github.com"
  )

  mockery::stub(
    test_gitstats$get_pull_requests,
    "private$get_pull_requests_from_hosts",
    new_prs
  )

  suppressMessages(
    result <- test_gitstats$get_pull_requests(
      since = "2024-01-01",
      until = "2024-12-31",
      state = "all",
      verbose = FALSE
    )
  )

  expect_equal(nrow(result), 2)
  expect_in(10L, result$number)
  expect_in(20L, result$number)
  expect_equal(attr(result, "date_range"), c("2024-01-01", "2024-12-31"))
})

# Incremental get_release_logs -------------------------------------------------

test_that("get_release_logs fetches only missing date range", {
  test_gitstats <- create_test_gitstats(hosts = 1)
  priv <- test_gitstats$.__enclos_env__$private

  stored_releases <- dplyr::tibble(
    repo_name = "test/repo",
    repo_url = "https://github.com/test/repo",
    release_name = "v1.0",
    release_tag = "v1.0.0",
    published_at = as.POSIXct("2024-03-15"),
    release_url = "https://github.com/test/repo/releases/v1.0.0",
    release_log = "Initial release"
  )
  class(stored_releases) <- c("gitstats_releases", class(stored_releases))
  attr(stored_releases, "date_range") <- c("2024-01-01", "2024-06-30")
  attr(stored_releases, "scope") <- "test_org"
  priv$storage_backend$save("release_logs", stored_releases)

  new_releases <- dplyr::tibble(
    repo_name = "test/repo",
    repo_url = "https://github.com/test/repo",
    release_name = "v2.0",
    release_tag = "v2.0.0",
    published_at = as.POSIXct("2024-09-15"),
    release_url = "https://github.com/test/repo/releases/v2.0.0",
    release_log = "Major update"
  )

  mockery::stub(
    test_gitstats$get_release_logs,
    "private$get_release_logs_from_hosts",
    new_releases
  )

  suppressMessages(
    result <- test_gitstats$get_release_logs(
      since = "2024-01-01",
      until = "2024-12-31",
      verbose = FALSE
    )
  )

  expect_equal(nrow(result), 2)
  expect_in("v1.0.0", result$release_tag)
  expect_in("v2.0.0", result$release_tag)
  expect_equal(attr(result, "date_range"), c("2024-01-01", "2024-12-31"))
})

test_that("get_commits_from_hosts works", {
  mockery::stub(
    test_gitstats_priv$get_commits_from_hosts,
    "host$get_commits",
    purrr::list_rbind(list(
      test_mocker$use("gh_commits_table"),
      test_mocker$use("gl_commits_table")
    ))
  )
  commits_table_from_hosts <- test_gitstats_priv$get_commits_from_hosts(
    since = "2023-01-01",
    until = "2025-03-06",
    verbose = FALSE,
    progress = FALSE
  )
  expect_commits_table(
    commits_table_from_hosts
  )
  test_mocker$cache(commits_table_from_hosts)
})

test_that("set_object_class works for commits", {
  commits_table <- test_gitstats_priv$set_object_class(
    object = test_mocker$use("commits_table_from_hosts"),
    class = "commits_data",
    attr_list = list(
      "date_range" = c("2023-06-15", "2023-06-30")
    )
  )
  expect_s3_class(
    commits_table,
    "commits_data"
  )
  expect_equal(
    attr(commits_table, "date_range"),
    c("2023-06-15", "2023-06-30")
  )
})

test_that("get_commits works properly", {
  mockery::stub(
    test_gitstats$get_commits,
    "private$get_commits_from_hosts",
    test_mocker$use("commits_table_from_hosts")
  )
  suppressMessages(
    commits_table <- test_gitstats$get_commits(
      since = "2023-06-15",
      until = "2023-06-30",
      verbose = FALSE
    )
  )
  expect_commits_table(
    commits_table
  )
  test_mocker$cache(commits_table)
})

test_that("get_commits() pulls data from storage for the second time", {
  expect_snapshot(
    commits_table <- test_gitstats$get_commits(
      since = "2023-06-15",
      until = "2023-06-30",
      verbose = FALSE
    )
  )  
  expect_commits_table(
    commits_table
  )
})

test_that("get_commits prints warning when commits table is empty", {
  mockery::stub(
    test_gitstats$get_commits,
    "private$get_commits_from_hosts",
    data.frame()
  )
  expect_snapshot(
    commits_table <- test_gitstats$get_commits(
      since = "2023-06-15",
      until = "2023-06-30",
      cache = FALSE,
      verbose = TRUE
    )
  )
  expect_equal(nrow(commits_table), 0)
})

test_that("get_commits() works", {
  mockery::stub(
    get_commits,
    "gitstats$get_commits",
    test_mocker$use("commits_table")
  )
  commits_data <- get_commits(
    test_gitstats,
    since = "2023-06-15",
    until = "2023-06-30",
    verbose = FALSE
  )
  expect_s3_class(
    commits_data,
    "gitstats_commits"
  )
  test_mocker$cache(commits_data)
})

test_that("get_commits() returns error when since is not defined", {
  mockery::stub(
    get_commits,
    "gitstats$get_commits",
    test_mocker$use("commits_table")
  )
  expect_snapshot_error(
    get_commits(
      test_gitstats,
      verbose = FALSE
    )
  )
})

test_that("get_commits() prints time used", {
  mockery::stub(
    get_commits,
    "gitstats$get_commits",
    test_mocker$use("commits_table")
  )
  expect_snapshot(
    commits <- get_commits(
      test_gitstats,
      since = "2023-06-15",
      until = "2023-06-30",
      verbose = TRUE
    )
  )
})

test_that("private get_from_storage works", {
  test_gitstats_priv <- test_gitstats$.__enclos_env__$private
  suppressMessages({
    commits <- test_gitstats_priv$get_from_storage(table = "commits")
  })
  expect_s3_class(
    commits,
    "tbl"
  )
  expect_commits_table(
    commits
  )
})

test_that("get_storage works", {
  gitstats_storage <- test_gitstats$get_storage(
    storage = NULL
  )
  expect_type(
    gitstats_storage,
    "list"
  )
  expect_s3_class(
    gitstats_storage[[1]],
    "tbl"
  )
  expect_gt(length(gitstats_storage), 0)
})
test_that("get_storage retrieves one table", {
  gitstats_storage <- test_gitstats$get_storage(
    storage = "commits"
  )
  expect_s3_class(
    gitstats_storage,
    "tbl"
  )
  expect_s3_class(
    gitstats_storage,
    "gitstats_commits"
  )
  expect_commits_table(
    gitstats_storage
  )
})

test_that("get_storage retrieves one table", {
  gitstats_storage <- get_storage(
    gitstats = test_gitstats,
    storage = "files"
  )
  expect_s3_class(
    gitstats_storage,
    "tbl"
  )
  expect_s3_class(
    gitstats_storage,
    "gitstats_files"
  )
})

test_that("get_orgs_from_hosts works", {
  mockery::stub(
    test_gitstats_priv$get_orgs_from_hosts,
    "host$get_orgs",
    purrr::list_rbind(list(
      test_mocker$use("github_orgs_table"),
      test_mocker$use("gitlab_orgs_table")
    ))
  )
  orgs_from_hosts <- test_gitstats_priv$get_orgs_from_hosts(
    output = "full_table",
    verbose = FALSE
  )
  expect_orgs_table(
    orgs_from_hosts,
    add_cols = c("host_url", "host_name")
  )
  test_mocker$cache(orgs_from_hosts)
})

test_that("set_object_class for repos_table works correctly", {
  orgs_table <- test_gitstats_priv$set_object_class(
    object = test_mocker$use("orgs_from_hosts"),
    class = "gitstats_orgs"
  )
  expect_s3_class(orgs_table, "gitstats_orgs")
})

test_that("get_orgs works", {
  mockery::stub(
    test_gitstats$get_orgs,
    "private$get_orgs_from_hosts",
    test_mocker$use("orgs_from_hosts")
  )
  orgs_table <- test_gitstats$get_orgs()
  expect_s3_class(orgs_table, "gitstats_orgs")
  expect_orgs_table(orgs_table, add_cols = c("host_url", "host_name"))
  test_mocker$cache(orgs_table)
})

test_that("get_orgs() pulls data from storage for the second time", {
  expect_snapshot(
    orgs_table <- test_gitstats$get_orgs()      
  )  
  expect_orgs_table(orgs_table, add_cols = c("host_url", "host_name"))
})

test_that("get_orgs works", {
  mockery::stub(
    get_orgs,
    "gitstats$get_orgs",
    test_mocker$use("orgs_table")
  )
  orgs_table <- get_orgs(
    test_gitstats,
    verbose = FALSE
  )
  expect_s3_class(orgs_table, "gitstats_orgs")
  test_mocker$cache(orgs_table)
})

test_that("get_orgs prints info on time used to pull data", {
  mockery::stub(
    get_orgs,
    "gitstats$get_orgs",
    test_mocker$use("orgs_table")
  )
  expect_snapshot(
    orgs_table <- get_orgs(
      test_gitstats,
      verbose = TRUE
    )
  )
})

test_gitstats <- create_gitstats()

test_that("Set connection returns appropriate messages", {
  if (!integration_tests_skipped) {
    expect_snapshot(
      set_github_host(
        gitstats = test_gitstats,
        token = Sys.getenv("GITHUB_PAT"),
        orgs = c("openpharma", "r-world-devs")
      )
    )
    expect_snapshot(
      test_gitstats |> set_gitlab_host(
        token = Sys.getenv("GITLAB_PAT_PUBLIC"),
        orgs = c("mbtests")
      )
    )
  }
})

test_that("When empty token for GitHub, GitStats pulls default token", {
  skip_on_cran()
  if (!integration_tests_skipped) {
    expect_snapshot(
      test_gitstats <- create_gitstats() |>
        set_github_host(
          orgs = c("openpharma", "r-world-devs")
        )
    )
  }
})

test_that("When empty token for GitLab, GitStats pulls default token", {
  if (!integration_tests_skipped) {
    expect_snapshot(
      withr::with_envvar(new = c("GITLAB_PAT" = Sys.getenv("GITLAB_PAT_PUBLIC")), {
        test_gitstats <- create_gitstats() |>
          set_gitlab_host(
            orgs = "mbtests"
          )
      })
    )
  }
})

test_that("Set GitHub host with particular repos vector instead of orgs", {
  if (!integration_tests_skipped) {
    test_gitstats <- create_gitstats()
    expect_snapshot(
      test_gitstats |>
        set_github_host(
          token = Sys.getenv("GITHUB_PAT"),
          repos = c("r-world-devs/GitStats", "r-world-devs/shinyCohortBuilder", "openpharma/GithubMetrics", "openpharma/DataFakeR")
        )
    )
    expect_length(
      test_gitstats$.__enclos_env__$private$hosts,
      1
    )
  }
})

test_that("Set GitLab host with particular repos vector instead of orgs", {
  if (!integration_tests_skipped) {
    test_gitstats <- create_gitstats()
    expect_snapshot(
      test_gitstats |>
        set_gitlab_host(
          token = Sys.getenv("GITLAB_PAT_PUBLIC"),
          repos = c("mbtests/gitstatstesting", "mbtests/gitstats-testing-2")
        )
    )
    expect_length(
      test_gitstats$.__enclos_env__$private$hosts,
      1
    )
  }
})

test_that("When no organizations or repositories are set information is printed", {
  skip_on_cran()
  test_gitstats <- create_gitstats()
  expect_snapshot(
    test_gitstats |>
      set_github_host(
        token = Sys.getenv("GITHUB_PAT"),
        verbose = TRUE
      )
  )
})

test_that("Error shows, when wrong input is passed when setting connection and host is not passed", {
  if (!integration_tests_skipped) {
    test_gitstats <- create_gitstats()
    expect_snapshot_error(
      set_gitlab_host(
        gitstats = test_gitstats,
        host = "https://avengers.com",
        token = Sys.getenv("GITLAB_PAT_PUBLIC")
      )
    )
    expect_error({
      create_gitstats() |>
        set_github_host(
          host = "wrong.url",
          orgs = c("openpharma", "r_world_devs")
        )
      },
      "Could not resolve host."
    )
  }
})

test_that("Error pops out, when two clients of the same url api are passed as input", {
  if (!integration_tests_skipped) {
    test_gitstats <- create_gitstats()
    expect_snapshot(
      error = TRUE,
      test_gitstats |>
        set_github_host(
          token = Sys.getenv("GITHUB_PAT"),
          orgs = "pharmaverse"
        ) |>
        set_github_host(
          token = Sys.getenv("GITHUB_PAT"),
          orgs = "openpharma"
        )
    )
  }
})

test_that("Error pops out when `org` does not exist", {
  if (!integration_tests_skipped) {
    expect_snapshot({
      test_gitstats <- create_gitstats() |>
        set_github_host(
          token = Sys.getenv("GITHUB_PAT"),
          orgs = c("openparma")
        )
      },
      error = TRUE
    )

    expect_snapshot({
      test_gitstats <- create_gitstats() |>
        set_gitlab_host(
          token = Sys.getenv("GITLAB_PAT_PUBLIC"),
          orgs = c("openparma", "mbtests")
        )
      },
      error = TRUE
    )

    expect_snapshot({
      test_gitstats <- create_gitstats() |>
        set_github_host(
          token = Sys.getenv("GITHUB_PAT"),
          orgs = c("openpharma", "r_world_devs")
        )
      },
      error = TRUE
    )
  }
})

test_that("When wrong orgs and repos are passed they are excluded but host is created", {
  if (!integration_tests_skipped) {
    expect_snapshot(
      test_gitstats <- create_gitstats() |>
        set_github_host(
          orgs = c("openpharma", "r_world_devs"),
          repos = c("r-world-devs/GitStats", "r-world-devs/GitMetrics"),
          verbose = TRUE,
          .error = FALSE
        )
    )
  }
})

test_that("Error pops out when repos are passed without org prefix", {
  skip_on_cran()
  expect_error(
    create_gitstats() |>
      set_github_host(
        token = Sys.getenv("GITHUB_PAT"),
        repos = c("GitStats", "shinyCohortBuilder")
      ),
    "org/repo_name"
  )
  expect_error(
    create_gitstats() |>
      set_gitlab_host(
        token = Sys.getenv("GITLAB_PAT_PUBLIC"),
        repos = c("gitstatstesting")
      ),
    "org/repo_name"
  )
})

test_that("Setting verbose for set_*_host() to FALSE works fine", {
  skip_on_cran()
  if (!integration_tests_skipped) {
    expect_no_error(
      create_gitstats() |>
        set_github_host(
          orgs = c("openpharma", "r-world-devs"),
          verbose = FALSE
        )
    )
    expect_no_error(
      create_gitstats() |>
        set_gitlab_host(
          orgs = "mbtests",
          verbose = FALSE
        )
    )
  }
})

test_that("get_release_logs_from_hosts works as expected", {
  test_gitstats_priv <- create_test_gitstats(hosts = 2, priv_mode = TRUE)
  mockery::stub(
    test_gitstats_priv$get_release_logs_from_hosts,
    "host$get_release_logs",
    test_mocker$use("releases_table")
  )
  release_logs_table <- test_gitstats_priv$get_release_logs_from_hosts(
    since = "2023-08-01",
    until = "2023-09-30",
    verbose = FALSE
  )
  expect_releases_table(release_logs_table)
  test_mocker$cache(release_logs_table)
})

test_that("get_release_logs works as expected", {
  mockery::stub(
    test_gitstats$get_release_logs,
    "private$get_release_logs_from_hosts",
    test_mocker$use("release_logs_table")
  )
  release_logs_table <- test_gitstats$get_release_logs(
    since = "2023-08-01",
    until = "2023-09-30",
    verbose = FALSE
  )
  expect_releases_table(release_logs_table)
  test_mocker$cache(release_logs_table)
})

test_that("get_issues gets data from storage for the second time", {
  expect_snapshot(
    release_logs_table <- test_gitstats$get_release_logs(
      since = "2023-08-01",
      until = "2023-09-30",
      verbose = FALSE
    )
  )  
  expect_releases_table(release_logs_table)
})

test_that("get_release_logs returns warning when releases table is empty", {
  mockery::stub(
    test_gitstats$get_release_logs,
    "private$get_release_logs_from_hosts",
    data.frame()
  )
  expect_snapshot(
    release_logs_data <- test_gitstats$get_release_logs(
      since = "2023-08-01",
      until = "2023-09-30",
      cache = FALSE,
      verbose = TRUE
    )
  )
  expect_equal(nrow(release_logs_data), 0)
})

test_that("get_release_logs works as expected", {
  mockery::stub(
    get_release_logs,
    "gitstats$get_release_logs",
    test_mocker$use("release_logs_table")
  )
  release_logs_table <- get_release_logs(
    gitstats = test_gitstats,
    since = "2023-08-01",
    until = "2023-09-30",
    verbose = FALSE
  )
  expect_releases_table(release_logs_table)
})

test_that("get_release_logs prints time used to pull data", {
  mockery::stub(
    get_release_logs,
    "gitstats$get_release_logs",
    test_mocker$use("release_logs_table")
  )
  expect_snapshot(
    release_logs_table <- get_release_logs(
      gitstats = test_gitstats,
      since = "2023-08-01",
      until = "2023-09-30",
      verbose = TRUE
    )
  )
})

test_that("get_repos_from_host_with_code works", {
  mockery::stub(
    test_gitstats_priv$get_repos_from_host_with_code,
    "host$get_repos",
    test_mocker$use("gh_repos_table_full")
  )
  repos_from_host_with_code <- test_gitstats_priv$get_repos_from_host_with_code(
    add_contributors = TRUE,
    add_languages = TRUE,
    with_code = "test_code",
    in_files = NULL,
    verbose = FALSE,
    progress = FALSE
  )
  expect_repos_table(
    repos_from_host_with_code,
    repo_cols = repo_host_colnames,
    with_cols = c("api_url", "githost", "contributors")
  )
  test_mocker$cache(repos_from_host_with_code)
})

test_that("get_repos_from_host_with_files works", {
  mockery::stub(
    test_gitstats_priv$get_repos_from_host_with_files,
    "host$get_repos",
    test_mocker$use("gl_repos_table_full")
  )
  repos_from_host_with_files <- test_gitstats_priv$get_repos_from_host_with_files(
    add_contributors = TRUE,
    add_languages = TRUE,
    with_files = "test_file",
    verbose = FALSE,
    progress = FALSE
  )
  expect_repos_table(
    repos_from_host_with_files,
    repo_cols = repo_host_colnames,
    with_cols = c("api_url", "githost", "contributors")
  )
  test_mocker$cache(repos_from_host_with_files)
})

test_that("get_repos_from_hosts works", {
  mockery::stub(
    test_gitstats_priv$get_repos_from_hosts,
    "private$get_repos_from_host_with_code",
    test_mocker$use("repos_from_host_with_code")
  )
  repos_from_hosts_with_code <- test_gitstats_priv$get_repos_from_hosts(
    add_contributors = TRUE,
    add_languages = TRUE,
    with_code = "test_code",
    in_files = NULL,
    with_files = NULL,
    verbose = FALSE,
    progress = FALSE
  )
  expect_repos_table(
    repos_from_hosts_with_code,
    repo_cols = repo_gitstats_colnames,
    with_cols = c("contributors", "contributors_n")
  )
  mockery::stub(
    test_gitstats_priv$get_repos_from_hosts,
    "private$get_repos_from_host_with_files",
    test_mocker$use("repos_from_host_with_files")
  )
  repos_from_hosts_with_files <- test_gitstats_priv$get_repos_from_hosts(
    add_contributors = TRUE,
    add_languages = TRUE,
    with_code = NULL,
    in_files = NULL,
    with_files = "test_file",
    verbose = FALSE,
    progress = FALSE
  )
  expect_repos_table(
    repos_from_hosts_with_files,
    repo_cols = repo_gitstats_colnames,
    with_cols = c("contributors", "contributors_n")
  )
  mockery::stub(
    test_gitstats_priv$get_repos_from_hosts,
    "host$get_repos",
    purrr::list_rbind(list(
      test_mocker$use("gh_repos_table_full"),
      test_mocker$use("gl_repos_table_with_api_url")
    ))
  )
  repos_from_hosts <- test_gitstats_priv$get_repos_from_hosts(
    add_contributors = TRUE,
    add_languages = TRUE,
    with_code = NULL,
    in_files = NULL,
    with_files = NULL,
    verbose = FALSE,
    progress = FALSE
  )
  expect_repos_table(
    repos_from_hosts,
    repo_cols = repo_gitstats_colnames,
    with_cols = c("contributors", "contributors_n")
  )
  test_mocker$cache(repos_from_hosts)
})

test_that("set_object_class for repos_table works correctly", {
  repos_table <- test_gitstats_priv$set_object_class(
    object = test_mocker$use("repos_from_hosts"),
    class = "repos_table",
    attr_list = list(
      "with_code" = NULL,
      "in_files" = NULL,
      "with_files" = "renv.lock"
    )
  )
  expect_s3_class(repos_table, "repos_table")
  expect_equal(attr(repos_table, "with_files"), "renv.lock")
  test_mocker$cache(repos_table)
})

test_that("get_repos works properly and for the second time uses cache", {
  mockery::stub(
    test_gitstats$get_repos,
    "private$get_repos_from_hosts",
    test_mocker$use("repos_table")
  )
  repos_data <- test_gitstats$get_repos(verbose = FALSE)
  expect_repos_table_object(
    repos_object = repos_data,
    with_cols = c("contributors", "contributors_n")
  )
  repos_data <- test_gitstats$get_repos(
    verbose = FALSE
  )
  expect_repos_table_object(
    repos_object = repos_data,
    with_cols = c("contributors", "contributors_n")
  )
  test_mocker$cache(repos_data)
})

test_that("get_repos returns warning when repositories table is empty", {
  mockery::stub(
    test_gitstats$get_repos,
    "private$get_repos_from_hosts",
    data.frame()
  )
  expect_snapshot(
    repos_data <- test_gitstats$get_repos(
      cache = FALSE,
      verbose = TRUE
    )
  )
  expect_equal(nrow(repos_data), 0)
})

test_that("get_repos works", {
  mockery::stub(
    get_repos,
    "gitstats$get_repos",
    test_mocker$use("repos_data")
  )
  repos_data <- get_repos(
    gitstats = test_gitstats,
    verbose = FALSE
  )
  expect_repos_table_object(
    repos_object = repos_data,
    with_cols = c("contributors", "contributors_n")
  )
})

test_that("get_repos prints time used to pull data", {
  mockery::stub(
    get_repos,
    "gitstats$get_repos",
    test_mocker$use("repos_data")
  )
  expect_snapshot(
    repos_data <- get_repos(
      gitstats = test_gitstats,
      verbose = TRUE
    )
  )
})

test_that("GitStats pulls repos files trees from hosts", {
  mockery::stub(
    test_gitstats_priv$get_repos_trees_from_hosts,
    "host$get_files_structure",
    test_mocker$use("gh_files_structure_from_orgs")
  )
  repos_trees_from_hosts <- test_gitstats_priv$get_repos_trees_from_hosts(
    pattern = NULL,
    depth = Inf,
    verbose = FALSE,
    progress = FALSE
  )
  expect_s3_class(repos_trees_from_hosts, "tbl")
  expect_named(repos_trees_from_hosts,
               c("repo_id", "repo_name", "organization", "files_tree", "githost"))
  test_mocker$cache(repos_trees_from_hosts)
})

test_that("GitStats pulls repos trees", {
  mockery::stub(
    test_gitstats$get_repos_trees,
    "private$get_repos_trees_from_hosts",
    test_mocker$use("repos_trees_from_hosts")
  )
  repos_trees <- test_gitstats$get_repos_trees(
    pattern = NULL,
    depth = Inf,
    cache = FALSE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_s3_class(repos_trees, "gitstats_repos_trees")
  test_mocker$cache(repos_trees)
})

test_that("GitStats makes use of stored data", {
  expect_snapshot(
    repos_trees <- test_gitstats$get_repos_trees(
      pattern = NULL,
      depth = Inf,
      cache = TRUE,
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("GitStats prints warning when no repos trees found", {
  mockery::stub(
    test_gitstats$get_repos_trees,
    "private$get_repos_trees_from_hosts",
    data.frame()
  )
  expect_snapshot(
    repos_trees <- test_gitstats$get_repos_trees(
      pattern = NULL,
      depth = Inf,
      cache = FALSE,
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("get_repos_trees pulls repos trees", {
  mockery::stub(
    get_repos_trees,
    "gitstats$get_repos_trees",
    test_mocker$use("repos_trees")
  )
  repos_trees <- get_repos_trees(
    gitstats = test_gitstats,
    pattern = NULL,
    depth = Inf,
    cache = TRUE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_s3_class(repos_trees, "gitstats_repos_trees")
})

test_that("get_files_from_hosts works properly", {
  mockery::stub(
    test_gitstats_priv$get_files_from_hosts,
    "host$get_files_content",
    purrr::list_rbind(
      list(
        test_mocker$use("gh_files_table"),
        test_mocker$use("gl_files_table")
      )
    )
  )
  files_table <- test_gitstats_priv$get_files_from_hosts(
    pattern = NULL,
    depth = Inf,
    file_path = "meta_data.yaml",
    verbose = FALSE,
    progress = FALSE
  )
  expect_files_table(
    files_table,
    with_cols = "api_url"
  )
  test_mocker$cache(files_table)
})

test_that("get_files_from_hosts works properly", {
  mockery::stub(
    test_gitstats_priv$get_files_from_hosts,
    "host$get_files_structure",
    test_mocker$use("gh_files_structure_from_orgs")
  )
  mockery::stub(
    test_gitstats_priv$get_files_from_hosts,
    "host$get_files_content",
    purrr::list_rbind(
      list(
        test_mocker$use("gh_files_table"),
        test_mocker$use("gl_files_table")
      )
    )
  )
  files_table <- test_gitstats_priv$get_files_from_hosts(
    pattern = "\\.md",
    depth = Inf,
    file_path = NULL,
    verbose = FALSE,
    progress = FALSE
  )
  expect_files_table(
    files_table,
    with_cols = "api_url"
  )
  test_mocker$cache(files_table)
})

test_that("get_files works properly", {
  mockery::stub(
    test_gitstats$get_files,
    "private$get_files_from_hosts",
    test_mocker$use("files_table")
  )
  files_table <- test_gitstats$get_files(
    pattern = NULL,
    depth = Inf,
    file_path = "meta_data.yaml",
    verbose = FALSE
  )
  expect_files_table(
    files_table,
    with_cols = "api_url"
  )
  test_mocker$cache(files_table)
})

test_that("get_files gets data from storage for the second time", {
  expect_snapshot(
    files_table <- test_gitstats$get_files(
      pattern = NULL,
      depth = Inf,
      file_path = "meta_data.yaml",
      verbose = FALSE
    )
  )
  expect_files_table(
    files_table,
    with_cols = "api_url"
  )
})

test_that("get_files prints message when files structure is empty", {
  skip_if(integration_tests_skipped)
  expect_snapshot(
    files_table <- test_gitstats$get_files(
      pattern = "\\.test",
      depth = 0L,
      verbose = FALSE,
      progress = FALSE
    )
  )
})

test_that("get_files function works", {
  mockery::stub(
    get_files,
    "gitstats$get_files",
    test_mocker$use("files_table")
  )
  files_table <- get_files(test_gitstats,
            file_path = "meta_data.yaml",
            verbose = FALSE,
            progress = FALSE)
  expect_files_table(
    files_table,
    with_cols = "api_url"
  )
})

test_that("error shows when file_path and pattern are defined at the same time", {
  expect_snapshot_error(
    get_files(test_gitstats,
              pattern = "\\.md",
              file_path = "meta_data.yaml",
              verbose = FALSE,
              progress = FALSE)
  )
})

test_that("get_files prints time used to pull data", {
  mockery::stub(
    get_files,
    "gitstats$get_files",
    test_mocker$use("files_table")
  )
  expect_snapshot(
    files_table <- get_files(
      test_gitstats,
      file_path = "meta_data.yaml",
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_gitstats <- create_test_gitstats(hosts = 0)

test_that("GitStats object is created", {
  expect_s3_class(test_gitstats, "GitStats")
})

test_that("GitStats prints empty fields.", {
  expect_snapshot(test_gitstats)
})

test_that("GitStats prints the proper info when connections are added.", {
  test_gitstats <- create_test_gitstats(hosts = 2)
  expect_snapshot(test_gitstats)
})

test_that("GitStats prints the proper info when repos are passed instead of orgs.", {
  skip_on_cran()
  suppressMessages(
    test_gitstats <- create_gitstats() |>
      set_github_host(
        token = Sys.getenv("GITHUB_PAT"),
        repos = c("r-world-devs/GitStats", "openpharma/GithubMetrics")
      ) |>
      set_gitlab_host(
        token = Sys.getenv("GITLAB_PAT_PUBLIC"),
        repos = c("mbtests/gitstatstesting", "mbtests/gitstats-testing-2")
      )
  )
  expect_snapshot(test_gitstats)
})

test_gitstats_priv <- create_test_gitstats(hosts = 0, priv_mode = TRUE)

test_that("check_for_host returns error when no hosts are passed", {
  expect_snapshot_error(
    test_gitstats_priv$check_for_host()
  )
})

test_that("check_params_conflict returns error", {
  expect_snapshot_error(
    test_gitstats_priv$check_params_conflict(
      with_code = NULL,
      with_files = NULL,
      in_files = "DESCRIPTION"
    )
  )
  expect_snapshot_error(
    test_gitstats_priv$check_params_conflict(
      with_code = "shiny",
      with_files = "DESCRIPTION",
      in_files = NULL
    )
  )
})

test_that("check_if_args_changed", {
  test_gitstats <- create_test_gitstats(
    hosts = 2,
    priv_mode = TRUE,
    test_mocker = test_mocker,
    inject_repos = "repos_table"
  )
  check <- test_gitstats$check_if_args_changed(
    storage = "repositories",
    args_list = list(
      "with_code" = "shiny",
      "in_files" = "DESCRIPTION",
      "with_files" = NULL
    )
  )
  # repos_table from mock should be set to with_files = 'renv.lock'
  expect_true(
    check
  )
  check <- test_gitstats$check_if_args_changed(
    storage = "repositories",
    args_list = list(
      "with_code" = NULL,
      "in_files" = NULL,
      "with_files" = "renv.lock"
    )
  )
  expect_false(
    check
  )
})

test_that("print_storage_attribute", {
  expect_snapshot(
    test_gitstats_priv$print_storage_attribute(
      storage_data = test_mocker$use("commits_table"),
      storage_name = "commits"
    )
  )
  expect_snapshot(
    test_gitstats_priv$print_storage_attribute(
      storage_data = test_mocker$use("release_logs_table"),
      storage_name = "release_logs"
    )
  )
  expect_snapshot(
    test_gitstats_priv$print_storage_attribute(
      storage_data = test_mocker$use("repos_trees"),
      storage_name = "repos_trees"
    )
  )
})

test_that("show_orgs print orgs properly", {
  test_gitstats <- create_test_gitstats(hosts = 2)
  expect_equal(
    test_gitstats$show_orgs(),
    c("github_test_org", "gitlab_test_group")
  )
})

test_that("show_orgs print subgroups properly", {
  skip_on_cran()
  suppressMessages(
    test_gitstats <- create_gitstats() |>
      set_gitlab_host(
        token = Sys.getenv("GITLAB_PAT_PUBLIC"),
        orgs = "mbtests/subgroup"
      )
  )
  expect_equal(
    test_gitstats$show_orgs(),
    "mbtests/subgroup"
  )
})

test_that("subgroups are cleanly printed in GitStats", {
  skip_on_cran()
  suppressMessages(
    test_gitstats <- create_gitstats() |>
      set_gitlab_host(
        token = Sys.getenv("GITLAB_PAT_PUBLIC"),
        orgs = "mbtests/subgroup"
      )
  )
  expect_snapshot(
    test_gitstats
  )
})

# ---- verbose_on / verbose_off ----

test_that("verbose is TRUE by default", {
  gs <- create_gitstats()
  expect_true(is_verbose(gs))
})

test_that("verbose_off sets verbose to FALSE", {
  gs <- create_gitstats()
  verbose_off(gs)
  expect_false(is_verbose(gs))
})

test_that("verbose_on sets verbose back to TRUE", {
  gs <- create_gitstats()
  verbose_off(gs)
  expect_false(is_verbose(gs))
  verbose_on(gs)
  expect_true(is_verbose(gs))
})

test_that("verbose_on returns self invisibly for piping", {
  gs <- create_gitstats()
  expect_invisible(verbose_on(gs))
  result <- verbose_on(gs)
  expect_identical(result, gs)
})

test_that("verbose_off returns self invisibly for piping", {
  gs <- create_gitstats()
  expect_invisible(verbose_off(gs))
  result <- verbose_off(gs)
  expect_identical(result, gs)
})

test_that("verbose_off and verbose_on work via R6 methods directly", {
  gs <- create_gitstats()
  gs$verbose_off()
  expect_false(gs$is_verbose())
  gs$verbose_on()
  expect_true(gs$is_verbose())
})

test_that("get_issues_from_hosts works", {
  mockery::stub(
    test_gitstats_priv$get_issues_from_hosts,
    "host$get_issues",
    purrr::list_rbind(list(
      test_mocker$use("gh_issues_table"),
      test_mocker$use("gl_issues_table")
    ))
  )
  issues_table_from_hosts <- test_gitstats_priv$get_issues_from_hosts(
    since = "2023-01-01",
    until = "2025-03-06",
    state = NULL,
    verbose = FALSE,
    progress = FALSE
  )
  expect_issues_table(
    issues_table_from_hosts
  )
  test_mocker$cache(issues_table_from_hosts)
})

test_that("set_object_class works for issues", {
  issues_table <- test_gitstats_priv$set_object_class(
    object = test_mocker$use("issues_table_from_hosts"),
    class = "issues_data",
    attr_list = list(
      "date_range" = c("2023-01-01", "2025-03-06")
    )
  )
  expect_s3_class(
    issues_table,
    "issues_data"
  )
  expect_equal(
    attr(issues_table, "date_range"),
    c("2023-01-01", "2025-03-06")
  )
})

test_that("get_issues works properly", {
  mockery::stub(
    test_gitstats$get_issues,
    "private$get_issues_from_hosts",
    test_mocker$use("issues_table_from_hosts")
  )
  issues_table <- test_gitstats$get_issues(
    since = "2023-01-01",
    until = "2025-03-06",
    state = NULL,
    verbose = FALSE
  )
  expect_issues_table(issues_table)
  test_mocker$cache(issues_table)
})

test_that("get_issues gets data from storage for the second time", {
  expect_snapshot(
    issues_table <- test_gitstats$get_issues(
      since = "2023-01-01",
      until = "2025-03-06",
      state = NULL,
      verbose = FALSE
    )
  )  
  expect_issues_table(issues_table)
})

test_that("get_issues() returns warning if issues table is empty", {
  mockery::stub(
    test_gitstats$get_issues,
    "private$get_issues_from_hosts",
    data.frame()
  )
  expect_snapshot(
    issues_table <- test_gitstats$get_issues(
      since = "2023-01-01",
      until = "2025-03-06",
      cache = FALSE,
      state = NULL,
      verbose = TRUE
    )
  )
  expect_equal(nrow(issues_table), 0)
})

test_that("get_issues() works", {
  mockery::stub(
    get_issues,
    "gitstats$get_issues",
    test_mocker$use("issues_table")
  )
  issues_data <- get_issues(
    test_gitstats,
    since = "2023-01-01",
    until = "2025-03-06",
    verbose = FALSE
  )
  expect_s3_class(
    issues_data,
    "gitstats_issues"
  )
  test_mocker$cache(issues_data)
})

test_that("get_issues() prints data on time used", {
  mockery::stub(
    get_issues,
    "gitstats$get_issues",
    test_mocker$use("issues_table")
  )
  expect_snapshot(
    issues_data <- get_issues(
      test_gitstats,
      since = "2023-01-01",
      until = "2025-03-06",
      verbose = TRUE
    )
  )
})

#' @title Set local (in-memory) storage
#' @name set_local_storage
#' @description Reset storage to the default in-memory backend.
#' @param gitstats A GitStats object.
#' @return A `GitStats` object (invisibly).
#' @examples
#' \dontrun{
#'   my_gitstats <- create_gitstats() |>
#'     set_local_storage()
#' }
#' @export
set_local_storage <- function(gitstats) {
  gitstats$set_local_storage()
}

#' @title Set PostgreSQL storage
#' @name set_postgres_storage
#' @description Persist GitStats data in a PostgreSQL database. R classes,
#'   custom attributes, and column types are preserved via a `_metadata` table.
#'   Requires `DBI`, `RPostgres`, and `jsonlite` packages.
#' @param gitstats A GitStats object.
#' @param host A character, database host.
#' @param port An integer, database port.
#' @param dbname A character, database name.
#' @param user A character, database user.
#' @param password A character, database password.
#' @param schema A character, database schema (default `"git_stats"`).
#' @param ... Additional arguments passed to
#'   `DBI::dbConnect(RPostgres::Postgres(), ...)`.
#' @return A `GitStats` object (invisibly).
#' @examples
#' \dontrun{
#'   my_gitstats <- create_gitstats() |>
#'     set_postgres_storage(
#'       dbname = "my_database",
#'       host = "localhost",
#'       user = "postgres",
#'       password = "secret"
#'     )
#' }
#' @export
set_postgres_storage <- function(gitstats,
                                 host = NULL,
                                 port = NULL,
                                 dbname = NULL,
                                 user = NULL,
                                 password = NULL,
                                 schema = "git_stats",
                                 ...) {
  gitstats$set_postgres_storage(
    host = host,
    port = port,
    dbname = dbname,
    user = user,
    password = password,
    schema = schema,
    ...
  )
}

#' @title Set SQLite storage
#' @name set_sqlite_storage
#' @description Persist GitStats data in a SQLite database. R classes,
#'   custom attributes, and column types are preserved via a `_metadata` table.
#'   Requires `DBI`, `RSQLite`, and `jsonlite` packages.
#' @param gitstats A GitStats object.
#' @param dbname A character, path to SQLite file. Defaults to `":memory:"`
#'   for an in-memory database.
#' @return A `GitStats` object (invisibly).
#' @examples
#' \dontrun{
#'   # File-based
#'   my_gitstats <- create_gitstats() |>
#'     set_sqlite_storage(dbname = "gitstats.sqlite")
#'
#'   # In-memory
#'   my_gitstats <- create_gitstats() |>
#'     set_sqlite_storage()
#' }
#' @export
set_sqlite_storage <- function(gitstats, dbname = ":memory:") {
  gitstats$set_sqlite_storage(dbname = dbname)
}

#' @title Get data from `GitStats` storage
#' @name get_storage
#' @description Retrieves whole or particular data (see `storage` parameter)
#'   pulled earlier with `GitStats`.
#' @param gitstats A GitStats object.
#' @param storage A character, type of the data you want to get from storage:
#'   `commits`, `repositories`, `release_logs`, `users`, `files`,
#'   `files_structure`, `R_package_usage` or `release_logs`.
#' @return A list of tibbles (if `storage` set to `NULL`) or a tibble (if
#'   `storage` defined).
#' @examples
#' \dontrun{
#'  my_gitstats <- create_gitstats() |>
#'   set_github_host(
#'     token = Sys.getenv("GITHUB_PAT"),
#'     orgs = c("r-world-devs", "openpharma")
#'   )
#'   get_release_logs(my_gistats, since = "2024-01-01")
#'   get_repos(my_gitstats)
#'
#'   release_logs <- get_storage(
#'     gitstats = my_gitstats,
#'     storage = "release_logs"
#'   )
#' }
#' @export
get_storage <- function(gitstats,
                        storage = NULL) {
  gitstats$get_storage(
    storage = storage
  )
}

#' @title Remove a table from `GitStats` storage
#' @name remove_from_storage
#' @description Removes a named table from the active storage backend.
#' @param gitstats A GitStats object.
#' @param storage A character, name of the table to remove (e.g. `"commits"`,
#'   `"repositories"`).
#' @return A `GitStats` object (invisibly).
#' @examples
#' \dontrun{
#'   my_gitstats <- create_gitstats() |>
#'     set_github_host(
#'       token = Sys.getenv("GITHUB_PAT"),
#'       orgs = "r-world-devs"
#'     )
#'   get_commits(my_gitstats, since = "2024-01-01")
#'   remove_from_storage(my_gitstats, storage = "commits")
#' }
#' @export
remove_from_storage <- function(gitstats, storage) {
  gitstats$remove_from_storage(storage = storage)
}

#' @title Remove PostgreSQL storage
#' @name remove_postgres_storage
#' @description Drops the GitStats schema (with all tables and metadata) from
#'   the PostgreSQL database, closes the connection, and reverts storage to the
#'   default in-memory local backend. Errors if no PostgreSQL backend is
#'   currently set.
#' @param gitstats A GitStats object.
#' @return A `GitStats` object (invisibly).
#' @examples
#' \dontrun{
#'   my_gitstats <- create_gitstats() |>
#'     set_postgres_storage(
#'       dbname = "my_database",
#'       host = "localhost",
#'       user = "postgres",
#'       password = "secret"
#'     )
#'   remove_postgres_storage(my_gitstats)
#' }
#' @export
remove_postgres_storage <- function(gitstats) {
  gitstats$remove_postgres_storage()
}

#' @title Remove SQLite storage
#' @name remove_sqlite_storage
#' @description Closes the SQLite connection and, for file-based databases,
#'   deletes the database file. For in-memory databases the data is simply
#'   discarded. Storage is reverted to the default in-memory local backend.
#'   Errors if no SQLite backend is currently set.
#' @param gitstats A GitStats object.
#' @return A `GitStats` object (invisibly).
#' @examples
#' \dontrun{
#'   my_gitstats <- create_gitstats() |>
#'     set_sqlite_storage(dbname = "gitstats.sqlite")
#'   remove_sqlite_storage(my_gitstats)
#' }
#' @export
remove_sqlite_storage <- function(gitstats) {
  gitstats$remove_sqlite_storage()
}

#' @title Get metadata for a storage table
#' @name get_storage_metadata
#' @description Retrieves metadata (R classes, custom attributes, column types)
#'   for a table stored in the active storage backend.
#' @param gitstats A GitStats object.
#' @param storage A character, name of the table (e.g. `"commits"`,
#'   `"repositories"`). If `NULL` (default), metadata for all tables in storage
#'   will be returned.
#' @return A list with metadata fields: `class`, `attributes`, and (for
#'   database backends) `column_types`.
#' @examples
#' \dontrun{
#'   my_gitstats <- create_gitstats() |>
#'     set_github_host(
#'       token = Sys.getenv("GITHUB_PAT"),
#'       orgs = "r-world-devs"
#'     )
#'   get_commits(my_gitstats, since = "2024-01-01")
#'   get_storage_metadata(my_gitstats, storage = "commits")
#' }
#' @export
get_storage_metadata <- function(gitstats, storage = NULL) {
  gitstats$get_storage_metadata(storage = storage)
}

#' @noRd
Storage <- R6::R6Class(
  classname = "Storage",
  public = list(
    save = function(name, data) {
      stop("Not implemented")
    },
    load = function(name) {
      stop("Not implemented")
    },
    exists = function(name) {
      stop("Not implemented")
    },
    remove = function(name) {
      stop("Not implemented")
    },
    drop_storage = function() {
      stop("Not implemented")
    },
    list = function() {
      stop("Not implemented")
    },
    get_metadata = function(name = NULL) {
      stop("Not implemented")
    },
    is_db = function() {
      FALSE
    }
  )
)

#' @noRd
StorageLocal <- R6::R6Class(
  classname = "StorageLocal",
  inherit = Storage,
  public = list(
    initialize = function() {
      private$data <- list()
    },
    save = function(name, data) {
      private$data[[name]] <- data
    },
    load = function(name) {
      private$data[[name]]
    },
    exists = function(name) {
      !is.null(private$data[[name]])
    },
    remove = function(name) {
      private$data[[name]] <- NULL
    },
    drop_storage = function() {
      private$data <- list()
    },
    list = function() {
      names(purrr::discard(private$data, is.null))
    },
    get_metadata = function(name = NULL) {
      if (is.null(name)) {
        names_list <- self$list()
        if (length(names_list) == 0) {
          return(private$empty_metadata_tibble())
        }
        build_row <- private$build_metadata_row
        rows <- purrr::map(names_list, function(n) build_row(n))
        return(dplyr::bind_rows(rows))
      }
      data <- private$data[[name]]
      if (is.null(data)) {
        return(private$empty_metadata_tibble())
      }
      private$build_metadata_row(name)
    }
  ),
  private = list(
    data = list(),
    build_metadata_row = function(name) {
      data <- private$data[[name]]
      custom_attrs <- setdiff(
        names(attributes(data)),
        c("names", "row.names", "class")
      )
      attrs <- list()
      for (attr_name in custom_attrs) {
        attrs[[attr_name]] <- attr(data, attr_name)
      }
      dplyr::tibble(
        table_name = name,
        class = list(class(data)),
        attributes = list(attrs)
      )
    },
    empty_metadata_tibble = function() {
      dplyr::tibble(
        table_name = character(),
        class = list(),
        attributes = list()
      )
    }
  )
)

#' @noRd
StoragePostgres <- R6::R6Class(
  classname = "StoragePostgres",
  inherit = Storage,
  public = list(
    initialize = function(schema = "git_stats", ...) {
      check_if_package_installed("DBI")
      check_if_package_installed("RPostgres")
      check_if_package_installed("jsonlite")
      private$conn <- DBI::dbConnect(RPostgres::Postgres(), ...)
      if (!DBI::dbIsValid(private$conn)) {
        cli::cli_abort("Failed to connect to the database.")
      }
      private$schema <- schema
      private$ensure_schema()
      private$ensure_metadata_table()
    },
    save = function(name, data) {
      qualified <- private$qualified_name(name)
      meta <- list(
        class = class(data),
        attributes = list()
      )
      custom_attrs <- setdiff(
        names(attributes(data)),
        c("names", "row.names", "class")
      )
      for (attr_name in custom_attrs) {
        meta$attributes[[attr_name]] <- attr(data, attr_name)
      }
      plain_data <- data
      class(plain_data) <- class(plain_data)[!grepl("^gitstats_", class(plain_data))]
      serialized <- private$serialize_columns(plain_data)
      plain_data <- serialized$data
      meta$column_types <- serialized$column_types
      DBI::dbWriteTable(
        private$conn,
        DBI::Id(schema = private$schema, table = name),
        plain_data,
        overwrite = TRUE
      )
      private$save_metadata(name, meta)
    },
    load = function(name) {
      if (!self$exists(name)) {
        return(NULL)
      }
      data <- DBI::dbReadTable(
        private$conn,
        DBI::Id(schema = private$schema, table = name)
      ) |>
        dplyr::as_tibble()
      meta <- private$load_metadata(name)
      if (!is.null(meta)) {
        if (!is.null(meta$column_types)) {
          data <- private$restore_columns(data, meta$column_types)
        }
        if (!is.null(meta$class)) {
          class(data) <- meta$class
        }
        if (!is.null(meta$attributes)) {
          for (attr_name in names(meta$attributes)) {
            attr(data, attr_name) <- meta$attributes[[attr_name]]
          }
        }
      }
      return(data)
    },
    exists = function(name) {
      DBI::dbExistsTable(
        private$conn,
        DBI::Id(schema = private$schema, table = name)
      )
    },
    remove = function(name) {
      if (self$exists(name)) {
        DBI::dbRemoveTable(
          private$conn,
          DBI::Id(schema = private$schema, table = name)
        )
        DBI::dbExecute(
          private$conn,
          glue::glue(
            "DELETE FROM {private$schema}._metadata
             WHERE dataset_name = $1"
          ),
          params = list(name)
        )
      }
      invisible(NULL)
    },
    drop_storage = function() {
      DBI::dbExecute(
        private$conn,
        glue::glue("DROP SCHEMA IF EXISTS {private$schema} CASCADE")
      )
      DBI::dbDisconnect(private$conn)
      private$conn <- NULL
    },
    list = function() {
      result <- DBI::dbGetQuery(
        private$conn,
        glue::glue(
          "SELECT table_name FROM information_schema.tables
           WHERE table_schema = $1 AND table_name != '_metadata'"
        ),
        params = list(private$schema)
      )
      result$table_name
    },
    get_metadata = function(name = NULL) {
      if (is.null(name)) {
        names_list <- self$list()
        if (length(names_list) == 0) {
          return(private$empty_metadata_tibble())
        }
        build_row <- private$build_metadata_row
        rows <- purrr::map(names_list, function(n) build_row(n))
        return(dplyr::bind_rows(rows))
      }
      if (!self$exists(name)) {
        return(private$empty_metadata_tibble())
      }
      private$build_metadata_row(name)
    },
    is_db = function() {
      TRUE
    }
  ),
  private = list(
    conn = NULL,
    schema = NULL,
    finalize = function() {
      if (!is.null(private$conn) && DBI::dbIsValid(private$conn)) {
        DBI::dbDisconnect(private$conn)
      }
    },
    ensure_schema = function() {
      DBI::dbExecute(
        private$conn,
        glue::glue("CREATE SCHEMA IF NOT EXISTS {private$schema}")
      )
    },
    ensure_metadata_table = function() {
      meta_table <- DBI::Id(schema = private$schema, table = "_metadata")
      if (!DBI::dbExistsTable(private$conn, meta_table)) {
        DBI::dbExecute(
          private$conn,
          glue::glue(
            "CREATE TABLE {private$schema}._metadata (
              dataset_name TEXT PRIMARY KEY,
              metadata JSONB NOT NULL
            )"
          )
        )
      }
    },
    qualified_name = function(name) {
      paste0(private$schema, ".", name)
    },
    save_metadata = function(name, meta) {
      meta_json <- jsonlite::toJSON(meta, auto_unbox = TRUE)
      DBI::dbExecute(
        private$conn,
        glue::glue(
          "INSERT INTO {private$schema}._metadata (dataset_name, metadata)
           VALUES ($1, $2)
           ON CONFLICT (dataset_name) DO UPDATE SET metadata = $2"
        ),
        params = list(name, as.character(meta_json))
      )
    },
    load_metadata = function(name) {
      result <- DBI::dbGetQuery(
        private$conn,
        glue::glue(
          "SELECT metadata FROM {private$schema}._metadata
           WHERE dataset_name = $1"
        ),
        params = list(name)
      )
      if (nrow(result) == 0) {
        return(NULL)
      }
      jsonlite::fromJSON(result$metadata[[1]])
    },
    serialize_columns = function(data) {
      column_types <- list()
      for (col in names(data)) {
        if (inherits(data[[col]], "difftime")) {
          column_types[[col]] <- list(
            type = "difftime",
            units = attr(data[[col]], "units")
          )
          data[[col]] <- as.numeric(data[[col]])
        } else if (inherits(data[[col]], "POSIXct")) {
          column_types[[col]] <- list(type = "POSIXct")
          data[[col]] <- format(data[[col]], "%Y-%m-%dT%H:%M:%S")
        }
      }
      list(data = data, column_types = column_types)
    },
    restore_columns = function(data, column_types) {
      for (col in names(column_types)) {
        if (!(col %in% names(data))) next
        col_type <- column_types[[col]]
        type <- if (is.list(col_type)) col_type$type else col_type
        if (type == "difftime") {
          units <- if (is.list(col_type)) col_type$units else "days"
          data[[col]] <- as.difftime(
            as.numeric(data[[col]]),
            units = units
          )
        } else if (type == "POSIXct") {
          data[[col]] <- as.POSIXct(data[[col]], format = "%Y-%m-%dT%H:%M:%S")
        }
      }
      data
    },
    build_metadata_row = function(name) {
      meta <- private$load_metadata(name)
      if (is.null(meta)) {
        return(dplyr::tibble(
          table_name = name,
          class = list(NULL),
          attributes = list(list())
        ))
      }
      dplyr::tibble(
        table_name = name,
        class = list(meta$class),
        attributes = list(meta$attributes %||% list())
      )
    },
    empty_metadata_tibble = function() {
      dplyr::tibble(
        table_name = character(),
        class = list(),
        attributes = list()
      )
    }
  )
)

#' @noRd
StorageSQLite <- R6::R6Class(
  classname = "StorageSQLite",
  inherit = Storage,
  public = list(
    initialize = function(dbname = ":memory:") {
      check_if_package_installed("DBI")
      check_if_package_installed("RSQLite")
      check_if_package_installed("jsonlite")
      private$dbname <- dbname
      private$conn <- DBI::dbConnect(RSQLite::SQLite(), dbname = dbname)
      private$ensure_metadata_table()
    },
    save = function(name, data) {
      meta <- list(
        class = class(data),
        attributes = list()
      )
      custom_attrs <- setdiff(
        names(attributes(data)),
        c("names", "row.names", "class")
      )
      for (attr_name in custom_attrs) {
        meta$attributes[[attr_name]] <- attr(data, attr_name)
      }
      plain_data <- data
      class(plain_data) <- class(plain_data)[!grepl("^gitstats_", class(plain_data))]
      serialized <- private$serialize_columns(plain_data)
      plain_data <- serialized$data
      meta$column_types <- serialized$column_types
      DBI::dbWriteTable(private$conn, name, plain_data, overwrite = TRUE)
      private$save_metadata(name, meta)
    },
    load = function(name) {
      if (!self$exists(name)) {
        return(NULL)
      }
      data <- DBI::dbReadTable(private$conn, name) |>
        dplyr::as_tibble()
      meta <- private$load_metadata(name)
      if (!is.null(meta)) {
        if (!is.null(meta$column_types)) {
          data <- private$restore_columns(data, meta$column_types)
        }
        if (!is.null(meta$class)) {
          class(data) <- meta$class
        }
        if (!is.null(meta$attributes)) {
          for (attr_name in names(meta$attributes)) {
            attr(data, attr_name) <- meta$attributes[[attr_name]]
          }
        }
      }
      return(data)
    },
    exists = function(name) {
      DBI::dbExistsTable(private$conn, name)
    },
    remove = function(name) {
      if (self$exists(name)) {
        DBI::dbRemoveTable(private$conn, name)
        DBI::dbExecute(
          private$conn,
          "DELETE FROM _metadata WHERE dataset_name = ?",
          params = list(name)
        )
      }
      invisible(NULL)
    },
    drop_storage = function() {
      if (!is.null(private$conn) && DBI::dbIsValid(private$conn)) {
        DBI::dbDisconnect(private$conn)
        private$conn <- NULL
      }
      if (!is.null(private$dbname) && private$dbname != ":memory:" &&
            file.exists(private$dbname)) {
        file.remove(private$dbname)
      }
    },
    list = function() {
      tables <- DBI::dbListTables(private$conn)
      tables[tables != "_metadata"]
    },
    get_metadata = function(name = NULL) {
      if (is.null(name)) {
        names_list <- self$list()
        if (length(names_list) == 0) {
          return(private$empty_metadata_tibble())
        }
        build_row <- private$build_metadata_row
        rows <- purrr::map(names_list, function(n) build_row(n))
        return(dplyr::bind_rows(rows))
      }
      if (!self$exists(name)) {
        return(private$empty_metadata_tibble())
      }
      private$build_metadata_row(name)
    },
    is_db = function() {
      TRUE
    }
  ),
  private = list(
    conn = NULL,
    dbname = NULL,
    finalize = function() {
      if (!is.null(private$conn) && DBI::dbIsValid(private$conn)) {
        DBI::dbDisconnect(private$conn)
      }
    },
    ensure_metadata_table = function() {
      if (!DBI::dbExistsTable(private$conn, "_metadata")) {
        DBI::dbExecute(
          private$conn,
          "CREATE TABLE _metadata (
            dataset_name TEXT PRIMARY KEY,
            metadata TEXT NOT NULL
          )"
        )
      }
    },
    save_metadata = function(name, meta) {
      meta_json <- jsonlite::toJSON(meta, auto_unbox = TRUE)
      DBI::dbExecute(
        private$conn,
        "DELETE FROM _metadata WHERE dataset_name = ?",
        params = list(name)
      )
      DBI::dbExecute(
        private$conn,
        "INSERT INTO _metadata (dataset_name, metadata) VALUES (?, ?)",
        params = list(name, as.character(meta_json))
      )
    },
    load_metadata = function(name) {
      result <- DBI::dbGetQuery(
        private$conn,
        "SELECT metadata FROM _metadata WHERE dataset_name = ?",
        params = list(name)
      )
      if (nrow(result) == 0) {
        return(NULL)
      }
      jsonlite::fromJSON(result$metadata[[1]])
    },
    serialize_columns = function(data) {
      column_types <- list()
      for (col in names(data)) {
        if (inherits(data[[col]], "difftime")) {
          column_types[[col]] <- list(
            type = "difftime",
            units = attr(data[[col]], "units")
          )
          data[[col]] <- as.numeric(data[[col]])
        } else if (inherits(data[[col]], "POSIXct")) {
          column_types[[col]] <- list(type = "POSIXct")
          data[[col]] <- format(data[[col]], "%Y-%m-%dT%H:%M:%S")
        }
      }
      list(data = data, column_types = column_types)
    },
    restore_columns = function(data, column_types) {
      for (col in names(column_types)) {
        if (!(col %in% names(data))) next
        col_type <- column_types[[col]]
        type <- if (is.list(col_type)) col_type$type else col_type
        if (type == "difftime") {
          units <- if (is.list(col_type)) col_type$units else "days"
          data[[col]] <- as.difftime(
            as.numeric(data[[col]]),
            units = units
          )
        } else if (type == "POSIXct") {
          data[[col]] <- as.POSIXct(data[[col]], format = "%Y-%m-%dT%H:%M:%S")
        }
      }
      data
    },
    build_metadata_row = function(name) {
      meta <- private$load_metadata(name)
      if (is.null(meta)) {
        return(dplyr::tibble(
          table_name = name,
          class = list(NULL),
          attributes = list(list())
        ))
      }
      dplyr::tibble(
        table_name = name,
        class = list(meta$class),
        attributes = list(meta$attributes %||% list())
      )
    },
    empty_metadata_tibble = function() {
      dplyr::tibble(
        table_name = character(),
        class = list(),
        attributes = list()
      )
    }
  )
)

# Storage base class -----------------------------------------------------------

test_that("Storage base class methods raise 'Not implemented'", {
  base <- Storage$new()
  expect_error(base$save("x", 1), "Not implemented")
  expect_error(base$load("x"), "Not implemented")
  expect_error(base$exists("x"), "Not implemented")
  expect_error(base$remove("x"), "Not implemented")
  expect_error(base$list(), "Not implemented")
  expect_error(base$get_metadata("x"), "Not implemented")
  expect_false(base$is_db())
})

# StorageLocal -----------------------------------------------------------------

test_that("StorageLocal save and load round-trips data", {
  storage <- StorageLocal$new()
  df <- dplyr::tibble(a = 1:3, b = letters[1:3])
  storage$save("test_data", df)
  loaded <- storage$load("test_data")
  expect_equal(loaded, df)
})

test_that("StorageLocal preserves class and attributes", {
  storage <- StorageLocal$new()
  df <- dplyr::tibble(a = 1:3)
  class(df) <- c("gitstats_repos", class(df))
  attr(df, "since") <- "2024-01-01"
  attr(df, "until") <- "2024-12-31"
  storage$save("repos", df)
  loaded <- storage$load("repos")
  expect_s3_class(loaded, "gitstats_repos")
  expect_equal(attr(loaded, "since"), "2024-01-01")
  expect_equal(attr(loaded, "until"), "2024-12-31")
})

test_that("StorageLocal exists returns TRUE/FALSE correctly", {
  storage <- StorageLocal$new()
  expect_false(storage$exists("missing"))
  storage$save("present", dplyr::tibble(x = 1))
  expect_true(storage$exists("present"))
})

test_that("StorageLocal list returns saved names", {
  storage <- StorageLocal$new()
  expect_null(storage$list())
  storage$save("alpha", dplyr::tibble(x = 1))
  storage$save("beta", dplyr::tibble(y = 2))
  expect_setequal(storage$list(), c("alpha", "beta"))
})

test_that("StorageLocal load returns NULL for missing data", {
  storage <- StorageLocal$new()
  expect_null(storage$load("nonexistent"))
})

test_that("StorageLocal overwrite replaces data", {
  storage <- StorageLocal$new()
  storage$save("data", dplyr::tibble(x = 1))
  storage$save("data", dplyr::tibble(x = 99))
  expect_equal(storage$load("data")$x, 99)
})

test_that("StorageLocal is_db returns FALSE", {
  storage <- StorageLocal$new()
  expect_false(storage$is_db())
})

test_that("StorageLocal remove deletes a table", {
  storage <- StorageLocal$new()
  storage$save("alpha", dplyr::tibble(x = 1))
  storage$save("beta", dplyr::tibble(y = 2))
  expect_true(storage$exists("alpha"))
  storage$remove("alpha")
  expect_false(storage$exists("alpha"))
  expect_true(storage$exists("beta"))
  expect_equal(storage$list(), "beta")
})

test_that("StorageLocal remove is silent for missing table", {
  storage <- StorageLocal$new()
  expect_no_error(storage$remove("nonexistent"))
})

test_that("StorageLocal get_metadata returns tibble with class and attributes", {
  storage <- StorageLocal$new()
  df <- dplyr::tibble(a = 1:3)
  class(df) <- c("gitstats_commits", class(df))
  attr(df, "since") <- "2024-01-01"
  attr(df, "until") <- "2024-12-31"
  storage$save("commits", df)
  meta <- storage$get_metadata("commits")
  expect_s3_class(meta, "tbl_df")
  expect_equal(nrow(meta), 1)
  expect_equal(meta$table_name, "commits")
  expect_in("gitstats_commits", meta$class[[1]])
  expect_equal(meta$attributes[[1]]$since, "2024-01-01")
  expect_equal(meta$attributes[[1]]$until, "2024-12-31")
})

test_that("StorageLocal get_metadata returns empty tibble for missing table", {
  storage <- StorageLocal$new()
  meta <- storage$get_metadata("nonexistent")
  expect_s3_class(meta, "tbl_df")
  expect_equal(nrow(meta), 0)
})

test_that("StorageLocal get_metadata returns empty attributes when none set", {
  storage <- StorageLocal$new()
  df <- dplyr::tibble(x = 1)
  storage$save("data", df)
  meta <- storage$get_metadata("data")
  expect_s3_class(meta, "tbl_df")
  expect_equal(meta$attributes[[1]], list())
})

test_that("StorageLocal get_metadata with NULL returns all tables", {
  storage <- StorageLocal$new()
  df1 <- dplyr::tibble(a = 1)
  class(df1) <- c("gitstats_commits", class(df1))
  attr(df1, "since") <- "2024-01-01"
  df2 <- dplyr::tibble(b = 2)
  class(df2) <- c("gitstats_repos", class(df2))
  storage$save("commits", df1)
  storage$save("repos", df2)
  meta <- storage$get_metadata()
  expect_s3_class(meta, "tbl_df")
  expect_equal(nrow(meta), 2)
  expect_in("commits", meta$table_name)
  expect_in("repos", meta$table_name)
})

test_that("StorageLocal get_metadata with NULL on empty storage returns empty tibble", {
  storage <- StorageLocal$new()
  meta <- storage$get_metadata()
  expect_s3_class(meta, "tbl_df")
  expect_equal(nrow(meta), 0)
})

# set_*_storage ----------------------------------------------------------------

test_that("default storage is StorageLocal", {
  gs <- create_gitstats()
  backend <- gs$.__enclos_env__$private$storage_backend
  expect_s3_class(backend, "StorageLocal")
  expect_false(backend$is_db())
})

test_that("set_local_storage resets to local", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  expect_true(gs$.__enclos_env__$private$storage_backend$is_db())
  gs$set_local_storage()
  expect_false(gs$.__enclos_env__$private$storage_backend$is_db())
})

test_that("set_local_storage returns self invisibly for piping", {
  gs <- create_gitstats()
  result <- gs$set_local_storage()
  expect_identical(result, gs)
})

test_that("set_sqlite_storage returns self invisibly for piping", {
  gs <- create_gitstats()
  result <- gs$set_sqlite_storage()
  expect_identical(result, gs)
})

test_that("print shows storage backend type for local", {
  gs <- create_gitstats()
  output <- capture.output(print(gs))
  expect_true(any(grepl("Storage \\[local\\]", output)))
})

test_that("print shows storage backend type for SQLite", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  output <- capture.output(print(gs))
  expect_true(any(grepl("Storage \\[SQLite\\]", output)))
})

test_that("print lists stored data on separate lines", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  priv <- gs$.__enclos_env__$private
  repos <- dplyr::tibble(repo = c("a/b", "c/d"), stars = c(1, 2))
  class(repos) <- c("gitstats_repos", class(repos))
  priv$storage_backend$save("repositories", repos)
  commits <- dplyr::tibble(sha = "abc123")
  class(commits) <- c("gitstats_commits", class(commits))
  attr(commits, "date_range") <- c("2024-01-01", "2024-06-30")
  priv$storage_backend$save("commits", commits)
  output <- capture.output(print(gs))
  storage_lines <- output[grepl("Storage|Repositories|Commits", output)]
  expect_true(any(grepl("Storage \\[SQLite\\]", storage_lines)))
  expect_true(any(grepl("Repositories: 2", storage_lines)))
  expect_true(any(grepl("Commits: 1", storage_lines)))
  storage_label_line <- grep("Storage \\[SQLite\\]", output)
  expect_false(grepl("Repositories|Commits", output[storage_label_line]))
})

# StorageSQLite ----------------------------------------------------------------

test_that("StorageSQLite save and load round-trips data", {
  storage <- StorageSQLite$new()
  df <- dplyr::tibble(a = 1:3, b = c("x", "y", "z"))
  storage$save("test_data", df)
  loaded <- storage$load("test_data")
  expect_equal(loaded$a, df$a)
  expect_equal(loaded$b, df$b)
})

test_that("StorageSQLite preserves class via metadata", {
  storage <- StorageSQLite$new()
  df <- dplyr::tibble(a = 1:3)
  class(df) <- c("gitstats_commits", class(df))
  storage$save("commits", df)
  loaded <- storage$load("commits")
  expect_s3_class(loaded, "gitstats_commits")
})

test_that("StorageSQLite preserves custom attributes via metadata", {
  storage <- StorageSQLite$new()
  df <- dplyr::tibble(a = 1:3)
  class(df) <- c("gitstats_repos", class(df))
  attr(df, "since") <- "2024-01-01"
  attr(df, "until") <- "2024-12-31"
  storage$save("repos", df)
  loaded <- storage$load("repos")
  expect_equal(attr(loaded, "since"), "2024-01-01")
  expect_equal(attr(loaded, "until"), "2024-12-31")
})

test_that("StorageSQLite preserves difftime columns", {
  storage <- StorageSQLite$new()
  df <- dplyr::tibble(
    repo = "test/repo",
    last_activity = as.difftime(c(38.5, 12.3), units = "days")
  )
  storage$save("repos", df)
  loaded <- storage$load("repos")
  expect_s3_class(loaded$last_activity, "difftime")
  expect_equal(attr(loaded$last_activity, "units"), "days")
  expect_equal(as.numeric(loaded$last_activity), c(38.5, 12.3))
})

test_that("StorageSQLite preserves POSIXct columns", {
  storage <- StorageSQLite$new()
  timestamps <- as.POSIXct(c("2024-01-15 10:30:00", "2024-06-20 14:00:00"))
  df <- dplyr::tibble(
    repo = "test/repo",
    created_at = timestamps,
    last_activity_at = timestamps
  )
  storage$save("repos", df)
  loaded <- storage$load("repos")
  expect_s3_class(loaded$created_at, "POSIXct")
  expect_s3_class(loaded$last_activity_at, "POSIXct")
  expect_equal(
    format(loaded$created_at, "%Y-%m-%d %H:%M:%S"),
    format(timestamps, "%Y-%m-%d %H:%M:%S")
  )
})

test_that("StorageSQLite preserves mixed difftime and POSIXct columns", {
  storage <- StorageSQLite$new()
  df <- dplyr::tibble(
    repo = "test/repo",
    created_at = as.POSIXct("2024-01-15 10:30:00"),
    last_activity_at = as.POSIXct("2024-06-20 14:00:00"),
    last_activity = as.difftime(38.5, units = "days")
  )
  class(df) <- c("gitstats_repos", class(df))
  storage$save("repos", df)
  loaded <- storage$load("repos")
  expect_s3_class(loaded, "gitstats_repos")
  expect_s3_class(loaded$created_at, "POSIXct")
  expect_s3_class(loaded$last_activity_at, "POSIXct")
  expect_s3_class(loaded$last_activity, "difftime")
  expect_equal(as.numeric(loaded$last_activity), 38.5)
})

test_that("StorageSQLite exists returns TRUE/FALSE correctly", {
  storage <- StorageSQLite$new()
  expect_false(storage$exists("missing"))
  storage$save("present", dplyr::tibble(x = 1))
  expect_true(storage$exists("present"))
})

test_that("StorageSQLite list excludes _metadata table", {
  storage <- StorageSQLite$new()
  storage$save("alpha", dplyr::tibble(x = 1))
  storage$save("beta", dplyr::tibble(y = 2))
  tables <- storage$list()
  expect_in("alpha", tables)
  expect_in("beta", tables)
  expect_false("_metadata" %in% tables)
})

test_that("StorageSQLite load returns NULL for missing data", {
  storage <- StorageSQLite$new()
  expect_null(storage$load("nonexistent"))
})

test_that("StorageSQLite overwrite replaces data", {
  storage <- StorageSQLite$new()
  storage$save("data", dplyr::tibble(x = 1))
  storage$save("data", dplyr::tibble(x = 99))
  expect_equal(storage$load("data")$x, 99)
})

test_that("StorageSQLite is_db returns TRUE", {
  storage <- StorageSQLite$new()
  expect_true(storage$is_db())
})

test_that("StorageSQLite remove deletes table and metadata", {
  storage <- StorageSQLite$new()
  df <- dplyr::tibble(a = 1:3)
  class(df) <- c("gitstats_commits", class(df))
  storage$save("commits", df)
  storage$save("repos", dplyr::tibble(x = 1))
  expect_true(storage$exists("commits"))
  storage$remove("commits")
  expect_false(storage$exists("commits"))
  expect_true(storage$exists("repos"))
  expect_equal(storage$list(), "repos")
  # metadata row should also be gone
  meta <- storage$get_metadata("commits")
  expect_equal(nrow(meta), 0)
})

test_that("StorageSQLite remove is silent for missing table", {
  storage <- StorageSQLite$new()
  expect_no_error(storage$remove("nonexistent"))
})

test_that("StorageSQLite get_metadata returns tibble with class and attributes", {
  storage <- StorageSQLite$new()
  df <- dplyr::tibble(a = 1:3)
  class(df) <- c("gitstats_repos", class(df))
  attr(df, "since") <- "2024-01-01"
  storage$save("repos", df)
  meta <- storage$get_metadata("repos")
  expect_s3_class(meta, "tbl_df")
  expect_equal(nrow(meta), 1)
  expect_equal(meta$table_name, "repos")
  expect_in("gitstats_repos", meta$class[[1]])
  expect_equal(meta$attributes[[1]]$since, "2024-01-01")
})

test_that("StorageSQLite get_metadata returns empty tibble for missing table", {
  storage <- StorageSQLite$new()
  meta <- storage$get_metadata("nonexistent")
  expect_s3_class(meta, "tbl_df")
  expect_equal(nrow(meta), 0)
})

test_that("StorageSQLite get_metadata with NULL returns all tables", {
  storage <- StorageSQLite$new()
  df1 <- dplyr::tibble(a = 1)
  class(df1) <- c("gitstats_commits", class(df1))
  df2 <- dplyr::tibble(b = 2)
  class(df2) <- c("gitstats_repos", class(df2))
  storage$save("commits", df1)
  storage$save("repos", df2)
  meta <- storage$get_metadata()
  expect_s3_class(meta, "tbl_df")
  expect_equal(nrow(meta), 2)
  expect_in("commits", meta$table_name)
  expect_in("repos", meta$table_name)
})

test_that("StorageSQLite finalize disconnects connection", {
  storage <- StorageSQLite$new()
  conn <- storage$.__enclos_env__$private$conn
  expect_true(DBI::dbIsValid(conn))
  storage$.__enclos_env__$private$finalize()
  expect_false(DBI::dbIsValid(conn))
})

test_that("StorageSQLite finalize handles NULL connection", {
  storage <- StorageSQLite$new()
  conn <- storage$.__enclos_env__$private$conn
  DBI::dbDisconnect(conn)
  storage$.__enclos_env__$private$conn <- NULL
  expect_no_error(storage$.__enclos_env__$private$finalize())
})

test_that("StorageSQLite load works when metadata is missing", {
  storage <- StorageSQLite$new()
  conn <- storage$.__enclos_env__$private$conn
  DBI::dbWriteTable(conn, "raw_table", data.frame(x = 1:3))
  loaded <- storage$load("raw_table")
  expect_equal(nrow(loaded), 3)
  expect_s3_class(loaded, "tbl_df")
})

test_that("StorageSQLite works with file-based database", {
  tmp <- tempfile(fileext = ".sqlite")
  on.exit(unlink(tmp), add = TRUE)
  storage <- StorageSQLite$new(dbname = tmp)
  storage$save("data", dplyr::tibble(val = 42))
  loaded <- storage$load("data")
  expect_equal(loaded$val, 42)
})

test_that("set_sqlite_storage creates SQLite backend with dbname", {
  tmp <- tempfile(fileext = ".sqlite")
  on.exit(unlink(tmp), add = TRUE)
  gs <- create_gitstats()
  gs$set_sqlite_storage(dbname = tmp)
  backend <- gs$.__enclos_env__$private$storage_backend
  expect_s3_class(backend, "StorageSQLite")
  expect_true(backend$is_db())
})

# GitStats + SQLite integration ------------------------------------------------

test_that("GitStats save_to_storage and get_from_storage work with SQLite", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  priv <- gs$.__enclos_env__$private
  df <- dplyr::tibble(repo = "test/repo", stars = 10)
  priv$storage_backend$save("repositories", df)
  expect_false(priv$storage_is_empty("repositories"))
  loaded <- priv$storage_backend$load("repositories")
  expect_equal(loaded$repo, "test/repo")
})

test_that("GitStats get_storage returns all data from SQLite", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  priv <- gs$.__enclos_env__$private
  priv$storage_backend$save("repos", dplyr::tibble(x = 1))
  priv$storage_backend$save("commits", dplyr::tibble(y = 2))
  all_data <- gs$get_storage(NULL)
  expect_in("repos", names(all_data))
  expect_in("commits", names(all_data))
})

test_that("GitStats get_storage returns single table from SQLite", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  priv <- gs$.__enclos_env__$private
  priv$storage_backend$save("files", dplyr::tibble(path = "R/main.R"))
  result <- gs$get_storage("files")
  expect_equal(result$path, "R/main.R")
})

# GitStats remove_from_storage -------------------------------------------------

test_that("GitStats remove_from_storage removes a table", {
  gs <- create_gitstats()
  priv <- gs$.__enclos_env__$private
  priv$storage_backend$save("commits", dplyr::tibble(sha = "abc"))
  expect_true(priv$storage_backend$exists("commits"))
  suppressMessages(gs$remove_from_storage("commits"))
  expect_false(priv$storage_backend$exists("commits"))
})

test_that("GitStats remove_from_storage errors for missing table", {
  gs <- create_gitstats()
  expect_error(
    gs$remove_from_storage("nonexistent"),
    "nonexistent"
  )
})

test_that("GitStats remove_from_storage returns self invisibly", {
  gs <- create_gitstats()
  priv <- gs$.__enclos_env__$private
  priv$storage_backend$save("repos", dplyr::tibble(x = 1))
  result <- suppressMessages(gs$remove_from_storage("repos"))
  expect_identical(result, gs)
})

test_that("remove_from_storage() exported wrapper works", {
  gs <- create_gitstats()
  priv <- gs$.__enclos_env__$private
  priv$storage_backend$save("commits", dplyr::tibble(sha = "abc"))
  suppressMessages(remove_from_storage(gs, storage = "commits"))
  expect_false(priv$storage_backend$exists("commits"))
})

# GitStats get_storage_metadata ------------------------------------------------

test_that("GitStats get_storage_metadata returns tibble for single table", {
  gs <- create_gitstats()
  priv <- gs$.__enclos_env__$private
  df <- dplyr::tibble(a = 1:3)
  class(df) <- c("gitstats_commits", class(df))
  attr(df, "since") <- "2024-01-01"
  priv$storage_backend$save("commits", df)
  meta <- gs$get_storage_metadata("commits")
  expect_s3_class(meta, "tbl_df")
  expect_equal(nrow(meta), 1)
  expect_equal(meta$table_name, "commits")
  expect_in("gitstats_commits", meta$class[[1]])
  expect_equal(meta$attributes[[1]]$since, "2024-01-01")
})

test_that("GitStats get_storage_metadata errors for missing table", {
  gs <- create_gitstats()
  expect_error(
    gs$get_storage_metadata("nonexistent"),
    "nonexistent"
  )
})

test_that("GitStats get_storage_metadata with NULL returns all tables", {
  gs <- create_gitstats()
  priv <- gs$.__enclos_env__$private
  df1 <- dplyr::tibble(a = 1)
  class(df1) <- c("gitstats_commits", class(df1))
  df2 <- dplyr::tibble(b = 2)
  class(df2) <- c("gitstats_repos", class(df2))
  priv$storage_backend$save("commits", df1)
  priv$storage_backend$save("repos", df2)
  meta <- gs$get_storage_metadata()
  expect_s3_class(meta, "tbl_df")
  expect_equal(nrow(meta), 2)
  expect_in("commits", meta$table_name)
  expect_in("repos", meta$table_name)
})

test_that("get_storage_metadata() exported wrapper works", {
  gs <- create_gitstats()
  priv <- gs$.__enclos_env__$private
  df <- dplyr::tibble(x = 1)
  class(df) <- c("gitstats_repos", class(df))
  priv$storage_backend$save("repos", df)
  meta <- get_storage_metadata(gs, storage = "repos")
  expect_s3_class(meta, "tbl_df")
  expect_in("gitstats_repos", meta$class[[1]])
})

test_that("get_storage_metadata() exported wrapper works with NULL", {
  gs <- create_gitstats()
  priv <- gs$.__enclos_env__$private
  priv$storage_backend$save("commits", dplyr::tibble(sha = "abc"))
  meta <- get_storage_metadata(gs)
  expect_s3_class(meta, "tbl_df")
  expect_equal(nrow(meta), 1)
})

test_that("GitStats get_storage_metadata works with SQLite backend", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  priv <- gs$.__enclos_env__$private
  df <- dplyr::tibble(a = 1:3)
  class(df) <- c("gitstats_commits", class(df))
  attr(df, "since") <- "2024-01-01"
  priv$storage_backend$save("commits", df)
  meta <- gs$get_storage_metadata("commits")
  expect_s3_class(meta, "tbl_df")
  expect_equal(meta$table_name, "commits")
  expect_in("gitstats_commits", meta$class[[1]])
  expect_equal(meta$attributes[[1]]$since, "2024-01-01")
})

test_that("GitStats get_storage_metadata with NULL works with SQLite backend", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  priv <- gs$.__enclos_env__$private
  priv$storage_backend$save("commits", dplyr::tibble(sha = "abc"))
  priv$storage_backend$save("repos", dplyr::tibble(x = 1))
  meta <- gs$get_storage_metadata()
  expect_s3_class(meta, "tbl_df")
  expect_equal(nrow(meta), 2)
})

test_that("GitStats remove_from_storage works with SQLite backend", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  priv <- gs$.__enclos_env__$private
  priv$storage_backend$save("commits", dplyr::tibble(sha = "abc"))
  expect_true(priv$storage_backend$exists("commits"))
  suppressMessages(gs$remove_from_storage("commits"))
  expect_false(priv$storage_backend$exists("commits"))
})

# check_if_package_installed ---------------------------------------------------

test_that("check_if_package_installed errors for missing package", {
  expect_error(
    check_if_package_installed("nonexistent_pkg_12345"),
    "nonexistent_pkg_12345"
  )
})

test_that("check_if_package_installed passes for installed package", {
  expect_no_error(check_if_package_installed("testthat"))
})

# Exported wrapper functions ----------------------------------------------------

test_that("set_local_storage() wrapper calls R6 method", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  result <- set_local_storage(gs)
  backend <- gs$.__enclos_env__$private$storage_backend
  expect_s3_class(backend, "StorageLocal")
})

test_that("set_sqlite_storage() wrapper calls R6 method", {
  gs <- create_gitstats()
  result <- set_sqlite_storage(gs)
  backend <- gs$.__enclos_env__$private$storage_backend
  expect_s3_class(backend, "StorageSQLite")
})

test_that("set_postgres_storage() R6 method executes with stubbed connection", {
  gs <- create_gitstats()
  mock_storage <- StorageLocal$new()
  mockery::stub(
    gs$set_postgres_storage,
    "do.call",
    mock_storage
  )
  gs$set_postgres_storage(
    host = "localhost",
    port = 5432,
    dbname = "test",
    user = "postgres",
    password = "secret",
    schema = "my_schema"
  )
  backend <- gs$.__enclos_env__$private$storage_backend
  expect_s3_class(backend, "StorageLocal")
})

test_that("set_postgres_storage() exported wrapper calls R6 method", {
  gs <- create_gitstats()
  mock_storage <- StorageLocal$new()
  mockery::stub(
    set_postgres_storage,
    "gitstats$set_postgres_storage",
    mock_storage
  )
  set_postgres_storage(
    gs,
    host = "localhost",
    port = 5432,
    dbname = "test",
    user = "postgres",
    password = "secret"
  )
  backend <- gs$.__enclos_env__$private$storage_backend
  expect_s3_class(backend, "StorageLocal")
})

test_that("print shows PostgreSQL backend type", {
  gs <- create_gitstats()
  mock_backend <- StorageLocal$new()
  class(mock_backend) <- c("StoragePostgres", class(mock_backend))
  gs$.__enclos_env__$private$storage_backend <- mock_backend
  output <- capture.output(print(gs))
  expect_true(any(grepl("Storage \\[PostgreSQL\\]", output)))
})

# Storage propagation to hosts -------------------------------------------------

test_that("set_sqlite_storage propagates backend to hosts", {
  gs <- create_test_gitstats(hosts = 1)
  suppressMessages(gs$set_sqlite_storage())
  host <- gs$.__enclos_env__$private$hosts[[1]]
  host_storage <- host$.__enclos_env__$private$storage_backend
  expect_s3_class(host_storage, "StorageSQLite")
})

test_that("set_local_storage propagates backend to hosts", {
  gs <- create_test_gitstats(hosts = 1)
  suppressMessages(gs$set_sqlite_storage())
  gs$set_local_storage()
  host <- gs$.__enclos_env__$private$hosts[[1]]
  host_storage <- host$.__enclos_env__$private$storage_backend
  expect_s3_class(host_storage, "StorageLocal")
})

test_that("add_new_host propagates current storage to new host", {
  gs <- create_gitstats()
  suppressMessages(gs$set_sqlite_storage())
  gs$.__enclos_env__$private$hosts[[1]] <- create_github_testhost(
    orgs = "test_org"
  )
  gs$.__enclos_env__$private$propagate_storage_to_hosts()
  host <- gs$.__enclos_env__$private$hosts[[1]]
  host_storage <- host$.__enclos_env__$private$storage_backend
  expect_s3_class(host_storage, "StorageSQLite")
})

# report_storage_contents ------------------------------------------------------

test_that("set_sqlite_storage reports empty database", {
  gs <- create_gitstats()
  output <- capture.output(
    gs$set_sqlite_storage(),
    type = "message"
  )
  expect_true(any(grepl("Database is empty", output)))
})

test_that("set_sqlite_storage reports existing data", {
  tmp <- tempfile(fileext = ".sqlite")
  on.exit(unlink(tmp), add = TRUE)
  storage <- StorageSQLite$new(dbname = tmp)
  df <- dplyr::tibble(repo = "test/repo", stars = 10L)
  class(df) <- c("gitstats_repos", class(df))
  storage$save("repositories", df)
  commits <- dplyr::tibble(id = "abc123")
  class(commits) <- c("gitstats_commits", class(commits))
  storage$save("commits", commits)
  rm(storage)
  gc()

  gs <- create_gitstats()
  output <- capture.output(
    gs$set_sqlite_storage(dbname = tmp),
    type = "message"
  )
  expect_true(any(grepl("Database contains data", output)))
  expect_true(any(grepl("repositories.*1", output)))
  expect_true(any(grepl("commits.*1", output)))
})

# GitHost set_storage_backend --------------------------------------------------

test_that("GitHost set_storage_backend sets storage", {
  host <- create_github_testhost(orgs = "test_org")
  storage <- StorageSQLite$new()
  host$set_storage_backend(storage)
  host_storage <- host$.__enclos_env__$private$storage_backend
  expect_s3_class(host_storage, "StorageSQLite")
})

# save_repos_to_storage --------------------------------------------------------

test_that("GitHub save_repos_to_storage saves repos to database", {
  host <- create_github_testhost(orgs = "test_org")
  storage <- StorageSQLite$new()
  host$set_storage_backend(storage)
  priv <- host$.__enclos_env__$private

  mock_repos_table <- dplyr::tibble(
    repo_id = "123",
    repo_name = "test-repo",
    default_branch = "main",
    organization = "test_org"
  )
  mock_engine <- list(
    prepare_repos_table = function(repos_from_org, org) mock_repos_table
  )

  priv$save_repos_to_storage(list(), "test_org", mock_engine)

  loaded <- storage$load("repositories")
  expect_equal(nrow(loaded), 1)
  expect_equal(loaded$repo_id, "123")
  expect_s3_class(loaded, "gitstats_repos")
})

test_that("GitHub save_repos_to_storage merges with existing repos", {
  host <- create_github_testhost(orgs = "test_org")
  storage <- StorageSQLite$new()
  host$set_storage_backend(storage)
  priv <- host$.__enclos_env__$private

  existing <- dplyr::tibble(
    repo_id = "100",
    repo_name = "old-repo",
    default_branch = "main",
    organization = "old_org"
  )
  class(existing) <- c("gitstats_repos", class(existing))
  storage$save("repositories", existing)

  new_repos <- dplyr::tibble(
    repo_id = "200",
    repo_name = "new-repo",
    default_branch = "main",
    organization = "test_org"
  )
  mock_engine <- list(
    prepare_repos_table = function(repos_from_org, org) new_repos
  )

  priv$save_repos_to_storage(list(), "test_org", mock_engine)

  loaded <- storage$load("repositories")
  expect_equal(nrow(loaded), 2)
  expect_in("100", loaded$repo_id)
  expect_in("200", loaded$repo_id)
})

test_that("GitHub save_repos_to_storage deduplicates by repo_id", {
  host <- create_github_testhost(orgs = "test_org")
  storage <- StorageSQLite$new()
  host$set_storage_backend(storage)
  priv <- host$.__enclos_env__$private

  existing <- dplyr::tibble(
    repo_id = "123",
    repo_name = "test-repo",
    default_branch = "main",
    organization = "test_org"
  )
  class(existing) <- c("gitstats_repos", class(existing))
  storage$save("repositories", existing)

  duplicate_repos <- dplyr::tibble(
    repo_id = "123",
    repo_name = "test-repo-updated",
    default_branch = "develop",
    organization = "test_org"
  )
  mock_engine <- list(
    prepare_repos_table = function(repos_from_org, org) duplicate_repos
  )

  priv$save_repos_to_storage(list(), "test_org", mock_engine)

  loaded <- storage$load("repositories")
  expect_equal(nrow(loaded), 1)
})

test_that("GitHub save_repos_to_storage skips when no db storage", {
  host <- create_github_testhost(orgs = "test_org")
  priv <- host$.__enclos_env__$private
  # storage_backend is NULL by default
  expect_no_error(
    priv$save_repos_to_storage(list(), "test_org", list())
  )
})

test_that("GitHub save_repos_to_storage skips when storage is local", {
  host <- create_github_testhost(orgs = "test_org")
  storage <- StorageLocal$new()
  host$set_storage_backend(storage)
  priv <- host$.__enclos_env__$private
  mock_engine <- list(
    prepare_repos_table = function(repos_from_org, org) {
      dplyr::tibble(repo_id = "1", repo_name = "r")
    }
  )
  priv$save_repos_to_storage(list(), "test_org", mock_engine)
  expect_null(storage$load("repositories"))
})

test_that("GitLab save_repos_to_storage saves repos to database", {
  host <- create_gitlab_testhost(orgs = "test_group")
  storage <- StorageSQLite$new()
  host$set_storage_backend(storage)
  priv <- host$.__enclos_env__$private

  mock_repos_table <- dplyr::tibble(
    repo_id = "456",
    repo_name = "gl-repo",
    default_branch = "main",
    organization = "test_group"
  )
  mock_engine <- list(
    prepare_repos_table = function(repos_from_org, org) mock_repos_table
  )

  priv$save_repos_to_storage(list(), "test_group", mock_engine)

  loaded <- storage$load("repositories")
  expect_equal(nrow(loaded), 1)
  expect_equal(loaded$repo_id, "456")
  expect_s3_class(loaded, "gitstats_repos")
})

# add_new_host propagates storage ----------------------------------------------

test_that("add_new_host propagates storage backend to host", {
  gs <- create_gitstats()
  suppressMessages(gs$set_sqlite_storage())
  priv <- gs$.__enclos_env__$private
  new_host <- create_github_testhost(orgs = "test_org")
  priv$add_new_host(new_host)
  host_storage <- priv$hosts[[1]]$.__enclos_env__$private$storage_backend
  expect_s3_class(host_storage, "StorageSQLite")
})

# Storage drop_storage ---------------------------------------------------------

test_that("Storage base class drop_storage raises 'Not implemented'", {
  base <- Storage$new()
  expect_error(base$drop_storage(), "Not implemented")
})

test_that("StorageLocal drop_storage clears all data", {
  storage <- StorageLocal$new()
  storage$save("alpha", dplyr::tibble(x = 1))
  storage$save("beta", dplyr::tibble(y = 2))
  expect_length(storage$list(), 2)
  storage$drop_storage()
  expect_null(storage$list())
})

test_that("StorageSQLite drop_storage removes file-based database", {
  tmp <- tempfile(fileext = ".sqlite")
  storage <- StorageSQLite$new(dbname = tmp)
  storage$save("commits", dplyr::tibble(sha = "abc"))
  expect_true(file.exists(tmp))
  storage$drop_storage()
  expect_false(file.exists(tmp))
})

test_that("StorageSQLite drop_storage works for in-memory database", {
  storage <- StorageSQLite$new()
  storage$save("commits", dplyr::tibble(sha = "abc"))
  expect_no_error(storage$drop_storage())
})

test_that("StorageSQLite stores dbname in private field", {
  tmp <- tempfile(fileext = ".sqlite")
  on.exit(unlink(tmp), add = TRUE)
  storage <- StorageSQLite$new(dbname = tmp)
  expect_equal(storage$.__enclos_env__$private$dbname, tmp)
})

# GitStats remove_sqlite_storage -----------------------------------------------

test_that("remove_sqlite_storage removes file-based SQLite and reverts to local", {
  tmp <- tempfile(fileext = ".sqlite")
  gs <- create_gitstats()
  gs$set_sqlite_storage(dbname = tmp)
  priv <- gs$.__enclos_env__$private
  priv$storage_backend$save("commits", dplyr::tibble(sha = "abc"))
  expect_true(file.exists(tmp))
  suppressMessages(gs$remove_sqlite_storage())
  expect_false(file.exists(tmp))
  expect_s3_class(priv$storage_backend, "StorageLocal")
})

test_that("remove_sqlite_storage removes in-memory SQLite and reverts to local", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  priv <- gs$.__enclos_env__$private
  expect_s3_class(priv$storage_backend, "StorageSQLite")
  suppressMessages(gs$remove_sqlite_storage())
  expect_s3_class(priv$storage_backend, "StorageLocal")
})

test_that("remove_sqlite_storage errors when no SQLite backend is set", {
  gs <- create_gitstats()
  expect_error(
    gs$remove_sqlite_storage(),
    "No SQLite storage"
  )
})

test_that("remove_sqlite_storage returns self invisibly", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  result <- suppressMessages(gs$remove_sqlite_storage())
  expect_identical(result, gs)
})

test_that("remove_sqlite_storage propagates local storage to hosts", {
  gs <- create_test_gitstats(hosts = 1)
  suppressMessages(gs$set_sqlite_storage())
  host <- gs$.__enclos_env__$private$hosts[[1]]
  expect_s3_class(host$.__enclos_env__$private$storage_backend, "StorageSQLite")
  suppressMessages(gs$remove_sqlite_storage())
  expect_s3_class(host$.__enclos_env__$private$storage_backend, "StorageLocal")
})

test_that("remove_sqlite_storage() exported wrapper works", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  suppressMessages(remove_sqlite_storage(gs))
  backend <- gs$.__enclos_env__$private$storage_backend
  expect_s3_class(backend, "StorageLocal")
})

test_that("remove_sqlite_storage prints file removal message", {
  tmp <- tempfile(fileext = ".sqlite")
  gs <- create_gitstats()
  gs$set_sqlite_storage(dbname = tmp)
  output <- capture.output(
    gs$remove_sqlite_storage(),
    type = "message"
  )
  expect_true(any(grepl("removed", output, ignore.case = TRUE)))
  expect_true(any(grepl("local", output)))
})

test_that("remove_sqlite_storage prints in-memory removal message", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  output <- capture.output(
    gs$remove_sqlite_storage(),
    type = "message"
  )
  expect_true(any(grepl("In-memory", output)))
  expect_true(any(grepl("local", output)))
})

# GitStats remove_postgres_storage ---------------------------------------------

test_that("remove_postgres_storage errors when no PostgreSQL backend is set", {
  gs <- create_gitstats()
  expect_error(
    gs$remove_postgres_storage(),
    "No PostgreSQL storage"
  )
})

test_that("remove_postgres_storage errors when SQLite is set instead", {
  gs <- create_gitstats()
  gs$set_sqlite_storage()
  expect_error(
    gs$remove_postgres_storage(),
    "No PostgreSQL storage"
  )
})

test_that("remove_postgres_storage R6 method works with stubbed backend", {
  gs <- create_gitstats()
  mock_backend <- StorageLocal$new()
  class(mock_backend) <- c("StoragePostgres", class(mock_backend))
  gs$.__enclos_env__$private$storage_backend <- mock_backend
  suppressMessages(gs$remove_postgres_storage())
  backend <- gs$.__enclos_env__$private$storage_backend
  expect_s3_class(backend, "StorageLocal")
  expect_false(inherits(backend, "StoragePostgres"))
})

test_that("remove_postgres_storage() exported wrapper works with stubbed backend", {
  gs <- create_gitstats()
  mock_backend <- StorageLocal$new()
  class(mock_backend) <- c("StoragePostgres", class(mock_backend))
  gs$.__enclos_env__$private$storage_backend <- mock_backend
  suppressMessages(remove_postgres_storage(gs))
  backend <- gs$.__enclos_env__$private$storage_backend
  expect_s3_class(backend, "StorageLocal")
})

# ---- GitHub REST get_files_tree ----

test_that("get_files_tree for GitHub returns file paths from tree response", {
  github_tree_response <- list(
    sha = "abc123",
    tree = list(
      list(path = "R/main.R", type = "blob"),
      list(path = "R/utils.R", type = "blob"),
      list(path = "R", type = "tree"),
      list(path = "tests/test-main.R", type = "blob"),
      list(path = "tests", type = "tree"),
      list(path = "README.md", type = "blob")
    )
  )
  mockery::stub(
    test_rest_github$get_files_tree,
    "self$response",
    github_tree_response
  )
  result <- test_rest_github$get_files_tree(
    org = "test-org",
    repo = "test-repo",
    def_branch = "main",
    pattern = NULL,
    depth = Inf,
    verbose = FALSE
  )
  expect_type(result, "character")
  expect_length(result, 4)
  expect_in(c("R/main.R", "R/utils.R", "tests/test-main.R", "README.md"), result)
  expect_false("R" %in% result)
})

test_that("get_files_tree for GitHub filters by depth", {
  github_tree_response <- list(
    sha = "abc123",
    tree = list(
      list(path = "R/main.R", type = "blob"),
      list(path = "R/sub/deep.R", type = "blob"),
      list(path = "README.md", type = "blob")
    )
  )
  mockery::stub(
    test_rest_github$get_files_tree,
    "self$response",
    github_tree_response
  )
  result <- test_rest_github$get_files_tree(
    org = "test-org",
    repo = "test-repo",
    def_branch = "main",
    pattern = NULL,
    depth = 1,
    verbose = FALSE
  )
  expect_equal(result, "README.md", ignore_attr = TRUE)
})

test_that("get_files_tree for GitHub filters by pattern", {
  github_tree_response <- list(
    sha = "abc123",
    tree = list(
      list(path = "R/main.R", type = "blob"),
      list(path = "DESCRIPTION", type = "blob"),
      list(path = "README.md", type = "blob")
    )
  )
  mockery::stub(
    test_rest_github$get_files_tree,
    "self$response",
    github_tree_response
  )
  result <- test_rest_github$get_files_tree(
    org = "test-org",
    repo = "test-repo",
    def_branch = "main",
    pattern = "\\.R$",
    depth = Inf,
    verbose = FALSE
  )
  expect_equal(result, "R/main.R", ignore_attr = TRUE)
})

test_that("get_files_tree for GitHub returns NULL on error", {
  mockery::stub(
    test_rest_github$get_files_tree,
    "self$response",
    function(...) stop("API error")
  )
  result <- test_rest_github$get_files_tree(
    org = "test-org",
    repo = "test-repo",
    def_branch = "main",
    pattern = NULL,
    depth = Inf,
    verbose = FALSE
  )
  expect_null(result)
})

# ---- GitLab REST get_files_tree ----

test_that("get_files_tree for GitLab returns file paths from tree response", {
  gitlab_tree_response <- list(
    list(id = "def456", path = "R/main.R", type = "blob"),
    list(id = "def457", path = "R/utils.R", type = "blob"),
    list(id = "def458", path = "R", type = "tree"),
    list(id = "def459", path = "README.md", type = "blob")
  )
  mockery::stub(
    test_rest_gitlab$get_files_tree,
    "private$paginate_results",
    gitlab_tree_response
  )
  result <- test_rest_gitlab$get_files_tree(
    org = "test-group",
    repo = "test-repo",
    pattern = NULL,
    depth = Inf,
    verbose = FALSE
  )
  expect_type(result, "character")
  expect_length(result, 3)
  expect_in(c("R/main.R", "R/utils.R", "README.md"), result)
})

test_that("get_files_tree for GitLab filters by depth", {
  gitlab_tree_response <- list(
    list(id = "def456", path = "R/main.R", type = "blob"),
    list(id = "def457", path = "R/sub/deep.R", type = "blob"),
    list(id = "def458", path = "README.md", type = "blob")
  )
  mockery::stub(
    test_rest_gitlab$get_files_tree,
    "private$paginate_results",
    gitlab_tree_response
  )
  result <- test_rest_gitlab$get_files_tree(
    org = "test-group",
    repo = "test-repo",
    pattern = NULL,
    depth = 1,
    verbose = FALSE
  )
  expect_equal(result, "README.md", ignore_attr = TRUE)
})

test_that("get_files_tree for GitLab returns NULL on error", {
  mockery::stub(
    test_rest_gitlab$get_files_tree,
    "private$paginate_results",
    function(...) stop("API error")
  )
  result <- test_rest_gitlab$get_files_tree(
    org = "test-group",
    repo = "test-repo",
    pattern = NULL,
    depth = Inf,
    verbose = FALSE
  )
  expect_null(result)
})

EngineRestGitLab <- R6::R6Class(
  classname = "EngineRestGitLab",
  inherit = EngineRest,
  public = list(

    get_orgs_count = function(verbose) {
      if (verbose) {
        cli::cli_alert("[Host:GitLab][Engine:{cli::col_green('REST')}] Pulling number of all organizations {cli_icons$org}...")
      }
      orgs_response <- self$perform_request(paste0(private$endpoints$organizations, "?all_available=True"),
                                            token = private$token,
                                            verbose = verbose)
      return(orgs_response$headers$`x-total`)
    },

    get_orgs = function(orgs_count, verbose, progress = verbose) {
      if (verbose) {
        cli::cli_alert("[Host:GitLab][Engine:{cli::col_green('REST')}] Pulling organizations {cli_icons$org}...")
      }
      iterations_number <- ceiling(orgs_count / 100)
      orgs_list <- purrr::map(1:iterations_number, function(page) {
        self$response(
          endpoint = paste0(
            private$endpoints[["organizations"]],
            "?all_available=true",
            "&pagination=keyset",
            "&per_page=100&page=",
            page
          ),
          verbose = verbose
        )
      }, .progress = progress) |>
        purrr::list_flatten()
      return(orgs_list)
    },

    get_org = function(org, verbose) {
      if (verbose) {
        cli::cli_alert("[Host:GitLab][Engine:{cli::col_green('REST')}] Pulling {url_decode(org)} organization {cli_icons$org}...")
      }
      self$response(
        endpoint = paste0(private$endpoints[["organizations"]], org),
        verbose = verbose
      )
    },

    prepare_orgs_table = function(orgs_list) {
      purrr::map(orgs_list, function(org) {
        data.frame(
          name = org$name,
          description = org$description,
          path = org$path,
          url = org$web_url,
          avatar_url = org$avatar_url %||% NA_character_,
          repos_count = NA_integer_,
          members_count = NA_integer_
        )
      }) |>
        purrr::list_rbind()
    },

    get_files = function(file_paths = NULL,
                         org = NULL,
                         clean_files_content = TRUE,
                         verbose = TRUE,
                         progress = TRUE) {
      files_list <- list()
      file_paths <- url_encode(file_paths)
      files_list <- purrr::map(file_paths, function(filename) {
        files_search_result <- self$search_for_code(
          code = filename,
          in_path = TRUE,
          org = org,
          verbose = verbose
        ) |>
          purrr::keep(~ .$path == filename)
        files_content <- private$add_file_info(
          files_search_result = files_search_result,
          clean_file_content = clean_files_content,
          filename = filename,
          verbose = verbose,
          progress = progress
        )
        return(files_content)
      }, .progress = progress) |>
        purrr::list_flatten()
      return(files_list)
    },

    get_repos_from_org = function(org, repos = NULL, add_languages = TRUE, output = "full_table", verbose = FALSE) {
      owner_type <- attr(org, "type") %||% "organization"
      owner_endpoint <- if (owner_type == "organization") {
        private$endpoints[["organizations"]]
      } else {
        private$endpoints[["users"]]
      }
      repos_response <- private$paginate_results(
        endpoint = paste0(owner_endpoint,
                          url_encode(org),
                          "/projects"),
        verbose = verbose
      )
      if (!is.null(repos)) {
        repos_response <- repos_response |>
          purrr::keep(~ .$path %in% repos)
      }
      if (output == "full_table" && add_languages) {
        repos_response <- repos_response |>
          private$get_repos_languages(
            verbose = verbose,
            progress = verbose
          )
      }
      return(repos_response)
    },

    get_repos_urls = function(type, org, repos, verbose = TRUE) {
      repos_response <- self$get_repos_from_org(
        org = org,
        repos = repos,
        output = "raw",
        verbose = verbose
      )
      repos_urls <- repos_response |>
        purrr::map_vec(function(project) {
          if (type == "api") {
            project$`_links`$self
          } else {
            project$web_url
          }
        })
      return(repos_urls)
    },

    prepare_repos_table = function(repos_list, org) {
      if (length(repos_list) > 0) {
        purrr::map(repos_list, function(repo) {
          if (length(repo$languages) == 0) {
            repo_languages <- ""
          } else {
            repo_languages <- paste0(repo$languages, collapse = ", ")
          }
          data.frame(
            repo_id = as.character(repo$id),
            repo_name = repo$path,
            repo_fullpath = repo$path_with_namespace,
            default_branch = repo$default_branch %||% "",
            stars = repo$star_count,
            forks = repo$forks_count %||% NA_integer_,
            created_at = gts_to_posixt(repo$created_at),
            last_activity_at = as.POSIXct(repo$last_activity_at),
            languages = repo_languages,
            issues_open = repo$open_issues_count %||% 0,
            issues_closed = NA_integer_,
            organization = org,
            repo_url = repo$web_url,
            commit_sha = NA_character_
          )
        }) |>
          purrr::list_rbind()
      } else {
        NULL
      }
    },

    get_files_tree = function(org, repo, pattern, depth, verbose) {
      repo_path <- paste0(url_encode(org), "%2F", repo)
      endpoint <- paste0(
        private$endpoints[["projects"]],
        repo_path,
        "/repository/tree?recursive=true&per_page=100"
      )
      response <- tryCatch(
        private$paginate_results(
          endpoint = endpoint,
          joining_sign = "&",
          verbose = verbose
        ),
        error = function(e) NULL
      )
      if (is.null(response) || length(response) == 0) {
        return(NULL)
      }
      files <- purrr::keep(response, ~ .$type == "blob") |>
        purrr::map_chr(~ .$path)
      if (!is.null(depth) && depth < Inf) {
        files <- purrr::keep(files, function(path) {
          stringr::str_count(path, "/") < depth
        })
      }
      if (!is.null(pattern)) {
        files <- purrr::keep(files, function(path) {
          any(stringr::str_detect(path, pattern))
        })
      }
      if (length(files) == 0) return(NULL)
      return(files)
    },

    get_repos_contributors = function(repos_table, verbose = TRUE, progress) {
      if (nrow(repos_table) > 0) {
        repo_urls <- repos_table$api_url
        user_name <- rlang::expr(.$name)
        repos_table$contributors <- gitstats_map_chr(repo_urls, function(repo_url) {
          contributors_endpoint <- paste0(
            repo_url,
            "/repository/contributors"
          )
          contributors_vec <- tryCatch({
            private$get_contributors_from_repo(
              contributors_endpoint = contributors_endpoint,
              user_name = user_name,
              verbose = verbose
            )
          },
          error = function(e) {
            NA
          })
          return(contributors_vec)
        }, .progress = progress)
      }
      return(repos_table)
    },

    prepare_files_table = function(files_list) {
      files_table <- NULL
      if (!is.null(files_list)) {
        files_table <- purrr::map(files_list, function(file_data) {
          org_repo <- stringr::str_split_1(file_data$repo_fullname, "/")
          org <- paste0(org_repo[1:(length(org_repo) - 1)], collapse = "/")
          data.frame(
            "repo_id" = as.character(file_data$repo_id),
            "repo_name" = file_data$repo_name,
            "organization" = org,
            "file_path" = file_data$file_path,
            "file_content" = file_data$content,
            "file_size" = file_data$size,
            "repo_url" = file_data$repo_url
          )
        }) |>
          purrr::list_rbind() |>
          unique()
      }
      return(files_table)
    },

    search_for_code = function(code,
                               filename = NULL,
                               in_path = FALSE,
                               org = NULL,
                               language = NULL,
                               page_max = 1e6,
                               verbose = TRUE) {
      page <- 1
      still_more_hits <- TRUE
      full_repos_list <- list()
      search_endpoint <- private$set_search_endpoint(org, verbose)
      if (verbose) cli::cli_alert("Searching for code [{code}]...")
      query <- private$build_search_query(
        code = code,
        filename = filename,
        in_path = in_path
      )
      while (still_more_hits | page < page_max) {
        search_result <- tryCatch({
          self$response(
            endpoint = paste0(
              search_endpoint,
              query,
              "&per_page=100&page=",
              page
            ),
            verbose = verbose
          )
        }, error = function(e) {
          if (length(full_repos_list) == 1e4) {
            if (verbose) {
              cli::cli_alert_warning("Reached 10 thousand response limit.")
            }
            class(full_repos_list) <<- c(class(full_repos_list), "rest_error", "rest_error_response_limit")
            return(NULL)
          } else {
            if (verbose) cli::cli_alert_danger(e$message)
            return(NULL)
          }
        })
        if (length(search_result) == 0 || inherits(full_repos_list, "rest_error")) {
          still_more_hits <- FALSE
          break
        } else {
          full_repos_list <- append(full_repos_list, search_result)
          page <- page + 1
        }
      }
      return(full_repos_list)
    },

    search_repos_for_code = function(code,
                                     repos,
                                     filename = NULL,
                                     in_path = FALSE,
                                     language = NULL,
                                     page_max = 1e6,
                                     verbose = TRUE) {
      if (verbose) cli::cli_alert("Searching for code [{code}]...")
      query <- private$build_search_query(code, filename = filename, in_path = in_path)
      code <- url_encode(code)
      search_response <- purrr::map(repos, function(repo) {
        page <- 1
        still_more_hits <- TRUE
        full_repos_list <- list()
        search_endpoint <- private$set_projects_search_endpoint(repo)
        while (still_more_hits | page < page_max) {
          search_result <- self$response(
            endpoint = paste0(
              search_endpoint,
              query,
              "&per_page=100&page=",
              page
            ),
            verbose = verbose
          )
          if (length(search_result) == 0) {
            still_more_hits <- FALSE
            break()
          } else {
            full_repos_list <- append(full_repos_list, search_result)
            page <- page + 1
          }
        }
        return(full_repos_list)
      }) |>
        purrr::list_flatten()
      return(search_response)
    },

    get_commits_from_repos = function(full_repos_names,
                                      since,
                                      until,
                                      verbose) {
      repos_list_with_commits <- gitstats_map(full_repos_names, function(repo_path) {
        commits_from_repo <- private$get_commits_from_one_repo(
          repo_path = repo_path,
          since = since,
          until = parse_until_param(until),
          verbose = verbose
        )
        return(commits_from_repo)
      })
      names(repos_list_with_commits) <- full_repos_names
      repos_list_with_commits <- repos_list_with_commits |>
        purrr::discard(~ length(.) == 0)
      return(repos_list_with_commits)
    },

    tailor_commits_info = function(repos_list_with_commits,
                                   org) {
      repos_list_with_commits_cut <- purrr::map(repos_list_with_commits, function(repo) {
        purrr::map(repo, function(commit) {
          list(
            "repo_name" = gsub(
              pattern = paste0("/-/commit/", commit$id),
              replacement = "",
              x = gsub(paste0("(.*)", org, "/"), "", commit$web_url)
            ),
            "id" = commit$id,
            "committed_date" = gts_to_posixt(commit$committed_date),
            "author" = commit$author_name,
            "additions" = commit$stats$additions,
            "deletions" = commit$stats$deletions,
            "organization" = org,
            "repo_url" = stringr::str_match(commit$web_url, "(.*)/-/commit/.*")[2]
          )
        })
      })
      return(repos_list_with_commits_cut)
    },

    prepare_commits_table = function(commits_list) {
      commits_dt <- purrr::map(commits_list, function(commit) {
        purrr::map(commit, ~ data.frame(.)) |>
          purrr::list_rbind()
      }) |>
        purrr::list_rbind()
      if (length(commits_dt) > 0) {
        commits_dt <- dplyr::mutate(
          commits_dt,
          api_url = self$rest_api_url
        )
      }
      return(commits_dt)
    },

    get_commits_authors_handles_and_names = function(commits_table,
                                                     verbose) {
      if (nrow(commits_table) > 0) {
        if (verbose) {
          cli::cli_alert("[Host:GitLab][Engine:{cli::col_green('REST')}] Looking up for authors' {cli_icons$user} names and logins...")
        }
        authors_dict <- private$get_authors_dict(
          commits_table = commits_table,
          verbose = verbose
        )
        commits_table <- commits_table |>
          dplyr::mutate(
            author_login = NA,
            author_name = NA
          ) |>
          dplyr::relocate(
            any_of(c("author_login", "author_name")),
            .after = author
          )

        empty_dict <- all(is.na(authors_dict[, c("author_login", "author_name")] |>
                                  unlist()))
        if (!empty_dict) {
          commits_table <- dplyr::mutate(
            commits_table,
            author_login = purrr::map_vec(author, ~ authors_dict$author_login[. == authors_dict$author]),
            author_name = purrr::map_vec(author, ~ authors_dict$author_name[. == authors_dict$author])
          )
        }
        return(commits_table)
      }
    },

    get_commit_sha_from_branch = function(project_id, default_branch) {
      if (is.null(default_branch) || nchar(default_branch) == 0) {
        return(NA_character_)
      }
      branch_response <- self$response(
        endpoint = paste0(
          private$endpoints[["projects"]],
          project_id,
          "/repository/branches/",
          utils::URLencode(default_branch, reserved = TRUE)
        ),
        verbose = FALSE
      )
      branch_response$commit$id %||% NA_character_
    }
  ),
  private = list(

    endpoints = list(
      organizations = NULL,
      projects = NULL,
      search = NULL
    ),

    build_search_query = function(code, filename = NULL, in_path = FALSE) {
      code <- url_encode(code)
      if (in_path) {
        query <- paste0("path:", code)
      } else {
        query <- code
      }
      if (!is.null(filename)) {
        query <- paste0(query, "%20filename:", filename)
      }
      return(query)
    },

    set_endpoints = function() {
      private$set_projects_endpoint()
    },

    set_projects_endpoint = function() {
      private$endpoints[["projects"]] <- paste0(
        self$rest_api_url,
        "/projects/"
      )
      private$endpoints[["organizations"]] <- paste0(
        self$rest_api_url,
        "/groups/"
      )
      private$endpoints[["users"]] <- paste0(
        self$rest_api_url,
        "/users/"
      )
    },

    set_search_endpoint = function(org = NULL, verbose = TRUE) {
      scope_endpoint <- if (!is.null(org)) {
        paste0("/groups/", private$get_group_id(url_encode(org), verbose))
      } else {
        ""
      }
      paste0(
        self$rest_api_url,
        scope_endpoint,
        "/search?scope=blobs&search="
      )
    },

    set_projects_search_endpoint = function(repo) {
      paste0(
        self$rest_api_url,
        "/projects/",
        url_encode(repo),
        "/search?scope=blobs&search="
      )
    },

    get_repos_languages = function(repos_list, verbose, progress) {
      if (verbose) {
        cli::cli_alert("Pulling languages {cli_icons$language}...")
      }
      repos_list_with_languages <- purrr::map(repos_list, function(repo) {
        id <- repo$id
        languages_response <- self$response(
          endpoint = paste0(private$endpoints[["projects"]], id, "/languages"),
          verbose = verbose
        )
        repo$languages <- names(languages_response)
        repo
      }, .progress = progress)
      return(repos_list_with_languages)
    },

    get_commits_from_one_repo = function(repo_path,
                                         since,
                                         until,
                                         verbose = TRUE) {
      commits_endpoint <- paste0(
        private$endpoints$projects,
        url_encode(repo_path),
        "/repository/commits?since='",
        as.Date(since),
        "'&until='",
        as.Date(until),
        "'&with_stats=true"
      )
      all_commits_in_repo <- tryCatch({
        private$paginate_results(
          endpoint = commits_endpoint,
          joining_sign = "&",
          verbose = verbose
        )
      }, error = function(e) {
        list()
      })
      return(all_commits_in_repo)
    },

    get_group_id = function(project_group, verbose) {
      self$response(
        endpoint = paste0(self$rest_api_url, "/groups/", project_group),
        verbose = verbose
      )[["id"]]
    },

    add_file_info = function(files_search_result,
                             filename,
                             clean_file_content = FALSE,
                             verbose = FALSE,
                             progress = FALSE) {
      purrr::map(files_search_result, function(file_data) {
        repo_data <- self$response(
          endpoint = paste0(
            private$endpoints$projects,
            file_data$project_id
          ),
          verbose = verbose
        )
        def_branch <- repo_data[["default_branch"]]
        file_data <- tryCatch({
          self$response(
            endpoint = paste0(
              private$endpoints$projects,
              file_data$project_id,
              "/repository/files/",
              filename,
              "?ref=",
              def_branch
            ),
            verbose = verbose
          )
        }, error = function(e) {
          NULL
        })
        if (!is.null(file_data)) {
          file_data$repo_name <- repo_data$path
          file_data$repo_fullname <- repo_data$path_with_namespace
          file_data$repo_id <- repo_data$id
          file_data$repo_url <- repo_data$web_url
          if (clean_file_content) {
            file_data$content <- NA
          }
        }
        return(file_data)
      }, .progress = if (progress) {
        glue::glue("Adding file [{filename}] info...")
      } else {
        FALSE
      }) |>
        purrr::discard(is.null)
    },

    get_authors_dict = function(commits_table, verbose) {
      authors_dict <- purrr::map(unique(commits_table$author), function(author) {
        author <- url_encode(author)
        search_endpoint <- paste0(
          self$rest_api_url,
          "/search?scope=users&search=%22", author, "%22"
        )
        user_response <- list()
        try({
          user_response <- self$response(
            endpoint = search_endpoint,
            verbose = verbose
          )
        }, silent = TRUE)
        if (length(user_response) == 0) {
          author <- stringi::stri_trans_general(author, "Latin-ASCII")
          search_endpoint <- paste0(
            self$rest_api_url,
            "/search?scope=users&search=%22", author, "%22"
          )
          try({
            user_response <- self$response(
              endpoint = search_endpoint,
              verbose = verbose
            )
          }, silent = TRUE)
        }
        if (!is.null(user_response) && length(user_response) > 1) {
          user_response <- purrr::keep(user_response, ~ grepl(author, .$name))
        }
        if (is.null(user_response) || length(user_response) == 0) {
          user_tbl <- tibble::tibble(
            author = author,
            author_login = NA,
            author_name = NA
          )
        } else {
          user_tbl <- tibble::tibble(
            author = author,
            author_login = user_response[[1]]$username,
            author_name = user_response[[1]]$name
          )
        }
        return(user_tbl)
      }) |>
        purrr::list_rbind()
      authors_dict <- authors_dict |>
        dplyr::mutate(author = author |>
            url_decode()
        ) |>
        private$clean_authors_dict()
      return(authors_dict)
    },

    clean_authors_dict = function(authors_dict) {
      authors_dict <- private$clean_authors_with_comma(
        authors_dict = authors_dict
      )
      authors_dict <- private$fill_empty_authors(
        authors_dict = authors_dict
      )
      return(authors_dict)
    },

    clean_authors_with_comma = function(authors_dict) {
      authors_to_clean <- authors_dict$author[is.na(authors_dict$author_name)]
      if (any(grepl(",", authors_to_clean))) {
        authors_with_comma <- authors_to_clean[grepl(",", authors_to_clean)]
        clean_authors <- purrr::map(authors_with_comma, function(author) {
          split_author <- stringr::str_split_1(author, ",")
          split_author <- purrr::map(split_author, function(x) {
            stringr::str_replace(x, "\\{.*?\\}", "") |>
              stringr::str_replace_all(" ", "")
          })
          source_author <- unlist(split_author)
          clean_author <- paste(source_author[2], source_author[1])
          dplyr::tibble(
            author = author,
            author_login = NA_character_,
            author_name = clean_author
          )
        }) |>
          purrr::list_rbind()
        authors_dict <- authors_dict |>
          dplyr::filter(!author %in% authors_with_comma)
        authors_dict <- rbind(authors_dict, clean_authors)
      }
      return(authors_dict)
    },

    fill_empty_authors = function(authors_dict) {
      authors_to_clean <- authors_dict$author[is.na(authors_dict$author_name)]
      authors_to_clean <- authors_to_clean[!grepl(",", authors_to_clean)]
      author_names <- purrr::keep(authors_to_clean, function(author) {
        length(stringr::str_split_1(author, " ")) > 1
      })
      author_logins <- purrr::keep(authors_to_clean, function(author) {
        length(stringr::str_split_1(author, " ")) == 1
      })
      authors_dict <- authors_dict |>
        dplyr::mutate(
          author_name = ifelse(author %in% author_names,
                               author,
                               author_name),
          author_login = ifelse(author %in% author_logins,
                                author,
                                author_login)
        )
      return(authors_dict)
    }
  )
)

test_that("add_file_info adds info to files search result", {
  mockery::stub(
    test_rest_gitlab_priv$add_file_info,
    "self$response",
    test_fixtures$gitlab_file_rest_response
  )
  file_info <- test_rest_gitlab_priv$add_file_info(
    files_search_result = test_fixtures$gitlab_search_response,
    filename = "test.R",
    clean_file_content = FALSE,
    verbose = TRUE,
    progress = FALSE
  )
  expect_type(file_info, "list")
  expect_length(file_info, 2)
  purrr::walk(file_info, ~ expect_equal(.$content, "test content"))
  test_mocker$cache(file_info)

  file_info_na <- test_rest_gitlab_priv$add_file_info(
    files_search_result = test_fixtures$gitlab_search_response,
    filename = "test.R",
    clean_file_content = TRUE,
    verbose = TRUE,
    progress = FALSE
  )
  purrr::walk(file_info_na, ~ expect_equal(.$content, NA))
})

test_that("get_files works as expected", {
  mockery::stub(
    test_rest_gitlab$get_files,
    "self$search_for_code",
    test_fixtures$gitlab_search_response
  )
  mockery::stub(
    test_rest_gitlab$get_files,
    "private$add_file_info",
    test_mocker$use("file_info")
  )
  files <- test_rest_gitlab$get_files(
    file_paths = "test.R",
    clean_files_content = FALSE,
    verbose = TRUE,
    progress = FALSE
  )
  expect_type(files, "list")
  expect_length(files, 2)
})

test_that("`get_commits_from_one_repo()` pulls commits from repository", {
  mockery::stub(
    test_rest_gitlab_priv$get_commits_from_one_repo,
    "private$paginate_results",
    test_fixtures$gitlab_commits_response
  )
  gl_commits_repo <- test_rest_gitlab_priv$get_commits_from_one_repo(
    repo_path = "TestRepo",
    since     = "2023-01-01",
    until     = "2023-04-20"
  )
  expect_gt(length(gl_commits_repo), 1)
  purrr::walk(gl_commits_repo, ~ expect_commit_gitlab_rest_response(.))
  test_mocker$cache(gl_commits_repo)
})

test_that("`get_commits_from_repos()` pulls commits from repositories", {
  mockery::stub(
    test_rest_gitlab$get_commits_from_repos,
    "private$get_commits_from_one_repo",
    test_mocker$use("gl_commits_repo")
  )
  full_repos_names <- c("test_org/TestRepo1", "test_org/TestRepo2")
  gl_commits_org <- test_rest_gitlab$get_commits_from_repos(
    full_repos_names = full_repos_names,
    since = "2023-01-01",
    until = "2023-04-20"
  )
  expect_equal(names(gl_commits_org), c("test_org/TestRepo1", "test_org/TestRepo2"))
  purrr::walk(gl_commits_org[[1]], ~ expect_commit_gitlab_rest_response(.))
  test_mocker$cache(gl_commits_org)
})

test_that("`tailor_commits_info()` retrieves only necessary info", {
  gl_commits_list <- test_mocker$use("gl_commits_org")

  gl_commits_list_cut <- test_rest_gitlab$tailor_commits_info(
    gl_commits_list,
    org = "mbtests"
  )
  expect_list_contains_only(
    gl_commits_list_cut[[1]][[1]],
    c(
      "repo_name", "id", "organization", "additions", "deletions",
      "committed_date", "author"
    )
  )
  test_mocker$cache(gl_commits_list_cut)
})

test_that("`prepare_commits_table()` prepares table of commits properly", {
  gl_commits_table <- test_rest_gitlab$prepare_commits_table(
    commits_list = test_mocker$use("gl_commits_list_cut")
  )
  expect_commits_table(
    gl_commits_table,
    exp_auth = FALSE
  )
  test_mocker$cache(gl_commits_table)
})

test_authors_dict <- data.frame(
  author = c("TestFamily, TestName {TestID}", "TestName TestFamily", "testlogin"),
  author_login = rep(NA, 3),
  author_name = rep(NA, 3)
)

test_that("clean_authors_with_comma parses properly authors data", {
  test_authors_dict <- test_rest_gitlab_priv$clean_authors_with_comma(test_authors_dict)
  expect_equal(
    test_authors_dict$author_name[test_authors_dict$author == "TestFamily, TestName {TestID}"],
    "TestName TestFamily"
  )
  test_mocker$cache(test_authors_dict)
})

test_that("fill_empty_authors fills properly authors data", {
  test_authors_dict <- test_rest_gitlab_priv$fill_empty_authors(
    authors_dict = test_mocker$use("test_authors_dict")
  )
  expect_equal(
    test_authors_dict$author_name[test_authors_dict$author == "TestName TestFamily"],
    "TestName TestFamily"
  )
  expect_equal(
    test_authors_dict$author_login[test_authors_dict$author == "testlogin"],
    "testlogin"
  )
  test_mocker$cache(test_authors_dict)
})

test_that("clean_authors_dict", {
  test_authors_dict <- test_rest_gitlab_priv$clean_authors_dict(
    authors_dict = test_authors_dict
  )
  expect_equal(
    test_authors_dict$author_name[test_authors_dict$author == "TestFamily, TestName {TestID}"],
    "TestName TestFamily"
  )
  expect_equal(
    test_authors_dict$author_name[test_authors_dict$author == "TestName TestFamily"],
    "TestName TestFamily"
  )
  expect_equal(
    test_authors_dict$author_login[test_authors_dict$author == "testlogin"],
    "testlogin"
  )
})

test_that("get_authors_dict() prepares dictionary with handles and names", {
  mockery::stub(
    test_rest_gitlab_priv$get_authors_dict,
    "self$response",
    test_fixtures$gitlab_user_search_response
  )
  authors_dict <- test_rest_gitlab_priv$get_authors_dict(
    commits_table = test_mocker$use("gl_commits_table")
  )
  expect_s3_class(authors_dict, "data.frame")
  expect_gt(nrow(authors_dict), 0)
  expect_length(authors_dict, 3)
  test_mocker$cache(authors_dict)
})

test_that("`get_commits_authors_handles_and_names()` adds author logis and names to commits table", {
  mockery::stub(
    test_rest_gitlab$get_commits_authors_handles_and_names,
    "private$get_authors_dict",
    test_mocker$use("authors_dict")
  )
  gl_commits_table <- test_rest_gitlab$get_commits_authors_handles_and_names(
    commits_table = test_mocker$use("gl_commits_table"),
    verbose = FALSE
  )
  expect_commits_table(
    gl_commits_table,
    exp_auth = TRUE
  )
  test_mocker$cache(gl_commits_table)
})

test_that("`get_commits_authors_handles_and_names()` prints messages", {
  mockery::stub(
    test_rest_gitlab$get_commits_authors_handles_and_names,
    "private$get_authors_dict",
    test_mocker$use("authors_dict")
  )
  expect_snapshot(
    gl_commits_table <- test_rest_gitlab$get_commits_authors_handles_and_names(
      commits_table = test_mocker$use("gl_commits_table"),
      verbose = TRUE
    )
  )
})

test_that("get_commits_from_orgs works", {
  mockery::stub(
    gitlab_testhost_priv$get_commits_from_orgs,
    "private$get_repos_names",
    c("test_org/TestRepo1", "test_org/TestRepo2")
  )
  mockery::stub(
    gitlab_testhost_priv$get_commits_from_orgs,
    "rest_engine$get_commits_authors_handles_and_names",
    test_mocker$use("gl_commits_table")
  )
  gitlab_testhost_priv$searching_scope <- "org"
  gitlab_testhost_priv$cached_repos <- list()
  expect_snapshot(
    gl_commits_table <- gitlab_testhost_priv$get_commits_from_orgs(
      since = "2023-03-01",
      until = "2023-04-01",
      verbose = TRUE,
      progress = FALSE
    )
  )
  expect_commits_table(
    gl_commits_table
  )
  test_mocker$cache(gl_commits_table)
})

test_that("get_commits_from_repos works", {
  gitlab_testhost_priv <- create_gitlab_testhost(
    repos = "TestRepo",
    mode = "private"
  )
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    gitlab_testhost_priv$get_repos_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  mockery::stub(
    gitlab_testhost_priv$get_commits_from_repos,
    "rest_engine$get_commits_authors_handles_and_names",
    test_mocker$use("gl_commits_table")
  )
  gitlab_testhost_priv$searching_scope <- "repo"
  gitlab_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
  expect_snapshot(
    gl_commits_table <- gitlab_testhost_priv$get_commits_from_repos(
      since    = "2023-03-01",
      until    = "2023-04-01",
      verbose  = TRUE,
      progress = FALSE
    )
  )
  expect_commits_table(
    gl_commits_table
  )
})

test_that("groups GitLab query is built properly", {
  gl_orgs_query <-
    test_gqlquery_gl$groups()
  expect_snapshot(
    gl_orgs_query
  )
})

test_that("group GitLab query is built properly", {
  gl_org_query <-
    test_gqlquery_gl$group()
  expect_snapshot(
    gl_org_query
  )
})

test_that("get_orgs_count works", {
  mockery::stub(
    test_rest_gitlab$get_orgs_count,
    "self$perform_request",
    test_fixtures$rest_gl_orgs_response
  )
  orgs_count <- test_rest_gitlab$get_orgs_count(verbose = FALSE)
  expect_type(
    orgs_count,
    "character"
  )
  expect_type(
    as.integer(orgs_count),
    "integer"
  )
})

test_that("get_orgs_count prints message", {
  mockery::stub(
    test_rest_gitlab$get_orgs_count,
    "self$perform_request",
    test_fixtures$rest_gl_orgs_response
  )
  expect_snapshot(
    orgs_count <- test_rest_gitlab$get_orgs_count(verbose = TRUE)
  )
})

test_that("get_orgs pulls responses from GraphQL", {
  mockery::stub(
    test_graphql_gitlab$get_orgs,
    "self$gql_response",
    test_fixtures$graphql_gl_orgs_response
  )
  gl_orgs_raw_response <- test_graphql_gitlab$get_orgs(
    orgs_count = 3L,
    output = "only_names",
    verbose = FALSE
  )
  expect_type(
    gl_orgs_raw_response,
    "character"
  )
  test_mocker$cache(gl_orgs_raw_response)
  gl_orgs_full_response <- test_graphql_gitlab$get_orgs(
    orgs_count = 3L,
    output = "full_table",
    verbose = FALSE
  )
  expect_type(
    gl_orgs_full_response,
    "list"
  )
  expect_orgs_gitlab_full_list(gl_orgs_full_response)
  test_mocker$cache(gl_orgs_full_response)
})

test_that("get_orgs prints message", {
  mockery::stub(
    test_graphql_gitlab$get_orgs,
    "self$gql_response",
    test_fixtures$graphql_gl_orgs_response
  )
  expect_snapshot(
    gl_orgs_full_response <- test_graphql_gitlab$get_orgs(
      orgs_count = 3L,
      output = "full_table",
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("if get_orgs runs into GraphQL error, it returns object of graphql_error class", {
  mockery::stub(
    test_graphql_gitlab$get_orgs,
    "self$gql_response",
    test_error_fixtures$graphql_error_no_groups |>
      test_graphql_gitlab_priv$set_graphql_error_class()
  )
  gl_orgs_error_response <- test_graphql_gitlab$get_orgs(
    orgs_count = 3L,
    output = "full_table",
    verbose = FALSE,
    progress = FALSE
  )
  expect_s3_class(gl_orgs_error_response, "graphql_error")
  test_mocker$cache(gl_orgs_error_response)
})

test_that("if get_orgs runs into GraphQL error, it prints warning", {
  mockery::stub(
    test_graphql_gitlab$get_orgs,
    "self$gql_response",
    test_error_fixtures$graphql_error_no_groups |>
      test_graphql_gitlab_priv$set_graphql_error_class()
  )
  expect_snapshot(
    gl_orgs_error_response <- test_graphql_gitlab$get_orgs(
      orgs_count = 3L,
      output = "full_table",
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("prepare_orgs_table works", {
  gitlab_orgs_table <- test_graphql_gitlab$prepare_orgs_table(
    full_orgs_list = test_mocker$use("gl_orgs_full_response")
  )
  expect_orgs_table(
    gitlab_orgs_table
  )
})

test_that("get_org pulls response for one org from GraphQL", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_gitlab$get_org,
      "self$gql_response",
      test_fixtures$graphql_gl_org_response
    )
  }
  org <- "mbtests"
  gl_org_response <- test_graphql_gitlab$get_org(
    org = org,
    verbose = FALSE
  )
  expect_type(gl_org_response, "list")
  test_mocker$cache(gl_org_response)
})

test_that("get_org prints proper message", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_gitlab$get_org,
      "self$gql_response",
      test_fixtures$graphql_gl_org_response
    )
  }
  org <- "mbtests"
  expect_snapshot(
    gl_org_response <- test_graphql_gitlab$get_org(
      org = org,
      verbose = TRUE
    )
  )
})

test_that("get_org sets error class to response", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_gitlab$get_org,
      "self$gql_response",
      list()
    )
  }
  org <- "mbtrests"
  gl_org_error_response <- test_graphql_gitlab$get_org(
    org = org,
    verbose = FALSE
  )
  expect_type(gl_org_error_response, "list")
  expect_s3_class(gl_org_error_response, "graphql_error")
  test_mocker$cache(gl_org_error_response)
})

test_that("get_org pulls response for one org from REST", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_rest_gitlab$get_org,
      "self$response",
      test_fixtures$rest_gl_org_response
    )
  }
  gl_org_rest_response <- test_rest_gitlab$get_org(
    org = "mbtests",
    verbose = FALSE
  )
  expect_type(gl_org_rest_response, "list")
  expect_true(
    all(c("id", "web_url", "name", "path", "description") %in% names(gl_org_rest_response))
  )
  test_mocker$cache(gl_org_rest_response)
})

test_that("get_org prints proper message", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_rest_gitlab$get_org,
      "self$response",
      test_fixtures$rest_gl_org_response
    )
  }
  expect_snapshot(
    gl_org_response <- test_rest_gitlab$get_org(
      org =  "mbtests",
      verbose = TRUE
    )
  )
})

test_that("REST method get_orgs works", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_rest_gitlab$get_orgs,
      "self$response",
      test_fixtures$rest_gl_org_response
    )
  }
  gl_orgs_rest_list <- test_rest_gitlab$get_orgs(
    orgs_count = 300L,
    verbose = FALSE
  )
  expect_type(gl_orgs_rest_list, "list")
  if (integration_tests_skipped) {
    gl_orgs_rest_list <- rep(list(gl_orgs_rest_list), 300L)
  }
  expect_length(gl_orgs_rest_list, 300L)
  expect_in(c("name", "id", "web_url", "path"), names(gl_orgs_rest_list[[1]]))
  test_mocker$cache(gl_orgs_rest_list)
})

test_that("REST method prints message", {
  skip_if(integration_tests_skipped)
  expect_snapshot(
    gl_orgs_rest_list <- test_rest_gitlab$get_orgs(
      orgs_count = 300L,
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("table is prepared from REST orgs response", {
  gitlab_org_rest_table <- test_rest_gitlab$prepare_orgs_table(
    orgs_list = test_mocker$use("gl_orgs_rest_list")
  )
  expect_s3_class(gitlab_org_rest_table, "data.frame")
  test_mocker$cache(gitlab_org_rest_table)
})

test_that("if get_orgs_from_host runs into GraphQL error, it switches to REST API", {
  skip_if(integration_tests_skipped)
  gitlab_test_host_priv_2 <- create_gitlab_testhost(
    orgs = "mbtests",
    mode = "private"
  )
  mockery::stub(
    gitlab_test_host_priv_2$get_orgs_from_host,
    "graphql_engine$get_orgs",
    test_mocker$use("gl_orgs_error_response")
  )
  mockery::stub(
    gitlab_test_host_priv_2$get_orgs_from_host,
    "rest_engine$get_orgs",
    test_mocker$use("gl_orgs_rest_list")
  )
  gitlab_test_host_priv_2$is_public <- FALSE
  expect_snapshot(
    gitlab_orgs_vec <- gitlab_test_host_priv_2$get_orgs_from_host(
      output = "only_names",
      verbose = TRUE
    )
  )
  expect_type(gitlab_orgs_vec, "character")
  expect_length(gitlab_orgs_vec, 300L)
})

test_that("if get_orgs_from_host runs into GraphQL error, it switches to REST API", {
  skip_if(integration_tests_skipped)
  gitlab_test_host_priv_2 <- create_gitlab_testhost(
    orgs = "mbtests",
    mode = "private"
  )
  gitlab_test_host_priv_2$is_public <- FALSE
  mockery::stub(
    gitlab_test_host_priv_2$get_orgs_from_host,
    "graphql_engine$get_orgs",
    test_mocker$use("gl_orgs_error_response")
  )
  mockery::stub(
    gitlab_test_host_priv_2$get_orgs_from_host,
    "rest_engine$get_orgs",
    test_mocker$use("gl_orgs_rest_list")
  )
  mockery::stub(
    gitlab_test_host_priv_2$get_orgs_from_host,
    "rest_engine$prepare_orgs_table",
    test_mocker$use("gitlab_org_rest_table")
  )
  expect_snapshot(
    gitlab_orgs_table <- gitlab_test_host_priv_2$get_orgs_from_host(
      output = "full_table",
      verbose = TRUE
    )
  )
  expect_s3_class(gitlab_orgs_table, "data.frame")
  expect_equal(nrow(gitlab_orgs_table), 300L)
})

test_that("get_orgs_from_host works on GitHost level", {
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_host,
    "graphql_engine$get_orgs",
    test_mocker$use("gl_orgs_full_response")
  )
  gitlab_testhost_priv$is_public <- FALSE
  gitlab_orgs_table <- gitlab_testhost_priv$get_orgs_from_host(
    output = "full_table",
    verbose = FALSE
  )
  expect_orgs_table(
    gitlab_orgs_table
  )
  test_mocker$cache(gitlab_orgs_table)
})

test_that("get_orgs_from_host prints message on number of organizations", {
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_host,
    "rest_engine$get_orgs_count",
    3L
  )
  gitlab_testhost_priv$is_public <- FALSE
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_host,
    "graphql_engine$get_orgs",
    test_mocker$use("gl_orgs_full_response")
  )
  expect_snapshot(
    gitlab_orgs_table <- gitlab_testhost_priv$get_orgs_from_host(
      output = "full_table",
      verbose = TRUE
    )
  )
})

test_that("get_orgs_from_orgs_and_repos works on GitHost level", {
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_orgs_and_repos,
    "graphql_engine$get_org",
    test_mocker$use("gl_org_response")
  )
  gitlab_orgs_from_orgs_table <- gitlab_testhost_priv$get_orgs_from_orgs_and_repos(
    verbose = FALSE
  )
  expect_orgs_table(
    gitlab_orgs_from_orgs_table
  )
  test_mocker$cache(gitlab_orgs_from_orgs_table)
})

test_that("get_orgs_from_orgs_and_repos works on GitHost level", {
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_orgs_and_repos,
    "graphql_engine$get_org",
    test_mocker$use("gl_org_response")
  )
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_orgs_and_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  gitlab_orgs_from_orgs_table <- gitlab_testhost_priv$get_orgs_from_orgs_and_repos(
    verbose = FALSE
  )
  expect_orgs_table(
    gitlab_orgs_from_orgs_table
  )
  test_mocker$cache(gitlab_orgs_from_orgs_table)
})

test_that("get_orgs_from_orgs_and_repos turns to REST if GraphQL fails", {
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_orgs_and_repos,
    "graphql_engine$get_org",
    test_mocker$use("gl_org_error_response")
  )
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_orgs_and_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_orgs_and_repos,
    "rest_engine$get_org",
    test_mocker$use("gl_org_rest_response")
  )
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_orgs_and_repos,
    "default_engine$prepare_orgs_table",
    test_mocker$use("gitlab_org_rest_table")
  )
  gitlab_orgs_from_orgs_table <- gitlab_testhost_priv$get_orgs_from_orgs_and_repos(
    verbose = FALSE
  )
  expect_orgs_table(
    gitlab_orgs_from_orgs_table
  )
  test_mocker$cache(gitlab_orgs_from_orgs_table)
})

test_that("get_orgs_from_orgs_and_repos prints message when turning to REST engine", {
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_orgs_and_repos,
    "graphql_engine$get_org",
    test_mocker$use("gl_org_error_response")
  )
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_orgs_and_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_orgs_and_repos,
    "rest_engine$get_org",
    test_mocker$use("gl_org_rest_response")
  )
  mockery::stub(
    gitlab_testhost_priv$get_orgs_from_orgs_and_repos,
    "default_engine$prepare_orgs_table",
    test_mocker$use("gitlab_org_rest_table")
  )
  expect_snapshot(
    gitlab_orgs_from_orgs_table <- gitlab_testhost_priv$get_orgs_from_orgs_and_repos(
      verbose = TRUE
    )
  )
})

test_that("get_orgs_from_host returns error when GitHost is public", {
  gitlab_testhost_priv$is_public <- TRUE
  expect_snapshot_error(
    gitlab_testhost_priv$get_orgs_from_host(
      output = "full_table",
      verbose = FALSE
    )
  )
})

test_that("get_orgs works on GitHost level", {
  mockery::stub(
    gitlab_testhost$get_orgs,
    "private$get_orgs_from_hosts",
    test_mocker$use("gitlab_orgs_table")
  )
  mockery::stub(
    gitlab_testhost$get_orgs,
    "private$get_orgs_from_orgs_and_repos",
    test_mocker$use("gitlab_orgs_from_orgs_table")
  )
  gitlab_orgs_table <- gitlab_testhost$get_orgs(
    output = "full_table",
    verbose = FALSE
  )
  expect_orgs_table(
    gitlab_orgs_table,
    add_cols = c("host_url", "host_name")
  )
  test_mocker$cache(gitlab_orgs_table)
})

GQLQueryGitLab <- R6::R6Class("GQLQueryGitLab",
  public = list(

    # groups GitLab >= 14.3
    groups = function() {
      paste0(
        'query GetGroups($groupCursor: String!) {
            groups (after: $groupCursor) {
              pageInfo {
                endCursor
                hasNextPage
              }
              edges {
                node {', private$group_fields, '}
              }
            }
        }'
      )
    },

    # group GitLab >= 12.2
    group = function() {
      paste0('
      query GetGroup($org: ID!) {
        group(fullPath: $org) {',
             private$group_fields
             , '}
      }
    ')
    },

    user_or_org_query =
      '
      query ($username: String! $grouppath: ID!) {
        user(username: $username) {
          __typename
          username
        }
        group(fullPath: $grouppath) {
          __typename
          fullPath
        }
      }'
    ,

    repos = function(repo_cursor) {
      paste0('
        query GetRepo($projects_ids: [ID!]!) {
          projects(ids: $projects_ids first:100', private$add_cursor(repo_cursor), ') {
          ', private$projects_field_content, '
          }
        }')
    },

    # projects in group GitLab >= 12.2
    repos_by_org = function() {
      paste0('
        query GetReposByOrg($org: ID! $repo_cursor: String!) {
          group(fullPath: $org) {
            projects(first: 100 after: $repo_cursor) {
            ', private$projects_field_content, '
            }
          }
        }')
    },

    repos_by_user = function() {
      paste0('
        query GetUserRepos ($username: String! $repo_cursor: String!) {
          projects(search: $username searchNamespaces: true after: $repo_cursor first: 100) {
            ', private$projects_field_content, '
          }
        }')
    },

    repo_by_fullpath = function() {
      paste0('
        query GetRepoByFullPath($fullPath: ID!) {
          project(fullPath: $fullPath) {
            ', private$project_node_fields, '
          }
        }')
    },

    issues_from_repo = function(issues_cursor = "") {
      paste0('
      query getIssuesFromRepo ($fullPath: ID!) {
          project(fullPath: $fullPath) {
            issues(first: 100
                   ', private$add_cursor(issues_cursor), ') {
              pageInfo {
                hasNextPage
      				  endCursor
      				}
              edges {
                node {
                  number: iid
                  title
                  description
                  created_at: createdAt
                  closed_at: closedAt
                  state
                  url: webUrl
                  author {
                    login: username
                  }
                }
              }
            }
          }
        }
      ')
    },

    pull_requests_from_repo = function(pr_cursor = "") {
      paste0('
      query getPullRequestsFromRepo ($fullPath: ID!) {
          project(fullPath: $fullPath) {
            mergeRequests(first: 100, ',
             private$add_cursor(pr_cursor), ') {
              pageInfo {
                endCursor
                hasNextPage
              }
              edges {
                node {
                  number: iid
                  created_at: createdAt
                  merged_at: mergedAt
                  state
                  author {
                    username
                  }
                  source_branch: sourceBranch
                  target_branch: targetBranch
                }
              }
            }
          }
        }
      ')
    },

    user = function() {
      paste0('
        query GetUser($user: String!) {
          user(username: $user) {
            id
            name
            login: username
            email: publicEmail
            location
            starred_repos: starredProjects {
              count
            }
            pull_requests: authoredMergeRequests {
              count
            }
            reviews: reviewRequestedMergeRequests {
              count
            }
            avatar_url: avatarUrl
            web_url: webUrl
          }
        }
      ')
    },

    files_by_org = function(end_cursor = "") {
      paste0(
        'query GetFilesByOrg($org: ID!, $file_paths: [String!]!) {
            group(fullPath: $org) {
              projects(first: 100',
        private$add_cursor(end_cursor),
        ') {
          count
          pageInfo {
            hasNextPage
            endCursor
          }
          edges {
            node {
              name
              path
              id
              webUrl
              repository {
                blobs(paths: $file_paths) {
                  nodes {
                    path
                    rawBlob
                    size
                    oid
                  }
                }
                lastCommit {
                  sha
                }
              }
            }
          }
        }
      }
    }'
      )
    },

    file_blob_from_repo = function() {
      '
      query GetFilesByRepo($fullPath: ID!, $file_paths: [String!]!) {
        project(fullPath: $fullPath) {
          name
          path
          id
          webUrl
          repository {
            blobs(paths: $file_paths) {
              nodes {
                path
                rawBlob
                size
                oid
              }
            }
            lastCommit {
              sha
            }
          }
        }
      }
      '
    },

    releases_from_repo = function() {
      'query GetReleasesFromRepo($project_path: ID!) {
              project(fullPath: $project_path) {
                name
                webUrl
    						releases {
                  nodes{
                    name
                    tagName
                    releasedAt
                    links {
                      selfUrl
                    }
                    description
                  }
                }
              }
          }'
    }
  ),
  private = list(
    add_cursor = function(cursor) {
      if (nchar(cursor) == 0) {
        cursor_argument <- cursor
      } else {
        cursor_argument <- paste0('after: "', cursor, '"')
      }
      return(cursor_argument)
    },

    group_fields =
      '
      name
      description
      fullPath
      webUrl
      projectsCount
      groupMembersCount
      avatarUrl
    ',

    project_node_fields =
      '
          repo_id: id
          repo_name: name
          repo_path: path
          repo_fullpath: fullPath
          ... on Project {
            repository {
              rootRef
              lastCommit {
                sha
              }
            }
          }
          stars: starCount
          forks: forksCount
          created_at: createdAt
          last_activity_at: lastActivityAt
          languages {
            name
          }
          issues: issueStatusCounts {
            all
            closed
            opened
          }
          namespace {
            path: fullPath
          }
          repo_url: webUrl',

    # count in ProjectConnection GitLab >= 13.0
    # languages in Project GitLab >= 12.9
    projects_field_content =
      '
      count
      pageInfo {
        hasNextPage
        endCursor
      }
      edges {
        node {
          repo_id: id
          repo_name: name
          repo_path: path
          repo_fullpath: fullPath
          ... on Project {
            repository {
              rootRef
              lastCommit {
                sha
              }
            }
          }
          stars: starCount
          forks: forksCount
          created_at: createdAt
          last_activity_at: lastActivityAt
          languages {
            name
          }
          issues: issueStatusCounts {
            all
            closed
            opened
          }
          namespace {
            path: fullPath
          }
          repo_url: webUrl
        }
      }'
  )
)

test_error_fixtures <- list()

test_error_fixtures$graphql_server_error <- list(
  "errors" = list(
    list(
      "message" = "Internal server error"
    )
  )
)

test_error_fixtures$graphql_complexity_error <- list(
  "errors" = list(
    list(
      "message" = "Query has complexity of ..."
    )
  )
)

test_error_fixtures$graphql_limit_error <- list(
  "errors" = list(
    list(
      "message" = "Resolvers::ProjectsResolver#ids cannot accept more than 1000 items",
      "locations" = list(
        list(
          "line" = 3L,
          "column" = 11L
        )
      ),
      "path" = list(
        "projects"
      )
    )
  ),
  "data" = list(
    "projects" = NULL
  )
)

test_error_fixtures$graphql_error_no_groups <- list(
  "errors" = list(
    list(
      "message" = "Field 'groups' doesn't exist on type 'Query'",
      "locations" = list(
        list(
          "line" = 2L,
          "column" = 13L
        )
      ),
      "path" = list(
        "query GetGroups",
        "groups"
      ),
      "extensions" = list(
        "code" = "undefinedField",
        "typeName" = "Query",
        "fieldName" = "groups"
      )
    ),
    list(
      "message" = "Variable $groupCursor is declared by GetGroups but not used",
      "locations" = list(
        list(
          "line" = 1L,
          "column" = 1L
        )
      ),
      "path" = list(
        "query GetGroups"
      ),
      "extensions" = list(
        "code" = "variableNotUsed",
        "variableName" = "groupCursor"
      )
    )
  )
)

test_error_fixtures$graphql_error_no_count_languages <- list(
  "errors" = list(
    list(
      "message" = "Field 'count' doesn't exist on type 'ProjectConnection'",
      "locations" = list(
        list(
          "line" = 6L,
          "column" = 7L
        )
      ),
      "path" = list(
        "query GetReposByOrg",
        "group",
        "projects",
        "count"
      ),
      "extensions" = list(
        "code" = "undefinedField",
        "typeName" = "ProjectConnection",
        "fieldName" = "count"
      )
    ),
    list(
      "message" = "Field 'languages' doesn't exist on type 'Project'",
      "locations" = list(
        list(
          "line" = 25L,
          "column" = 11L
        )
      ),
      "path" = list(
        "query GetReposByOrg",
        "group",
        "projects",
        "edges",
        "node",
        "languages"
      ),
      "extensions" = list(
        "code" = "undefinedField",
        "typeName" = "Project",
        "fieldName" = "languages"
      )
    )
  )
)

#' @title Get data on organizations
#' @name get_orgs
#' @description  Pulls data on all organizations from a Git host and parses it
#'   into table format.
#' @param gitstats A GitStats object.
#' @param cache A logical, if set to `TRUE` GitStats will retrieve the last
#'   result from its storage.
#' @param verbose A logical, `TRUE` by default. If `FALSE` messages and printing
#'   output is switched off.
#' @return A data.frame.
#' @examples
#' \dontrun{
#' my_gitstats <- create_gitstats() |>
#'   set_github_host(
#'     orgs = c("r-world-devs", "openpharma"),
#'     token = Sys.getenv("GITHUB_PAT")
#'   ) |>
#'   set_gitlab_host(
#'     orgs = "mbtests",
#'     token = Sys.getenv("GITLAB_PAT_PUBLIC")
#'   )
#' get_orgs(my_gitstats)
#' }
#' @export
get_orgs <- function(gitstats,
                     cache = TRUE,
                     verbose = FALSE) {
  start_time <- Sys.time()
  orgs <- gitstats$get_orgs(
    cache = cache,
    verbose = verbose
  )
  end_time <- Sys.time()
  time_taken <- end_time - start_time
  cli::cli_alert_success("Data pulled in {round(time_taken, 1)} {attr(time_taken, 'units')}")
  return(orgs)
}

test_fixtures$github_file_response <- list(
  "data" = list(
    "repository" = list(
      "repo_id"   = "01010101",
      "repo_name" = "TestProject",
      "repo_url"  = "https://github.com/r-world-devs/GitStats",
      "file" = list(
        "text" = "Some interesting text.",
        "byteSize" = 50L,
        "oid" = "12345"
      ),
      "defaultBranchRef" = list(
        "target" = list(
          "oid" = "54321abcde"
        )
      )
    )
  )
)

test_fixtures$gitlab_file_org_response <- list(
  "data" = list(
    "group" = list(
      "projects" = list(
        "count" = 3,
        "pageInfo" = list(
          "hasNextPage" = FALSE,
          "endCursor" = "xxxxxx"
        ),
        "edges" = list(
          list(
            "node" = list(
              "name" = "graphql_tests",
              "path" = "graphql_tests",
              "id" = "gid://gitlab/Project/61399846",
              "webUrl" = "https://gitlab.com/mbtests/graphql_tests",
              "repository" = list(
                "blobs" = list(
                  "nodes" = list() # empty response for query, no files found in org
                )
              )
            )
          ),
          list(
            "node" = list(
              "name" = "RM Tests 3",
              "path" = "rm_tests_3",
              "id" = "gid://gitlab/Project/44346961",
              "webUrl" = "https://gitlab.com/mbtests/rm-tests-3",
              "repository" = list(
                "blobs" = list(
                  "nodes" = list(
                    list(
                      "path" = "meta_data.yaml",
                      "rawBlob" = "Some interesting text",
                      "size" = 4,
                      "oid" = "1a2b3c"
                    )
                  )
                ),
                "lastCommit" = list(
                  "sha" = "e5f6d7"
                )
              )
            )
          ),
          list(
            "node" = list(
              "name" = "RM Tests 2",
              "path" = "rm_tests_2",
              "id" = "gid://gitlab/Project/44293594",
              "webUrl" = "https://gitlab.com/mbtests/rm-tests-2",
              "repository" = list(
                "blobs" = list(
                  "nodes" = list(
                    list(
                      "path" = "meta_data.yaml",
                      "rawBlob" = "Some interesting text",
                      "size" = 5,
                      "oid" = "1a2b3c"
                    )
                  )
                ),
                "lastCommit" = list(
                  "sha" = "e5f6d7"
                )
              )
            )
          )
        )
      )
    )
  )
)

test_fixtures$gitlab_file_repo_response <- list(
  "data" = list(
    "project" = list(
      "name" = "TestProject",
      "path" = "TestProject",
      "id" = "1010101",
      "webUrl"  = "https://gitlab.com/mbtests/graphql_tests",
      "repository" = list(
        "blobs" = list(
          "nodes" = list(
            list(
              "path" = "README.md",
              "rawBlob" = "This project is for testing GraphQL capabilities.",
              "size" =  "67",
              "oid" = "1a2b3c"
            ),
            list(
              "path" = "project_metadata.yaml",
              "rawBlob" = "GraphQL Tests",
              "size" = "19",
              "oid" = "1a2b3c"
            )
          )
        ),
        "lastCommit" = list(
          "sha" = "e5f6d7"
        )
      )
    )
  )
)

test_fixtures$gitlab_search_response <- list(
  list(
    "basename"= "test",
    "data" = "some text with searched phrase",
    "path" = "test.R",
    "filename" = "test.R",
    "id" = NULL,
    "ref" = "main",
    "startline" = 10,
    "project_id" = 61399846
  ),
  list(
    "basename" = "test",
    "data" = "some text with searched phrase",
    "path" = "test.R",
    "filename" = "test.R",
    "id" = NULL,
    "ref" = "main",
    "startline" = 15,
    "project_id" = 43400864
  )
)

github_search_item <- list(
  "name" = "test1.R",
  "path" = "examples/test1.R",
  "sha" = "0d42b9d23ddfc0bca1",
  "url" = "test_url",
  "git_url" = "test_git_url",
  "html_url" = "test_html_url",
  "repository" = list(
    "node_id" = "test_node_id",
    "id" = 627452680,
    "url" = "https://api.github.com/repos/r-world-devs/GitStats",
    "html_url" = "https://github.com/r-world-devs/GitStats"
  ),
  "score" = 1
)

test_fixtures$github_file_rest_response <- github_search_item

test_fixtures$github_search_response <- list(
  "total_count" = 250,
  "incomplete_results" = FALSE,
  "items" = list(
    rep(github_search_item, 250)
  )
)

test_fixtures$github_search_response_large <- list(
  "total_count" = 1001,
  "incomplete_results" = FALSE,
  "items" = list(
    rep(github_search_item, 1001)
  )
)

test_fixtures$gitlab_files_tree_response <- list(
  "data" = list(
    "project" = list(
      "id" = "123456",
      "repository" = list(
        "tree" = list(
          "trees" = list(
            "nodes" = list(
              list(
                "name" = "R"
              ),
              list(
                "name" = "tests"
              )
            )
          ),
          "blobs" = list(
            "nodes" = list(
              list(
                "name" = "DESCRIPTION"
              ),
              list(
                "name" = "README.md"
              ),
              list(
                "name" = "project_metadata.yaml"
              ),
              list(
                "name" = "report.md"
              )
            )
          )
        )
      )
    )
  )
)

test_fixtures$github_files_tree_response <- list(
  "data" = list(
    "repository" = list(
      "id" = "R_kgD0Ivtxsg",
      "name" = "TestRepo",
      "url" = "https://github.com/test_org/TestRepo",
      "object" = list(
        "entries" = list(
          list(
            "name" = ".Rbuildignore",
            "type" = "blob"
          ),
          list(
            "name" = ".Renviron",
            "type" = "blob"
          ),
          list(
            "name" = ".Rprofile",
            "type" = "blob"
          ),
          list(
            "name" = ".covrignore",
            "type" = "blob"
          ),
          list(
            "name" = "test_file.R",
            "type" = "blob"
          ),
          list(
            "name" = "test.md",
            "type" = "blob"
          ),
          list(
            "name" = "test.qmd",
            "type" = "blob"
          ),
          list(
            "name" = "test.Rmd",
            "type" = "blob"
          ),
          list(
            "name" = "renv",
            "type" = "tree"
          ),
          list(
            "name" = "tests",
            "type" = "tree"
          ),
          list(
            "name" = "vignettes",
            "type" = "tree"
          )
        )
      )
    )
  )
)

test_fixtures$gitlab_file_rest_response <- list(
  "file_name" = "test.R",
  "file_path" = "test.R",
  "size" = 19L,
  "encoding" = "base64",
  "content" = "test content",
  "ref" = "main",
  "blob_id" = "1a2b3c",
  "commit_id" = "e5f6d7",
  "last_commit_id" = "e5f6d7"
)

EngineRestGitHub <- R6::R6Class(
  classname = "EngineRestGitHub",
  inherit = EngineRest,
  public = list(

    get_files = function(file_paths, verbose = TRUE) {
      files_list <- list()
      for (filename in file_paths) {
        search_file_endpoint <- paste0(private$endpoints[["search"]], "filename:", filename)
        total_n <- self$response(
          endpoint = search_file_endpoint,
          verbose = verbose
        )[["total_count"]]
        if (length(total_n) > 0) {
          search_result <- private$search_response(
            search_endpoint = search_file_endpoint,
            total_n = total_n,
            verbose = verbose
          ) |>
            purrr::keep(~ .$path == filename)
          files_content <- private$get_files_content(search_result, filename)
          files_list <- append(files_list, files_content)
        }
      }
      return(files_list)
    },

    prepare_files_table = function(files_list) {
      files_table <- NULL
      if (!is.null(files_list)) {
        files_table <- purrr::map(files_list, function(file_data) {
          repo_fullname <- private$get_repo_fullname(file_data$url)
          org_repo <- stringr::str_split_1(repo_fullname, "/")
          data.frame(
            "repo_id" = NA_character_,
            "repo_name" = org_repo[2],
            "organization" = org_repo[1],
            "file_path" = file_data$path,
            "file_content" = file_data$content,
            "file_size" = file_data$size,
            "repo_url" = private$set_repo_url(file_data$url)
          )
        }) |>
          purrr::list_rbind()
      }
      return(files_table)
    },

    search_for_code = function(code,
                               org = NULL,
                               filename = NULL,
                               in_path = FALSE,
                               language = NULL,
                               verbose = TRUE) {
      query <- private$build_search_query(
        code = code, org = org, filename = filename,
        in_path = in_path, language = language
      )
      search_endpoint <- paste0(private$endpoints[["search"]], query)
      if (verbose) cli::cli_alert("Searching for code [{code}]...")
      total_n <- self$response(
        endpoint = search_endpoint,
        verbose = verbose
      )[["total_count"]]
      search_result <- if (length(total_n) > 0) {
        private$search_response(
          search_endpoint = search_endpoint,
          total_n = total_n,
          verbose = verbose
        )
      } else {
        list()
      }
      return(search_result)
    },

    search_repos_for_code = function(code,
                                     repos,
                                     filename = NULL,
                                     in_path = FALSE,
                                     language = NULL,
                                     verbose = TRUE) {
      if (verbose) cli::cli_alert("Searching for code [{code}]...")
      search_result <- purrr::map(repos, function(repo) {
        query <- private$build_search_query(
          code = code, repo = repo, filename = filename,
          in_path = in_path, language = language
        )
        search_endpoint <- paste0(private$endpoints[["search"]], query)
        total_n <- self$response(
          endpoint = search_endpoint,
          verbose = verbose
        )[["total_count"]]
        result <- if (length(total_n) > 0) {
          private$search_response(
            search_endpoint = search_endpoint,
            total_n = total_n,
            verbose = verbose
          )
        } else {
          list()
        }
        return(result)
      }) |>
        purrr::list_flatten()
      return(search_result)
    },

    get_repos_urls = function(type, org, repos, verbose = TRUE) {
      owner_type <- attr(org, "type") %||% "organization"
      if (owner_type == "user") {
        repo_endpoint <- paste0(private$endpoints[["users"]], url_decode(org), "/repos")
      } else {
        repo_endpoint <- paste0(private$endpoints[["organizations"]], url_decode(org), "/repos")
      }
      repos_response <- private$paginate_results(
        endpoint = repo_endpoint,
        verbose = verbose
      )
      if (!is.null(repos)) {
        repos_response <- repos_response |>
          purrr::keep(~ .$name %in% repos)
      }
      repos_urls <- repos_response |>
        purrr::map_vec(function(repository) {
          if (type == "api") {
            repository$url
          } else {
            repository$html_url
          }
        })
      return(repos_urls)
    },

    get_files_tree = function(org, repo, def_branch, pattern, depth, verbose) {
      endpoint <- paste0(
        private$endpoints[["repositories"]],
        org, "/", repo,
        "/git/trees/", def_branch,
        "?recursive=1"
      )
      response <- tryCatch(
        self$response(endpoint = endpoint, verbose = verbose),
        error = function(e) NULL
      )
      if (is.null(response) || inherits(response, "rest_api_error")) {
        return(NULL)
      }
      tree <- response$tree
      files <- purrr::keep(tree, ~ .$type == "blob") |>
        purrr::map_chr(~ .$path)
      if (!is.null(depth) && depth < Inf) {
        files <- purrr::keep(files, function(path) {
          stringr::str_count(path, "/") < depth
        })
      }
      if (!is.null(pattern)) {
        files <- purrr::keep(files, function(path) {
          any(stringr::str_detect(path, pattern))
        })
      }
      if (length(files) == 0) return(NULL)
      return(files)
    },

    get_repos_contributors = function(repos_table, verbose = TRUE, progress) {
      if (nrow(repos_table) > 0) {
        repo_iterator <- paste0(repos_table$organization, "/", repos_table$repo_name)
        user_name <- rlang::expr(.$login)
        repos_table$contributors <- gitstats_map_chr(repo_iterator, function(repos_id) {
          tryCatch({
            contributors_endpoint <- paste0(private$endpoints[["repositories"]], repos_id, "/contributors")
            contributors_vec <- private$get_contributors_from_repo(
              contributors_endpoint = contributors_endpoint,
              user_name = user_name,
              verbose = verbose
            )
            return(contributors_vec)
          },
          error = function(e) {
            NA
          })
        }, .progress = progress)
      }
      return(repos_table)
    }
  ),
  private = list(

    endpoints = list(
      search = NULL,
      organizations = NULL,
      repositories = NULL
    ),

    build_search_query = function(code, org = NULL, repo = NULL,
                                  filename = NULL, in_path = FALSE,
                                  language = NULL) {
      query <- utils::URLencode(code)
      if (!is.null(repo)) query <- paste0(query, '+repo:', repo)
      if (!is.null(org)) query <- paste0(query, '+user:', org)
      if (in_path) query <- paste0(query, '+in:path')
      if (!is.null(filename)) query <- paste0(query, '+in:file+filename:', filename)
      if (!is.null(language)) query <- paste0(query, '+language:', language)
      return(query)
    },

    set_endpoints = function() {
      private$endpoints[["search"]] <- paste0(
        self$rest_api_url,
        '/search/code?q='
      )
      private$endpoints[["organizations"]] <- paste0(
        self$rest_api_url,
        "/orgs/"
      )
      private$endpoints[["users"]] <- paste0(
        self$rest_api_url,
        "/users/"
      )
      private$endpoints[["repositories"]] <- paste0(
        self$rest_api_url,
        "/repos/"
      )
    },

    # @param byte_max According to GitHub documentation only files smaller than
    #   384 KB are searchable. See
    #   \link{https://docs.github.com/en/rest/search?apiVersion=2022-11-28#search-code}
    search_response = function(search_endpoint,
                               total_n,
                               byte_max = "384000",
                               verbose = TRUE) {
      if (total_n >= 0 & total_n < 1e3) {
        page_iterator <- 1:(total_n %/% 100)
        items_list <- purrr::map(page_iterator, function(page) {
          response <- self$response(
            endpoint = paste0(
              search_endpoint,
              "+size:0..",
              byte_max,
              "&page=",
              page,
              "&per_page=100"
            ),
            verbose = verbose
          )
          return(response[["items"]])
        }) |>
          purrr::list_flatten()
        return(items_list)
      } else if (total_n >= 1e3) {
        items_list <- list()
        index <- c(0, 50)
        spinner <- cli::make_spinner(
          which = "timeTravel",
          template = cli::col_grey(
            "GitHub search limit (1000 results) exceeded. Results will be divided. {spin}"
          )
        )
        while (index[2] < as.numeric(byte_max)) {
          size_formula <- paste0("+size:", as.character(index[1]), "..", as.character(index[2]))
          spinner$spin()
          n_count <- tryCatch({
            self$response(
              endpoint = paste0(search_endpoint, size_formula),
              verbose = verbose
            )[["total_count"]]
          }, error = function(e) {
            NULL
          })
          if (is.null(n_count)) {
            NULL
          } else if ((n_count - 1) %/% 100 > 0) {
            total_pages <- 1:(n_count %/% 100) + 1
            items_part_list <- purrr::map(total_pages, function(page) {
              self$response(
                endpoint = paste0(search_endpoint,
                                  size_formula,
                                  "&page=",
                                  page,
                                  "&per_page=100"),
                verbose = verbose
              )[["items"]]
            }) |>
              purrr::list_flatten()
            items_list <- append(items_list, items_part_list)
          } else if ((n_count - 1) %/% 100 == 0) {
            response <- self$response(
              endpoint = paste0(search_endpoint,
                                size_formula,
                                "&page=1&per_page=100"),
              verbose = verbose
            )
            items_list <- append(items_list, response[["items"]])
          }
          index[1] <- index[2]
          if (index[2] < 1e3) {
            index[2] <- index[2] + 50
          }
          if (index[2] >= 1e3 && index[2] < 1e4) {
            index[2] <- index[2] + 100
          }
          if (index[2] >= 1e4 && index[2] < 1e5) {
            index[2] <- index[2] + 1000
          }
          if (index[2] >= 1e5 && index[2] < 1e6) {
            index[2] <- index[2] + 10000
          }
        }
        spinner$finish()
        return(items_list)
      }
    },

    get_files_content = function(search_result, filename) {
      purrr::map(search_result, ~ self$response(.$url),
                 .progress = glue::glue("Adding file [{filename}] info...")) |>
        unique()
    },

    get_repo_fullname = function(file_url) {
      stringr::str_remove_all(file_url,
                              paste0(private$endpoints$repositories, "/")) |>
        stringr::str_replace_all("/contents.*", "")
    }
  )
)

test_that("get_files_content works", {
  mockery::stub(
    test_rest_github_priv$get_files_content,
    "self$response",
    test_fixtures$github_file_rest_response
  )
  github_files_content <- test_rest_github_priv$get_files_content(
    search_result = test_fixtures$github_search_response,
    filename = "test.R"
  )
  expect_type(github_files_content, "list")
  test_mocker$cache(github_files_content)
})

test_that("get_files works as expected", {
  mockery::stub(
    test_rest_github$get_files,
    "private$get_files_content",
    test_mocker$use("github_files_content")
  )
  mockery::stub(
    test_rest_github$get_files,
    "private$search_response",
    test_fixtures$github_search_response$items
  )
  mockery::stub(
    test_rest_github$get_files,
    "self$response",
    test_fixtures$github_search_response
  )
  files <- test_rest_github$get_files(
    file_paths = "examples/test1.R",
    verbose = TRUE
  )
  expect_type(files, "list")
  expect_length(files, 1)
})

#' @importFrom stringr str_length str_replace

#' @noRd
#' @description Transform R date object into git time stamp
#' @param date A date.
#' @return A character (git time stamp format).
date_to_gts <- function(date) {
  date_format <- lubridate::as_date(date)
  posixt_format <- lubridate::as_datetime(date)
  if (lubridate::as_datetime(date_format) == lubridate::as_datetime(posixt_format)) {
    paste0(date, "T00:00:00Z")
  } else {
    stringr::str_replace(
      date,
      "^(.{10})(.*)$",
      "\\1T\\2"
    ) |>
      stringr::str_replace(" ", "") |>
      stringr::str_replace(
        "^(.{19})(.*)$",
        "\\1Z"
      )
  }
}

#' @noRd
#' @description Transform git time stamp format into Posixt
#' @param date_vector A character vector of dates obtained from Git API.
#' @return A vector of dates in Posixt format.
gts_to_posixt <- function(date_vector) {
  date_vector <- gsub("T", " ", gsub("Z", "", date_vector))
  lubridate::as_datetime(date_vector)
}

#' @noRd
retrieve_githost <- function(api_url) {
  stringr::str_remove_all(
    string = api_url,
    pattern = "(?<=com).*|(https://)|(api.)|(.com)"
  )
}

#' @noRd
standardize_dates <- function(dates) {
  purrr::discard(dates, is.null) |>
    purrr::map_vec(lubridate::as_datetime)
}

#' @noRd
#' @description Apply url encoding to string
url_encode <- function(url) {
  utils::URLencode(url, reserved = TRUE)
}

#' @noRd
#' @description Apply url decoding to string
url_decode <- function(url) {
  utils::URLdecode(url)
}

#' @noRd
is_name <- function(author) {
  length(stringr::str_split_1(author, " ")) > 1
}

#' @noRd
is_login <- function(author) {
  length(stringr::str_split_1(author, " ")) == 1 && identical(author, tolower(author))
}

get_gitlab_repo_id <- function(gitlab_repo_id) {
  sub(".*/(\\d+)$", "\\1", gitlab_repo_id)
}

set_progress_bar <- function(progress, private) {
  if (progress) {
    glue::glue("{private$host_name}")
  } else {
    FALSE
  }
}

#' @noRd
#' @description Check whether mirai daemons are active.
mirai_active <- function() {
  tryCatch(
    mirai::daemons_set(),
    error = function(e) FALSE
  )
}

#' @noRd
#' @description Parallel-aware map over elements. Uses mirai::mirai_map when
#'   daemons are active (via set_parallel), otherwise falls back to purrr::map.
gitstats_map <- function(.x, .f, ..., .progress = FALSE) {
  if (!mirai_active()) {
    purrr::map(.x, .f, ..., .progress = .progress)
  } else {
    if (!isFALSE(.progress)) {
      mirai::mirai_map(.x, .f, ...)[.progress]
    } else {
      mirai::mirai_map(.x, .f, ...)[]
    }
  }
}

#' @noRd
gitstats_map2 <- function(.x, .y, .f, ...) {
  if (!mirai_active()) {
    purrr::map2(.x, .y, .f, ...)
  } else {
    # mirai_map has no map2 variant. Zip inputs and use indexed access.
    # .f is passed as a named arg so the worker can resolve it.
    pairs <- purrr::map2(.x, .y, function(a, b) list(a = a, b = b))
    mirai::mirai_map(pairs, function(pair, .fn) .fn(pair$a, pair$b),
                     .fn = .f)[]
  }
}

#' @noRd
gitstats_map_chr <- function(.x, .f, ..., .progress = FALSE) {
  if (!mirai_active()) {
    purrr::map_chr(.x, .f, ..., .progress = .progress)
  } else {
    result <- if (!isFALSE(.progress)) {
      mirai::mirai_map(.x, .f, ...)[.progress]
    } else {
      mirai::mirai_map(.x, .f, ...)[]
    }
    purrr::map_chr(result, identity)
  }
}

parse_until_param <- function(until) {
  lubridate::as_datetime(until) + lubridate::days(1)
}

check_if_package_installed <- function(package) {
  if (!requireNamespace(package, quietly = TRUE)) {
    cli::cli_abort(
      "Package {.pkg {package}} is required but not installed. Install it with {.code install.packages(\"{package}\")}."
    )
  }
}

test_that("when dates format passed `date_to_gts()` prepares git time stamps with zeros", {
  dates <- as.Date(c("2022-12-19", "2023-01-15", "2020-06-05"))
  gts <- c("2022-12-19T00:00:00Z", "2023-01-15T00:00:00Z", "2020-06-05T00:00:00Z")

  purrr::walk2(dates, gts, ~ expect_equal(date_to_gts(.x), .y))
})

test_that("when posixt format passed `date_to_gts()` prepares git time stamps with precision to seconds", {
  posixt <- as.POSIXct(c("2022-12-19 12:27:15", "2022-12-19 12:21:55", "2022-12-19 11:55:39"))
  gts <- c("2022-12-19T12:27:15Z", "2022-12-19T12:21:55Z", "2022-12-19T11:55:39Z")

  purrr::walk2(posixt, gts, ~ expect_equal(date_to_gts(.x), .y))
})

test_that("`gts_to_posixt()` transforms git timestamp format into posixt format", {
  gts <- c("2022-12-19T12:27:15Z", "2022-12-19T12:21:55Z", "2022-12-19T11:55:39Z")
  posixt <- lubridate::as_datetime(c("2022-12-19 12:27:15", "2022-12-19 12:21:55", "2022-12-19 11:55:39"))

  expect_equal(gts_to_posixt(gts), posixt)
})

test_that("retrieve_githost works", {
  expect_equal(
    retrieve_githost("https://api.github.com/repositories/r-world-devs/GitStats"),
    "github"
  )
  expect_equal(
    retrieve_githost("https://gitlab.com/api/v4"),
    "gitlab"
  )
})

test_that("parse_until_param returns date + 1 day", {
  expect_equal(
    parse_until_param("2025-12-08"),
    lubridate::as_datetime("2025-12-09")
  )
})

test_that("url_encode encodes reserved characters", {
  expect_equal(url_encode("my/path"), "my%2Fpath")
  expect_equal(url_encode("a b"), "a%20b")
})

test_that("url_decode decodes encoded strings", {
  expect_equal(url_decode("my%2Fpath"), "my/path")
  expect_equal(url_decode("a%20b"), "a b")
})

test_that("`standardize_dates` converts to datetime and drops NULLs", {
  dates <- list("2024-01-15", NULL, "2024-06-01")
  result <- standardize_dates(dates)
  expect_equal(
    result,
    c(lubridate::as_datetime("2024-01-15"), lubridate::as_datetime("2024-06-01"))
  )
})

test_that("`standardize_dates` returns empty vector for all NULLs", {
  result <- standardize_dates(list(NULL, NULL))
  expect_length(result, 0)
})

test_that("`is_name` returns TRUE for multi-word strings", {
  expect_true(is_name("John Doe"))
  expect_true(is_name("Jane Mary Smith"))
})

test_that("`is_name` returns FALSE for single-word strings", {
  expect_false(is_name("johndoe"))
})

test_that("`is_login` returns TRUE for single lowercase word", {
  expect_true(is_login("johndoe"))
  expect_true(is_login("user123"))
})

test_that("`is_login` returns FALSE for multi-word or uppercase strings", {
  expect_false(is_login("John Doe"))
  expect_false(is_login("JohnDoe"))
})

test_that("`get_gitlab_repo_id` extracts numeric id from URL", {
  expect_equal(get_gitlab_repo_id("https://gitlab.com/api/v4/projects/12345"), "12345")
  expect_equal(get_gitlab_repo_id("projects/99"), "99")
})

test_that("`set_progress_bar` returns host name when progress is TRUE", {
  mock_private <- list(host_name = "GitHub")
  result <- set_progress_bar(TRUE, mock_private)
  expect_equal(as.character(result), "GitHub")
})

test_that("`set_progress_bar` returns FALSE when progress is FALSE", {
  mock_private <- list(host_name = "GitHub")
  result <- set_progress_bar(FALSE, mock_private)
  expect_false(result)
})

test_that("`check_if_package_installed` does not error for installed package", {
  expect_no_error(check_if_package_installed("testthat"))
})

test_that("`check_if_package_installed` errors for missing package", {
  expect_error(
    check_if_package_installed("nonexistent_pkg_xyz"),
    "nonexistent_pkg_xyz"
  )
})

test_that("`show_message` prints message with scope", {
  expect_message(
    show_message(
      host = "GitHub",
      engine = "graphql",
      scope = "r-world-devs",
      information = "Pulling repos"
    ),
    "GitHub"
  )
})

test_that("`show_message` prints message without scope", {
  expect_message(
    show_message(
      host = "GitLab",
      engine = "rest",
      information = "Pulling commits"
    ),
    "GitLab"
  )
})

test_that("`show_message` handles 'both' engine type", {
  expect_message(
    show_message(
      host = "GitHub",
      engine = "both",
      information = "Pulling data"
    ),
    "GitHub"
  )
})

test_that("`cut_item_to_print` returns all items when fewer than 10", {
  items <- c("a", "b", "c")
  result <- cut_item_to_print(items)
  expect_equal(result, "a, b, c")
})

test_that("`cut_item_to_print` truncates items when 10 or more", {
  items <- letters[1:15]
  result <- cut_item_to_print(items)
  expect_match(result, "and 5 more")
  expect_match(result, "^a, b, c")
})

test_that("`set_repo_scope` formats scope string", {
  mock_private <- list(orgs_repos = list("my-org" = c("repo1", "repo2")))
  result <- set_repo_scope("my-org", mock_private)
  expect_equal(result, "my-org: 2 repos")
})

# ---- mirai_active ----

test_that("`mirai_active` returns FALSE when no daemons are set", {
  mirai::daemons(0)
  expect_false(mirai_active())
})

# ---- gitstats_map (sequential fallback) ----

test_that("`gitstats_map` falls back to purrr::map when no daemons are set", {
  mirai::daemons(0)
  result <- gitstats_map(1:3, function(x) x * 2)
  expect_equal(result, list(2L, 4L, 6L))
})

test_that("`gitstats_map` passes .progress to purrr::map in sequential mode", {
  mirai::daemons(0)
  result <- gitstats_map(1:3, function(x) x + 1, .progress = FALSE)
  expect_equal(result, list(2L, 3L, 4L))
})

# ---- gitstats_map2 (sequential fallback) ----

test_that("`gitstats_map2` falls back to purrr::map2 when no daemons are set", {
  mirai::daemons(0)
  result <- gitstats_map2(1:3, 4:6, function(a, b) a + b)
  expect_equal(result, list(5L, 7L, 9L))
})

# ---- gitstats_map_chr (sequential fallback) ----

test_that("`gitstats_map_chr` falls back to purrr::map_chr when no daemons are set", {
  mirai::daemons(0)
  result <- gitstats_map_chr(c("a", "b", "c"), function(x) toupper(x))
  expect_equal(result, c("A", "B", "C"))
})

# ---- gitstats_map with .progress = TRUE (sequential) ----

test_that("`gitstats_map` works with character .progress in sequential mode", {
  mirai::daemons(0)
  result <- gitstats_map(1:3, function(x) x * 10, .progress = "GitHub")
  expect_equal(result, list(10L, 20L, 30L))
})

# ---- gitstats_map_chr with .progress (sequential) ----

test_that("`gitstats_map_chr` works with .progress = TRUE in sequential mode", {
  mirai::daemons(0)
  result <- gitstats_map_chr(1:3, function(x) as.character(x), .progress = TRUE)
  expect_equal(result, c("1", "2", "3"))
})

# ---- gitstats_map with empty input (sequential) ----

test_that("`gitstats_map` handles empty input in sequential mode", {
  mirai::daemons(0)
  result <- gitstats_map(list(), function(x) x)
  expect_equal(result, list())
})

test_that("`gitstats_map2` handles empty input in sequential mode", {
  mirai::daemons(0)
  result <- gitstats_map2(list(), list(), function(a, b) a + b)
  expect_equal(result, list())
})

test_that("`gitstats_map_chr` handles empty input in sequential mode", {
  mirai::daemons(0)
  result <- gitstats_map_chr(character(0), function(x) x)
  expect_equal(result, character(0))
})

# ---- set_parallel / is_parallel input validation ----

test_that("`set_parallel` errors when called without a GitStats object", {
  expect_error(set_parallel(4), "GitStats")
  expect_error(set_parallel(TRUE), "GitStats")
  expect_error(set_parallel(FALSE), "GitStats")
  expect_error(set_parallel("not_gitstats", 4), "GitStats")
  expect_error(set_parallel(list(), 2), "GitStats")
})

test_that("`is_parallel` errors when called without a GitStats object", {
  expect_error(is_parallel(), "GitStats")
  expect_error(is_parallel("not_gitstats"), "GitStats")
  expect_error(is_parallel(42), "GitStats")
})

# ---- set_parallel ----

test_that("`set_parallel` enables and disables daemons", {
  skip_if(Sys.getenv("R_COVR") == "true", "mirai daemons conflict with covr tracing")
  withr::defer(mirai::daemons(0))
  gs <- create_gitstats()

  expect_message(set_parallel(gs, 2), "Parallel processing enabled with 2 workers")
  expect_true(is_parallel(gs))

  expect_message(set_parallel(gs, FALSE), "Parallel processing disabled")
  expect_false(is_parallel(gs))

  expect_message(set_parallel(gs, 0), "Parallel processing disabled")
  expect_false(is_parallel(gs))

  expect_message(set_parallel(gs, 0L), "Parallel processing disabled")
  expect_false(is_parallel(gs))
})

test_that("`set_parallel(TRUE)` auto-detects workers", {
  skip_if(Sys.getenv("R_COVR") == "true", "mirai daemons conflict with covr tracing")
  withr::defer(mirai::daemons(0))
  gs <- create_gitstats()
  expect_message(set_parallel(gs, TRUE), "Parallel processing enabled with \\d+ workers")
  expect_true(is_parallel(gs))
})

# ---- set_parallel (mocked, runs under covr) ----

test_that("`set_parallel` disable path executes under covr", {
  gs <- create_gitstats()
  mock_daemons <- mockery::mock("status_off")
  mockery::stub(gs$set_parallel, "mirai::daemons", mock_daemons)

  gs$set_parallel(FALSE)
  mockery::expect_called(mock_daemons, 1)
  expect_equal(mockery::mock_args(mock_daemons)[[1]], list(0))
  expect_false(gs$is_parallel())
})

test_that("`set_parallel(0)` disable path executes under covr", {
  gs <- create_gitstats()
  mock_daemons <- mockery::mock("status_off")
  mockery::stub(gs$set_parallel, "mirai::daemons", mock_daemons)

  gs$set_parallel(0)
  mockery::expect_called(mock_daemons, 1)
  expect_false(gs$is_parallel())
})

test_that("`set_parallel(0L)` disable path executes under covr", {
  gs <- create_gitstats()
  mock_daemons <- mockery::mock("status_off")
  mockery::stub(gs$set_parallel, "mirai::daemons", mock_daemons)

  gs$set_parallel(0L)
  mockery::expect_called(mock_daemons, 1)
  expect_false(gs$is_parallel())
})

test_that("`set_parallel` enable path with explicit workers executes under covr", {
  gs <- create_gitstats()
  mock_daemons <- mockery::mock("status_on")
  mockery::stub(gs$set_parallel, "mirai::daemons", mock_daemons)
  mockery::stub(gs$set_parallel, "do.call", NULL)

  gs$set_parallel(4)
  mockery::expect_called(mock_daemons, 1)
  expect_equal(mockery::mock_args(mock_daemons)[[1]], list(4))
})

test_that("`set_parallel(TRUE)` auto-detects and falls back to 2 when cores < 2", {
  gs <- create_gitstats()
  mock_daemons <- mockery::mock("status_on")
  mockery::stub(gs$set_parallel, "mirai::daemons", mock_daemons)
  mockery::stub(gs$set_parallel, "parallel::detectCores", 1L)
  mockery::stub(gs$set_parallel, "do.call", NULL)

  gs$set_parallel(TRUE)
  # Should have been called with 2 (the fallback)
  expect_equal(mockery::mock_args(mock_daemons)[[1]], list(2))
})

test_that("`set_parallel(TRUE)` falls back to 2 when detectCores returns NA", {
  gs <- create_gitstats()
  mock_daemons <- mockery::mock("status_on")
  mockery::stub(gs$set_parallel, "mirai::daemons", mock_daemons)
  mockery::stub(gs$set_parallel, "parallel::detectCores", NA_integer_)
  mockery::stub(gs$set_parallel, "do.call", NULL)

  gs$set_parallel(TRUE)
  expect_equal(mockery::mock_args(mock_daemons)[[1]], list(2))
})

test_that("`set_parallel(TRUE)` uses detected cores when >= 2", {
  gs <- create_gitstats()
  mock_daemons <- mockery::mock("status_on")
  mockery::stub(gs$set_parallel, "mirai::daemons", mock_daemons)
  mockery::stub(gs$set_parallel, "parallel::detectCores", 8L)
  mockery::stub(gs$set_parallel, "do.call", NULL)

  gs$set_parallel(TRUE)
  expect_equal(mockery::mock_args(mock_daemons)[[1]], list(8))
})

test_that("`set_parallel` returns GitStats object invisibly", {
  gs <- create_gitstats()
  mock_daemons <- mockery::mock("status", "status")
  mockery::stub(gs$set_parallel, "mirai::daemons", mock_daemons)

  result <- set_parallel(gs, FALSE)
  expect_invisible(set_parallel(gs, FALSE))
  expect_s3_class(result, "GitStats")
})

test_that("`set_parallel` can be used in a pipeline", {
  skip_if(Sys.getenv("R_COVR") == "true", "mirai daemons conflict with covr tracing")
  withr::defer(mirai::daemons(0))

  gs <- create_gitstats() |>
    set_parallel(2)
  expect_true(is_parallel(gs))
  expect_s3_class(gs, "GitStats")
})

# ---- is_parallel ----

test_that("`is_parallel` returns FALSE for a fresh GitStats object", {
  gs <- create_gitstats()
  expect_false(is_parallel(gs))
})

test_that("`is_parallel` reflects parallel state per object", {
  skip_if(Sys.getenv("R_COVR") == "true", "mirai daemons conflict with covr tracing")
  withr::defer(mirai::daemons(0))

  gs1 <- create_gitstats()
  gs2 <- create_gitstats()

  set_parallel(gs1, 2)
  # gs1 was set to parallel, gs2 was not
  expect_true(is_parallel(gs1))
  expect_false(is_parallel(gs2))
})

# ---- parallel execution paths (single daemon startup) ----

test_that("gitstats_map/map_chr work via mirai_map when daemons are active", {
  skip_if(Sys.getenv("R_COVR") == "true", "mirai daemons conflict with covr tracing")
  withr::defer(mirai::daemons(0))
  gs <- create_gitstats()
  set_parallel(gs, 2)

  # gitstats_map without progress
  result <- gitstats_map(1:3, function(x) x * 2)
  expect_equal(result, list(2L, 4L, 6L))

  # gitstats_map with progress
  result <- gitstats_map(1:3, function(x) x + 1, .progress = TRUE)
  expect_equal(result, list(2L, 3L, 4L))

  # gitstats_map_chr without progress
  result <- gitstats_map_chr(c("a", "b"), function(x) toupper(x))
  expect_equal(result, c("A", "B"))

  # gitstats_map_chr with progress
  result <- gitstats_map_chr(c("x", "y"), function(x) toupper(x), .progress = TRUE)
  expect_equal(result, c("X", "Y"))
})

test_fixtures$github_repository_rest_response <- list(
  "id" = 627452680,
  "node_id" = "R_kgDOJWYrCA",
  "name" = "testRepo",
  "full_name" = "test_org/TestRepo",
  "private" = FALSE,
  "owner" = list(
    "login" = "test_org",
    "id" = 103638913,
    "node_id" = "O_kgDOBi1ngQ",
    "avatar_url" = "https://avatars.githubusercontent.com/u/103638913?v=4"
  ),
  "html_url" = "https://testhost.com/test-org/TestRepo",
  "description" = NULL,
  "fork" = FALSE,
  "url" = "https://testhost.com/api/v4/repos/test-org/TestRepo",
  "forks_url" = "https://testhost.com/api/v4/repos/test-org/TestRepo/forks",
  "collaborators_url" = "https://testhost.com/api/v4/repos/test-org/TestRepo/collaborators{/collaborator}",
  "teams_url" = "https://testhost.com/api/v4/repos/test-org/TestRepo/teams",
  "events_url" = "https://testhost.com/api/v4/repos/test-org/TestRepo/events",
  "branches_url" = "https://testhost.com/api/v4/repos/test-org/TestRepo/branches{/branch}",
  "created_at" = "2023-04-13T13:52:24Z",
  "pushed_at" = "2023-12-21T20:36:23Z",
  "size" = 211L,
  "stargazers_count" = 2,
  "watchers_count" = 2,
  "language" = "R",
  "forks_count" = 0,
  "open_issues_count" = 3L
)

test_fixtures$github_repositories_rest_response <- list(
  test_fixtures$github_repository_rest_response,
  list(
    "id" = 2222222222222,
    "name" = "testRepo2",
    "html_url" = "https://testhost.com/test-org/TestRepo",
    "url" = "https://testhost.com/api/v4/repos/test-org/TestRepo"
  ),
  list(
    "id" = 2222222222222,
    "name" = "testRepo3",
    "html_url" = "https://testhost.com/test-org/TestRepo",
    "url" = "https://testhost.com/api/v4/repos/test-org/TestRepo"
  )
)


test_fixtures$gitlab_repositories_rest_response <- list(
  list(
    "id" = "1111",
    "name" = "test repo 1",
    "path" = "test_repo_1",
    "path_with_namespace" = "test_org/test_repo_1",
    "default_branch" = "main",
    "star_count" = 5L,
    "forks_count" = 2L,
    "created_at" = "2023-04-13T13:52:24Z",
    "last_activity_at" = "2025-04-09T17:04:00Z",
    "open_issues_count" = 10L,
    "path" = "testRepo1",
    "_links" = list("self" = "https://gitlab.com/api/v4/projects/43400864"),
    "web_url" = "https://gitlab.com/mbtests/gitstats-testing-2"
  ),
  list(
    "id" = "2222",
    "name" = "test repo 2",
    "path" = "test_repo_2",
    "path_with_namespace" = "test_org/test_repo_2",
    "default_branch" = "devel",
    "star_count" = 5L,
    "forks_count" = 2L,
    "created_at" = "2023-04-14T13:52:24Z",
    "last_activity_at" = "2025-04-09T17:04:00Z",
    "open_issues_count" = 10L,
    "path" = "testRepo2",
    "_links" = list("self" = "https://gitlab.com/api/v4/projects/43398933"),
    "web_url" = "https://gitlab.com/mbtests/gitstatstesting"
  ),
  list(
    "id" = "3333",
    "name" = "test repo 3",
    "path" = "test_repo_3",
    "path_with_namespace" = "test_org/test_repo_3",
    "default_branch" = "test",
    "star_count" = 5L,
    "forks_count" = 2L,
    "created_at" = "2023-05-23T13:52:24Z",
    "last_activity_at" = "2025-04-09T17:04:00Z",
    "open_issues_count" = 10L,
    "path" = "testRepo3",
    "_links" = list("self" = "https://gitlab.com/api/v4/projects/43398933"),
    "web_url" = "https://gitlab.com/mbtests/gitstatstesting"
  )
)

github_repository_node <- function(org_name, repo_name) {
  list(
    "repo_id" = "test_node_id",
    "repo_name" = repo_name,
    "repo_path" = repo_name,
    "repo_fullpath" = paste0(org_name, "/", repo_name),
    "default_branch" = list(
      "name" = "main"
    ),
    "stars" = 10,
    "forks" = 2,
    "created_at" = "2022-04-20T00:00:00Z",
    "last_activity_at" = "2023-04-20T00:00:00Z",
    "languages" = list(
      "nodes" = list(
        list(
          "name" = "R"
        ),
        list(
          "name" = "CSS"
        ),
        list(
          "name" = "JavaScript"
        )
      )
    ),
    "issues_open" = list(
      "totalCount" = 10
    ),
    "issues_closed" = list(
      "totalCount" = 5
    ),
    "organization" = list(
      "login" = "r-world-devs"
    ),
    "repo_url" = "https://github.test.com/api/test_url",
    "defaultBranchRef" = list(
      "target" = list(
        "oid" = "1a2b3c4e5f"
      )
    )
  )
}

test_fixtures$github_repos_by_ids_response <- list(
  "data" = list(
    "nodes" = list(
      github_repository_node("TestOrg", "TestRepo"),
      github_repository_node("TestOrg", "TestRepo1"),
      github_repository_node("TestOrg", "TestRepo2"),
      github_repository_node("TestOrg", "TestRepo3"),
      github_repository_node("TestOrg", "TestRepo4")
    )
  )
)

test_fixtures$github_repos_by_org_response <- list(
  "data" = list(
    "repositoryOwner" = list(
      "repositories" = list(
        "totalCount" = 5,
        "pageInfo" = list(
          "endCursor" = "xyx",
          "hasNextPage" = FALSE
        ),
        "nodes" = list(
          github_repository_node("r-world-devs", "GitStats"),
          github_repository_node("r-world-devs", "GitAI"),
          github_repository_node("r-world-devs", "cohortBuilder"),
          github_repository_node("r-world-devs", "shinyCohortBuilder"),
          github_repository_node("r-world-devs", "shinyGizmo")
        )
      )
    )
  )
)

test_fixtures$github_repos_by_user_response <- list(
  "data" = list(
    "user" = list(
      "repositories" = list(
        "totalCount" = 5,
        "pageInfo" = list(
          "endCursor" = "xyx",
          "hasNextPage" = FALSE
        ),
        "nodes" = list(
          github_repository_node("TestOrg", "TestRepo"),
          github_repository_node("TestOrg", "TestRepo1"),
          github_repository_node("TestOrg", "TestRepo2"),
          github_repository_node("TestOrg", "TestRepo3"),
          github_repository_node("TestOrg", "TestRepo4")
        )
      )
    )
  )
)

gitlab_project_node <- list(
  "node" = list(
    "repo_id" = "gid://gitlab/Project/61399846",
    "repo_name" = "gitstatstesting",
    "repo_path" = "gitstatstesting",
    "repo_fullpath" = "mbtests/gitstatstesting",
    "repository" = list(
      "rootRef" = "main",
      "lastCommit" = list(
        "sha" = "1a2bc3d4e5"
      )
    ),
    "stars" = 8,
    "forks" = 3,
    "created_at" = "2023-09-18T00:00:00Z",
    "last_activity_at" = "2024-09-18T00:00:00Z",
    "languages" = list(
      list(
        "name" = "Python"
      ),
      list(
        "name" = "R"
      )
    ),
    "issues" = list(
      "all" = 10,
      "closed" = 8,
      "opened" = 2
    ),
    "namespace" = list(
      "path" = "mbtests"
    ),
    "repo_url" = "https://gitlab.com/mbtests/gitstatstesting"
  )
)

gitlab_archived_project_node <- list(
  "node" = list(
    "repo_id" = "gid://gitlab/Project/99999999",
    "repo_name" = "archived-project",
    "repo_path" = "archived-project",
    "repo_fullpath" = "mbtests/archived-project",
    "repository" = list(
      "rootRef" = "main",
      "lastCommit" = NULL
    ),
    "stars" = 2,
    "forks" = 0,
    "created_at" = "2022-01-10T00:00:00Z",
    "last_activity_at" = "2023-06-15T00:00:00Z",
    "languages" = list(
      list(
        "name" = "R"
      )
    ),
    "issues" = list(
      "all" = 3,
      "closed" = 3,
      "opened" = 0
    ),
    "namespace" = list(
      "path" = "mbtests"
    ),
    "repo_url" = "https://gitlab.com/mbtests/archived-project"
  )
)

test_fixtures$gitlab_branch_response <- list(
  "name" = "main",
  "merged" = FALSE,
  "protected" = TRUE,
  "default" = TRUE,
  "commit" = list(
    "id" = "abcdef1234567890",
    "short_id" = "abcdef12",
    "created_at" = "2023-06-15T10:00:00Z",
    "title" = "Last commit on archived project",
    "message" = "Last commit on archived project",
    "author_name" = "Test User",
    "author_email" = "test@example.com"
  )
)

test_fixtures$gitlab_repos_by_org_with_archived_response <- list(
  "data" = list(
    "group" = list(
      "projects" = list(
        "count" = 2,
        "pageInfo" = list(
          "hasNextPage" = FALSE,
          "endCursor" = "xyz"
        ),
        "edges" = list(
          gitlab_project_node,
          gitlab_archived_project_node
        )
      )
    )
  )
)

test_fixtures$gitlab_repos_by_org_response <- list(
  "data" = list(
    "group" = list(
      "projects" = list(
        "count" = 5,
        "pageInfo" = list(
          "hasNextPage" = FALSE,
          "endCursor" = "xyz"
        ),
        "edges" = list(
          gitlab_project_node,
          gitlab_project_node,
          gitlab_project_node,
          gitlab_project_node,
          gitlab_project_node
        )
      )
    )
  )
)

test_fixtures$gitlab_repos_by_user_response <- list(
  "data" = list(
    "projects" = list(
      "count" = 5,
      "pageInfo" = list(
        "hasNextPage" = FALSE,
        "endCursor" = "xyz"
      ),
      "edges" = list(
        gitlab_project_node,
        gitlab_project_node,
        gitlab_project_node,
        gitlab_project_node,
        gitlab_project_node
      )
    )
  )
)

test_fixtures$gitlab_repo_by_fullpath_response <- list(
  "data" = list(
    "project" = gitlab_project_node$node
  )
)

test_fixtures$gitlab_repos_response_flawed <- list(
  "data" = list(
    "projects" = list(
      "count" = 5,
      "pageInfo" = list(
        "hasNextPage" = NULL,
        "endCursor" = "xyz"
      ),
      "edges" = NULL
    )
  )
)

test_fixtures$gitlab_languages_response <- list(
  "Python" = 100,
  "R" = 50
)

test_fixtures$github_user_login <- list(
  "data" = list(
    "user" = list(
      "__typename" = "User",
      "login" = "test_user"
    ),
    "organization" = NULL
  )
)

test_fixtures$github_org_login <- list(
  "data" = list(
    "user" = NULL,
    "organization" = list(
      "__typename" = "Organization",
      "login" = "test_org"
    )
  )
)

test_fixtures$github_contributor_response <- list(
  "login" = "test_login",
  "id" = 11111L,
  "node_id" = "xxxxx",
  "avatar_url" = "",
  "gravatar_id" = "",
  "url" = "",
  "html_url" = "",
  "followers_url" = ""
)

test_fixtures$github_contributors_response <- list(
  test_fixtures$github_contributor_response,
  test_fixtures$github_contributor_response,
  test_fixtures$github_contributor_response
)

#' @title Get users data
#' @name get_users
#' @param gitstats A GitStats object.
#' @param logins A character vector of logins.
#' @param cache A logical, if set to `TRUE` GitStats will retrieve the last
#'   result from its storage.
#' @param verbose A logical, `TRUE` by default. If `FALSE` messages and
#'   printing output is switched off.
#' @examples
#' \dontrun{
#'  my_gitstats <- create_gitstats() |>
#'   set_github_host(
#'     token = Sys.getenv("GITHUB_PAT"),
#'     orgs = c("r-world-devs")
#'   ) |>
#'   set_gitlab_host(
#'     token = Sys.getenv("GITLAB_PAT_PUBLIC"),
#'     orgs = "mbtests"
#'   )
#'  get_users(my_gitstats, c("maciekabanas", "marcinkowskak"))
#' }
#' @return A data.frame.
#' @export
get_users <- function(gitstats,
                      logins,
                      cache = TRUE,
                      verbose = FALSE) {
  start_time <- Sys.time()
  users <- gitstats$get_users(
    logins = logins,
    cache = cache,
    verbose = verbose
  )
  end_time <- Sys.time()
  time_taken <- end_time - start_time
  cli::cli_alert_success("Data pulled in {round(time_taken, 1)} {attr(time_taken, 'units')}")
  return(users)
}

#' @importFrom R6 R6Class
#' @importFrom cli cli_alert_info cli_alert_success cli_alert_warning col_yellow cli_alert_danger cli_abort col_green
#' @importFrom dplyr glimpse distinct mutate relocate filter
#' @importFrom lubridate floor_date
#' @importFrom mirai daemons mirai_map everywhere
#' @importFrom purrr map map_chr keep
#' @importFrom rlang expr %||%
#' @importFrom httr2 request req_headers req_perform resp_body_json
#' @importFrom glue glue
#' @importFrom stats runif
NULL

#' @title Get release logs
#' @name get_release_logs
#' @description Pull release logs from repositories.
#' @inheritParams get_commits
#' @return A data.frame.
#' @examples
#' \dontrun{
#'  my_gitstats <- create_gitstats() |>
#'   set_github_host(
#'     token = Sys.getenv("GITHUB_PAT"),
#'     orgs = c("r-world-devs", "openpharma")
#'   )
#'   get_release_logs(my_gistats, since = "2024-01-01")
#' }
#' @export
get_release_logs <- function(gitstats,
                             since = NULL,
                             until = Sys.Date(),
                             cache = TRUE,
                             verbose = FALSE,
                             progress = TRUE) {
  start_time <- Sys.time()
  if (is.null(since)) {
    cli::cli_abort(cli::col_red("You need to pass date to `since` parameter."), call = NULL)
  }
  release_logs <- gitstats$get_release_logs(
    since = since,
    until = until,
    cache = cache,
    verbose = verbose,
    progress = progress
  )
  end_time <- Sys.time()
  time_taken <- end_time - start_time
  cli::cli_alert_success("Data pulled in {round(time_taken, 1)} {attr(time_taken, 'units')}")
  return(release_logs)
}

Engine <- R6::R6Class("Engine",
  private = list(
    token = NULL,
    scan_all = FALSE,
    engine = NULL,
    verbose = TRUE,
    set_verbose = function(verbose) {
      private$verbose <- verbose
    }
  )
)

test_fixtures$github_user_response <- list(
  "data" = list(
    "user" = list(
      "id" = "xxx",
      "name" = "Mr Test",
      "login" = "testlogin",
      "email" = "",
      "location" = NULL,
      "starred_repos" = list(
        "totalCount" = 15L
      ),
      "contributions" = list(
        "totalCommitContributions" = 100L,
        "totalIssueContributions" = 50L,
        "totalPullRequestContributions" = 25L,
        "totalPullRequestReviewContributions" = 5L
      ),
      "avatar_url" = "",
      "web_url" = NULL
    )
  )
)

test_fixtures$gitlab_user_response <- list(
  "data" = list(
    "user" = list(
      "id" = "xxx",
      "name" = "Mr Test",
      "login" = "testlogin",
      "email" = "",
      "location" = NULL,
      "starred_repos" = list(
        "count" = 15L
      ),
      "pull_requests" = list(
        "count" = 2L
      ),
      "reviews" = list(
        "count" = 5L
      ),
      "avatar_url" = "",
      "web_url" = "https://gitlab.com/maciekbanas"
    )
  )
)

test_fixtures$gitlab_user_search_response <- list(
  list(
    "id" = "xxx",
    "username" = "maciekbanas",
    "name" = "Maciej Banas",
    "state" = "active",
    "locked" = FALSE,
    "avatar_url" = "",
    "web_url" = ""
  )
)

#' @title Set GitHub host
#' @name set_github_host
#' @param gitstats A GitStats object.
#' @param host A character, optional, URL name of the host. If not passed, a
#'   public host will be used.
#' @param token A token.
#' @param orgs An optional character vector of organisations
#' @param repos An optional character vector of repositories full names
#'   (organization and repository name, e.g. "r-world-devs/GitStats").
#' @param verbose A logical, `TRUE` by default. If `FALSE` messages and printing
#'   output is switched off.
#' @param .error A logical to control if passing wrong input
#'   (`repositories` and `organizations`) should end with an error or not.
#' @details If you do not define `orgs` and `repos`, `GitStats` will be set to
#'   scan whole Git platform (such as enterprise version of GitHub or GitLab),
#'   unless it is a public platform. In case of a public one (like GitHub) you
#'   need to define `orgs` or `repos` as scanning through all organizations may
#'   take large amount of time.
#' @return A `GitStats` object with added information on host.
#' @examples
#' \dontrun{
#' my_gitstats <- create_gitstats() |>
#'   set_github_host(
#'     orgs = c("r-world-devs", "openpharma", "pharmaverse")
#'   )
#' }
#' @export
set_github_host <- function(gitstats,
                            host = NULL,
                            token = NULL,
                            orgs = NULL,
                            repos = NULL,
                            verbose = is_verbose(gitstats),
                            .error = TRUE) {
  gitstats$set_github_host(
    host = host,
    token = token,
    orgs = orgs,
    repos = repos,
    verbose = verbose,
    .error = .error
  )

  return(invisible(gitstats))
}

#' @title Set GitLab host
#' @name set_gitlab_host
#' @inheritParams set_github_host
#' @details If you do not define `orgs` and `repos`, `GitStats` will be set to
#'   scan whole Git platform (such as enterprise version of GitHub or GitLab),
#'   unless it is a public platform. In case of a public one (like GitHub) you
#'   need to define `orgs` or `repos` as scanning through all organizations may
#'   take large amount of time.
#' @return A `GitStats` object with added information on host.
#' @examples
#' \dontrun{
#' my_gitstats <- create_gitstats() |>
#'   set_gitlab_host(
#'     token = Sys.getenv("GITLAB_PAT_PUBLIC"),
#'     orgs = "mbtests"
#'   )
#' }
#' @export
set_gitlab_host <- function(gitstats,
                            host = NULL,
                            token = NULL,
                            orgs = NULL,
                            repos = NULL,
                            verbose = is_verbose(gitstats),
                            .error = TRUE) {
  gitstats$set_gitlab_host(
    host = host,
    token = token,
    orgs = orgs,
    repos = repos,
    verbose = verbose,
    .error = .error
  )

  return(invisible(gitstats))
}

#' @title Get data on repositories
#' @name get_repos
#' @description  Pulls data on all repositories for an organization, individual
#'   user or those with a given text in code blobs (`with_code` parameter) or a
#'   file (`with_files` parameter) and parse it into table format.
#' @param gitstats A GitStats object.
#' @param add_contributors A logical parameter to decide whether to add
#'   information about repositories' contributors to the repositories output
#'   (table). If set to `FALSE` it makes function run faster as, in the case of
#'   `org` search mode, it reaches only `GraphQL` endpoint with a query on
#'   repositories, and in the case of `code` search mode it reaches only
#'   `repositories REST API` endpoint. However, the pitfall is that the result
#'   does not convey information on contributors. \cr\cr When set to `TRUE` (by
#'   default), `GitStats` iterates additionally over pulled repositories and
#'   reaches to the `contributors APIs`, which makes it slower, but gives
#'   additional information.
#' @param add_languages A logical, `TRUE` by default. If set to `FALSE`,
#'   languages data will not be included in the repositories output which
#'   may speed up the process for GitLab REST engine.
#' @param with_code A character vector, if defined, GitStats will pull
#'   repositories with specified code phrases in code blobs.
#' @param in_files A character vector of file names. Works when `with_code` is
#'   set - then it searches code blobs only in files passed to `in_files`
#'   parameter.
#' @param with_files A character vector, if defined, GitStats will pull
#'   repositories with specified files.
#' @param language A character. If defined, GitStats will return only
#'   repositories with given language.
#' @param cache A logical, if set to `TRUE` GitStats will retrieve the last
#'   result from its storage.
#' @param verbose A logical, `TRUE` by default. If `FALSE` messages and printing
#'   output is switched off.
#' @param progress A logical, by default set to `verbose` value. If `FALSE` no
#'   `cli` progress bar will be displayed.
#' @param fill_empty_sha A logical, `FALSE` by default. If `TRUE`, GitStats
#'   will try to fetch missing `commit_sha` values (e.g. for archived GitLab
#'   projects) via the REST Branches API. This may slow down the call for
#'   large numbers of repositories.
#' @return A data.frame.
#' @examples
#' \dontrun{
#' my_gitstats <- create_gitstats() |>
#'   set_github_host(
#'     token = Sys.getenv("GITHUB_PAT"),
#'     orgs = c("r-world-devs", "openpharma")
#'   ) |>
#'   set_gitlab_host(
#'     token = Sys.getenv("GITLAB_PAT_PUBLIC"),
#'     orgs = "mbtests"
#'   )
#' get_repos(my_gitstats)
#' get_repos(my_gitstats, add_contributors = FALSE)
#' get_repos(my_gitstats, add_languages = FALSE)
#' get_repos(my_gitstats, with_code = "Shiny", in_files = "renv.lock")
#' get_repos(my_gitstats, with_files = "DESCRIPTION")
#' }
#' @export
get_repos <- function(gitstats,
                      add_contributors = TRUE,
                      add_languages = TRUE,
                      with_code = NULL,
                      in_files = NULL,
                      with_files = NULL,
                      language = NULL,
                      cache = TRUE,
                      verbose = FALSE,
                      progress = TRUE,
                      fill_empty_sha = FALSE) {
  start_time <- Sys.time()
  repos <- gitstats$get_repos(
    add_contributors = add_contributors,
    add_languages = add_languages,
    with_code = with_code,
    in_files = in_files,
    with_files = with_files,
    language = language,
    cache = cache,
    verbose = verbose,
    progress = progress,
    fill_empty_sha = fill_empty_sha
  )
  end_time <- Sys.time()
  time_taken <- end_time - start_time
  cli::cli_alert_success("Data pulled in {round(time_taken, 1)} {attr(time_taken, 'units')}")
  return(repos)
}

#' @title Get repository URLS
#' @name get_repos_urls
#' @description Pulls a vector of repositories URLs (web or API): either all for
#'   an organization or those with a given text in code blobs (`with_code`
#'   parameter) or a file (`with_files` parameter).
#' @param gitstats A GitStats object.
#' @param type A character, choose if `api` or `web` (`html`) URLs should be
#'   returned. `api` type is set by default as setting `web` results in parsing
#'   which may be time consuming in case of large number of repositories.
#' @param with_code A character vector, if defined, `GitStats` will pull
#'   repositories with specified code phrases in code blobs.
#' @param in_files A character vector of file names. Works when `with_code` is
#'   set - then it searches code blobs only in files passed to `in_files`
#'   parameter.
#' @param with_files A character vector, if defined, GitStats will pull
#'   repositories with specified files.
#' @param cache A logical, if set to `TRUE` GitStats will retrieve the last
#'   result from its storage.
#' @param verbose A logical, `TRUE` by default. If `FALSE` messages and printing
#'   output is switched off.
#' @param progress A logical, by default set to `verbose` value. If `FALSE` no
#'   `cli` progress bar will be displayed.
#' @return A character vector.
#' @examples
#' \dontrun{
#' my_gitstats <- create_gitstats() |>
#'   set_github_host(
#'     token = Sys.getenv("GITHUB_PAT"),
#'     orgs = c("r-world-devs", "openpharma")
#'   ) |>
#'   set_gitlab_host(
#'     token = Sys.getenv("GITLAB_PAT_PUBLIC"),
#'     orgs = "mbtests"
#'   )
#' get_repos_urls(my_gitstats, with_files = c("DESCRIPTION", "LICENSE"))
#' }
#' @export
get_repos_urls <- function(gitstats,
                           type = "api",
                           with_code = NULL,
                           in_files = NULL,
                           with_files = NULL,
                           cache = TRUE,
                           verbose = FALSE,
                           progress = TRUE) {
  start_time <- Sys.time()
  repos_urls <- gitstats$get_repos_urls(
    type = type,
    with_code = with_code,
    in_files = in_files,
    with_files = with_files,
    cache = cache,
    verbose = verbose,
    progress = progress
  )
  end_time <- Sys.time()
  time_taken <- end_time - start_time
  cli::cli_alert_success("Data pulled in {round(time_taken, 1)} {attr(time_taken, 'units')}")
  return(repos_urls)
}

#' @export
print.gitstats_repos_urls <- function(x, ...) {
  if (length(x) == 0) {
    cat("No repository URLs found.\n")
  } else {
    hosts <- sub("^https?://([^/]+).*", "\\1", x)
    host_counts <- table(hosts)

    if (length(x) > 5) {
      cat(cli::col_grey("# Repository URLs (showing first 5 of ", length(x), "):\n"))
      cat(paste0("- ", head(x, 5)), sep = "\n")
    } else {
      cat(cli::col_grey("# Repository URLs:\n"))
      cat(paste0("- ", x), sep = "\n")
    }

    cat(cli::col_grey("\n# Host Summary:\n"))
    for (host in names(host_counts)) {
      cat(cli::col_grey("- ", host, ": ", host_counts[[host]], " URL(s)\n"))
    }
  }
}

#' @title Get data on files trees across repositories
#' @name get_repos_trees
#' @description Pulls files tree (structure) per repository. Files trees are
#'   then stored as character vectors in `files_tree` column of output table.
#' @param gitstats A GitStats object.
#' @param pattern A regular expression. If defined, it pulls structure of files
#'   in a repository matching this pattern reaching to the level of directories
#'   defined by `depth` parameter.
#' @param depth Defines level of directories to reach for files structure from.
#'   E.g. if set to `0`, it will pull files tree only from `root`, if `1L`, will
#'   take data from `root` directory and directories visible in `root`
#'   directory. If left with no argument, will pull files tree down to every
#'   directory in a repo.
#' @param cache A logical, if set to `TRUE` GitStats will retrieve the last
#'   result from its storage.
#' @param verbose A logical, `TRUE` by default. If `FALSE` messages and printing
#'   output is switched off.
#' @param progress A logical, by default set to `verbose` value. If `FALSE` no
#'   `cli` progress bar will be displayed.
#' @return A `tibble`.
#' @examples
#' \dontrun{
#'  my_gitstats <- create_gitstats() |>
#'   set_github_host(
#'     token = Sys.getenv("GITHUB_PAT"),
#'     orgs = c("r-world-devs", "openpharma")
#'   )
#'
#'  get_repos_trees(
#'    gitstats = my_gitstats,
#'    pattern = "\\.md"
#'  )
#' }
#' @export
get_repos_trees <- function(gitstats,
                            pattern = NULL,
                            depth = Inf,
                            cache = TRUE,
                            verbose = FALSE,
                            progress = TRUE) {
  start_time <- Sys.time()
  repos_trees <- gitstats$get_repos_trees(
    pattern = pattern,
    depth = depth,
    cache = cache,
    verbose = verbose,
    progress = progress
  )
  end_time <- Sys.time()
  time_taken <- end_time - start_time
  cli::cli_alert_success("Data pulled in {round(time_taken, 1)} {attr(time_taken, 'units')}")
  return(repos_trees)
}

GitHostGitLab <- R6::R6Class("GitHostGitLab",
  inherit = GitHost,
  public = list(
    initialize = function(orgs = NA,
                          repos = NA,
                          token = NA,
                          host = NA,
                          verbose = NA,
                          .error = TRUE) {
      repos <- if (!is.null(repos)) {
        url_encode(repos)
      }
      orgs <- if (!is.null(orgs)) {
        url_encode(orgs)
      }
      super$initialize(orgs = orgs,
                       repos = repos,
                       token = token,
                       host = host,
                       verbose = verbose,
                       .error = .error)
      if (verbose) {
        cli::cli_alert_success("Set connection to GitLab.")
      }
    }
  ),
  private = list(

    host_name = "GitLab",
    api_version = 4,
    token_name = "GITLAB_PAT",
    min_access_scopes = c("read_api"),
    access_scopes = c("api", "read_api"),

    engine_methods = list(
      "graphql" = list(
        "repos",
        "release_logs"
      ),
      "rest" = list(
        "code",
        "commits",
        "contributors"
      )
    ),

    set_api_url = function(host) {
      if (is.null(host)) {
        private$api_url <- glue::glue(
          "https://gitlab.com/api/v{private$api_version}"
        )
      } else {
        private$set_custom_api_url(host)
      }
    },

    set_web_url = function(host) {
      if (is.null(host)) {
        private$web_url <- glue::glue(
          "https://gitlab.com"
        )
      } else {
        private$set_custom_web_url(host)
      }
    },

    check_if_public = function(host) {
      private$is_public <- is.null(host) ||
        host == "gitlab.com" ||
        host == "https://gitlab.com"
    },

    set_test_endpoint = function() {
      private$test_endpoint <- glue::glue("{private$api_url}/projects")
    },

    set_tokens_endpoint = function() {
      private$endpoints$tokens <- glue::glue("{private$api_url}/personal_access_tokens")
    },

    set_orgs_endpoint = function() {
      private$endpoints$orgs <- glue::glue("{private$api_url}/groups")
      private$endpoints$users <- glue::glue("{private$api_url}/users?username=")
    },

    set_repositories_endpoint = function() {
      private$endpoints$repositories <- glue::glue("{private$api_url}/projects")
    },

    setup_engines = function() {
      private$engines$rest <- EngineRestGitLab$new(
        rest_api_url = private$api_url,
        token = private$token,
        scan_all = private$scan_all
      )
      private$engines$graphql <- EngineGraphQLGitLab$new(
        gql_api_url = private$graphql_api_url,
        token = private$token,
        scan_all = private$scan_all
      )
    },

    check_token_scopes = function(response = NULL, token) {
      TRUE
    },

    add_repo_api_url = function(repos_table) {
      if (!is.null(repos_table) && nrow(repos_table) > 0) {
        repos_table <- dplyr::mutate(
          repos_table,
          api_url = paste0(private$endpoints$repositories,
                           "/",
                           stringr::str_match(repo_id, "[0-9].*"))
        )
      }
      return(repos_table)
    },

    get_orgs_from_host = function(output, verbose) {
      if (private$is_public) {
        cli::cli_abort("This feature is not applicable for public hosts.",
                       call = NULL)
      }
      rest_engine <- private$engines$rest
      orgs_count <- rest_engine$get_orgs_count(verbose)
      if (verbose) {
        cli::cli_alert_info("{orgs_count} organizations found.")
      }
      graphql_engine <- private$engines$graphql
      orgs <- graphql_engine$get_orgs(
        orgs_count = as.integer(orgs_count),
        output = output,
        verbose = verbose
      )
      if (inherits(orgs, "graphql_error")) {
        if (verbose) {
          cli::cli_alert_info("Switching to REST API")
        }
        orgs <- rest_engine$get_orgs(
          orgs_count = as.integer(orgs_count),
          verbose = verbose
        )
        if (output == "full_table") {
          orgs <- orgs |>
            rest_engine$prepare_orgs_table()
          private$orgs <- dplyr::pull(orgs, path)
        } else {
          orgs <- purrr::map_vec(orgs, ~.$full_path)
          private$orgs <- orgs
        }
      } else {
        if (output == "full_table") {
          orgs <- orgs |>
            graphql_engine$prepare_orgs_table()
          private$orgs <- dplyr::pull(orgs, path)
        } else {
          private$orgs <- orgs
        }
      }
      return(orgs)
    },

    get_orgs_from_orgs_and_repos = function(output, verbose) {
      graphql_engine <- private$engines$graphql
      default_engine <- graphql_engine
      rest_engine <- private$engines$rest
      orgs_names <- NULL
      orgs_names_from_repos <- NULL
      if ("org" %in% private$searching_scope) {
        orgs_names <- purrr::keep(private$orgs, function(org) {
          type <- attr(org, "type") %||% "organization"
          type == "organization"
        })
      }
      if ("repo" %in% private$searching_scope) {
        orgs_names_from_repos <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        ) |>
          purrr::keep(function(org) {
            type <- attr(org, "type") %||% "organization"
            type == "organization"
          })
      }
      total_orgs_names <- c(orgs_names, orgs_names_from_repos)
      orgs_list <- purrr::map(total_orgs_names, function(org) {
        type <- attr(org, "type") %||% "organization"
        org_response <- graphql_engine$get_org(
          org = url_decode(org),
          verbose = verbose
        )
        if (inherits(org_response, "graphql_error")) {
          if (verbose) {
            cli::cli_alert_info("Switching to REST API")
          }
          org_response <- rest_engine$get_org(
            org = url_encode(org),
            verbose = verbose
          )
          default_engine <<- rest_engine
        }
        return(org_response)
      })
      orgs_table <- default_engine$prepare_orgs_table(orgs_list)
      return(orgs_table)
    },

    get_repos_ids = function(search_response) {
      purrr::map_vec(search_response, ~.$project_id) |> unique()
    },

    get_repo_url_from_response = function(search_response,
                                          type,
                                          repos_fullnames = NULL,
                                          verbose = TRUE,
                                          progress = TRUE) {
      repo_urls <- purrr::map_vec(search_response, function(response) {
        api_url <- paste0(private$api_url, "/projects/", gsub("gid://gitlab/Project/", "", response$node$repo_id))
        if (type == "api") {
          return(api_url)
        } else {
          rest_engine <- private$engines$rest
          project_response <- rest_engine$response(
            endpoint = api_url,
            verbose = verbose
          )
          web_url <- project_response$web_url
          return(web_url)
        }
      }, .progress = if (progress && type != "api") {
        "Mapping api URL to web URL..."
      } else {
        FALSE
      })
      return(repo_urls)
    },

    get_commits_from_orgs = function(since,
                                     until,
                                     verbose,
                                     progress) {
      if ("org" %in% private$searching_scope) {
        rest_engine <- private$engines$rest
        commits_table <- gitstats_map(private$orgs, function(org) {
          commits_table_org <- NULL
          repos_data <- private$get_repos_data(
            org = org,
            verbose = verbose
          )
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "rest",
              scope = url_decode(org),
              information = paste0("Pulling commits ", cli_icons$commit)
            )
          }
          full_repos_encoded <- paste0(
            url_encode(org),
            "%2f",
            repos_data[["paths"]]
          )
          commits_table_org <- rest_engine$get_commits_from_repos(
            full_repos_names = full_repos_encoded,
            since = since,
            until = until,
            verbose = verbose
          ) |>
            rest_engine$tailor_commits_info(org = org) |>
            rest_engine$prepare_commits_table() |>
            rest_engine$get_commits_authors_handles_and_names(
              verbose = verbose
            )
          return(commits_table_org)
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind()
        return(commits_table)
      }
    },

    get_commits_from_repos = function(since,
                                      until,
                                      verbose,
                                      progress) {
      if ("repo" %in% private$searching_scope) {
        rest_engine <- private$engines$rest
        graphql_engine <- private$engines$graphql
        orgs <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        )
        commits_table <- gitstats_map(orgs, function(org) {
          commits_table_org <- NULL
          repos <- private$orgs_repos[[org]]
          full_repos_names <- paste0(url_encode(org), "%2f", repos)
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "rest",
              scope = set_repo_scope(org, private),
              information = paste0("Pulling commits ", cli_icons$commit)
            )
          }
          commits_table_org <- rest_engine$get_commits_from_repos(
            full_repos_names = full_repos_names,
            since = since,
            until = until,
            verbose = verbose
          ) |>
            rest_engine$tailor_commits_info(org = org) |>
            rest_engine$prepare_commits_table() |>
            rest_engine$get_commits_authors_handles_and_names(
              verbose = verbose
            )
          return(commits_table_org)
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind()
        return(commits_table)
      }
    },

    get_repos_data = function(org, repos = NULL, verbose) {
      cached_repos <- private$get_cached_repos(org)
      if (is.null(cached_repos)) {
        if (verbose) cli::cli_alert("[{org}] Pulling repositories {cli_icons$repo} data...")
        graphql_engine <- private$engines$graphql
        owner_type <- attr(org, "type") %||% "organization"
        repos_from_org <- graphql_engine$get_repos_from_org(
          org = url_decode(org),
          owner_type = owner_type,
          verbose = verbose
        )
        engine_used <- graphql_engine
        if (inherits(repos_from_org, "graphql_error")) {
          if (verbose) {
            cli::cli_alert_info("Switching to REST API...")
          }
          rest_engine <- private$engines$rest
          repos_from_org <- rest_engine$get_repos_from_org(
            org = url_encode(org),
            output = "raw",
            verbose = verbose
          )
          engine_used <- rest_engine
        } else {
          repos_from_org <- purrr::map(repos_from_org, function(repos_data) {
            repos_data$path <- repos_data$node$repo_path
            repos_data
          })
        }
        private$set_cached_repos(repos_from_org, org, verbose)
        tryCatch(
          private$save_repos_to_storage(
            repos_from_org, org, engine_used
          ),
          error = function(e) NULL
        )
      } else {
        if (verbose) cli::cli_alert("Using cached repositories data...")
        repos_from_org <- cached_repos
      }
      if (!is.null(repos)) {
        repos_from_org <- purrr::keep(repos_from_org, ~ .$path %in% repos)
      }
      repos_names <- repos_from_org |>
        purrr::map_vec(~ .$path)
      repo_ids <- purrr::map_chr(repos_from_org, function(repo) {
        if (!is.null(repo$node$repo_id)) {
          get_gitlab_repo_id(repo$node$repo_id)
        } else {
          as.character(repo$id)
        }
      })
      repos_data <- list(
        "paths" = repos_names,
        "repo_ids" = repo_ids
      )
      return(repos_data)
    },

    save_repos_to_storage = function(repos_from_org, org, engine) {
      storage <- private$storage_backend
      if (!is.null(storage) && storage$is_db()) {
        repos_table <- engine$prepare_repos_table(repos_from_org, org)
        if (!is.null(repos_table) && nrow(repos_table) > 0) {
          repos_table <- dplyr::as_tibble(repos_table)
          existing <- storage$load("repositories")
          if (!is.null(existing)) {
            repos_table <- dplyr::bind_rows(existing, repos_table) |>
              dplyr::distinct(repo_id, .keep_all = TRUE)
          }
          class(repos_table) <- c("gitstats_repos", class(repos_table))
          storage$save("repositories", repos_table)
        }
      }
    },

    are_non_text_files = function(file_path, host_files_structure) {
      if (!is.null(file_path)) {
        any(grepl(non_text_files_pattern, file_path))
      } else if (!is.null(host_files_structure)) {
        any(grepl(non_text_files_pattern, unlist(host_files_structure, use.names = FALSE)))
      } else {
        FALSE
      }
    },

    get_files_structure_from_repos = function(pattern,
                                              depth,
                                              verbose  = TRUE,
                                              progress = TRUE) {
      if ("repo" %in% private$searching_scope) {
        rest_engine <- private$engines$rest
        graphql_engine <- private$engines$graphql
        orgs <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        )
        files_structure_list <- gitstats_map(orgs, function(org) {
          if (verbose) {
            user_info <- if (!is.null(pattern)) {
              glue::glue(
                "Pulling repos {cli_icons$tree} [files matching pattern: '{paste0(pattern, collapse = '|')}']"
              )
            } else {
              glue::glue("Pulling repos {cli_icons$tree}")
            }
            show_message(
              host = private$host_name,
              engine = "rest",
              scope = set_repo_scope(org, private),
              information = user_info
            )
          }
          private$get_files_structure_from_repos_data(
            rest_engine = rest_engine,
            org = org,
            repos_data = list("paths" = private$orgs_repos[[org]]),
            pattern = pattern,
            depth = depth,
            verbose = verbose
          )
        }, .progress = set_progress_bar(progress, private))
        names(files_structure_list) <- orgs
        files_structure_list <- files_structure_list |>
          purrr::discard(~ length(.) == 0)
        if (length(files_structure_list) == 0 && verbose) {
          cli::cli_alert_warning(
            cli::col_yellow(
              "For {private$host_name} no files {cli_icons$tree} structure found."
            )
          )
        }
        return(files_structure_list)
      }
    },

    get_files_content_from_repos = function(file_path,
                                            verbose = TRUE,
                                            progress = TRUE) {
      if ("repo" %in% private$searching_scope) {
        graphql_engine <- private$engines$graphql
        orgs <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        )
        files_table <- gitstats_map(orgs, function(org) {
          if (verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = set_repo_scope(org, private),
              information = glue::glue("Pulling files {cli_icons$file} content: [{paste0(file_path, collapse = ', ')}]")
            )
          }
          owner_type <- attr(org, "type") %||% "organization"
          graphql_engine$get_files_from_org_per_repo(
            org = org,
            owner_type = owner_type,
            repos_data = list("paths" = private$orgs_repos[[org]]),
            file_paths = file_path,
            verbose = verbose
          ) |>
            graphql_engine$prepare_files_table(
              org = org
            )
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind() |>
          private$add_repo_api_url()
        return(files_table)
      }
    },

    get_files_content_from_files_structure = function(files_structure,
                                                      verbose = TRUE,
                                                      progress = TRUE) {
      if (length(files_structure) > 0) {
        graphql_engine <- private$engines$graphql
        orgs <- names(files_structure)
        files_table <- purrr::map(orgs, function(org) {
          if (verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = org,
              information = paste0("Pulling files ", cli_icons$file, " from files structure")
            )
          }
          owner_type <- attr(org, "type") %||% "organization"
          repos <- files_structure[[org]] |>
            names()
          graphql_engine$get_files_from_org_per_repo(
            org = org,
            owner_type = owner_type,
            repos_data = list("paths" = repos),
            host_files_structure = files_structure,
            verbose = verbose
          ) |>
            graphql_engine$prepare_files_table(
              org = org
            )
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind() |>
          private$add_repo_api_url()
        return(files_table)
      } else {
        cli::cli_alert_warning("[GitLab] No files {cli_icons$file} found. Skipping pulling files content.")
        return(NULL)
      }
    },

    get_commit_sha = function(project_id, default_branch) {
      rest_engine <- private$engines$rest
      rest_engine$get_commit_sha_from_branch(
        project_id = project_id,
        default_branch = default_branch
      )
    },

    fill_repos_commit_sha = function(repos_table, verbose = FALSE) {
      if (is.null(repos_table) || nrow(repos_table) == 0) {
        return(repos_table)
      }
      missing_sha <- is.na(repos_table[["commit_sha"]]) & nchar(repos_table[["default_branch"]]) > 0
      if (any(missing_sha)) {
        n_missing <- sum(missing_sha)
        if (verbose) {
          show_message(
            host = private$host_name,
            engine = "rest",
            information = glue::glue(
              "Fetching missing commit SHAs for {n_missing} repo{ifelse(n_missing > 1, 's', '')}"
            )
          )
        }
        for (i in which(missing_sha)) {
          repos_table[["commit_sha"]][i] <- private$get_commit_sha(
            project_id = repos_table[["repo_id"]][i],
            default_branch = repos_table[["default_branch"]][i]
          )
        }
      }
      return(repos_table)
    },

    # Override parent to query repos directly by fullpath instead of
    # fetching all repos from org/user and filtering client-side.
    # The parent's approach is slow for GitLab because the repos_by_user
    # query searches the entire GitLab instance.
    get_repos_from_repos = function(add_languages, verbose, progress) {
      if ("repo" %in% private$searching_scope) {
        graphql_engine <- private$engines$graphql
        orgs <- names(private$orgs_repos)
        full_paths <- purrr::map(orgs, function(org) {
          repos <- private$orgs_repos[[org]]
          paste0(url_decode(org), "/", repos)
        }) |>
          unlist()
        if (!private$scan_all && verbose) {
          show_message(
            host = private$host_name,
            engine = "graphql",
            scope = paste0(full_paths, collapse = ", "),
            information = paste0("Pulling repositories ", cli_icons$repo)
          )
        }
        repos_list <- graphql_engine$get_repos_by_fullpath(
          full_paths = full_paths,
          verbose = verbose
        )
        if (length(repos_list) > 0) {
          repos_table <- graphql_engine$prepare_repos_table(repos_list)
        } else {
          repos_table <- NULL
        }
        return(repos_table)
      }
    },

    get_all_repos = function(add_languages = TRUE, verbose = TRUE, progress = TRUE, fill_empty_sha = FALSE) {
      if (private$scan_all && is.null(private$orgs)) {
        private$orgs <- private$get_orgs_from_host(
          output = "only_names",
          verbose = verbose
        )
      }
      repos_from_orgs <- private$get_repos_from_orgs(add_languages, verbose, progress)
      repos_from_repos <- private$get_repos_from_repos(add_languages, verbose, progress)
      repos_table <- purrr::list_rbind(
        list(repos_from_orgs, repos_from_repos)
      )
      if (fill_empty_sha) {
        repos_table <- private$fill_repos_commit_sha(repos_table, verbose = verbose)
      }
      return(repos_table)
    }
  )
)

test_that("repos queries are built properly", {
  gl_repos_by_org_query <-
    test_gqlquery_gl$repos_by_org()
  expect_snapshot(
    gl_repos_by_org_query
  )
  test_mocker$cache(gl_repos_by_org_query)
})

test_that("repos queries are built properly", {
  gl_repos_query <-
    test_gqlquery_gl$repos("")
  expect_snapshot(
    gl_repos_query
  )
})

test_that("repo_by_fullpath query is built properly", {
  gl_repo_by_fullpath_query <-
    test_gqlquery_gl$repo_by_fullpath()
  expect_snapshot(
    gl_repo_by_fullpath_query
  )
})

test_that("`get_repos_page()` pulls repos page from GitLab group", {
  mockery::stub(
    test_graphql_gitlab_priv$get_repos_page,
    "self$gql_response",
    test_fixtures$gitlab_repos_by_org_response
  )
  gl_repos_page <- test_graphql_gitlab_priv$get_repos_page(
    org = "mbtests",
    type = "organization"
  )
  expect_repos_gitlab_gql_response(
    gl_repos_page
  )
  test_mocker$cache(gl_repos_page)
})

test_that("`get_repos_from_org()` prepares formatted list", {
  mockery::stub(
    test_graphql_gitlab$get_repos_from_org,
    "private$get_repos_page",
    test_mocker$use("gl_repos_page")
  )
  gl_repos_from_org <- test_graphql_gitlab$get_repos_from_org(
    org = "mbtests",
    owner_type = "organization"
  )
  expect_equal(
    names(gl_repos_from_org[[1]]$node),
    c(
      "repo_id", "repo_name", "repo_path", "repo_fullpath", "repository",
      "stars", "forks", "created_at", "last_activity_at",
      "languages", "issues", "namespace", "repo_url"
    )
  )
  test_mocker$cache(gl_repos_from_org)
})

test_that("`get_repos_page()` pulls repos page from GitLab user", {
  mockery::stub(
    test_graphql_gitlab_priv$get_repos_page,
    "self$gql_response",
    test_fixtures$gitlab_repos_by_user_response
  )
  gl_repos_user_page <- test_graphql_gitlab_priv$get_repos_page(
    org = "test_user",
    type = "user"
  )
  expect_repos_gitlab_gql_response(
    gl_repos_user_page,
    type = "user"
  )
  test_mocker$cache(gl_repos_user_page)
})

test_that("`get_repos_from_org()` prepares formatted list", {
  mockery::stub(
    test_graphql_gitlab$get_repos_from_org,
    "private$get_repos_page",
    test_mocker$use("gl_repos_user_page")
  )
  gl_repos_from_user <- test_graphql_gitlab$get_repos_from_org(
    org = "test_user",
    owner_type = "user"
  )
  expect_equal(
    names(gl_repos_from_user[[1]]$node),
    c(
      "repo_id", "repo_name", "repo_path", "repo_fullpath", "repository",
      "stars", "forks", "created_at", "last_activity_at",
      "languages", "issues", "namespace", "repo_url"
    )
  )
  test_mocker$cache(gl_repos_from_user)
})

test_that("`get_repos_by_fullpath()` queries repos directly by path", {
  mockery::stub(
    test_graphql_gitlab$get_repos_by_fullpath,
    "self$gql_response",
    test_fixtures$gitlab_repo_by_fullpath_response
  )
  gl_repos_by_path <- test_graphql_gitlab$get_repos_by_fullpath(
    full_paths = c("mbtests/gitstatstesting")
  )
  expect_type(gl_repos_by_path, "list")
  expect_length(gl_repos_by_path, 1)
  expect_equal(
    names(gl_repos_by_path[[1]]$node),
    c(
      "repo_id", "repo_name", "repo_path", "repo_fullpath", "repository",
      "stars", "forks", "created_at", "last_activity_at",
      "languages", "issues", "namespace", "repo_url"
    )
  )
  test_mocker$cache(gl_repos_by_path)
})

test_that("`get_repos_by_fullpath()` skips repos that return NULL", {
  null_response <- list("data" = list("project" = NULL))
  mockery::stub(
    test_graphql_gitlab$get_repos_by_fullpath,
    "self$gql_response",
    null_response
  )
  gl_repos_by_path <- test_graphql_gitlab$get_repos_by_fullpath(
    full_paths = c("nonexistent/repo")
  )
  expect_type(gl_repos_by_path, "list")
  expect_length(gl_repos_by_path, 0)
})

test_that("`get_repos_from_org()` does not fail when GraphQL response is not complete", {
  mockery::stub(
    test_graphql_gitlab$get_repos_from_org,
    "private$get_repos_page",
    test_fixtures$empty_gql_response
  )
  gl_repos_from_org <- test_graphql_gitlab$get_repos_from_org(
    org = "mbtests",
    owner_type = "organization"
  )
  expect_type(
    gl_repos_from_org,
    "list"
  )
  expect_length(
    gl_repos_from_org,
    0
  )
  mockery::stub(
    test_graphql_gitlab$get_repos_from_org,
    "private$get_repos_page",
    test_fixtures$half_empty_gql_response
  )
  gl_repos_from_org <- test_graphql_gitlab$get_repos_from_org(
    org = "mbtests",
    owner_type = "organization"
  )
  expect_type(
    gl_repos_from_org,
    "list"
  )
  expect_length(
    gl_repos_from_org,
    0
  )
})

test_that("get_repos_page handles properly a GraphQL query error", {
  mockery::stub(
    test_graphql_gitlab_priv$get_repos_page,
    "self$gql_response",
    test_error_fixtures$graphql_error_no_count_languages |>
      test_graphql_gitlab_priv$set_graphql_error_class()
  )
  repos_graphql_error <- test_graphql_gitlab_priv$get_repos_page()
  expect_s3_class(repos_graphql_error, "graphql_error")
  test_mocker$cache(repos_graphql_error)
})


test_that("get_repos_page handles properly a GraphQL query error", {
  mockery::stub(
    test_graphql_gitlab_priv$get_repos_page,
    "self$gql_response",
    test_error_fixtures$graphql_server_error |>
      test_graphql_gitlab_priv$set_graphql_error_class()
  )
  internal_server_error <- test_graphql_gitlab_priv$get_repos_page()
  expect_s3_class(internal_server_error, "graphql_error")
  test_mocker$cache(internal_server_error)
})

test_that("error handler works correctly", {
  output <- test_graphql_gitlab_priv$handle_graphql_error(
    responses_list = test_mocker$use("repos_graphql_error"),
    verbose = FALSE
  )
  expect_s3_class(output, "graphql_error")
})

test_that("error handler prints proper messages", {
  expect_snapshot(
    output <- test_graphql_gitlab_priv$handle_graphql_error(
      responses_list = test_mocker$use("repos_graphql_error"),
      verbose = TRUE
    )
  )
})

test_that("get_repos returns empty list when data is flawed (edges and nextPage is NULL)", {
  flawed_repos_response <- test_fixtures$gitlab_repos_response_flawed
  mockery::stub(
    test_graphql_gitlab$get_repos,
    "private$get_repos_page",
    flawed_repos_response
  )
  repos_response <- test_graphql_gitlab$get_repos(
    repos_ids = c("test_id_1", "test_id_2"),
    verbose = TRUE
  )
  expect_equal(
    repos_response,
    list()
  )
})

test_that("get_repos tries one more time pull data when encounters GraphQL query error", {
  standard_graphql_error <- test_fixtures$gitlab_repos_by_user_response
  class(standard_graphql_error) <- c(class(standard_graphql_error), "graphql_error")
  mockery::stub(
    test_graphql_gitlab$get_repos,
    "private$get_repos_page",
    standard_graphql_error
  )
  repos_response <- test_graphql_gitlab$get_repos(
    repos_ids = c("test_id_1", "test_id_2"),
    verbose = TRUE
  )
  expect_named(
    repos_response[[1]]$node,
    c("repo_id", "repo_name", "repo_path", "repo_fullpath", "repository", "stars",
      "forks", "created_at", "last_activity_at", "languages", "issues", "namespace",
      "repo_url")
  )
})

test_that("get_repos breaks when response is a GraphQL 'no fields' error", {
  mockery::stub(
    test_graphql_gitlab$get_repos,
    "private$get_repos_page",
    test_mocker$use("repos_graphql_error")
  )
  response <- test_graphql_gitlab$get_repos(
    repos_ids = c("test_id_1", "test_id_2"),
    verbose = FALSE
  )
  expect_s3_class(response, "graphql_error")
  expect_s3_class(response, "graphql_no_fields_error")
})

test_that("get_repos falls back to batching on GraphQL limit error", {
  limit_error <- test_error_fixtures$graphql_limit_error
  class(limit_error) <- c(class(limit_error), "graphql_error", "graphql_limit_error")
  mockery::stub(
    test_graphql_gitlab$get_repos,
    "private$get_repos_page",
    limit_error
  )
  mockery::stub(
    test_graphql_gitlab$get_repos,
    "private$get_repos_in_batches",
    test_fixtures$gitlab_repos_by_user_response$data$projects$edges
  )
  repos_response <- test_graphql_gitlab$get_repos(
    repos_ids = c("test_id_1", "test_id_2"),
    verbose = FALSE
  )
  expect_type(repos_response, "list")
  expect_gt(length(repos_response), 0)
  expect_named(
    repos_response[[1]]$node,
    c("repo_id", "repo_name", "repo_path", "repo_fullpath", "repository", "stars",
      "forks", "created_at", "last_activity_at", "languages", "issues", "namespace",
      "repo_url")
  )
})

test_that("get_repos prints messages when falling back to batching", {
  limit_error <- test_error_fixtures$graphql_limit_error
  class(limit_error) <- c(class(limit_error), "graphql_error", "graphql_limit_error")
  mockery::stub(
    test_graphql_gitlab$get_repos,
    "private$get_repos_page",
    limit_error
  )
  mockery::stub(
    test_graphql_gitlab$get_repos,
    "private$get_repos_in_batches",
    test_fixtures$gitlab_repos_by_user_response$data$projects$edges
  )
  expect_snapshot(
    gitlab_repos <- test_graphql_gitlab$get_repos(
      repos_ids = c("test_id_1", "test_id_2"),
      verbose = TRUE
    )
  )
  expect_type(gitlab_repos, "list")
  expect_gt(length(gitlab_repos), 0)
  expect_list_contains(
    gitlab_repos[[1]]$node,
    c("repo_id", "repo_name", "repository", "stars", "forks", "created_at", "last_activity_at")
  )
})

test_that("get_repos_in_batches pulls repos in smaller chunks", {
  mock_response <- httr2::response_json(
    status_code = 200,
    body = test_fixtures$gitlab_repos_by_user_response
  )
  mockery::stub(
    test_graphql_gitlab_priv$get_repos_in_batches,
    "graphql_response",
    test_fixtures$gitlab_repos_by_user_response
  )
  repos_response <- test_graphql_gitlab_priv$get_repos_in_batches(
    repos_ids = as.character(1:120),
    batch_size = 50,
    verbose = FALSE
  )
  expect_type(repos_response, "list")
  expect_gt(length(repos_response), 0)
})

test_that("get_repos_in_batches dispatches via gitstats_map", {
  mockery::stub(
    test_graphql_gitlab_priv$get_repos_in_batches,
    "graphql_response",
    test_fixtures$gitlab_repos_by_user_response
  )
  gitstats_map_called <- FALSE
  mockery::stub(
    test_graphql_gitlab_priv$get_repos_in_batches,
    "gitstats_map",
    function(.x, .f, ...) {
      gitstats_map_called <<- TRUE
      purrr::map(.x, .f, ...)
    }
  )
  repos_response <- test_graphql_gitlab_priv$get_repos_in_batches(
    repos_ids = as.character(1:120),
    batch_size = 50,
    verbose = FALSE
  )
  expect_true(gitstats_map_called)
  expect_type(repos_response, "list")
  expect_gt(length(repos_response), 0)
})

test_that("get_repos_from_org handles properly a GraphQL query error", {
  mockery::stub(
    test_graphql_gitlab$get_repos_from_org,
    "private$get_repos_page",
    test_mocker$use("repos_graphql_error")
  )
  gitlab_repos_error <- test_graphql_gitlab$get_repos_from_org(
    org = "mbtests",
    owner_type = "organization",
    verbose = FALSE
  )
  expect_s3_class(gitlab_repos_error, "graphql_error")
  test_mocker$cache(gitlab_repos_error)
})

test_that("`get_repos_languages()` works", {
  repos_list <- test_fixtures$gitlab_repositories_rest_response
  mockery::stub(
    test_rest_gitlab_priv$get_repos_languages,
    "self$response",
    test_fixtures$gitlab_languages_response
  )
  gl_repos_list_with_languages <- test_rest_gitlab_priv$get_repos_languages(
    repos_list = repos_list,
    verbose = FALSE,
    progress = FALSE
  )
  purrr::walk(gl_repos_list_with_languages, ~ expect_list_contains(., "languages"))
  expect_equal(gl_repos_list_with_languages[[1]]$languages, c("Python", "R"))
  test_mocker$cache(gl_repos_list_with_languages)
})

test_that("REST engine pulls repositories from organization", {
  mockery::stub(
    test_rest_gitlab$get_repos_from_org,
    "private$paginate_results",
    test_fixtures$gitlab_repositories_rest_response
  )
  mockery::stub(
    test_rest_gitlab$get_repos_from_org,
    "private$get_repos_languages",
    test_mocker$use("gl_repos_list_with_languages")
  )
  test_org <- "mbtests"
  attr(test_org, "type") <- "organization"
  gitlab_rest_repos_from_org_raw <- test_rest_gitlab$get_repos_from_org(
    org = test_org,
    repos =  c("test_repo_1", "test_repo_2"),
    output = "raw"
  )
  expect_length(gitlab_rest_repos_from_org_raw, 2L)
  test_mocker$cache(gitlab_rest_repos_from_org_raw)
  gitlab_rest_repos_from_org <- test_rest_gitlab$get_repos_from_org(
    org = test_org,
    output = "full_table"
  )
  purrr::walk(gitlab_rest_repos_from_org, ~ expect_in("languages", names(.)))
  test_mocker$cache(gitlab_rest_repos_from_org)
})

test_that("REST engine prepares repositories table", {
  gitlab_rest_repos_table <- test_rest_gitlab$prepare_repos_table(
    repos_list = test_mocker$use("gitlab_rest_repos_from_org"),
    org = "mbtests"
  )
  expect_repos_table(gitlab_rest_repos_table)
  test_mocker$cache(gitlab_rest_repos_table)
})

test_that("GitLab build_search_query builds query with code only", {
  query <- test_rest_gitlab_priv$build_search_query(code = "test")
  expect_equal(query, utils::URLencode("test", reserved = TRUE))
})

test_that("GitLab build_search_query builds query with in_path", {
  query <- test_rest_gitlab_priv$build_search_query(code = "src/main", in_path = TRUE)
  expect_match(query, "^path:")
})

test_that("GitLab build_search_query builds query with filename", {
  query <- test_rest_gitlab_priv$build_search_query(
    code = "test",
    filename = "DESCRIPTION"
  )
  expect_match(query, "filename:DESCRIPTION$")
})

test_that("`search_for_code()` works", {
  mockery::stub(
    test_rest_gitlab$search_for_code,
    "self$response",
    list() # output not mocked as falls into infinite loop
  )
  expect_snapshot(
    gl_search_repos_by_code <- test_rest_gitlab$search_for_code(
      code = "test",
      filename = "TESTFILE",
      verbose = TRUE,
      page_max = 2
    )
  )
})

test_that("`search_for_code()` handles the 1000 response limit", {
  limit_rest_error <- function() stop("test")
  mockery::stub(
    test_rest_gitlab$search_for_code,
    "self$response",
    limit_rest_error
  )
  mockery::stub(
    test_rest_gitlab$search_for_code,
    "list",
    rep(1, 1e4)
  )
  expect_snapshot(
    gl_search_repos_by_code <- test_rest_gitlab$search_for_code(
      code = "test",
      filename = "TESTFILE",
      verbose = TRUE,
      page_max = 2
    )
  )
  expect_s3_class(gl_search_repos_by_code, "rest_error")
  expect_s3_class(gl_search_repos_by_code, "rest_error_response_limit")
})

test_that("`search_repos_for_code()` works", {
  mockery::stub(
    test_rest_gitlab$search_repos_for_code,
    "self$response",
    list()
  )
  expect_snapshot(
    gl_search_repos_by_code <- test_rest_gitlab$search_repos_for_code(
      code = "test",
      repos = "TestRepo",
      filename = "TESTFILE",
      verbose = TRUE,
      page_max = 2
    )
  )
})

test_that("GraphQL engine prepares repos table", {
  gl_repos_table <- test_graphql_gitlab$prepare_repos_table(
    repos_list = test_mocker$use("gl_repos_from_org"),
    org = "mbtests"
  )
  expect_repos_table(
    gl_repos_table
  )
  test_mocker$cache(gl_repos_table)
})

test_that("GraphQL engine prepares repos table with no org specified", {
  gl_repos_table <- test_graphql_gitlab$prepare_repos_table(
    repos_list = test_mocker$use("gl_repos_from_org"),
    org = NULL
  )
  expect_repos_table(
    gl_repos_table
  )
})

test_that("GraphQL engine prepares repos table with no info on org at all", {
  repos_from_org <- test_mocker$use("gl_repos_from_org") |>
    purrr::map(function(x) {
      x$node$namespace$path <- NULL
      x
    })
  gl_repos_table <- test_graphql_gitlab$prepare_repos_table(
    repos_list = repos_from_org,
    org = NULL
  )
  expect_repos_table(
    gl_repos_table
  )
  expect_true(
    all(gl_repos_table$org == "mbtests")
  )
})

test_that("get_repos_from_org prints proper message", {
  if (integration_tests_skipped) {
    mockery::stub(
      gitlab_testhost_priv$get_repos_from_orgs,
      "graphql_engine$get_repos_from_org",
      test_mocker$use("gl_repos_from_org")
    )
  }
  gitlab_testhost_priv$orgs <- "mbtests"
  expect_snapshot(
    gl_repos_from_orgs <- gitlab_testhost_priv$get_repos_from_orgs(
      add_languages = TRUE,
      verbose = TRUE,
      progress = FALSE
    )
  )
  expect_repos_table(
    gl_repos_from_orgs
  )
  test_mocker$cache(gl_repos_from_orgs)
})

test_that("GitLab Host turns to REST if GraphQL fails with error (org setup)", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_from_orgs,
    "graphql_engine$get_repos_from_org",
    test_mocker$use("gitlab_repos_error")
  )
  mockery::stub(
    gitlab_testhost_priv$get_repos_from_orgs,
    "rest_engine$prepare_repos_table",
    test_mocker$use("gitlab_rest_repos_table")
  )
  gl_repos_from_orgs <- gitlab_testhost_priv$get_repos_from_orgs(
    add_languages = TRUE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_repos_table(gl_repos_from_orgs)
})

test_that("`get_repos_from_repos()` queries repos directly by fullpath", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_from_repos,
    "graphql_engine$get_repos_by_fullpath",
    test_mocker$use("gl_repos_by_path")
  )
  mockery::stub(
    gitlab_testhost_priv$get_repos_from_repos,
    "graphql_engine$prepare_repos_table",
    test_mocker$use("gl_repos_table")
  )
  gitlab_testhost_priv$searching_scope <- "repo"
  gitlab_testhost_priv$orgs_repos <- list("mbtests" = "gitstatstesting")
  gl_repos_from_repos <- gitlab_testhost_priv$get_repos_from_repos(
    add_languages = TRUE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_repos_table(gl_repos_from_repos)
  test_mocker$cache(gl_repos_from_repos)
})

test_that("`get_repos_from_repos()` returns NULL when no repos found", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_from_repos,
    "graphql_engine$get_repos_by_fullpath",
    list()
  )
  gitlab_testhost_priv$searching_scope <- "repo"
  gitlab_testhost_priv$orgs_repos <- list("mbtests" = "nonexistent")
  gl_repos_from_repos <- gitlab_testhost_priv$get_repos_from_repos(
    add_languages = TRUE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_null(gl_repos_from_repos)
})

test_that("GitLab Host prints message when turning to REST engine (from orgs)", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_from_orgs,
    "graphql_engine$get_repos_from_org",
    test_mocker$use("gitlab_repos_error")
  )
  mockery::stub(
    gitlab_testhost_priv$get_repos_from_orgs,
    "rest_engine$prepare_repos_table",
    test_mocker$use("gitlab_rest_repos_table")
  )
  gitlab_testhost_priv$searching_scope <- "org"
  expect_snapshot(
    gl_repos_from_orgs <- gitlab_testhost_priv$get_repos_from_orgs(
      add_languages = TRUE,
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("`get_repos_from_repos()` prints proper message", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_from_repos,
    "graphql_engine$get_repos_by_fullpath",
    test_mocker$use("gl_repos_by_path")
  )
  mockery::stub(
    gitlab_testhost_priv$get_repos_from_repos,
    "graphql_engine$prepare_repos_table",
    test_mocker$use("gl_repos_table")
  )
  gitlab_testhost_priv$searching_scope <- "repo"
  gitlab_testhost_priv$orgs_repos <- list("mbtests" = "gitstatstesting")
  expect_snapshot(
    gl_repos_from_repos <- gitlab_testhost_priv$get_repos_from_repos(
      add_languages = TRUE,
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("get_repos_ids", {
  repos_ids <- gitlab_testhost_priv$get_repos_ids(
    search_response = test_fixtures$gitlab_search_response
  )
  expect_type(
    repos_ids,
    "double"
  )
  expect_gt(
    length(repos_ids), 0
  )
})

test_that("parse_search_response works", {
  mockery::stub(
    gitlab_testhost_priv$parse_search_response,
    "api_engine$get_repos_from_org",
    test_mocker$use("gl_repos_from_org")
  )
  gl_repos_raw_output <- gitlab_testhost_priv$parse_search_response(
    search_response = test_fixtures$gitlab_search_response,
    org = "test_group",
    output = "raw",
    verbose = FALSE
  )
  expect_type(
    gl_repos_raw_output,
    "list"
  )
  expect_true(
    all(names(gl_repos_raw_output[[1]]$node) %in% c("repo_id", "repo_name", "repo_path",
                                                    "repo_fullpath",
                                                    "repository", "stars", "forks", "created_at",
                                                    "last_activity_at", "languages", "issues",
                                                    "namespace", "repo_url"))
  )
  test_mocker$cache(gl_repos_raw_output)
})

test_that("GitHost adds `repo_api_url` column to GitLab repos table", {
  repos_table <- test_mocker$use("gl_repos_table")
  gl_repos_table_with_api_url <- gitlab_testhost_priv$add_repo_api_url(repos_table)
  expect_true(all(grepl("gitlab.com/api/v4", gl_repos_table_with_api_url$api_url)))
  test_mocker$cache(gl_repos_table_with_api_url)
})


test_that("add_githost_info adds data on Git platform to repos table", {
  gl_repos_table_with_platform <- gitlab_testhost_priv$add_githost_info(
    repos_table = test_mocker$use("gl_repos_table_with_api_url")
  )
  expect_repos_table(
    gl_repos_table_with_platform,
    with_cols = c("api_url", "githost")
  )
  test_mocker$cache(gl_repos_table_with_platform)
})

test_that("`get_repos_contributors()` adds contributors to repos table", {
  mockery::stub(
    test_rest_gitlab$get_repos_contributors,
    "private$get_contributors_from_repo",
    "Maciej Banas"
  )
  gl_repos_table_full <- test_rest_gitlab$get_repos_contributors(
    test_mocker$use("gl_repos_table_with_platform"),
    progress = FALSE
  )
  expect_repos_table(
    gl_repos_table_full,
    with_cols = c("api_url", "githost", "contributors")
  )
  expect_gt(
    length(gl_repos_table_full$contributors),
    0
  )
  test_mocker$cache(gl_repos_table_full)
})

test_that("`get_repos_data` pulls data from org", {
  if (integration_tests_skipped) {
    mockery::stub(
      gitlab_testhost_priv$get_repos_data,
      "graphql_engine$get_repos_from_org",
      test_mocker$use("gl_repos_from_org")
    )
  }
  gitlab_testhost_priv$searching_scope <- "org"
  gl_repos_data <- gitlab_testhost_priv$get_repos_data(
    org = "mbtests",
    verbose = FALSE
  )
  expect_type(gl_repos_data, "list")
  expect_type(gl_repos_data[["paths"]], "character")
  expect_gt(length(gl_repos_data[["paths"]]), 0)
  test_mocker$cache(gl_repos_data)
})

test_that("`get_repos_data` uses cached data from org", {
  if (integration_tests_skipped) {
    mockery::stub(
      gitlab_testhost_priv$get_repos_data,
      "graphql_engine$get_repos_from_org",
      test_mocker$use("gl_repos_from_org")
    )
  }
  expect_gt(
    length(gitlab_testhost_priv$cached_repos),
    0
  )
  gitlab_testhost_priv$searching_scope <- "org"
  expect_snapshot(
    gl_repos_data <- gitlab_testhost_priv$get_repos_data(
      org = "mbtests",
      verbose = TRUE
    )
  )
  expect_type(gl_repos_data, "list")
  expect_type(gl_repos_data[["paths"]], "character")
  expect_gt(length(gl_repos_data[["paths"]]), 0)
})

test_that("`get_repos_data` pulls data from repos", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_data,
    "graphql_engine$get_repos_from_org",
    test_mocker$use("gl_repos_from_org")
  )
  gitlab_testhost_priv$searching_scope <- "repo"
  gitlab_testhost_priv$cached_repos <- list()
  gl_repos_data <- gitlab_testhost_priv$get_repos_data(
    org = "mbtests",
    repos = "gitstatstesting",
    verbose = FALSE
  )
  expect_type(gl_repos_data, "list")
  expect_type(gl_repos_data[["paths"]], "character")
  expect_gt(length(gl_repos_data[["paths"]]), 0)
})

test_that("get_repos_data turns to REST if GraphQL fails with error", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_data,
    "graphql_engine$get_repos_from_org",
    test_mocker$use("gitlab_repos_error")
  )
  mockery::stub(
    gitlab_testhost_priv$get_repos_data,
    "rest_engine$get_repos_from_org",
    test_mocker$use("gitlab_rest_repos_from_org_raw")
  )
  gitlab_testhost_priv$searching_scope <- "org"
  gl_repos_data <- gitlab_testhost_priv$get_repos_data(
    org = "test_org",
    verbose = FALSE
  )
  expect_type(gl_repos_data, "list")
  expect_type(gl_repos_data[["paths"]], "character")
  expect_gt(length(gl_repos_data[["paths"]]), 0)
})

test_that("get_repos_data prints message when turns to REST engine", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_data,
    "graphql_engine$get_repos_from_org",
    test_mocker$use("gitlab_repos_error")
  )
  mockery::stub(
    gitlab_testhost_priv$get_repos_data,
    "rest_engine$get_repos_from_org",
    test_mocker$use("gitlab_rest_repos_from_org_raw")
  )
  gitlab_testhost_priv$searching_scope <- "org"
  gitlab_testhost_priv$cached_repos <- list()
  expect_snapshot(
    gl_repos_data <- gitlab_testhost_priv$get_repos_data(
      org = "test_org",
      verbose = TRUE
    )
  )
})

# ---- commit_sha fallback for archived projects ----

test_that("GraphQL engine returns NA commit_sha for archived projects with null lastCommit", {
  repos_with_archived <- list(
    gitlab_project_node,
    gitlab_archived_project_node
  )
  gl_repos_table <- test_graphql_gitlab$prepare_repos_table(
    repos_list = repos_with_archived,
    org = "mbtests"
  )
  expect_repos_table(gl_repos_table)
  expect_equal(gl_repos_table$commit_sha[1], "1a2bc3d4e5")
  expect_true(is.na(gl_repos_table$commit_sha[2]))
})

test_that("`get_commit_sha_from_branch()` retrieves SHA from branches API", {
  mockery::stub(
    test_rest_gitlab$get_commit_sha_from_branch,
    "self$response",
    test_fixtures$gitlab_branch_response
  )
  commit_sha <- test_rest_gitlab$get_commit_sha_from_branch(
    project_id = "99999999",
    default_branch = "main"
  )
  expect_equal(commit_sha, "abcdef1234567890")
  test_mocker$cache(commit_sha)
})

test_that("`get_commit_sha_from_branch()` returns NA when branch is empty", {
  sha <- test_rest_gitlab$get_commit_sha_from_branch(
    project_id = "99999999",
    default_branch = ""
  )
  expect_true(is.na(sha))
})

test_that("`get_commit_sha_from_branch()` returns NA when branch is NULL", {
  sha <- test_rest_gitlab$get_commit_sha_from_branch(
    project_id = "99999999",
    default_branch = NULL
  )
  expect_true(is.na(sha))
})

test_that("`get_commit_sha_from_branch()` returns NA on API error", {
  rest_error_response <- list("message" = "404 Not Found")
  class(rest_error_response) <- c("rest_error", "404_not_found", class(rest_error_response))
  mockery::stub(
    test_rest_gitlab$get_commit_sha_from_branch,
    "self$response",
    rest_error_response
  )
  sha <- test_rest_gitlab$get_commit_sha_from_branch(
    project_id = "99999999",
    default_branch = "main"
  )
  expect_true(is.na(sha))
})

test_that("get_commit_sha works on host level", {
  gitlab_testhost_fill <- create_gitlab_testhost(
    orgs = "mbtests",
    mode = "private"
  )
  mockery::stub(
    gitlab_testhost_fill$get_commit_sha,
    "rest_engine$get_commit_sha_from_branch",
    test_mocker$use("commit_sha")
  )
  commit_sha <- gitlab_testhost_fill$get_commit_sha(
    project_id = "99999999",
    default_branch = "main"
  )
  expect_equal(commit_sha, "abcdef1234567890")
})

test_that("fill_repos_commit_sha() returns empty table",{
  gitlab_testhost_fill <- create_gitlab_testhost(
    orgs = "mbtests",
    mode = "private"
  )
  repos_table <- data.frame()
  expect_equal(
    gitlab_testhost_fill$fill_repos_commit_sha(repos_table, verbose = FALSE),
    data.frame()
  )
})

test_that("`fill_repos_commit_sha()` fills missing commit_sha via REST", {
  gitlab_testhost_fill <- create_gitlab_testhost(
    orgs = "mbtests",
    mode = "private"
  )
  repos_table <- data.frame(
    repo_id = c("61399846", "99999999"),
    repo_name = c("gitstatstesting", "archived-project"),
    repo_fullpath = c("mbtests/gitstatstesting", "mbtests/archived-project"),
    default_branch = c("main", "main"),
    stars = c(8L, 2L),
    forks = c(3L, 0L),
    created_at = as.POSIXct(c("2023-09-18", "2022-01-10")),
    last_activity_at = as.POSIXct(c("2024-09-18", "2023-06-15")),
    languages = c("Python, R", "R"),
    issues_open = c(2L, 0L),
    issues_closed = c(8L, 3L),
    organization = c("mbtests", "mbtests"),
    repo_url = c("https://gitlab.com/mbtests/gitstatstesting",
                 "https://gitlab.com/mbtests/archived-project"),
    commit_sha = c("1a2bc3d4e5", NA_character_),
    stringsAsFactors = FALSE
  )
  mockery::stub(
    gitlab_testhost_fill$fill_repos_commit_sha,
    "private$get_commit_sha",
    test_mocker$use("commit_sha")
  )
  expect_snapshot(
    repos_commit_sha <- gitlab_testhost_fill$fill_repos_commit_sha(repos_table, verbose = TRUE)
  )
  expect_equal(repos_commit_sha$commit_sha[1], "1a2bc3d4e5")
  expect_equal(repos_commit_sha$commit_sha[2], "abcdef1234567890")
  test_mocker$cache(repos_commit_sha)
})

test_that("`fill_repos_commit_sha()` skips repos with empty default_branch", {
  gitlab_testhost_fill <- create_gitlab_testhost(
    orgs = "mbtests",
    mode = "private"
  )
  repos_table <- data.frame(
    repo_id = c("99999999"),
    repo_name = c("empty-repo"),
    repo_fullpath = c("mbtests/empty-repo"),
    default_branch = c(""),
    stars = 0L,
    forks = 0L,
    created_at = as.POSIXct("2022-01-10"),
    last_activity_at = as.POSIXct("2023-06-15"),
    languages = "",
    issues_open = 0L,
    issues_closed = 0L,
    organization = "mbtests",
    repo_url = "https://gitlab.com/mbtests/empty-repo",
    commit_sha = NA_character_,
    stringsAsFactors = FALSE
  )
  result <- gitlab_testhost_fill$fill_repos_commit_sha(repos_table, verbose = FALSE)
  expect_true(is.na(result$commit_sha[1]))
})

test_that("get_all_repos() works", {
  gitlab_testhost_fill <- create_gitlab_testhost(
    orgs = "mbtests",
    mode = "private"
  )
  mockery::stub(
    gitlab_testhost_fill$get_all_repos,
    "private$get_repos_from_orgs",
    test_mocker$use("gl_repos_from_orgs")
  )
  mockery::stub(
    gitlab_testhost_fill$get_all_repos,
    "private$get_repos_from_repos",
    test_mocker$use("gl_repos_from_repos")
  )
  mockery::stub(
    gitlab_testhost_fill$get_all_repos,
    "private$fill_repos_commit_sha",
    test_mocker$use("repos_commit_sha")
  )
  gitlab_repos <- gitlab_testhost_fill$get_all_repos(add_languages = TRUE, fill_empty_sha = TRUE)
  expect_repos_table(gitlab_repos)
  expect_equal(gitlab_repos$commit_sha[1], "1a2bc3d4e5")
  expect_equal(gitlab_repos$commit_sha[2], "abcdef1234567890")
})

# ---- EngineGraphQLGitLab: get_repo_name_from_url ----

test_that("`get_repo_name_from_url()` extracts repo name from URL", {
  expect_equal(
    test_graphql_gitlab_priv$get_repo_name_from_url(
      "https://gitlab.com/mbtests/gitstatstesting"
    ),
    "gitstatstesting"
  )
})

test_that("`get_repo_name_from_url()` handles nested group paths", {
  expect_equal(
    test_graphql_gitlab_priv$get_repo_name_from_url(
      "https://gitlab.com/group/subgroup/my-repo"
    ),
    "my-repo"
  )
})

# ---- EngineGraphQLGitLab: prepare_files_table_row ----

test_that("`prepare_files_table_row()` builds files data.frame from project", {
  project <- test_fixtures$gitlab_file_repo_response$data$project
  files_row <- test_graphql_gitlab_priv$prepare_files_table_row(
    project = project,
    org = "mbtests"
  )
  expect_s3_class(files_row, "data.frame")
  expect_gt(nrow(files_row), 0)
  expect_in(c("repo_name", "file_path", "file_content", "repo_url"), names(files_row))
  expect_equal(files_row$organization[1], "mbtests")
})

# ---- EngineGraphQLGitLab: set_owner_type ----

test_that("`set_owner_type()` identifies organization (group) type", {
  gl_group_response <- list(
    "data" = list(
      "user" = NULL,
      "group" = list("__typename" = "Group", "fullPath" = "mbtests")
    )
  )
  mockery::stub(
    test_graphql_gitlab$set_owner_type,
    "self$gql_response",
    gl_group_response
  )
  result <- test_graphql_gitlab$set_owner_type(
    owners = "mbtests_org_test",
    verbose = FALSE
  )
  expect_equal(attr(result[[1]], "type"), "organization")
})

test_that("`set_owner_type()` identifies user type", {
  gl_user_response <- list(
    "data" = list(
      "user" = list("__typename" = "User", "username" = "test_user"),
      "group" = NULL
    )
  )
  mockery::stub(
    test_graphql_gitlab$set_owner_type,
    "self$gql_response",
    gl_user_response
  )
  result <- test_graphql_gitlab$set_owner_type(
    owners = "test_user_type_test",
    verbose = FALSE
  )
  expect_equal(attr(result[[1]], "type"), "user")
})

test_that("`set_owner_type()` returns 'not found' when owner does not exist", {
  mockery::stub(
    test_graphql_gitlab$set_owner_type,
    "self$gql_response",
    list("data" = list("user" = NULL, "group" = NULL))
  )
  result <- test_graphql_gitlab$set_owner_type(
    owners = "nonexistent_owner_test",
    verbose = FALSE
  )
  expect_equal(attr(result[[1]], "type"), "not found")
})

# ---- GitHostGitLab: get_repo_url_from_response with type "api" ----

test_that("`get_repo_url_from_response()` returns api URLs", {
  gl_repo_api_urls <- gitlab_testhost_priv$get_repo_url_from_response(
    search_response = test_mocker$use("gl_repos_raw_output"),
    type = "api",
    progress = FALSE
  )
  expect_gt(length(gl_repo_api_urls), 0)
  expect_type(gl_repo_api_urls, "character")
  expect_true(all(grepl("api", gl_repo_api_urls)))
})

# ---- GitHostGitLab: check_if_public ----

test_that("`check_if_public()` sets TRUE for public gitlab.com", {
  gitlab_testhost_priv$check_if_public(NULL)
  expect_true(gitlab_testhost_priv$is_public)
  gitlab_testhost_priv$check_if_public("https://gitlab.com")
  expect_true(gitlab_testhost_priv$is_public)
})

test_that("`check_if_public()` sets FALSE for custom host", {
  gitlab_testhost_priv$check_if_public("https://internal-gitlab.company.com")
  expect_false(gitlab_testhost_priv$is_public)
  # Reset to default
  gitlab_testhost_priv$check_if_public(NULL)
})

github_org_edge <- list(
  "node" = list(
    "name" = "test_name",
    "description" = "test_description",
    "login" = "test_login",
    "url" = "test_url",
    "repositories" = list(
      "totalCount" = 5L
    ),
    "membersWithRole" = list(
      "totalCount" = 3L
    ),
    "avatarUrl" = "test_url"
  )
)

test_fixtures$graphql_gh_orgs_response <- list(
  "data" = list(
    "search" = list(
      "pageInfo" = list(
        "hasNextPage" = FALSE,
        "endCursor" = ""
      ),
      "edges" = list(
        github_org_edge,
        github_org_edge,
        github_org_edge
      )
    )
  )
)

test_fixtures$graphql_gh_org_response <- list(
  "data" = list(
    "organization" = github_org_edge$node
  )
)

test_fixtures$rest_gl_orgs_response <- list(
  "headers" = list(
    "x-total" = "3"
  )
)

gitlab_org_edge <- list(
  "node" = list(
    "name" = "mbtests",
    "description" = "test_description",
    "fullPath" = "mbtests",
    "webUrl" = "web_url",
    "projectsCount" = 5L,
    "groupMembersCount" = 3L,
    "avatarUrl" = "test_url"
  )
)

test_fixtures$graphql_gl_orgs_response <- list(
  "data" = list(
    "groups" = list(
      "pageInfo" = list(
        "endCursor" = "",
        "hasNextPage" = FALSE
      ),
      "edges" = list(
        gitlab_org_edge,
        gitlab_org_edge,
        gitlab_org_edge
      )
    )
  )
)

test_fixtures$graphql_gl_org_response <- list(
  "data" = list(
    "group" = gitlab_org_edge$node
  )
)

test_fixtures$rest_gl_org_response <- list(
  "id" = 11111,
  "web_url" = "url",
  "name" = "mbtests",
  "path" = "mbtests",
  "description" = "test_description"
)

#' @title Get data on commits
#' @name get_commits
#' @description List all commits from all repositories for an organization or a
#'   vector of repositories.
#' @param gitstats A `GitStats` object.
#' @param since A starting date.
#' @param until An end date.
#' @param cache A logical, if set to `TRUE` GitStats will retrieve the last
#'   result from its storage.
#' @param verbose A logical, `TRUE` by default. If `FALSE` messages and printing
#'   output is switched off.
#' @param progress A logical, by default set to `verbose` value. If `FALSE` no
#'   `cli` progress bar will be displayed.
#' @return A table of `tibble` and `gitstats_commits` classes.
#' @examples
#' \dontrun{
#' my_gitstats <- create_gitstats() |>
#'   set_github_host(
#'     token = Sys.getenv("GITHUB_PAT"),
#'     repos = c("openpharma/DataFakeR", "openpharma/visR")
#'   ) |>
#'   set_gitlab_host(
#'     token = Sys.getenv("GITLAB_PAT_PUBLIC"),
#'     orgs = "mbtests"
#'   )
#'  get_commits(my_gitstats, since = "2018-01-01")
#' }
#' @export
get_commits <- function(gitstats,
                        since = NULL,
                        until = Sys.Date(),
                        cache = TRUE,
                        verbose = FALSE,
                        progress = TRUE) {
  start_time <- Sys.time()
  if (is.null(since)) {
    cli::cli_abort(cli::col_red("You need to pass date to `since` parameter."), call = NULL)
  }
  commits <- gitstats$get_commits(
    since = since,
    until = until,
    cache = cache,
    verbose = verbose,
    progress = progress
  )
  end_time <- Sys.time()
  time_taken <- end_time - start_time
  cli::cli_alert_success("Data pulled in {round(time_taken, 1)} {attr(time_taken, 'units')}")
  return(commits)
}

#' @title Get commits statistics
#' @name get_commits_stats
#' @description Prepare statistics from the pulled commits data.
#' @details To make function work, you need first to get commits data with
#'   `GitStats`. See examples section.
#' @param commits A `gitstats_commits` S3 class table object (output of `get_commits()`).
#' @param time_aggregation A character, specifying time aggregation of
#'   statistics.
#' @param group_var Other grouping variable to be passed to `dplyr::group_by()`
#'   function apart from `stats_date` and `githost`. Could be: `author`,
#'   `author_login`, `author_name` or `organization`. Should be passed without
#'   quotation marks.
#' @return A table of `commits_stats` class.
#' @examples
#' \dontrun{
#'  my_gitstats <- create_gitstats() |>
#'    set_github_host(
#'      token = Sys.getenv("GITHUB_PAT"),
#'      repos = c("r-world-devs/GitStats", "openpharma/visR")
#'    ) |>
#'    get_commits(my_gitstats, since = "2022-01-01") |>
#'    get_commits_stats(
#'      time_aggregation = "year",
#'      group_var = author
#'    )
#' }
#' @export
get_commits_stats <- function(commits,
                              time_aggregation = c("year", "month", "week", "day"),
                              group_var) {
  if (!inherits(commits, "gitstats_commits")) {
    cli::cli_abort(c(
      "x" = "`commits` must be a `gitstats_commits` object.",
      "i" = "Pull first your commits with `get_commits()` function."
    ))
  }
  commits <- commits |>
    dplyr::mutate(
      stats_date = lubridate::floor_date(
        committed_date,
        unit = time_aggregation
      ),
      githost = retrieve_githost(api_url)
    )
  commits_stats <- commits |>
    dplyr::group_by(stats_date, githost, {{ group_var }}) |>
    dplyr::summarise(
      stats = dplyr::n()
    ) |>
    dplyr::arrange(
      stats_date
    )
  commits_stats <- set_commits_stats_class(
    object = commits_stats,
    time_aggregation = time_aggregation
  )
  return(commits_stats)
}

set_commits_stats_class <- function(object, time_aggregation) {
  stopifnot(inherits(object, "grouped_df"))
  object <- dplyr::ungroup(object)
  class(object) <- append(class(object), "gitstats_commits_stats")
  attr(object, "time_aggregation") <- time_aggregation
  object
}

test_that("get_commits_stats method works", {
  commits_stats <- get_commits_stats(
    commits = test_mocker$use("commits_data"),
    time_aggregation = "month",
    group_var = organization
  )
  expect_s3_class(commits_stats, "gitstats_commits_stats")
  expect_equal(
    colnames(commits_stats),
    c("stats_date", "githost", "organization", "stats")
  )
  expect_true(
    all(c("gitlab", "github") %in% commits_stats$githost)
  )
  commits_stats_yearly <- get_commits_stats(
    commits = test_mocker$use("commits_data"),
    time_aggregation = "year"
  )
  expect_equal(commits_stats_yearly$stats_date,
               as.POSIXct(c(rep("2023-01-01", 2), "2024-01-01"), tz = 'UTC'))
  expect_s3_class(commits_stats_yearly, "gitstats_commits_stats")
  expect_equal(
    colnames(commits_stats_yearly),
    c("stats_date", "githost", "stats")
  )
})

test_that("get_commits_stats returns error when commits is not commits_data object", {
  expect_snapshot_error(
    get_commits_stats(
      commits = test_mocker$use("gh_commits_table")
    )
  )
})

EngineRest <- R6::R6Class("EngineRest",
  inherit = Engine,
  public = list(
    rest_api_url = NULL,
    initialize = function(rest_api_url = NA,
                          token = NA,
                          scan_all = FALSE) {
      private$engine <- "rest"
      private$token <- token
      self$rest_api_url <- rest_api_url
      private$scan_all <- scan_all
      private$set_endpoints()
    },

    response = function(endpoint,
                        token = private$token,
                        verbose) {
      response <- self$perform_request(
        endpoint = endpoint,
        token = token,
        verbose = verbose
      )
      result <- response |>
        httr2::resp_body_json(check_type = FALSE) |>
        private$set_rest_error_class()
      return(result)
    },

    perform_request = function(endpoint, token, verbose) {
      resp <- httr2::request(endpoint) |>
        httr2::req_headers("Authorization" = paste0("Bearer ", token)) |>
        httr2::req_error(is_error = function(resp) FALSE) |>
        httr2::req_perform()
      if (resp$status_code == 401 && verbose) {
        cli::cli_alert_danger("HTTP 401 Unauthorized.")
      }
      if (resp$status_code == 404 && verbose) {
        cli::cli_alert_danger("HTTP 404 Not Found.")
      }
      if (resp$status_code %in% c(400, 500, 403)) {
        resp <- httr2::request(endpoint) |>
          httr2::req_headers("Authorization" = paste0("Bearer ", token)) |>
          httr2::req_retry(
            is_transient = ~ httr2::resp_status(.x) %in% c(400, 500, 403),
            max_seconds = 60
          ) |>
          httr2::req_perform()
      }
      return(resp)
    }

  ),
  private = list(
    set_rest_error_class = function(resp) {
      if (!is.null(resp$message) && grepl("404", resp$message)) {
        class(resp) <- c("rest_error", "404_not_found", class(resp))
      }
      resp
    },

    get_contributors_from_repo = function(contributors_endpoint, user_name, verbose) {
      contributors_response <- private$paginate_results(
        endpoint = contributors_endpoint,
        verbose = verbose
      )
      contributors_vec <- contributors_response |>
        purrr::map_chr(~ eval(user_name)) |>
        paste0(collapse = ", ")
      return(contributors_vec)
    },

    paginate_results = function(endpoint, joining_sign = "?", verbose = TRUE) {
      full_response <- list()
      page <- 1
      repeat {
        endpoint_with_pagination <- paste0(endpoint, joining_sign, "per_page=100&page=", page)
        response_page <- self$response(
          endpoint = endpoint_with_pagination,
          verbose = verbose
        )
        if (inherits(response_page, "rest_error")) {
          response_page <- NULL
        }
        if (length(response_page) > 0) {
          full_response <- append(full_response, response_page)
          page <- page + 1
        } else {
          break
        }
      }
      return(full_response)
    }
  )
)

test_that("`perform_request()` returns proper status when token is empty or invalid", {
  testthat::skip_on_cran()
  wrong_tokens <- c("", "bad_token")
  purrr::walk(
    wrong_tokens,
    ~ expect_message(
      test_rest_github$perform_request(
        endpoint = "https://api.github.com",
        token = .,
        verbose = TRUE
      ),
      "HTTP 401 Unauthorized."
    )
  )
})

test_that("`perform_request()` throws error on bad host", {
  bad_host <- "https://github.bad_host.com"
  expect_error(
    suppressMessages(
      test_rest_github$perform_request(
        endpoint = paste0(bad_host),
        token = Sys.getenv("GITHUB_PAT"),
        verbose = TRUE
      )
    ),
    "Could not resolve host"
  )
})

test_that("403 error is handled", {
  testthat::skip_on_cran()
  mockery::stub(
    test_rest_github$perform_request,
    "httr2::req_perform",
    httr2::response(status_code = 403)
  )
  expect_no_error(
    response <- test_rest_github$perform_request(
      endpoint =  "https://example.com",
      token = Sys.getenv("GITHUB_PAT")
    )
  )
})

test_that("`perform_request()` returns proper status", {
  skip_if(Sys.getenv("GITHUB_PAT") == "")
  bad_endpoint <- "https://api.github.com/orgs/everybody_loves_somebody"
  expect_message(
    test_rest_github$perform_request(
      endpoint = bad_endpoint,
      token = Sys.getenv("GITHUB_PAT"),
      verbose = TRUE
    ),
    "HTTP 404 Not Found"
  )
})

test_that("`paginate_results()` works properly", {
  if (!integration_tests_skipped) {
    expect_no_error(
      test_rest_github_priv$paginate_results(
        endpoint = "https://api.github.com/orgs/pharmaverse/repos"
      )
    )
  }
})

test_that("`perform_request()` returns status 200", {
  skip_if(Sys.getenv("GITHUB_PAT") == "")
  response <- test_rest_github$perform_request(
    endpoint = "https://api.github.com/repos/r-world-devs/GitStats",
    token = Sys.getenv("GITHUB_PAT"),
    verbose = FALSE
  )
  expect_equal(
    response$status_code,
    200
  )
})

test_that("`perform_request()` for GraphQL returns status 200", {
  skip_if(Sys.getenv("GITHUB_PAT") == "")
  response <- test_graphql_github_priv$perform_request(
    gql_query = "{
      viewer {
        login
      }
    }",
    vars = NULL,
    token = Sys.getenv("GITHUB_PAT")
  )
  expect_equal(
    response$status_code,
    200
  )
})

test_that("`perform_request()` for GraphQL handles error 400", {
  skip_if(Sys.getenv("GITHUB_PAT") == "")
  testthat::skip_on_cran()
  mockery::stub(
    test_graphql_github_priv$perform_request,
    "graphql_perform_request",
    httr2::response(status_code = 400)
  )
  response <- test_graphql_github_priv$perform_request(
    gql_query = "{
      viewer {
        login
      }
    }",
    vars = NULL,
    token = Sys.getenv("GITHUB_PAT")
  )
  expect_equal(
    response$status_code,
    400
  )
})

show_message <- function(host,
                         engine,
                         scope = NULL,
                         code = NULL,
                         information) {
  msg_graphql <- cli::col_yellow('GraphQl')
  msg_rest <- cli::col_green('REST')
  engine_msg <- if (engine == "graphql") {
    msg_graphql
  } else if (engine == "rest") {
    msg_rest
  } else if (engine == "both") {
    paste0(msg_rest, "&", msg_graphql)
  }
  information <- cli::col_br_blue(information)
  message <- if (is.null(scope)) {
    glue::glue("[Host:{host}][Engine:{engine_msg}] {information}...")
  } else {
    glue::glue("[Host:{host}][Engine:{engine_msg}][Scope:{scope}] {information}...")
  }
  cli::cli_alert(message)
}

set_repo_scope <- function(org, private) {
  paste0(org, ": ", length(private$orgs_repos[[org]]), " repos")
}

cut_item_to_print <- function(item_to_print) {
  if (length(item_to_print) < 10) {
    list_items <- paste0(item_to_print, collapse = ", ")
  } else {
    item_to_print_cut <- item_to_print[1:10]
    list_items <- paste0(item_to_print_cut, collapse = ", ") |>
      paste0("... and ", length(item_to_print) - 10, " more")
  }
  return(list_items)
}

github_open_issue_edge <- function(created_at) {
  list(
    "node" = list(
      "number" = 1,
      "title" = "test",
      "description" = "test",
      "created_at" = created_at,
      "closed_at" = "",
      "state" = "open",
      "url" = "test",
      "author" = list(
        "login" = "test"
      )
    )
  )
}

github_closed_issue_edge <- function(created_at) {
  list(
    "node" = list(
      "number" = 1,
      "title" = "test",
      "description" = "test",
      "created_at" = created_at,
      "closed_at" = "2025-01-01T00:00:00Z",
      "state" = "closed",
      "url" = "test",
      "author" = list(
        "login" = "test"
      )
    )
  )
}

open_issue_timestamps <- generate_random_timestamps(25, 2023, 2025)
set.seed(234)
closed_issue_timestamps <- generate_random_timestamps(25, 2023, 2025)

open_issues <- purrr::map(open_issue_timestamps, github_open_issue_edge)
closed_issues <- purrr::map(closed_issue_timestamps, github_closed_issue_edge)

test_fixtures$github_graphql_issues_response <- list(
  "data" = list(
    "repository" = list(
      "issues" = list(
        "pageInfo" = list(
          "hasNextPage" = FALSE,
          "endCursor" = ""
        ),
        "edges" = append(open_issues, closed_issues)
      )
    )
  )
)


test_fixtures$gitlab_graphql_issues_response <- list(
  "data" = list(
    "project" = list(
      "issues" = list(
        "pageInfo" = list(
          "hasNextPage" = FALSE,
          "endCursor" = ""
        ),
        "edges" = append(open_issues, closed_issues)
      )
    )
  )
)

test_fixtures$github_open_issue_response <- list(
  "url" = "test",
  "repository_url" = "test",
  "labels_url" = "test",
  "comments_url" = "test",
  "events_url" = "test",
  "html_url" = "test",
  "id" = 111111L,
  "node_id" = "xxxxxx",
  "number" = 1L,
  "title" = "Test title",
  "user" = list(),
  "labels" = list(),
  "state" = "open"
)

test_fixtures$github_closed_issue_response <- list(
  "url" = "test",
  "repository_url" = "test",
  "labels_url" = "test",
  "comments_url" = "test",
  "events_url" = "test",
  "html_url" = "test",
  "id" = 111111L,
  "node_id" = "xxxxxx",
  "number" = 1L,
  "title" = "Test title",
  "user" = list(),
  "labels" = list(),
  "state" = "closed"
)

test_fixtures$github_issues_response <- list(
  test_fixtures$github_open_issue_response,
  test_fixtures$github_open_issue_response,
  test_fixtures$github_closed_issue_response
)

test_fixtures$gitlab_issues_response <- list(
  "statistics" = list(
    "counts" = list(
      "all" = 3,
      "closed" = 2,
      "opened" = 1
    )
  )
)

#' @title Get files
#' @name get_files
#' @description Pulls text files and their content.
#' @param gitstats A `GitStats` object.
#' @param pattern A regular expression. If defined, it pulls content of all
#'   files in a repository matching this pattern reaching to the level of
#'   directories defined by `depth` parameter. Works only if `file_path` stays
#'   `NULL`.
#' @param depth Defines level of directories to retrieve files from. E.g. if set
#'   to `0`, it will pull files only from `root`, if `1L`, will take data from
#'   `root` directory and directories visible in `root` directory. If left with
#'   no argument, will pull files from all directories.
#' @param file_path A specific path to file(s) in repositories. May be a
#'   character vector if multiple files are to be pulled. If defined, the
#'   function pulls content of this specific `file_path`. Can be defined only if
#'   `pattern` stays `NULL`.
#' @param cache A logical, if set to `TRUE` `GitStats` will retrieve the last
#'   result from its storage.
#' @param verbose A logical, `TRUE` by default. If `FALSE` messages and printing
#'   output is switched off.
#' @param progress A logical, by default set to `verbose` value. If `FALSE` no
#'   `cli` progress bar will be displayed.
#' @details `get_files()` may be used in two ways: either with `pattern` (with
#'   optional `depth`) or `file_path` argument defined.
#'
#'   In the first scenario, `GitStats` will pull first a files structure
#'   responding to the passed `pattern` and `depth` arguments and afterwards
#'   files content for all of these files. In the second scenario, `GitStats`
#'   will pull only the content of files for the specific `file_path` of the
#'   repository.
#'
#'   If user wants to pull a particular file or files, a `file_path` approach
#'   seems more reasonable, as it is a faster way since it omits pulling the
#'   whole file structure from the repo.
#'
#'   For example, if user wants to pull content of `README.md` and/or `NEWS.md`
#'   files placed in the `root` directories of the repositories, he should take
#'   the `file_path` approach as he already knows precisely paths of the files.
#'
#'   On the other hand, if user wants to pull specific type of files (e.g. all
#'   `.md` or `.Rmd` files in the repository), without knowing their path, it is
#'   recommended to use a `pattern` approach, which will trigger `GitStats` to
#'   find all the files in the repository on the given level of directories
#'   (`pattern` argument) and afterwards pull their content.
#'
#'   The latter approach is slower than the former but may be more useful
#'   depending on users' goals. Both approaches return data in the same format:
#'   `tibble` with data on `files`, namely their `path` and their `content`.
#'
#' @examples
#' \dontrun{
#'  git_stats <- create_gitstats() |>
#'   set_github_host(
#'     token = Sys.getenv("GITHUB_PAT"),
#'     orgs = c("r-world-devs")
#'   ) |>
#'   set_gitlab_host(
#'     token = Sys.getenv("GITLAB_PAT_PUBLIC"),
#'     orgs = "mbtests"
#'   )
#'
#'  rmd_files <- get_files(
#'    gitstats = git_stats,
#'    pattern = "\\.Rmd",
#'    depth = 2L
#'  )
#'
#'  app_files <- get_files(
#'    gitstats = git_stats,
#'    file_path = c("R/app.R", "R/ui.R", "R/server.R")
#'  )
#'
#' }
#' @return A data.frame.
#' @export
get_files <- function(gitstats,
                      pattern = NULL,
                      depth = Inf,
                      file_path = NULL,
                      cache = TRUE,
                      verbose = FALSE,
                      progress = TRUE) {
  start_time <- Sys.time()
  if (!is.null(pattern) && !is.null(file_path)) {
    cli::cli_abort(
      "Please choose either `pattern` or `file_path`.",
      call = NULL
    )
  }
  files <- gitstats$get_files(
    pattern = pattern,
    depth = depth,
    file_path = file_path,
    cache = cache,
    verbose = verbose,
    progress = progress
  )
  end_time <- Sys.time()
  time_taken <- end_time - start_time
  cli::cli_alert_success("Data pulled in {round(time_taken, 1)} {attr(time_taken, 'units')}")
  return(files)
}

GQLQueryGitHub <- R6::R6Class("GQLQueryGitHub",
  public = list(

    orgs = function(end_cursor) {
      paste0(
        'query {
          search(first: 100, type: USER, query: "type:org" ', private$add_cursor(end_cursor), ') {
            pageInfo {
               hasNextPage
               endCursor
            }
            edges {
              node{
                ... on Organization {', private$org_fields, '}
              }
            }
          }
        }'
      )
    },

    org = function() {
      paste0(
        '
        query GetOrg($org: String!) {
          organization(login: $org) {', private$org_fields, '}
        }'
      )
    },

    user_or_org_query =
      '
      query ($login: String!) {
        user(login: $login) {
          __typename
          login
        }
        organization(login: $login) {
          __typename
          login
        }
      }'
    ,

    repos_by_ids = function() {
      paste0(
        '
        query GetReposByIds($ids: [ID!]!) {
          nodes(ids: $ids) {
            ... on Repository {', private$repo_node_data, '
            }
          }
        }
        '
      )
    },

    repos_by_org = function(repo_cursor) {
      paste0('
        query GetReposByOrg($login: String!) {
          repositoryOwner(login: $login) {
            ... on Organization {
              ', private$repositories_field(repo_cursor), '
            }
          }
        }')
    },

    repos_by_user = function(repo_cursor) {
      paste0('
        query GetUsersRepos($login: String!){
          user(login: $login) {
            ', private$repositories_field(repo_cursor), '
          }
        }'
      )
    },

    user = function() {
      paste0('
        query GetUser($user: String!) {
          user(login: $user) {
            id
            name
            login
            email
            location
            starred_repos: starredRepositories {
              totalCount
            }
            contributions: contributionsCollection {
              totalIssueContributions
              totalCommitContributions
              totalPullRequestContributions
              totalPullRequestReviewContributions
            }
            avatar_url: avatarUrl
            web_url: websiteUrl
          }
        }')
    },

    commits_from_repo = function(commits_cursor = "") {
      paste0('
      query GetCommitsFromRepo($repo: String!
                               $org: String!
                               $since: GitTimestamp
                               $until: GitTimestamp){
          repository(name: $repo, owner: $org) {
            defaultBranchRef {
              target {
                ... on Commit {
                  history(since: $since
                          until: $until
                          ', private$add_cursor(commits_cursor), ') {
                    pageInfo {
                      hasNextPage
                      endCursor
                    }
                    edges {
                      node {
                        ... on Commit {
                          id
                          committed_date: committedDate
                          author {
                            name
                            user {
                              name
                              login
                            }
                          }
                          additions
                          deletions
                          repository {
                            url
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }')
    },

    issues_from_repo = function(issues_cursor = "") {
      paste0('
      query getIssuesFromRepo ($repo: String!
                               $org: String!) {
          repository(name: $repo, owner: $org) {
            issues(first: 100
                   ', private$add_cursor(issues_cursor), ') {
              pageInfo {
                hasNextPage
      				  endCursor
      				}
              edges {
                node {
                  number
                  title
                  description: body
                  created_at: createdAt
                  closed_at: closedAt
                  state
                  url
                  author {
                    login
                  }
                }
              }
            }
          }
        }
      ')
    },

    pull_requests_from_repo = function(pr_cursor = "") {
      paste0(
        'query GetPullRequestsFromRepo($org: String!, $repo: String!) {
          repository(owner: $org, name: $repo) {
            pullRequests(first: 100, ', private$add_cursor(pr_cursor), ') {
              pageInfo {
                 hasNextPage
                endCursor
              }
              edges {
                node {
                  number
                  created_at: createdAt
                  merged_at: mergedAt
                  state
                  author {
                    login
                  }
                  source_branch: headRefName
                  target_branch: baseRefName
                }
              }
            }
          }
        }'
      )
    },

    file_blob_from_repo = function() {
      'query GetFileBlobFromRepo($org: String!, $repo: String!, $expression: String!) {
          repository(owner: $org, name: $repo) {
            repo_id: id
            repo_name: name
            repo_url: url
            defaultBranchRef {
              target {
                ... on Commit {
                  oid
                }
              }
            }
            file: object(expression: $expression) {
              ... on Blob {
                text
                byteSize
                oid
              }
            }
          }
      }'
    },

    releases_from_repo = function() {
      'query GetReleasesFromRepo ($org: String!, $repo: String!) {
          repository(owner:$org, name:$repo){
            name
            url
            releases (last: 100) {
              nodes {
                name
                tagName
                publishedAt
                url
                description
              }
            }
          }
        }'
    }
  ),
  private = list(
    add_cursor = function(cursor) {
      if (nchar(cursor) == 0) {
        cursor_argument <- cursor
      } else {
        cursor_argument <- paste0('after: "', cursor, '"')
      }
      return(cursor_argument)
    },

    org_fields = '
      name
      description
      login
      url
      repositories (first: 100) {
        totalCount
      }
      membersWithRole(first: 100) {
        totalCount
      }
      avatarUrl
    ',

    repositories_field = function(repo_cursor) {
      paste0('
      repositories(first: 100', private$add_cursor(repo_cursor), ') {
        totalCount
        pageInfo {
          endCursor
          hasNextPage
        }
        nodes {', private$repo_node_data, '}
      }')
    },

    repo_node_data = '
      repo_id: id
      repo_name: name
      repo_path: name
      repo_fullpath: nameWithOwner
      default_branch: defaultBranchRef {
        name
      }
      stars: stargazerCount
      forks: forkCount
      created_at: createdAt
      last_activity_at: pushedAt
      languages(first: 5) {
        nodes {
          name
        }
      }
      issues_open: issues(first: 100, states: [OPEN]) {
        totalCount
      }
      issues_closed: issues(first: 100, states: [CLOSED]) {
        totalCount
      }
      organization: owner {
        login
      }
      repo_url: url
      defaultBranchRef {
        target {
          ... on Commit {
            oid
          }
        }
      }
    '
  )
)

# This is a solution for warnings in r-cmd checks of global variables
globalVariables(c(
  ".", "fullname", "organization", "repo_url",
  "repo_name", "created_at", "last_activity_at", "last_activity", "stats_date",
  "committed_date", "commits_n", "api_url", "row_no", ".N", ".data",
  "repository", "stars", "forks", "languages", "issues_open", "issues_closed",
  "contributors_n", "githost", "head", "state", "merged_at"
))

cli_icons <- list(
  org = "\U1F3DB\UFE0F",
  repo = "\U1F310",
  commit = "\U1F552",
  issue = "\U1F41B",
  language = "\U1F3AD",
  pull_request = "\U1F500",
  user = "\U1F64D",
  file = "\U1F4C4",
  tree = "\U1F333",
  release = "\U1F3F7\UFE0F"
)

text_ext_files <- "\\.(txt|md|qmd|Rmd|markdown|yaml|yml|csv|json|xml|html|htm|css|js|R|r|py|sh|bat|ini|conf|log|sql|tsv|mdx)$"
no_ext_files <- "^[^\\.]+$"

text_files_pattern <- paste0("(", text_ext_files, "|", no_ext_files, ")")

mirai::daemons(0)

integration_tests_skipped <- Sys.getenv("GITSTATS_INTEGRATION_TEST_SKIPPED", unset = "true") |>
  as.logical()
gitlab_group <- "mbtests"
github_org <- "r-world-devs"
if (integration_tests_skipped) {
  github_token <- NULL
} else {
  github_token <- Sys.getenv("GITHUB_PAT")
}

test_mocker <- Mocker$new()

test_gitstats <- create_test_gitstats(hosts = 2)
test_gitstats_priv <- create_test_gitstats(hosts = 2, priv_mode = TRUE)

test_gqlquery_gh <- GQLQueryGitHub$new()
test_gqlquery_gl <- GQLQueryGitLab$new()

test_rest_github <- TestEngineRestGitHub$new(
  rest_api_url = "https://api.github.com",
  token = github_token
)
test_rest_github_priv <- environment(test_rest_github$initialize)$private

test_graphql_github <- EngineGraphQLGitHub$new(
  gql_api_url = "https://api.github.com/graphql",
  token = github_token
)
test_graphql_github_priv <- environment(test_graphql_github$initialize)$private

test_rest_gitlab <- TestEngineRestGitLab$new(
  rest_api_url = "https://gitlab.com/api/v4",
  token = NULL
)
test_rest_gitlab_priv <- environment(test_rest_gitlab$initialize)$private

test_graphql_gitlab <- EngineGraphQLGitLab$new(
  gql_api_url = "https://gitlab.com/api/graphql",
  token = NULL
)
test_graphql_gitlab_priv <- environment(test_graphql_gitlab$initialize)$private

github_testhost <- create_github_testhost(
  orgs = github_org,
  token = github_token
)

github_testhost_priv <- create_github_testhost(
  orgs = github_org,
  token = github_token,
  mode = "private"
)

gitlab_testhost <- create_gitlab_testhost(
  orgs = gitlab_group
)

gitlab_testhost_priv <- create_gitlab_testhost(
  orgs = gitlab_group,
  mode = "private"
)

# ---- Integration test utils ----
skip_integration_tests <- function() {
  Sys.setenv(GITSTATS_INTEGRATION_TEST_SKIPPED = "true")
}

unskip_integration_tests <- function() {
  Sys.setenv(GITSTATS_INTEGRATION_TEST_SKIPPED = "false")
}

are_integrations_tests_skipped <- function() {
  as.logical(Sys.getenv("GITSTATS_INTEGRATION_TEST_SKIPPED"))
}

# ---- Mocker ----
Mocker <- R6::R6Class("Mocker",
  public = list(

    storage = list(),

    cache = function(object = NULL) {
      object_name <- deparse(substitute(object))
      self$storage[[paste0(object_name)]] <- object
    },

    use = function(object_name) {
      self$storage[[paste0(object_name)]]
    }
  )
)

# ---- Test host classes and factories ----
GitHostGitHubTest <- R6::R6Class(
  classname = "GitHostGitHubTest",
  inherit = GitHostGitHub,
  public = list(
    initialize = function(orgs = NA,
                          repos = NA,
                          token = NA,
                          host = NA) {
      private$set_api_url(host)
      private$set_web_url(host)
      private$set_endpoints()
      private$is_public <- FALSE
      private$token <- token
      private$set_graphql_url()
      private$set_orgs_and_repos_mocked(orgs, repos)
      private$setup_test_engines()
      private$set_searching_scope(orgs, repos, verbose = FALSE)
    }
  ),
  private = list(
    set_orgs_and_repos_mocked = function(orgs, repos) {
      if (is.null(orgs) && is.null(repos)) {
        private$scan_all <- TRUE
      } else {
        private$orgs <- orgs
      }
      if (!is.null(repos)) {
        private$repos <- repos
        orgs_repos <- private$extract_repos_and_orgs(repos)
        private$orgs <- names(orgs_repos)
      }
    },
    setup_test_engines = function() {
      private$engines$rest <- TestEngineRestGitHub$new(
        token = private$token,
        rest_api_url = private$api_url
      )
      private$engines$graphql <- EngineGraphQLGitHub$new(
        token = private$token,
        gql_api_url = private$set_graphql_url()
      )
    }
  )
)

GitHostGitLabTest <- R6::R6Class(
  classname = "GitHostGitLabTest",
  inherit = GitHostGitLab,
  public = list(
    initialize = function(orgs = NA,
                          repos = NA,
                          token = NA,
                          host = NA) {
      private$set_api_url(host)
      private$set_web_url(host)
      private$set_endpoints()
      private$check_if_public(host)
      private$token <- token
      private$set_graphql_url()
      private$set_orgs_and_repos_mocked(orgs, repos)
      private$setup_test_engines()
      private$set_searching_scope(orgs, repos, verbose = FALSE)
    }
  ),
  private = list(
    set_orgs_and_repos_mocked = function(orgs, repos) {
      private$orgs <- orgs
      if (!is.null(repos)) {
        private$repos <- repos
        orgs_repos <- private$extract_repos_and_orgs(repos)
        private$orgs <- names(orgs_repos)
      }
    },
    setup_test_engines = function() {
      private$engines$rest <- TestEngineRestGitLab$new(
        token = private$token,
        rest_api_url = private$api_url
      )
      private$engines$graphql <- EngineGraphQLGitLab$new(
        token = private$token,
        gql_api_url = private$set_graphql_url()
      )
    }
  )
)

create_github_testhost <- function(host  = NULL,
                                   orgs  = NULL,
                                   repos = NULL,
                                   token = NULL,
                                   mode = "") {
  suppressMessages(
    test_host <- GitHostGitHubTest$new(
      host  = host,
      token = token,
      orgs  = orgs,
      repos = repos
    )
  )
  if (mode == "private") {
    test_host <- environment(test_host$initialize)$private
  }
  return(test_host)
}

create_github_testhost_all <- function(host  = NULL,
                                       orgs  = NULL,
                                       repos = NULL,
                                       token = NULL,
                                       mode = "") {
  suppressMessages(
    test_host <- GitHostGitHubTest$new(
      host  = NULL,
      token = token,
      orgs  = orgs,
      repos = repos
    )
  )
  test_host$.__enclos_env__$private$orgs <- NULL
  test_host$.__enclos_env__$private$scan_all <- TRUE
  if (mode == "private") {
    test_host <- environment(test_host$initialize)$private
  }
  return(test_host)
}

create_gitlab_testhost <- function(host  = NULL,
                                   orgs  = NULL,
                                   repos = NULL,
                                   token = NULL,
                                   mode = "") {
  suppressMessages(
    test_host <- GitHostGitLabTest$new(
      host  = NULL,
      token = token,
      orgs  = orgs,
      repos = repos
    )
  )
  if (mode == "private") {
    test_host <- environment(test_host$initialize)$private
  }
  return(test_host)
}

# ---- Test engine classes and factories ----
TestEngineRest <- R6::R6Class("TestEngineRest",
  inherit = EngineRest,
  public = list(
    initialize = function(token,
                          rest_api_url) {
      private$token <- token
      self$rest_api_url <- rest_api_url
    }
  )
)

TestEngineRestGitHub <- R6::R6Class("TestEngineRestGitHub",
  inherit = EngineRestGitHub,
  public = list(
    initialize = function(token,
                          rest_api_url) {
      private$token <- token
      self$rest_api_url <- rest_api_url
      private$set_endpoints()
    }
  )
)

TestEngineRestGitLab <- R6::R6Class("TestEngineRestGitLab",
  inherit = EngineRestGitLab,
  public = list(
    initialize = function(token,
                          rest_api_url) {
      private$token <- token
      self$rest_api_url <- rest_api_url
      private$set_endpoints()
    }
  )
)

create_testrest <- function(rest_api_url = "https://api.github.com",
                            token,
                            mode = "") {
  test_rest <- TestEngineRest$new(
    token = token,
    rest_api_url = rest_api_url
  )
  if (mode == "private") {
    test_rest <- environment(test_rest$initialize)$private
  }
  return(test_rest)
}

# ---- Random data generators ----
generate_random_timestamps <- function(n, start_year, end_year) {
  start_date <- as.POSIXct(paste0(start_year, "-01-01 00:00:00"), tz = "UTC")
  end_date <- as.POSIXct(paste0(end_year, "-12-31 23:59:59"), tz = "UTC")

  random_times <- runif(n, min = as.numeric(start_date), max = as.numeric(end_date))
  random_datetimes <- as.POSIXct(random_times, origin = "1970-01-01", tz = "UTC")

  formatted_dates <- format(random_datetimes, "%Y-%m-%dT%H:%M:%SZ")

  return(formatted_dates)
}

generate_random_names <- function(n, names) {
  random_names <- sample(names, size = n, replace = TRUE)
  return(random_names)
}

test_that("get_issues_stats method works", {
  issues_stats <- get_issues_stats(
    issues = test_mocker$use("issues_data"),
    time_aggregation = "month",
    group_var = state
  )
  expect_s3_class(issues_stats, "gitstats_issues_stats")
  expect_equal(
    colnames(issues_stats),
    c("stats_date", "githost", "state", "stats")
  )
  expect_true(
    all(c("gitlab", "github") %in% issues_stats$githost)
  )
  issues_stats_yearly <- get_issues_stats(
    issues = test_mocker$use("issues_data"),
    time_aggregation = "year"
  )
  expect_in(as.POSIXct(c("2023-01-01", "2024-01-01", "2025-01-01"), tz = 'UTC'), issues_stats_yearly$stats_date)
  expect_s3_class(issues_stats_yearly, "gitstats_issues_stats")
  expect_equal(
    colnames(issues_stats_yearly),
    c("stats_date", "githost", "stats")
  )
})

test_that("get_issues_stats returns error when issues is not issues_data object", {
  expect_snapshot_error(
    get_issues_stats(
      issues = test_mocker$use("gh_issues_table")
    )
  )
})

gh_org <- "r-world-devs"
gh_repo <- "GitStats"

gh_md_files_structure <- list(
  "GitStats" = c("README.md", "CONTRIBUTING.md", "NEWS.md"),
  "GitAI" = c("README.md"),
  "cohortBuilder" = c("README.md", "vignettes/overview.Rmd"),
  "shinyCohortBuilder" = c("README.md"),
  "shinyGizmo" = c("README.md")
)
test_mocker$cache(gh_md_files_structure)

test_that("get_files_structure_from_orgs() works", {
  gh_md_files_with_ids <- purrr::imap(test_mocker$use("gh_md_files_structure"), function(files, repo_name) {
    attr(files, "repo_id") <- paste0("R_", repo_name)
    files
  })
  mockery::stub(
    github_testhost_priv$get_files_structure_from_orgs,
    "private$get_files_structure_from_repos_data",
    gh_md_files_with_ids
  )
  github_testhost_priv$searching_scope <- "org"
  gh_files_structure_from_orgs <- github_testhost_priv$get_files_structure_from_orgs(
    pattern = "\\.md|\\.qmd|\\.Rmd",
    depth = Inf,
    verbose = FALSE
  )
  expect_gt(
    length(gh_files_structure_from_orgs),
    0
  )
  expect_equal(
    names(gh_files_structure_from_orgs),
    gh_org
  )
  test_mocker$cache(gh_files_structure_from_orgs)
})

test_that("get_files_structure_from_orgs() prints message", {
  mockery::stub(
    github_testhost_priv$get_files_structure_from_orgs,
    "private$get_files_structure_from_repos_data",
    test_mocker$use("gh_md_files_structure")
  )
  github_testhost_priv$searching_scope <- "org"
  github_testhost_priv$cached_repos <- list()
  expect_snapshot(
    gh_files_structure_from_orgs <- github_testhost_priv$get_files_structure_from_orgs(
      pattern = "\\.md|\\.qmd|\\.Rmd",
      depth = Inf,
      verbose = TRUE
    )
  )
  github_testhost_priv$cached_repos <- list()
  expect_snapshot(
    gh_files_structure_from_orgs <- github_testhost_priv$get_files_structure_from_orgs(
      pattern = NULL,
      depth = Inf,
      verbose = TRUE
    )
  )
})

test_that("get_files_structure_from_repos() works", {
  test_org <- "r-world-devs"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    github_testhost_priv$get_files_structure_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  mockery::stub(
    github_testhost_priv$get_files_structure_from_repos,
    "private$get_files_structure_from_repos_data",
    test_mocker$use("gh_md_files_structure")
  )
  github_testhost_priv$searching_scope <- "repo"
  gh_files_structure_from_repos <- github_testhost_priv$get_files_structure_from_repos(
    pattern = "\\.md|\\.qmd|\\.Rmd",
    depth = Inf,
    verbose = FALSE
  )
  expect_gt(
    length(gh_files_structure_from_repos),
    0
  )
  expect_equal(
    names(gh_files_structure_from_repos),
    gh_org
  )
  test_mocker$cache(gh_files_structure_from_repos)
})

test_that("get_files_structure_from_repos() prints message", {
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    github_testhost_priv$get_files_structure_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  mockery::stub(
    github_testhost_priv$get_files_structure_from_repos,
    "private$get_files_structure_from_repos_data",
    test_mocker$use("gh_md_files_structure")
  )
  github_testhost_priv$searching_scope <- "repo"
  github_testhost_priv$cached_repos <- list()
  expect_snapshot(
    gh_files_structure_from_repos <- github_testhost_priv$get_files_structure_from_repos(
      pattern = "\\.md|\\.qmd|\\.Rmd",
      depth = Inf,
      verbose = TRUE
    )
  )
  github_testhost_priv$cached_repos <- list()
  expect_snapshot(
    gh_files_structure_from_repos <- github_testhost_priv$get_files_structure_from_repos(
      pattern = NULL,
      depth = Inf,
      verbose = TRUE
    )
  )
})

test_that("when files_structure is empty, appropriate message is returned", {
  github_testhost_priv <- create_github_testhost(
    repos = c("r-world-devs/GitStats", "openpharma/DataFakeR", "openpharma/VisR"),
    mode = "private"
  )
  mockery::stub(
    github_testhost_priv$get_files_structure_from_repos,
    "private$get_files_structure_from_repos_data",
    list() |>
      purrr::set_names()
  )
  github_testhost_priv$searching_scope <- "repo"
  expect_snapshot(
    github_testhost_priv$get_files_structure_from_repos(
      pattern = "\\.png",
      depth = 1L,
      verbose = TRUE
    )
  )
})

test_that("get_path_from_files_structure gets file path from files structure", {
  test_graphql_github <- EngineGraphQLGitHub$new(
    gql_api_url = "https://api.github.com/graphql",
    token = Sys.getenv("GITHUB_PAT")
  )
  test_graphql_github <- environment(test_graphql_github$initialize)$private
  file_path <- test_graphql_github$get_path_from_files_structure(
    host_files_structure = test_mocker$use("gh_files_structure_from_orgs"),
    org = gh_org,
    repo = gh_repo
  )
  expect_type(file_path, "character")
  expect_gt(length(file_path), 0)
})

test_that("get_files_structure pulls files structure for repositories in orgs", {
  mockery::stub(
    github_testhost$get_files_structure,
    "private$get_files_structure_from_orgs",
    test_mocker$use("gh_files_structure_from_orgs")
  )
  gh_files_structure_from_orgs <- github_testhost$get_files_structure(
    pattern = "\\.md|\\.qmd",
    depth = 1L,
    verbose = FALSE
  )
  expect_equal(
    names(gh_files_structure_from_orgs),
    gh_org
  )
  purrr::walk(gh_files_structure_from_orgs[[1]], function(repo_files) {
    expect_true(any(grepl("\\.md|\\.Rmd", repo_files)))
  })
  test_mocker$cache(gh_files_structure_from_orgs)
})

test_that("get_files_structure runs when scope is set to scan whole host", {
  github_testhost$.__enclos_env__$private$scan_all <- TRUE
  github_testhost$.__enclos_env__$private$orgs <- NULL
  mockery::stub(
    github_testhost$get_files_structure,
    "private$get_orgs_from_host",
    c("test_org")
  )
  mockery::stub(
    github_testhost$get_files_structure,
    "private$get_files_structure_from_orgs",
    test_mocker$use("gh_files_structure_from_orgs")
  )
  files_structure <- github_testhost$get_files_structure(
    pattern = "\\.md|\\.qmd",
    depth = 1L,
    verbose = TRUE
  )
  expect_equal(
    names(files_structure),
    "r-world-devs"
  )
})

test_that("get_files_structure pulls empty files structure", {
  gh_empty_files_stucture <- github_testhost$get_files_structure(
    pattern = "\\.test",
    depth = 1L,
    verbose = FALSE,
    progress = FALSE
  )
  expect_type(
    gh_empty_files_stucture,
    "list"
  )
  expect_length(
    gh_empty_files_stucture,
    0L
  )
  test_mocker$cache(gh_empty_files_stucture)
})

test_that("releases query is built properly", {
  gh_releases_query <-
    test_gqlquery_gh$releases_from_repo()
  expect_snapshot(
    gh_releases_query
  )
})

test_that("`get_releases_from_org()` pulls releases from the repositories", {
  mockery::stub(
    test_graphql_github$get_release_logs_from_org,
    "self$gql_response",
    test_fixtures$github_releases_from_repo
  )
  releases_from_repos <- test_graphql_github$get_release_logs_from_org(
    repos_names = c("TestProject1", "TestProject2"),
    org = "test_org"
  )
  expect_releases_github_response(releases_from_repos)
  test_mocker$cache(releases_from_repos)
})

test_that("`prepare_releases_table()` prepares releases table", {
  releases_table <- test_graphql_github$prepare_releases_table(
    releases_response = test_mocker$use("releases_from_repos"),
    org = "r-world-devs"
  )
  expect_releases_table(releases_table)
  test_mocker$cache(releases_table)
})

test_that("`get_release_logs_from_orgs()` works", {
  mockery::stub(
    github_testhost_priv$get_release_logs_from_orgs,
    "graphql_engine$prepare_releases_table",
    test_mocker$use("releases_table")
  )
  mockery::stub(
    github_testhost_priv$get_release_logs_from_orgs,
    "private$get_repos_data",
    test_mocker$use("gh_repos_data")
  )
  github_testhost_priv$searching_scope <- "org"
  releases_from_orgs <- github_testhost_priv$get_release_logs_from_orgs(
    verbose = FALSE,
    progress = FALSE
  )
  expect_releases_table(releases_from_orgs)
  test_mocker$cache(releases_from_orgs)
})

test_that("`get_release_logs_from_repos()` works", {
  mockery::stub(
    github_testhost_priv$get_release_logs_from_repos,
    "graphql_engine$prepare_releases_table",
    test_mocker$use("releases_table")
  )
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    github_testhost_priv$get_release_logs_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  github_testhost_priv$searching_scope <- "repo"
  github_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
  releases_from_repos <- github_testhost_priv$get_release_logs_from_repos(
    verbose = FALSE,
    progress = FALSE
  )
  expect_releases_table(releases_from_repos)
  test_mocker$cache(releases_from_repos)
})

test_that("`get_release_logs()` pulls release logs in the table format", {
  mockery::stub(
    github_testhost$get_release_logs,
    "private$get_release_logs_from_repos",
    test_mocker$use("releases_from_repos")
  )
  mockery::stub(
    github_testhost$get_release_logs,
    "private$get_release_logs_from_orgs",
    test_mocker$use("releases_from_orgs")
  )
  releases_table <- github_testhost$get_release_logs(
    since = "2023-05-01",
    until = "2023-09-30",
    verbose = FALSE,
    progress = FALSE
  )
  expect_releases_table(releases_table)
  expect_gt(min(releases_table$published_at), as.POSIXct("2023-05-01"))
  expect_lt(max(releases_table$published_at), as.POSIXct("2023-09-30"))
  test_mocker$cache(releases_table)
})

test_that("get_release_logs_from_orgs() pulls data from one day", {
  skip_if(integration_tests_skipped)
  github_testhost <- create_github_testhost(
    host = "github.com", 
    orgs = "r-world-devs",
    token = github_token
  )
  gitstats_release <- github_testhost$get_release_logs(
    since = "2025-12-08",
    until = "2025-12-08",
    verbose = FALSE,
    progress = FALSE
  )
  expect_releases_table(gitstats_release)
  expect_equal(nrow(gitstats_release), 1L)
  expect_equal(as.Date(gitstats_release$published_at), as.Date("2025-12-08"))
})

test_that("`get_release_logs()` is set to scan whole git host", {
  github_testhost_all <- create_github_testhost_all(orgs = "test_org")
  mockery::stub(
    github_testhost_all$get_release_logs,
    "private$get_orgs_from_host",
    "test_org"
  )
  mockery::stub(
    github_testhost_all$get_release_logs,
    "private$get_release_logs_from_repos",
    test_mocker$use("releases_from_repos")
  )
  mockery::stub(
    github_testhost_all$get_release_logs,
    "private$get_release_logs_from_orgs",
    test_mocker$use("releases_from_orgs")
  )
  expect_releases_table(
    github_testhost_all$get_release_logs(
      since = "2023-01-01",
      until = "2024-01-01",
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("releases query is built properly", {
  gl_releases_query <-
    test_gqlquery_gl$releases_from_repo()
  expect_snapshot(
    gl_releases_query
  )
})

test_that("`get_releases_from_org()` pulls releases from the repositories", {
  mockery::stub(
    test_graphql_gitlab$get_release_logs_from_org,
    "self$gql_response",
    test_fixtures$gitlab_releases_from_repo
  )
  releases_from_repos <- test_graphql_gitlab$get_release_logs_from_org(
    repos_names = c("TestProject1", "TestProject2"),
    org = "test_org"
  )
  expect_releases_gitlab_response(releases_from_repos)
  test_mocker$cache(releases_from_repos)
})


test_that("`prepare_releases_table()` prepares releases table", {
  releases_table <- test_graphql_gitlab$prepare_releases_table(
    releases_response = test_mocker$use("releases_from_repos"),
    org = "test_org"
  )
  expect_releases_table(releases_table)
  test_mocker$cache(releases_table)
})

test_that("`get_release_logs_from_orgs()` works", {
  mockery::stub(
    gitlab_testhost_priv$get_release_logs_from_orgs,
    "graphql_engine$prepare_releases_table",
    test_mocker$use("releases_table")
  )
  mockery::stub(
    gitlab_testhost_priv$get_release_logs_from_orgs,
    "private$get_repos_data",
    test_mocker$use("gl_repos_data")
  )
  gitlab_testhost_priv$searching_scope <- "org"
  releases_from_orgs <- gitlab_testhost_priv$get_release_logs_from_orgs(
    verbose = FALSE,
    progress = FALSE
  )
  expect_releases_table(releases_from_orgs)
  test_mocker$cache(releases_from_orgs)
})

test_that("`get_release_logs_from_orgs()` prints proper message", {
  mockery::stub(
    gitlab_testhost_priv$get_release_logs_from_orgs,
    "graphql_engine$prepare_releases_table",
    test_mocker$use("releases_table")
  )
  mockery::stub(
    gitlab_testhost_priv$get_release_logs_from_orgs,
    "private$get_repos_data",
    test_mocker$use("gl_repos_data")
  )
  gitlab_testhost_priv$searching_scope <- "org"
  expect_snapshot(
    releases_from_orgs <- gitlab_testhost_priv$get_release_logs_from_orgs(
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("`get_release_logs_from_repos()` works", {
  gitlab_testhost_priv <- create_gitlab_testhost(
    repos = "test_org/TestRepo",
    mode = "private"
  )
  mockery::stub(
    gitlab_testhost_priv$get_release_logs_from_repos,
    "graphql_engine$prepare_releases_table",
    test_mocker$use("releases_table")
  )
  gitlab_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
  expect_snapshot(
    releases_from_repos <- gitlab_testhost_priv$get_release_logs_from_repos(
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("`get_release_logs_from_repos()` works", {
  gitlab_testhost_priv <- create_gitlab_testhost(
    repos = "test_org/TestRepo",
    mode = "private"
  )
  mockery::stub(
    gitlab_testhost_priv$get_release_logs_from_repos,
    "graphql_engine$prepare_releases_table",
    test_mocker$use("releases_table")
  )
  gitlab_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
  releases_from_repos <- gitlab_testhost_priv$get_release_logs_from_repos(
    verbose = FALSE,
    progress = FALSE
  )
  expect_releases_table(releases_from_repos)
  test_mocker$cache(releases_from_repos)
})

test_that("`get_release_logs()` pulls release logs in the table format", {
  mockery::stub(
    gitlab_testhost$get_release_logs,
    "private$get_release_logs_from_repos",
    test_mocker$use("releases_from_repos")
  )
  mockery::stub(
    gitlab_testhost$get_release_logs,
    "private$get_release_logs_from_orgs",
    test_mocker$use("releases_from_orgs")
  )
  releases_table <- gitlab_testhost$get_release_logs(
    since = "2023-08-01",
    until = "2024-06-30",
    verbose = FALSE,
    progress = FALSE
  )
  expect_releases_table(releases_table)
  expect_gt(min(releases_table$published_at), as.POSIXct("2023-08-01"))
  expect_lt(max(releases_table$published_at), as.POSIXct("2024-06-30"))
  test_mocker$cache(releases_table)
})

test_that("user query is built properly", {
  gh_user_query <-
    test_gqlquery_gh$user()
  expect_snapshot(
    gh_user_query
  )
  test_mocker$cache(gh_user_query)
})

test_that("get_user pulls GitHub user response", {
  mockery::stub(
    test_graphql_github$get_user,
    "self$gql_response",
    test_fixtures$github_user_response
  )
  gh_user_response <- test_graphql_github$get_user(username = "testuser")
  expect_user_gql_response(
    gh_user_response
  )
  test_mocker$cache(gh_user_response)
})

test_that("GitHub prepares user table", {
  gh_user_table <- test_graphql_github$prepare_user_table(
    user_response = test_mocker$use("gh_user_response")
  )
  expect_users_table(
    gh_user_table,
    one_user = TRUE
  )
  test_mocker$cache(gh_user_table)
})

test_that("GitHost gets users tables", {
  mockery::stub(
    github_testhost$get_users,
    "graphql_engine$prepare_user_table",
    test_mocker$use("gh_user_table")
  )
  github_users <- github_testhost$get_users(
    users = c("testuser1", "testuser2")
  )
  expect_users_table(github_users)
  test_mocker$cache(github_users)
})

gl_group <- "mbtests"

gl_files_structure_shallow <- list(
  "gitstatstesting" = c("README.md"),
  "graphql_tests" = c("README.md", "CHANGELOG.md")
)
test_mocker$cache(gl_files_structure_shallow)

test_that("get_files_structure_from_orgs pulls files structure for repositories in orgs", {
  if (integration_tests_skipped) {
    mockery::stub(
      gitlab_testhost_priv$get_files_structure_from_orgs,
      "private$get_files_structure_from_repos_data",
      test_mocker$use("gl_files_structure_shallow")
    )
  }
  gl_files_structure_from_orgs <- gitlab_testhost_priv$get_files_structure_from_orgs(
    pattern = "\\.md",
    depth = 1L,
    verbose = FALSE,
    progress = FALSE
  )
  expect_equal(
    names(gl_files_structure_from_orgs),
    gl_group
  )
  purrr::walk(gl_files_structure_from_orgs[[1]], function(repo_files) {
    expect_true(all(grepl("\\.md", repo_files)))
  })
  test_mocker$cache(gl_files_structure_from_orgs)
})

test_that("get_files_structure_from_orgs prints messages without pattern", {
  mockery::stub(
    gitlab_testhost_priv$get_files_structure_from_orgs,
    "private$get_files_structure_from_repos_data",
    test_mocker$use("gl_files_structure_shallow")
  )
  gitlab_testhost_priv$searching_scope <- "org"
  expect_snapshot(
    gl_files_structure <- gitlab_testhost_priv$get_files_structure_from_orgs(
      pattern = NULL,
      depth = Inf,
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("get_files_structure_from_orgs warns when no structure found", {
  mockery::stub(
    gitlab_testhost_priv$get_files_structure_from_orgs,
    "private$get_files_structure_from_repos_data",
    list()
  )
  gitlab_testhost_priv$searching_scope <- "org"
  expect_snapshot(
    gl_files_structure <- gitlab_testhost_priv$get_files_structure_from_orgs(
      pattern = NULL,
      depth = Inf,
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("get_files_structure_from_repos pulls files structure for repositories", {
  mockery::stub(
    gitlab_testhost_priv$get_files_structure_from_repos,
    "private$get_files_structure_from_repos_data",
    test_mocker$use("gl_files_structure_shallow")
  )
  test_org <- "mbtests"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    gitlab_testhost_priv$get_files_structure_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  gitlab_testhost_priv$searching_scope <- "repo"
  gl_files_structure_from_repos <- gitlab_testhost_priv$get_files_structure_from_repos(
    pattern = "\\.md",
    depth = 1L,
    verbose = FALSE,
    progress = FALSE
  )
  expect_equal(
    names(gl_files_structure_from_repos),
    gl_group
  )
  purrr::walk(gl_files_structure_from_repos[[1]], function(repo_files) {
    expect_true(all(grepl("\\.md", repo_files)))
  })
  gitlab_testhost_priv$searching_scope <- "org"
})

test_that("get_files_structure_from_repos prints message", {
  mockery::stub(
    gitlab_testhost_priv$get_files_structure_from_repos,
    "private$get_files_structure_from_repos_data",
    test_mocker$use("gl_files_structure_shallow")
  )
  test_org <- "test_group"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    gitlab_testhost_priv$get_files_structure_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  gitlab_testhost_priv$searching_scope <- "repo"
  expect_snapshot(
    gl_files_structure_from_repos <- gitlab_testhost_priv$get_files_structure_from_repos(
      pattern = "\\.md",
      depth = 1L,
      verbose = TRUE,
      progress = FALSE
    )
  )
  gitlab_testhost_priv$searching_scope <- "org"
})

test_that("get_path_from_files_structure gets file path from files structure", {
  test_graphql_gitlab <- EngineGraphQLGitLab$new(
    gql_api_url = "https://gitlab.com/api/v4/graphql",
    token = Sys.getenv("GITHLAB_PAT_PUBLIC")
  )
  test_graphql_gitlab <- environment(test_graphql_gitlab$initialize)$private
  file_path <- test_graphql_gitlab$get_path_from_files_structure(
    host_files_structure = test_mocker$use("gl_files_structure_from_orgs"),
    org = gl_group # this will need fixing and repo parameter must come back
  )
  expect_type(file_path, "character")
  expect_gt(length(file_path), 0)
})

test_that("get_files_structure pulls files structure for repositories in orgs", {
  mockery::stub(
    gitlab_testhost$get_files_structure,
    "private$get_files_structure_from_orgs",
    test_mocker$use("gl_files_structure_from_orgs")
  )
  gl_files_structure_from_orgs <- gitlab_testhost$get_files_structure(
    pattern = "\\.md",
    depth = 1L,
    verbose = FALSE,
    progress = FALSE
  )
  expect_equal(
    names(gl_files_structure_from_orgs),
    gl_group
  )
  purrr::walk(gl_files_structure_from_orgs[[1]], function(repo_files) {
    expect_true(all(grepl("\\.md", repo_files)))
  })
  test_mocker$cache(gl_files_structure_from_orgs)
})

test_that("get_files_structure pulls empty files structure", {
  gl_empty_files_stucture <- gitlab_testhost$get_files_structure(
    pattern = "\\.test",
    depth = 1L,
    verbose = FALSE,
    progress = FALSE
  )
  expect_type(
    gl_empty_files_stucture,
    "list"
  )
  expect_length(
    gl_empty_files_stucture,
    0L
  )
  test_mocker$cache(gl_empty_files_stucture)
})

test_that("get_pull_requests_stats method works", {

  pr_stats <- get_pull_requests_stats(
    pull_requests = test_mocker$use("pr_data"),
    time_aggregation = "month",
    group_var = author
  )
  expect_s3_class(pr_stats, "gitstats_pr_stats")
  expect_equal(
    colnames(pr_stats),
    c("state", "created_at", "author", "stats")
  )
  pr_stats_yearly <- get_pull_requests_stats(
    pull_requests = test_mocker$use("pr_data"),
    time_aggregation = "year"
  )
  expect_in(as.POSIXct(c("2024-01-01", "2025-01-01", "2026-01-01"), tz = 'UTC'), pr_stats_yearly$created_at)
  expect_s3_class(pr_stats_yearly, "gitstats_pr_stats")
  expect_equal(
    colnames(pr_stats_yearly),
    c("state", "created_at", "stats")
  )
})

test_that("get_pull_requests_stats returns error when issues is not pr_data object", {
  expect_snapshot_error(
    get_pull_requests_stats(
      pull_requests = test_mocker$use("gh_prs_table")
    )
  )
})

#' Create GitStats object for tests
create_test_gitstats <- function(hosts = 0,
                                 priv_mode = FALSE,
                                 test_mocker = NULL,
                                 inject_repos = NULL,
                                 inject_commits = NULL,
                                 inject_files = NULL,
                                 inject_files_structure = NULL,
                                 inject_users = NULL,
                                 inject_package_usage = NULL) {
  test_gitstats <- create_gitstats()
  if (hosts == 1) {
    test_gitstats$.__enclos_env__$private$hosts[[1]] <- create_github_testhost(
      orgs = "test_org"
    )
  } else if (hosts == 2) {
    test_gitstats$.__enclos_env__$private$hosts[[1]] <- create_github_testhost(
      orgs = "github_test_org"
    )
    test_gitstats$.__enclos_env__$private$hosts[[2]] <- create_gitlab_testhost(
      orgs = "gitlab_test_group"
    )
  }
  storage <- test_gitstats$.__enclos_env__$private$storage_backend
  if (!is.null(inject_repos)) {
    storage$save("repositories", test_mocker$use(inject_repos))
  }
  if (!is.null(inject_commits)) {
    storage$save("commits", test_mocker$use(inject_commits))
  }
  if (!is.null(inject_files)) {
    storage$save("files", test_mocker$use(inject_files))
  }
  if (!is.null(inject_files_structure)) {
    storage$save("files_structure", test_mocker$use(inject_files_structure))
  }
  if (!is.null(inject_users)) {
    storage$save("users", test_mocker$use(inject_users))
  }
  if (!is.null(inject_package_usage)) {
    storage$save("R_package_usage", test_mocker$use(inject_package_usage))
  }
  if (priv_mode) {
    test_gitstats <- environment(test_gitstats$print)$private
  }
  return(test_gitstats)
}

github_commit_edge <- function(timestamp, author) {
  list(
    "node" = list(
      "id" = "xxx",
      "committed_date" = timestamp,
      "author" = list(
        "name" = author,
        "user" = list(
          "name" = "Maciej Banas",
          "login" = "maciekbanas"
        )
      ),
      "additions" = 5L,
      "deletions" = 8L,
      "repository" = list(
        "url" = "test_url"
      )
    )
  )
}

set.seed(123)
commit_timestamps <- generate_random_timestamps(25, 2023, 2024)
commit_authors <- generate_random_names(25, c("John Test", "Barbara Check", "Bob Test"))

test_fixtures$github_commits_response <- list(
  "data" = list(
    "repository" = list(
      "defaultBranchRef" = list(
        "target" = list(
          "history" = list(
            "edges" = purrr::map2(commit_timestamps, commit_authors, github_commit_edge)
          )
        )
      )
    )
  )
)

gitlab_commit <- list(
  "id" = "xxxxxxxxxxxxxxxxx",
  "short_id" = "xxx",
  "created_at" = "2023-04-05T12:07:50.000+00:00",
  "parent_ids" = list(
    "iiiiiiiiiiiiiii"
  ),
  "title" = "Test title",
  "message" = "Test title",
  "author_name" = "Maciej Banas",
  "author_email" = "testmail@test.com",
  "authored_date" = "2023-04-05T12:07:50.000+00:00",
  "committer_name" = "Maciej Banas",
  "committer_email" = "testmail@test.com",
  "committed_date" = "2023-04-05T12:07:50.000+00:00",
  "trailers" = list(),
  "extedned_trailers" = list(),
  "web_url" = "https://test_url.com/-/commit/sxsxsxsx",
  "stats" = list(
    "additions" = 1L,
    "deletions" = 0L,
    "total" = 1L
  )
)

test_fixtures$gitlab_commits_response <- rep(list(gitlab_commit), 5)

GitHost <- R6::R6Class(
  classname = "GitHost",
  public = list(

    initialize = function(orgs = NA,
                          repos = NA,
                          token = NA,
                          host = NA,
                          verbose = NA,
                          .error = TRUE) {
      private$set_api_url(host)
      private$set_web_url(host)
      private$set_endpoints()
      private$check_if_public(host)
      private$set_token(
        token = token,
        verbose = verbose
      )
      private$set_graphql_url()
      private$set_searching_scope(
        orgs = orgs,
        repos = repos,
        verbose = verbose
      )
      private$setup_engines()
      private$set_orgs_and_repos(
        orgs = orgs,
        repos = repos,
        verbose = verbose,
        .error = .error
      )
    },

    get_orgs = function(output, verbose) {
      if (private$scan_all) {
        orgs <- private$get_orgs_from_host(
          output = output,
          verbose = verbose
        )
      } else {
        orgs <- private$get_orgs_from_orgs_and_repos(
          output = output,
          verbose = verbose
        )
      }
      if (output == "full_table") {
        orgs <- orgs |>
          dplyr::mutate(
            host_url = private$api_url,
            host_name = private$host_name
          )
      }
      return(orgs)
    },

    get_repos = function(add_contributors = TRUE,
                         add_languages = TRUE,
                         with_code = NULL,
                         in_files = NULL,
                         with_file = NULL,
                         language = NULL,
                         output = "table",
                         verbose = TRUE,
                         progress = TRUE,
                         fill_empty_sha = FALSE) {
      if (is.null(with_code) && is.null(with_file)) {
        repos_table <- private$get_all_repos(
          add_languages = add_languages,
          verbose  = verbose,
          progress = progress,
          fill_empty_sha = fill_empty_sha
        )
      }
      if (!is.null(with_code)) {
        repos_table <- private$get_repos_with_code(
          code = with_code,
          in_files = in_files,
          language = language,
          output = output,
          verbose = verbose,
          progress = progress
        )
      } else if (!is.null(with_file)) {
        repos_table <- private$get_repos_with_code(
          code = with_file,
          in_path = TRUE,
          language = language,
          output = output,
          verbose = verbose,
          progress = progress
        )
      }
      if (output == "table") {
        repos_table <- private$add_repo_api_url(repos_table) |>
          private$add_githost_info()
        if (!is.null(language)) {
          repos_table <- private$filter_repos_table_by_language(
            repos_table = repos_table,
            language_filter = language
          )
        }
        if (!add_languages && "languages" %in% colnames(repos_table)) {
          repos_table <- repos_table |>
            dplyr::select(-languages)
        }
        if (add_contributors) {
          repos_table <- private$get_repos_contributors(
            repos_table = repos_table,
            verbose = verbose,
            progress = progress
          )
        }
      }
      return(repos_table)
    },

    get_repos_urls = function(type = "web",
                              with_code = NULL,
                              in_files = NULL,
                              with_file = NULL,
                              verbose = TRUE,
                              progress = TRUE) {
      if (!is.null(with_code)) {
        repo_urls <- private$get_repos_urls_with_code(
          type = type,
          code = with_code,
          in_files = in_files,
          verbose = verbose,
          progress = progress
        )
      } else if (!is.null(with_file)) {
        repo_urls <- private$get_repos_urls_with_code(
          type = type,
          code = with_file,
          in_path = TRUE,
          verbose = verbose,
          progress = progress
        )
      } else {
        repo_urls <- private$get_all_repos_urls(
          type = type,
          verbose = verbose,
          progress = progress
        )
      }
      return(repo_urls)
    },

    get_commits = function(since,
                           until,
                           verbose = TRUE,
                           progress = TRUE) {
      if (private$scan_all && is.null(private$orgs)) {
        private$orgs <- private$get_orgs_from_host(
          output = "only_names",
          verbose = verbose
        )
      }
      commits_from_orgs <- private$get_commits_from_orgs(
        since = since,
        until = until,
        verbose = verbose,
        progress = progress
      )
      commits_from_repos <- private$get_commits_from_repos(
        since = since,
        until = until,
        verbose = verbose,
        progress = progress
      )
      commits_table <- list(
        commits_from_orgs,
        commits_from_repos
      ) |>
        purrr::list_rbind()
      return(commits_table)
    },

    get_issues = function(since,
                          until,
                          state = NULL,
                          verbose = TRUE,
                          progress = TRUE) {
      if (private$scan_all && is.null(private$orgs)) {
        private$orgs <- private$get_orgs_from_host(
          output = "only_names",
          verbose = verbose
        )
      }
      issues_from_orgs <- private$get_issues_from_orgs(
        verbose = verbose,
        progress = progress
      )
      issues_from_repos <- private$get_issues_from_repos(
        verbose = verbose,
        progress = progress
      )
      issues_table <- list(
        issues_from_orgs,
        issues_from_repos
      ) |>
        purrr::list_rbind() |>
        dplyr::distinct()
      if (nrow(issues_table) > 0) {
        issues_table <- issues_table |>
          dplyr::filter(
            created_at >= since & created_at <= parse_until_param(until)
          )
        if (!is.null(state)) {
          type <- state
          issues_table <- issues_table |>
            dplyr::filter(
              state == type
            )
        }
      }
      return(issues_table)
    },

    get_pull_requests = function(since,
                                 until,
                                 state = NULL,
                                 verbose = TRUE,
                                 progress = TRUE) {
      if (private$scan_all && is.null(private$orgs)) {
        private$orgs <- private$get_orgs_from_host(
          output = "only_names",
          verbose = verbose
        )
      }
      pr_from_orgs <- private$get_pr_from_orgs(
        verbose = verbose,
        progress = progress
      )
      pr_from_repos <- private$get_pr_from_repos(
        verbose = verbose,
        progress = progress
      )
      pr_table <- list(
        pr_from_orgs,
        pr_from_repos
      ) |>
        purrr::list_rbind() |>
        dplyr::distinct()
      if (nrow(pr_table) > 0) {
        pr_table <- pr_table |>
          dplyr::filter(
            created_at >= since & created_at <= parse_until_param(until)
          )
        if (!is.null(state)) {
          type <- state
          pr_table <- pr_table |>
            dplyr::filter(
              state == type
            )
        }
      }
      return(pr_table)
    },

    get_users = function(users) {
      graphql_engine <- private$engines$graphql
      users_table <-  gitstats_map(users, function(user) {
        graphql_engine$get_user(user) |>
          graphql_engine$prepare_user_table()
      }) |>
        purrr::list_rbind()
      return(users_table)
    },

    get_files_content = function(file_path,
                                 files_structure = NULL,
                                 verbose = TRUE,
                                 progress = TRUE) {
      if (is.null(files_structure)) {
        if (!private$scan_all) {
          files_content_from_orgs <- private$get_files_content_from_orgs(
            file_path = file_path,
            verbose = verbose,
            progress = progress
          )
          files_content_from_repos <- private$get_files_content_from_repos(
            file_path = file_path,
            verbose = verbose,
            progress = progress
          )
          files_table <- rbind(
            files_content_from_orgs,
            files_content_from_repos
          )
        } else {
          files_table <- private$get_files_content_from_host(
            file_path = file_path,
            verbose = verbose,
            progress = progress
          )
        }
      }
      if (!is.null(files_structure)) {
        files_table <- private$get_files_content_from_files_structure(
          files_structure = files_structure,
          verbose = verbose,
          progress = progress
        )
      }
      return(files_table)
    },

    get_files_structure = function(pattern,
                                   depth,
                                   verbose  = TRUE,
                                   progress = TRUE) {
      if (private$scan_all && is.null(private$orgs)) {
        private$orgs <- private$get_orgs_from_host(
          output = "only_names",
          verbose = verbose
        )
      }
      files_structure_from_orgs <- private$get_files_structure_from_orgs(
        pattern = pattern,
        depth = depth,
        verbose = verbose,
        progress = progress
      )
      files_structure_from_repos <- private$get_files_structure_from_repos(
        pattern = pattern,
        depth = depth,
        verbose = verbose,
        progress = progress
      )
      files_structure <- append(
        files_structure_from_orgs %||% list(),
        files_structure_from_repos %||% list()
      )
      return(files_structure)
    },

    get_release_logs = function(since, until, verbose, progress) {
      if (private$scan_all && is.null(private$orgs)) {
        private$orgs <- private$get_orgs_from_host(
          output = "only_names",
          verbose = verbose
        )
      }
      release_logs_from_orgs <- private$get_release_logs_from_orgs(
        verbose = verbose,
        progress = progress
      )
      release_logs_from_repos <- private$get_release_logs_from_repos(
        verbose = verbose,
        progress = progress
      )
      release_logs_table <- list(
        release_logs_from_orgs,
        release_logs_from_repos
      ) |>
        purrr::list_rbind()
      if (nrow(release_logs_table) > 0) {
        release_logs_table <- release_logs_table |>
          dplyr::filter(
            published_at >= since & published_at <= parse_until_param(until)
          )
      }
      return(release_logs_table)
    },

    set_storage_backend = function(backend) {
      private$storage_backend <- backend
    }
  ),
  private = list(

    api_url = NULL,
    web_url = NULL,
    graphql_api_url = NULL,
    searching_scope = NULL,
    test_endpoint = NULL,
    endpoints = list(tokens = NULL, orgs = NULL, repositories = NULL),
    token = NULL,
    token_scopes = NULL,
    is_public = NULL,
    orgs = NULL,
    repos = NULL,
    repos_fullnames = NULL,
    orgs_repos = NULL,
    scan_all = FALSE,
    verbose = TRUE,
    engines = list(),
    cached_repos = list(),
    storage_backend = NULL,

    get_cached_repos = function(org) {
      private$cached_repos[[org]]
    },

    set_cached_repos = function(repos, org, verbose) {
      if (verbose) {
        cli::cli_alert("Caching repositories for [{org}]...")
      }
      private$cached_repos[[org]] <- repos
    },

    set_custom_api_url = function(host) {
      private$api_url <- if (!grepl("https|http", host)) {
        glue::glue(
          "https://{host}/api/v{private$api_version}"
        )
      } else {
        if (grepl("http(?!s)", host, perl = TRUE)) {
          host <- gsub("http", "https", host)
        }
        glue::glue(
          "{host}/api/v{private$api_version}"
        )
      }
    },

    set_custom_web_url = function(host) {
      private$web_url <- if (!grepl("https://", host)) {
        glue::glue(
          "https://{host}"
        )
      } else {
        host
      }
    },

    set_endpoints = function() {
      private$set_test_endpoint()
      private$set_tokens_endpoint()
      private$set_orgs_endpoint()
      private$set_repositories_endpoint()
    },

    set_token = function(token, verbose) {
      if (is.null(token)) {
        token <- private$set_default_token(
          verbose = verbose
        )
      } else {
        token <- private$check_token(token)
      }
      private$token <- token
    },

    check_token = function(token) {
      if (nchar(token) == 0) {
        cli::cli_abort(c(
          "i" = "No token provided."
        ))
      } else {
        if (!private$test_token(token)) {
          cli::cli_abort(c(
            "x" = "Token exists but does not grant access.",
            "i" = "Check if you use correct token.",
            "!" = "Scope that is needed: [{paste0(private$min_access_scopes, collapse = ', ')}]."
          ),
          call = NULL)
        } else {
          return(token)
        }
      }
    },

    set_searching_scope = function(orgs, repos, verbose) {
      if (is.null(repos) && is.null(orgs)) {
        if (verbose) {
          cli::cli_alert_info(cli::col_grey(
            "No `orgs` nor `repos` specified."
          ))
          cli::cli_alert_info(cli::col_grey(
            "Searching scope set to [all]."
          ))
        }
        private$searching_scope <- "all"
        private$scan_all <- TRUE
      }
      if (!is.null(repos)) {
        private$searching_scope <- c(private$searching_scope, "repo")
      }
      if (!is.null(orgs)) {
        private$searching_scope <- c(private$searching_scope, "org")
      }
    },

    set_orgs_and_repos = function(orgs, repos, verbose, .error) {
      if (!private$scan_all) {
        if (!is.null(orgs)) {
          private$orgs <- private$check_organizations(
            orgs = orgs,
            verbose = verbose,
            .error = .error
          )
        }
        if (!is.null(repos)) {
          repos <- private$check_repositories(
            repos = repos,
            verbose = verbose,
            .error = .error
          )
          private$repos_fullnames <- repos
          orgs_repos <- private$extract_repos_and_orgs(private$repos_fullnames)
          private$repos <- unname(unlist(orgs_repos))
          private$orgs_repos <- orgs_repos
        }
      }
    },

    check_repositories = function(repos, verbose, .error) {
      repos_without_org <- repos[!grepl("/", url_decode(repos))]
      if (length(repos_without_org) > 0) {
        cli::cli_abort(
          c(
            "x" = "Repository name must be provided as a full path: {.val org/repo_name}.",
            "i" = "Got: {.val {url_decode(repos_without_org)}}",
            "!" = "Pass the full path as seen in the web URL, e.g. {.val r-world-devs/GitStats}."
          ),
          call = NULL
        )
      }
      if (verbose) {
        cli::cli_alert(cli::col_grey("Checking repositories..."))
      }
      repos <- gitstats_map(repos, function(repo) {
        repo_endpoint <- glue::glue("{private$endpoints$repositories}/{repo}")
        check <- private$check_endpoint(
          endpoint = repo_endpoint,
          type = "Repository",
          verbose = verbose,
          .error = .error
        )
        if (!check) {
          repo <- NULL
        }
        return(repo)
      }, .progress = verbose) |>
        purrr::keep(~ length(.) > 0) |>
        unlist()
      if (length(repos) == 0) {
        return(NULL)
      }
      repos
    },

    check_organizations = function(orgs, verbose, .error) {
      if (verbose) {
        cli::cli_alert(cli::col_grey("Checking owners..."))
      }
      orgs <- private$engines$graphql$set_owner_type(
        owners = url_decode(orgs),
        verbose = verbose
      ) |>
        purrr::map(function(org) {
          if (attr(org, "type") == "not found") {
            if (.error) {
              cli::cli_abort(
                c(
                  "x" = "Org/user you provided does not exist or its name was passed
                  in a wrong way: {cli::col_red({url_decode(org)})}",
                  "!" = "Please type your org/user name the way you see it in
                  web URL."
                ),
                call = NULL
              )
            } else {
              if (verbose) {
                cli::cli_alert_warning(
                  cli::col_yellow(
                    "Org/user you provided does not exist: {cli::col_red({org})}"
                  )
                )
              }
              org <- NULL
            }
          }
          return(org)
        }, .progress = verbose) |>
        purrr::keep(~ length(.) > 0)
      if (length(orgs) == 0) {
        return(NULL)
      }
      orgs
    },

    check_endpoint = function(endpoint, type, verbose, .error) {
      check <- TRUE
      endpoint_response <- private$engines$rest$perform_request(
        endpoint = endpoint,
        token = private$token,
        verbose = verbose
      )
      if (endpoint_response$status_code == 404) {
        if (.error) {
          cli::cli_abort(
            c(
              "x" = "{type} you provided does not exist or its name was passed
                in a wrong way: {cli::col_red({url_decode(endpoint)})}",
              "!" = "Please type your {tolower(type)} name as you see it in
                web URL.",
              "i" = "E.g. do not use spaces. {type} names as you see on the
                page may differ from their web 'address'.",
              "i" = "Repository names should be provided as full paths: {.val org/repo_name}."
            ),
            call = NULL
          )
        } else {
          if (verbose) {
            cli::cli_alert_warning(
              cli::col_yellow(
                "{type} you provided does not exist: {cli::col_red({url_decode(endpoint)})}"
              )
            )
          }
          check <- FALSE
        }
      }
      return(check)
    },

    set_graphql_url = function() {
      clean_api_url <- gsub("/v+.*", "", private$api_url)
      private$graphql_api_url <- glue::glue("{clean_api_url}/graphql")
    },

    set_default_token = function(verbose) {
      primary_token_name <- private$token_name
      token <- Sys.getenv(primary_token_name)
      if (private$test_token(token)) {
        if (verbose) {
          cli::cli_alert_info("Using PAT from {primary_token_name} envar.")
        }
      } else {
        pat_names <- names(Sys.getenv()[grepl(primary_token_name, names(Sys.getenv()))])
        possible_tokens <- pat_names[pat_names != primary_token_name]
        token_checks <- purrr::map_lgl(possible_tokens, function(token_name) {
          private$test_token(Sys.getenv(token_name))
        })
        if (!any(token_checks)) {
          cli::cli_abort(c(
            "x" = "No sufficient token found among: [{paste0(pat_names, collapse = ', ')}].",
            "i" = "Check if you have correct token.",
            "!" = "Scope that is needed: [{paste0(private$min_access_scopes, collapse = ', ')}]."
          ),
          call = NULL)
        } else {
          token_name <- possible_tokens[token_checks][1]
          if (verbose) {
            cli::cli_alert_info("Using PAT from {token_name} envar.")
          }
          token <- Sys.getenv(token_name)
        }
      }
      return(token)
    },

    test_token = function(token) {
      response <- NULL
      test_endpoint <- private$test_endpoint
      response <- tryCatch({
        httr2::request(test_endpoint) |>
          httr2::req_headers("Authorization" = paste0("Bearer ", token)) |>
          httr2::req_perform()
      },
      error = function(e) {
        if (!is.null(e$parent) && grepl("Could not resolve host", e$parent$message)) {
          cli::cli_abort(e$parent$message, call = NULL)
        } else {
          NULL
        }
      })
      if (!is.null(response)) {
        check <- private$check_token_scopes(response, token)
      } else {
        check <- FALSE
      }
      return(check)
    },

    extract_repos_and_orgs = function(repos_fullnames = NULL) {
      repos_fullnames <- url_decode(repos_fullnames)
      repos_vec <- stringr::str_split(repos_fullnames, "/") |>
        purrr::map(~ paste0(.[length(.)], collapse = "/")) |>
        unlist()
      orgs_vec <- stringr::str_split(repos_fullnames, "/") |>
        purrr::map(~ paste0(.[-length(.)], collapse = "/")) |>
        unlist()
      names(repos_vec) <- orgs_vec
      orgs_names <- unique(orgs_vec)
      orgs_repo_list <- purrr::map(orgs_names, function(org) {
        unname(repos_vec[which(names(repos_vec) == org)])
      })
      names(orgs_repo_list) <- orgs_names
      return(orgs_repo_list)
    },

    filter_repos_by_host = function(repos_table) {
      dplyr::filter(
        repos_table,
        grepl(gsub("/v+.*", "", private$api_url), api_url)
      )
    },

    get_orgs_from_orgs_and_repos = function(output, verbose) {
      graphql_engine <- private$engines$graphql
      orgs_names <- NULL
      orgs_names_from_repos <- NULL
      if ("org" %in% private$searching_scope) {
        orgs_names <- purrr::keep(private$orgs, function(org) {
          type <- attr(org, "type") %||% "organization"
          type == "organization"
        })
      }
      if ("repo" %in% private$searching_scope) {
        orgs_names_from_repos <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        ) |>
          purrr::keep(function(org) {
            type <- attr(org, "type") %||% "organization"
            type == "organization"
          })
      }
      total_orgs_names <- c(orgs_names, orgs_names_from_repos)
      orgs_table <- gitstats_map(total_orgs_names, function(org) {
        type <- attr(org, "type") %||% "organization"
        org <- url_decode(org)
        graphql_engine$get_org(
          org = org,
          verbose = verbose
        )
      }) |>
        graphql_engine$prepare_orgs_table()
      return(orgs_table)
    },

    get_all_repos = function(add_languages = TRUE, verbose = TRUE, progress = TRUE, fill_empty_sha = FALSE) {
      if (private$scan_all && is.null(private$orgs)) {
        private$orgs <- private$get_orgs_from_host(
          output = "only_names",
          verbose = verbose
        )
      }
      repos_table <- purrr::list_rbind(
        list(
          private$get_repos_from_orgs(add_languages, verbose, progress),
          private$get_repos_from_repos(add_languages, verbose, progress)
        )
      )
      return(repos_table)
    },

    get_repos_from_orgs = function(add_languages, verbose, progress) {
      if (any(c("all", "org") %in% private$searching_scope)) {
        graphql_engine <- private$engines$graphql
        gitstats_map(private$orgs, function(org) {
          owner_type <- attr(org, "type") %||% "organization"
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = url_decode(org),
              information = paste0("Pulling repositories ", cli_icons$repo)
            )
          }
          repos_from_org <- graphql_engine$get_repos_from_org(
            org = url_decode(org),
            owner_type = owner_type,
            verbose = verbose
          )
          if (!inherits(repos_from_org, "graphql_error")) {
            if (length(repos_from_org) > 0) {
              repos_table <- repos_from_org |>
                graphql_engine$prepare_repos_table(
                  org = unclass(url_decode(org))
                ) |>
                dplyr::filter(organization == unclass(url_decode(org)))
            } else {
              repos_table <- NULL
            }
          } else {
            if (verbose) {
              cli::cli_alert_info("Switching to REST API")
              show_message(
                host = private$host_name,
                engine = "rest",
                scope = org,
                information = paste0("Pulling repositories ", cli_icons$repo)
              )
            }
            rest_engine <- private$engines$rest
            repos_table <- rest_engine$get_repos_from_org(
              org = org,
              add_languages = add_languages,
              output = "full_table",
              verbose = verbose
            ) |>
              rest_engine$prepare_repos_table(
                org = org
              )
          }
          return(repos_table)
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind()
      }
    },

    get_repos_from_repos = function(add_languages, verbose, progress) {
      if ("repo" %in% private$searching_scope) {
        graphql_engine <- private$engines$graphql
        orgs <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        )
        gitstats_map(orgs, function(org) {
          owner_type <- attr(org, "type") %||% "organization"
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = set_repo_scope(url_decode(org), private),
              information = paste0("Pulling repositories ", cli_icons$repo)
            )
          }
          repos_from_org <- graphql_engine$get_repos_from_org(
            org = url_decode(org),
            owner_type = owner_type,
            verbose = verbose
          )
          if (!inherits(repos_from_org, "graphql_error")) {
            if (length(repos_from_org) > 0) {
              repos_table <- repos_from_org |>
                graphql_engine$prepare_repos_table() |>
                dplyr::filter(repo_name %in% private$orgs_repos[[url_decode(org)]])
            } else {
              repos_table <- NULL
            }
          } else {
            if (verbose) {
              cli::cli_alert_info("Switching to REST API")
              show_message(
                host = private$host_name,
                engine = "rest",
                scope = org,
                information = paste0("Pulling repositories ", cli_icons$repo)
              )
            }
            rest_engine <- private$engines$rest
            repos_table <- rest_engine$get_repos_from_org(
              org = org,
              repos = private$orgs_repos[[org]],
              add_languages = add_languages,
              output = "full_table",
              verbose = verbose
            ) |>
              rest_engine$prepare_repos_table(
                org = org
              )
          }
          return(repos_table)
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind()
      }
    },

    get_repos_with_code = function(code,
                                   in_files = NULL,
                                   in_path = FALSE,
                                   language = NULL,
                                   output = "table_full",
                                   verbose = TRUE,
                                   progress = TRUE) {
      if (private$scan_all) {
        repos_table <- private$get_repos_with_code_from_host(
          code = code,
          in_files = in_files,
          in_path = in_path,
          language = language,
          output = output,
          verbose = verbose
        )
      } else {
        repos_from_org <- private$get_repos_with_code_from_orgs(
          code = code,
          in_files = in_files,
          in_path = in_path,
          language = language,
          output = output,
          verbose = verbose,
          progress = progress
        )
        repos_from_repos <- private$get_repos_with_code_from_repos(
          code = code,
          in_files = in_files,
          in_path = in_path,
          language = language,
          output = output,
          verbose = verbose
        )
        repos_table <- rbind(repos_from_org, repos_from_repos)
      }
      return(repos_table)
    },

    get_all_repos_urls = function(type, verbose = TRUE, progress = TRUE) {
      if (private$scan_all && is.null(private$orgs)) {
        private$orgs <- private$get_orgs_from_host(
          output = "only_names",
          verbose = verbose
        )
      }
      repos_urls_from_orgs <- private$get_repos_urls_from_orgs(
        type = type,
        verbose = verbose,
        progress = progress
      )
      repos_urls_from_repos <- private$get_repos_urls_from_repos(
        type = type,
        verbose = verbose,
        progress = progress
      )
      repos_vector <- c(
        repos_urls_from_orgs,
        repos_urls_from_repos
      )
      return(repos_vector)
    },

    get_repos_urls_from_orgs = function(type, verbose, progress) {
      if (any(c("all", "org") %in% private$searching_scope)) {
        rest_engine <- private$engines$rest
        repos_vector <- gitstats_map(private$orgs, function(org) {
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "rest",
              scope = url_decode(org),
              information = paste0("Pulling repositories ", cli_icons$repo, " (URLs)")
            )
          }
          repos_urls <- rest_engine$get_repos_urls(
            type = type,
            org = org,
            repos = NULL
          )
          return(repos_urls)
        }, .progress = set_progress_bar(progress, private)) |>
          unlist()
      }
    },

    get_repos_urls_from_repos = function(type, verbose, progress) {
      if ("repo" %in% private$searching_scope) {
        rest_engine <- private$engines$rest
        graphql_engine <- private$engines$graphql
        orgs <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        )
        repos_vector <- gitstats_map(orgs, function(org) {
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "rest",
              scope = paste0(org, "/", private$orgs_repos[[org]], collapse = ", "),
              information = paste0("Pulling repositories ", cli_icons$repo, " (URLs)")
            )
          }
          repos_urls <- rest_engine$get_repos_urls(
            type = type,
            org = org,
            repos = private$orgs_repos[[org]]
          )
          return(repos_urls)
        }, .progress = set_progress_bar(progress, private)) |>
          unlist()
      }
    },

    # Run a search function with optional in_files iteration, then parse the response.
    search_and_parse = function(search_fn, code, in_files, in_path, language,
                                output, verbose, org = NULL) {
      if (is.null(in_files)) {
        search_response <- search_fn(
          code = code,
          in_path = in_path,
          language = language,
          verbose = verbose
        )
      } else {
        search_response <- purrr::map(in_files, function(filename) {
          search_fn(
            code = code,
            filename = filename,
            in_path = in_path,
            language = language,
            verbose = verbose
          )
        }) |>
          purrr::list_flatten()
      }
      private$parse_search_response(
        search_response = search_response,
        org = org,
        output = output,
        verbose = verbose
      )
    },

    get_repos_with_code_from_host = function(code,
                                             in_files = NULL,
                                             in_path = FALSE,
                                             language = NULL,
                                             output = "table_full",
                                             verbose = TRUE) {
      if (verbose) {
        show_message(
          host = private$host_name,
          engine = "rest",
          information = paste0("Pulling repositories ", cli_icons$repo)
        )
      }
      rest_engine <- private$engines$rest
      search_fn <- function(...) rest_engine$search_for_code(...)
      repos_output <- private$search_and_parse(
        search_fn = search_fn,
        code = code, in_files = in_files, in_path = in_path,
        language = language, output = output, verbose = verbose
      )
      return(repos_output)
    },

    get_repos_with_code_from_orgs = function(code,
                                             in_files = NULL,
                                             in_path = FALSE,
                                             language = NULL,
                                             output = "table",
                                             verbose = TRUE,
                                             progress = TRUE) {
      if (any(private$searching_scope %in% c("org", "all"))) {
        repos_list <- gitstats_map(private$orgs, function(org) {
          if (verbose) {
            show_message(
              host = private$host_name,
              engine = "rest",
              scope = url_decode(org),
              code = code,
              information = paste0("Pulling repositories ", cli_icons$repo)
            )
          }
          rest_engine <- private$engines$rest
          search_fn <- function(...) rest_engine$search_for_code(org = org, ...)
          private$search_and_parse(
            search_fn = search_fn,
            code = code, in_files = in_files, in_path = in_path,
            language = language, output = output, verbose = verbose,
            org = org
          )
        }, .progress = set_progress_bar(progress, private))
        if (output != "raw") {
          repos_output <- purrr::list_rbind(repos_list)
        } else {
          repos_output <- purrr::list_flatten(repos_list)
        }
        return(repos_output)
      }
    },

    get_repos_with_code_from_repos = function(code,
                                              in_files = NULL,
                                              in_path = FALSE,
                                              language = NULL,
                                              output = "table_full",
                                              verbose = TRUE) {
      orgs <- names(private$orgs_repos)
      if ("repo" %in% private$searching_scope) {
        if (verbose) {
          show_message(
            host = private$host_name,
            engine = "rest",
            scope = url_decode(paste0(orgs, collapse = "|")),
            code = code,
            information = paste0("Pulling repositories ", cli_icons$repo)
          )
        }
        rest_engine <- private$engines$rest
        search_fn <- function(...) {
          rest_engine$search_repos_for_code(repos = private$repos_fullnames, ...)
        }
        repos_output <- private$search_and_parse(
          search_fn = search_fn,
          code = code, in_files = in_files, in_path = in_path,
          language = language, output = output, verbose = verbose
        )
        return(repos_output)
      }
    },

    parse_search_response = function(search_response, org = NULL, output, verbose = TRUE) {
      if (length(search_response) > 0) {
        repos_ids <- private$get_repos_ids(search_response)
        api_engine <- private$engines$graphql
        if (verbose) cli::cli_alert("Parsing search response with GraphQL...")
        repos_response <- api_engine$get_repos(
          repos_ids = repos_ids,
          verbose = verbose
        )
        if (output != "raw") {
          repos_output <- repos_response |>
            api_engine$prepare_repos_table(
              org = org
            )
        } else {
          repos_output <- repos_response
        }
      } else {
        repos_output <- NULL
      }
      return(repos_output)
    },

    get_repos_urls_with_code = function(type,
                                        code,
                                        in_files = NULL,
                                        in_path = FALSE,
                                        verbose,
                                        progress) {
      if (private$scan_all) {
        repos_urls <- private$get_repos_urls_with_code_from_host(
          type = type,
          code = code,
          in_files = in_files,
          in_path = in_path,
          verbose = verbose,
          progress = progress
        )
      } else {
        repos_urls_from_orgs <- private$get_repos_urls_with_code_from_orgs(
          type = type,
          code = code,
          in_files = in_files,
          in_path = in_path,
          verbose = verbose,
          progress = progress
        )
        repos_urls_from_repos <- private$get_repos_urls_with_code_from_repos(
          type = type,
          code = code,
          in_files = in_files,
          in_path = in_path,
          verbose = verbose,
          progress = progress
        )
        repos_urls <- c(repos_urls_from_orgs, repos_urls_from_repos)
      }
      return(repos_urls)
    },

    get_repos_urls_with_code_from_host = function(type, code, in_files, in_path, verbose, progress) {
      private$get_repos_with_code_from_host(
        code = code,
        in_files = in_files,
        in_path = in_path,
        output = "raw",
        verbose = verbose
      ) |>
        private$get_repo_url_from_response(
          type = type,
          verbose = verbose,
          progress = progress
        )
    },

    get_repos_urls_with_code_from_orgs = function(type, code, in_files, in_path, verbose, progress) {
      private$get_repos_with_code_from_orgs(
        code = code,
        in_files = in_files,
        in_path = in_path,
        output = "raw",
        verbose = verbose
      ) |>
        purrr::discard(~ is.null(.)) |>
        private$get_repo_url_from_response(
          type = type,
          verbose = verbose,
          progress = progress
        )
    },

    get_repos_urls_with_code_from_repos = function(type, code, in_files, in_path, verbose, progress) {
      private$get_repos_with_code_from_repos(
        code = code,
        in_files = in_files,
        in_path = in_path,
        output = "raw",
        verbose = verbose
      ) |>
        purrr::discard(~ is.null(.)) |>
        private$get_repo_url_from_response(
          type = type,
          repos_fullnames = private$repos_fullnames,
          verbose = verbose,
          progress = progress
        )
    },

    add_githost_info = function(repos_table) {
      if (!is.null(repos_table) && nrow(repos_table) > 0) {
        dplyr::mutate(
          repos_table,
          githost = retrieve_githost(api_url)
        )
      }
    },

    get_repos_contributors = function(repos_table, verbose, progress) {
      if (!is.null(repos_table) && nrow(repos_table) > 0) {
        if (!private$scan_all && verbose) {
          show_message(
            host = private$host_name,
            engine = "rest",
            information = paste0("Pulling contributors ", cli_icons$user)
          )
        }
        repos_table <- private$filter_repos_by_host(repos_table)
        rest_engine <- private$engines$rest
        repos_table <- rest_engine$get_repos_contributors(
          repos_table = repos_table,
          verbose = verbose,
          progress = progress
        )
        return(repos_table)
      }
    },

    get_issues_from_orgs = function(verbose, progress) {
      if ("org" %in% private$searching_scope) {
        graphql_engine <- private$engines$graphql
        issues_table <- gitstats_map(private$orgs, function(org) {
          issues_table_org <- NULL
          repos_data <- private$get_repos_data(
            org = org,
            verbose = verbose
          )
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = org,
              information = paste0("Getting issues ", cli_icons$issue)
            )
          }
          issues_table_org <- graphql_engine$get_issues_from_repos(
            org = org,
            repos_names = repos_data[["paths"]],
            verbose = verbose
          ) |>
            graphql_engine$prepare_issues_table(
              org = org
            )
          return(issues_table_org)
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind()
        return(issues_table)
      }
    },

    get_issues_from_repos = function(verbose, progress) {
      if ("repo" %in% private$searching_scope) {
        graphql_engine <- private$engines$graphql
        orgs <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        )
        issues_table <- gitstats_map(orgs, function(org) {
          issues_table_org <- NULL
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = set_repo_scope(org, private),
              information = paste0("Getting issues ", cli_icons$issue)
            )
          }
          issues_table_org <- graphql_engine$get_issues_from_repos(
            org = org,
            repos_names = private$orgs_repos[[org]],
            verbose = verbose
          ) |>
            graphql_engine$prepare_issues_table(
              org = org
            )
          return(issues_table_org)
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind()
        return(issues_table)
      }
    },

    get_pr_from_orgs = function(verbose, progress) {
      if ("org" %in% private$searching_scope) {
        graphql_engine <- private$engines$graphql
        pr_table <- gitstats_map(private$orgs, function(org) {
          pr_table_org <- NULL
          repos_data <- private$get_repos_data(
            org = org,
            verbose = verbose
          )
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = org,
              information = paste0("Pulling pull requests ", cli_icons$pull_request)
            )
          }
          pr_table_org <- graphql_engine$get_pr_from_repos(
            org = org,
            repos_names = repos_data[["paths"]],
            verbose = verbose
          ) |>
            graphql_engine$prepare_pr_table(
              org = org
            )
          return(pr_table_org)
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind()
        return(pr_table)
      }
    },

    get_pr_from_repos = function(verbose, progress) {
      if ("repo" %in% private$searching_scope) {
        graphql_engine <- private$engines$graphql
        orgs <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        )
        pr_table <- gitstats_map(orgs, function(org) {
          pr_table_org <- NULL
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = set_repo_scope(org, private),
              information = paste0("Getting pull requests ", cli_icons$pull_request)
            )
          }
          pr_table_org <- graphql_engine$get_pr_from_repos(
            org = org,
            repos_names = private$orgs_repos[[org]],
            verbose = verbose
          ) |>
            graphql_engine$prepare_pr_table(
              org = org
            )
          return(pr_table_org)
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind()
        return(pr_table)
      }
    },

    get_files_content_from_orgs = function(file_path,
                                           verbose = TRUE,
                                           progress = TRUE) {
      if ("org" %in% private$searching_scope) {
        graphql_engine <- private$engines$graphql
        files_table <- gitstats_map(private$orgs, function(org) {
          owner_type <- attr(org, "type") %||% "organization"
          repos_data <- private$get_repos_data(
            org = org,
            verbose = verbose
          )
          if (verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = org,
              information = glue::glue("Pulling files {cli_icons$file} content: [{paste0(file_path, collapse = ', ')}]")
            )
          }
          graphql_engine$get_files_from_org(
            org = org,
            owner_type = owner_type,
            repos_data = repos_data,
            file_paths = file_path,
            verbose = verbose
          ) |>
            graphql_engine$prepare_files_table(
              org = org
            )
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind() |>
          private$add_repo_api_url()
        return(files_table)
      }
    },

    get_files_structure_from_orgs = function(pattern,
                                             depth,
                                             verbose  = TRUE,
                                             progress = TRUE) {
      if (any(c("all", "org") %in% private$searching_scope)) {
        rest_engine <- private$engines$rest
        files_structure_list <- gitstats_map(private$orgs, function(org) {
          repos_data <- private$get_repos_data(
            org = org,
            verbose = verbose
          )
          if (verbose) {
            user_info <- if (!is.null(pattern)) {
              glue::glue(
                "Pulling repos {cli_icons$tree} [files matching pattern: '{paste0(pattern, collapse = '|')}']"
              )
            } else {
              glue::glue("Pulling repos {cli_icons$tree}")
            }
            show_message(
              host = private$host_name,
              engine = "rest",
              scope = org,
              information = user_info
            )
          }
          private$get_files_structure_from_repos_data(
            rest_engine = rest_engine,
            org = org,
            repos_data = repos_data,
            pattern = pattern,
            depth = depth,
            verbose = verbose
          )
        }, .progress = set_progress_bar(progress, private))
        names(files_structure_list) <- private$orgs
        files_structure_list <- files_structure_list |>
          purrr::discard(~ length(.) == 0)
        if (length(files_structure_list) == 0 && verbose) {
          cli::cli_alert_warning(
            cli::col_yellow(
              "For {private$host_name} no files structure found."
            )
          )
        }
        return(files_structure_list)
      }
    },

    get_files_structure_from_repos_data = function(rest_engine,
                                                   org,
                                                   repos_data,
                                                   pattern,
                                                   depth,
                                                   verbose) {
      repositories <- repos_data[["paths"]]
      def_branches <- repos_data[["def_branches"]]
      repo_ids <- repos_data[["repo_ids"]]
      if (!is.null(def_branches)) {
        files_structure <- gitstats_map2(repositories, def_branches,
          function(repo, def_branch) {
            rest_engine$get_files_tree(
              org = org,
              repo = repo,
              def_branch = def_branch,
              pattern = pattern,
              depth = depth,
              verbose = verbose
            )
          }
        )
      } else {
        files_structure <- gitstats_map(repositories,
          function(repo) {
            rest_engine$get_files_tree(
              org = org,
              repo = repo,
              pattern = pattern,
              depth = depth,
              verbose = verbose
            )
          }
        )
      }
      names(files_structure) <- repositories
      if (!is.null(repo_ids)) {
        for (i in seq_along(files_structure)) {
          if (!is.null(files_structure[[i]])) {
            attr(files_structure[[i]], "repo_id") <- repo_ids[[i]]
          }
        }
      }
      files_structure <- purrr::discard(files_structure, ~ length(.) == 0)
      return(files_structure)
    },

    get_files_content_from_host = function(file_path,
                                           verbose  = TRUE,
                                           progress = TRUE) {
      rest_engine <- private$engines$rest
      if (verbose) {
        show_message(
          host = private$host_name,
          engine = "rest",
          information = glue::glue("Pulling files {cli_icons$file}: [{paste0(file_path, collapse = ', ')}]")
        )
      }
      files_table <- rest_engine$get_files(
        file_paths = file_path,
        verbose = verbose
      ) |>
        rest_engine$prepare_files_table() |>
        private$add_repo_api_url()
      return(files_table)
    },

    get_release_logs_from_orgs = function(verbose, progress) {
      if ("org" %in% private$searching_scope) {
        release_logs_table <- gitstats_map(private$orgs, function(org) {
          org <- url_decode(org)
          release_logs_table_org <- NULL
          repos_data <- private$get_repos_data(
            org = org,
            verbose = verbose
          )
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = org,
              information = paste0("Pulling release logs ", cli_icons$release)
            )
          }
          graphql_engine <- private$engines$graphql
          if (length(repos_data[["paths"]]) > 0) {
            release_logs_table_org <- graphql_engine$get_release_logs_from_org(
              repos_names = repos_data[["paths"]],
              org = org,
              verbose = verbose
            ) |>
              graphql_engine$prepare_releases_table(
                org = org
              )
          } else {
            releases_logs_table_org <- NULL
          }
          return(release_logs_table_org)
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind()
        return(release_logs_table)
      }
    },

    get_release_logs_from_repos = function(verbose, progress) {
      if ("repo" %in% private$searching_scope) {
        graphql_engine <- private$engines$graphql
        orgs <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        )
        release_logs_table <- gitstats_map(orgs, function(org) {
          org <- url_decode(org)
          release_logs_table_org <- NULL
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = set_repo_scope(org, private),
              information = paste0("Pulling release logs ", cli_icons$release)
            )
          }
          release_logs_table_org <- graphql_engine$get_release_logs_from_org(
            repos_names = private$orgs_repos[[org]],
            org = org,
            verbose = verbose
          ) |>
            graphql_engine$prepare_releases_table(
              org = org
            )
          return(release_logs_table_org)
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind()
        return(release_logs_table)
      }
    },

    filter_repos_table_by_language = function(repos_table, language_filter) {
      if (!is.null(repos_table)) {
        repos_table |>
          dplyr::filter(grepl(paste0("\\b", language_filter, "\\b"), languages))
      }
    }
  )
)

test_that("repos_by_org query is built properly", {
  gh_repos_by_org_query <-
    test_gqlquery_gh$repos_by_org(repo_cursor = "")
  expect_snapshot(
    gh_repos_by_org_query
  )
  test_mocker$cache(gh_repos_by_org_query)
})

test_that("repos_by_user query is built properly", {
  gh_repos_by_user_query <-
    test_gqlquery_gh$repos_by_user(repo_cursor = "")
  expect_snapshot(
    gh_repos_by_user_query
  )
  test_mocker$cache(gh_repos_by_user_query)
})

test_that("repos_by_ids query is built properly", {
  gh_repos_by_ids_query <-
    test_gqlquery_gh$repos_by_ids()
  expect_snapshot(
    gh_repos_by_ids_query
  )
  test_mocker$cache(gh_repos_by_ids_query)
})

gh_org <- "r-world-devs"
if (integration_tests_skipped) {
  gh_user <- "test_user"
  gh_repo <- "TestRepo"
  gh_code <- "test_code"
  gh_file <- "test_file"
} else {
  gh_repo <- "GitAI"
  gh_user <- "maciekbanas"
  gh_code <- "GitStats"
  gh_file <- "DESCRIPTION"
}

test_that("`get_repos_page()` pulls repos page from GitHub organization", {
  mockery::stub(
    test_graphql_github_priv$get_repos_page,
    "self$gql_response",
    test_fixtures$github_repos_by_org_response
  )
  gh_repos_page <- test_graphql_github_priv$get_repos_page(
    login = gh_org,
    type = "organization"
  )
  expect_repos_github_gql_response(
    gh_repos_page$data$repositoryOwner$repositories$nodes[[1]]
  )
  test_mocker$cache(gh_repos_page)
})

test_that("get_repos() pulls repositories by ids", {
  mockery::stub(
    test_graphql_github$get_repos,
    "self$gql_response",
    test_fixtures$github_repos_by_ids_response
  )
  gh_repos_by_ids <- test_graphql_github$get_repos(
    repos_ids = c("repo_1", "repo_2", "repo_3")
  )
  expect_repos_github_gql_response(
    gh_repos_by_ids[[1]]
  )
  test_mocker$cache(gh_repos_by_ids)
})

test_that("`get_repos_from_org()` prepares formatted list", {
  mockery::stub(
    test_graphql_github$get_repos_from_org,
    "private$get_repos_page",
    test_mocker$use("gh_repos_page")
  )
  gh_repos_from_org <- test_graphql_github$get_repos_from_org(
    org = gh_org,
    owner_type = "organization"
  )
  expect_list_contains(
    gh_repos_from_org[[1]],
    c(
      "id", "name", "stars", "forks", "created_at",
      "last_activity_at", "languages", "issues_open", "issues_closed",
      "contributors", "repo_url"
    )
  )
  test_mocker$cache(gh_repos_from_org)
})

test_that("`get_repos_page()` pulls repos page from GitHub user", {
  mockery::stub(
    test_graphql_github_priv$get_repos_page,
    "self$gql_response",
    test_fixtures$github_repos_by_user_response
  )
  gh_repos_user_page <- test_graphql_github_priv$get_repos_page(
    login = gh_user,
    type = "user"
  )
  expect_repos_github_gql_response(
    gh_repos_user_page$data$user$repositories$nodes[[1]]
  )
  test_mocker$cache(gh_repos_user_page)
})

test_that("`get_repos_from_org()` prepares formatted list", {
  mockery::stub(
    test_graphql_github$get_repos_from_org,
    "private$get_repos_page",
    test_mocker$use("gh_repos_user_page")
  )
  gh_repos_from_user <- test_graphql_github$get_repos_from_org(
    org = gh_user,
    owner_type = "user"
  )
  expect_list_contains(
    gh_repos_from_user[[1]],
    c(
      "id", "name", "stars", "forks", "created_at",
      "last_activity_at", "languages", "issues_open", "issues_closed",
      "contributors", "repo_url"
    )
  )
  test_mocker$cache(gh_repos_from_user)
})

# REST Engine search repos by code

test_that("`search_response()` performs search with limit under 1000", {
  total_n <- test_fixtures$github_search_response[["total_count"]]
  mockery::stub(
    test_rest_github_priv$search_response,
    "self$response",
    test_fixtures$github_search_response
  )
  gh_search_repos_response <- test_rest_github_priv$search_response(
    search_endpoint = test_mocker$use("search_endpoint"),
    total_n = total_n
  )
  expect_search_github_response(gh_search_repos_response)
  test_mocker$cache(gh_search_repos_response)
})

test_that("`search_response()` performs search with limit over 1000", {
  total_n <- test_fixtures$github_search_response_large[["total_count"]]
  mockery::stub(
    test_rest_github_priv$search_response,
    "self$response",
    test_fixtures$github_search_response_large
  )
  gh_search_repos_response_large <- test_rest_github_priv$search_response(
    search_endpoint = test_mocker$use("search_endpoint"),
    total_n = total_n
  )
  expect_search_github_response(gh_search_repos_response_large)
})

test_that("GitHub build_search_query builds query with code only", {
  query <- test_rest_github_priv$build_search_query(code = "shiny")
  expect_equal(query, "shiny")
})

test_that("GitHub build_search_query builds query with all params", {
  query <- test_rest_github_priv$build_search_query(
    code = "shiny",
    org = "r-world-devs",
    repo = "r-world-devs/GitStats",
    filename = "DESCRIPTION",
    in_path = TRUE,
    language = "R"
  )
  expect_match(query, "\\+repo:r-world-devs/GitStats")
  expect_match(query, "\\+user:r-world-devs")
  expect_match(query, "\\+in:path")
  expect_match(query, "\\+in:file\\+filename:DESCRIPTION")
  expect_match(query, "\\+language:R")
})

test_that("`search_for_code()` returns repos output for code search in files", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_rest_github$search_for_code,
      "self$response",
      list("total_count" = 3L)
    )
    mockery::stub(
      test_rest_github$search_for_code,
      "private$search_response",
      test_fixtures$github_search_response$items
    )
  }
  gh_search_for_code <- test_rest_github$search_for_code(
    code = gh_code,
    filename = gh_file,
    in_path = FALSE,
    org = gh_org,
    verbose = FALSE
  )
  expect_search_github_response(gh_search_for_code)
  test_mocker$cache(gh_search_for_code)
})

test_that("`search_repos_for_code()` returns repos output for code search in files", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_rest_github$search_repos_for_code,
      "self$response",
      list("total_count" = 3L)
    )
    mockery::stub(
      test_rest_github$search_repos_for_code,
      "private$search_response",
      test_fixtures$github_search_response$items
    )
  }
  gh_search_repos_for_code <- test_rest_github$search_repos_for_code(
    code = gh_code,
    filename = gh_file,
    in_path = FALSE,
    repos = paste0(gh_org, "/", gh_repo),
    verbose = FALSE
  )
  expect_search_github_response(gh_search_repos_for_code)
  test_mocker$cache(gh_search_repos_for_code)
})

test_that("`search_for_code()` works with language filter", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_rest_github$search_for_code,
      "self$response",
      list("total_count" = 3L)
    )
    mockery::stub(
      test_rest_github$search_for_code,
      "private$search_response",
      test_fixtures$github_search_response$items
    )
    gh_search_for_code_language <- test_rest_github$search_for_code(
      code = gh_code,
      filename = gh_file,
      in_path = FALSE,
      language = "R",
      verbose = FALSE
    )
    expect_search_github_response(gh_search_for_code_language)
  }
})

test_that("`prepare_repos_table()` prepares repos table", {
  gh_repos_by_code_table <- test_graphql_github$prepare_repos_table(
      repos_list = test_mocker$use("gh_repos_from_org"),
      org = gh_org
  )
  expect_repos_table(
    gh_repos_by_code_table
  )
  test_mocker$cache(gh_repos_by_code_table)
})

test_that("get_repos_ids", {
  repos_ids <- github_testhost_priv$get_repos_ids(
    search_response = test_mocker$use("gh_search_repos_for_code")
  )
  expect_type(
    repos_ids,
    "character"
  )
  expect_gt(
    length(repos_ids), 0
  )
})

test_that("parse_search_response parses search response into repositories output", {
  mockery::stub(
    github_testhost_priv$parse_search_response,
    "api_engine$get_repos",
    test_mocker$use("gh_repos_by_ids")
  )
  gh_repos_raw_output <- github_testhost_priv$parse_search_response(
    search_response = test_mocker$use("gh_search_repos_for_code"),
    output = "raw",
    verbose = FALSE
  )
  expect_repos_github_gql_response(gh_repos_raw_output[[1]])
  test_mocker$cache(gh_repos_raw_output)
})

test_that("parse_search_response parses search response into repositories output", {
  mockery::stub(
    github_testhost_priv$parse_search_response,
    "api_engine$get_repos",
    test_mocker$use("gh_repos_by_ids")
  )
  gh_repos_by_code_table <- github_testhost_priv$parse_search_response(
    search_response = test_mocker$use("gh_search_repos_for_code"),
    org = gh_org,
    output = "table",
    verbose = FALSE
  )
  expect_repos_table(gh_repos_by_code_table)
  test_mocker$cache(gh_repos_by_code_table)
})

test_that("parse_search_response prints message", {
  mockery::stub(
    github_testhost_priv$parse_search_response,
    "api_engine$get_repos_from_org",
    test_mocker$use("gh_repos_by_ids")
  )
  expect_snapshot(
    gh_repos_raw_output <- github_testhost_priv$parse_search_response(
      search_response = test_mocker$use("gh_search_repos_for_code"),
      org = gh_org,
      output = "raw",
      verbose = TRUE
    )
  )
})

test_that("search_and_parse calls search_fn and parse_search_response without in_files", {
  mock_search_fn <- mockery::mock(test_mocker$use("gh_search_repos_for_code"))
  mockery::stub(
    github_testhost_priv$search_and_parse,
    "private$parse_search_response",
    test_mocker$use("gh_repos_by_code_table")
  )
  result <- github_testhost_priv$search_and_parse(
    search_fn = mock_search_fn,
    code = "shiny",
    in_files = NULL,
    in_path = FALSE,
    language = NULL,
    output = "table",
    verbose = FALSE
  )
  mockery::expect_called(mock_search_fn, 1)
  expect_repos_table(result)
})

test_that("search_and_parse iterates over in_files", {
  mock_search_fn <- mockery::mock(
    test_mocker$use("gh_search_repos_for_code"),
    test_mocker$use("gh_search_repos_for_code")
  )
  mockery::stub(
    github_testhost_priv$search_and_parse,
    "private$parse_search_response",
    test_mocker$use("gh_repos_by_code_table")
  )
  result <- github_testhost_priv$search_and_parse(
    search_fn = mock_search_fn,
    code = "shiny",
    in_files = c("DESCRIPTION", "NAMESPACE"),
    in_path = FALSE,
    language = "R",
    output = "table",
    verbose = FALSE
  )
  mockery::expect_called(mock_search_fn, 2)
  expect_repos_table(result)
})

test_that("`get_repos_with_code_from_orgs()` works", {
  mockery::stub(
    github_testhost_priv$get_repos_with_code_from_orgs,
    "private$search_and_parse",
    test_mocker$use("gh_repos_by_code_table")
  )
  repos_with_code_from_orgs_full <- github_testhost_priv$get_repos_with_code_from_orgs(
    code = "shiny",
    output = "table",
    verbose = FALSE
  )
  expect_repos_table(repos_with_code_from_orgs_full)
  test_mocker$cache(repos_with_code_from_orgs_full)
})

test_that("`get_repos_with_code_from_orgs()` pulls raw response", {
  mockery::stub(
    github_testhost_priv$get_repos_with_code_from_orgs,
    "private$search_and_parse",
    test_mocker$use("gh_repos_raw_output")
  )
  github_testhost_priv$orgs <- gh_org
  expect_snapshot(
    repos_with_code_from_orgs_raw <- github_testhost_priv$get_repos_with_code_from_orgs(
      code = "shiny",
      in_files = c("DESCRIPTION", "NAMESPACE"),
      language = "R",
      output = "raw",
      verbose = TRUE
    )
  )
  expect_type(repos_with_code_from_orgs_raw, "list")
  expect_gt(length(repos_with_code_from_orgs_raw), 0)
})

test_that("`get_repos_with_code_from_host()` pulls and parses output into table", {
  mockery::stub(
    github_testhost_priv$get_repos_with_code_from_host,
    "private$search_and_parse",
    test_mocker$use("gh_repos_by_code_table")
  )
  github_testhost_priv$searching_scope <- "all"
  expect_snapshot(
    repos_with_code_from_host_table <- github_testhost_priv$get_repos_with_code_from_host(
      code = "DESCRIPTION",
      in_path = TRUE,
      output = "table_full",
      verbose = TRUE
    )
  )
  expect_repos_table(repos_with_code_from_host_table)
})

test_that("`get_repos_with_code_from_repos()` works", {
  github_testhost_priv <- create_github_testhost(
    repos = c("TestRepo1", "TestRepo2"),
    mode = "private"
  )
  mockery::stub(
    github_testhost_priv$get_repos_with_code_from_repos,
    "private$search_and_parse",
    test_mocker$use("gh_repos_by_code_table")
  )
  github_testhost_priv$searching_scope <- c("repo")
  expect_snapshot(
    repos_with_code_from_repos_full <- github_testhost_priv$get_repos_with_code_from_repos(
      code = "tests",
      output = "table",
      verbose = TRUE
    )
  )
  expect_repos_table(repos_with_code_from_repos_full)
})

test_that("`get_repos_with_code_from_host()` pulls raw response", {
  mockery::stub(
    github_testhost_priv$get_repos_with_code_from_host,
    "private$search_and_parse",
    test_mocker$use("gh_repos_raw_output")
  )
  repos_with_code_from_host_raw <- github_testhost_priv$get_repos_with_code_from_host(
    code = "shiny",
    in_files = c("DESCRIPTION", "NAMESPACE"),
    output = "raw",
    verbose = FALSE
  )
  expect_type(repos_with_code_from_host_raw, "list")
  expect_gt(length(repos_with_code_from_host_raw), 0)
  test_mocker$cache(repos_with_code_from_host_raw)
})

test_that("get_repos_with_code() works", {
  mockery::stub(
    github_testhost_priv$get_repos_with_code,
    "private$get_repos_with_code_from_orgs",
    test_mocker$use("repos_with_code_from_orgs_full")
  )
  github_repos_with_code <- github_testhost_priv$get_repos_with_code(
    code = "test-code",
    verbose = FALSE,
    progress = FALSE
  )
  expect_repos_table(
    github_repos_with_code
  )
})

test_that("get_repos_with_code() scans whole host", {
  github_testhost_priv <- create_github_testhost(
    mode = "private"
  )
  mockery::stub(
    github_testhost_priv$get_repos_with_code,
    "private$get_repos_with_code_from_host",
    test_mocker$use("repos_with_code_from_host_raw")
  )
  github_repos_with_code_raw <- github_testhost_priv$get_repos_with_code(
    code = "test-code",
    output = "raw",
    verbose = FALSE,
    progress = FALSE
  )
  expect_type(github_repos_with_code_raw, "list")
  expect_gt(length(github_repos_with_code_raw), 0)
  github_testhost_priv$scan_all <- FALSE
})

test_that("GitHub prepares repos table from repositories response", {
  gh_repos_table <- test_graphql_github$prepare_repos_table(
    repos_list = test_mocker$use("gh_repos_from_org")
  )
  expect_repos_table(
    gh_repos_table
  )
  test_mocker$cache(gh_repos_table)
})

test_that("get_repos_from_orgs works", {
  mockery::stub(
    github_testhost_priv$get_repos_from_orgs,
    "graphql_engine$get_repos_from_org",
    test_mocker$use("gh_repos_from_org")
  )
  github_testhost_priv$orgs <- gh_org
  gh_repos_from_orgs <- github_testhost_priv$get_repos_from_orgs(
    add_languages = TRUE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_repos_table(
    gh_repos_from_orgs
  )
  test_mocker$cache(gh_repos_from_orgs)
})

test_that("get_repos_from_repos works", {
  test_org <- "r-world-devs"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    github_testhost_priv$get_repos_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  mockery::stub(
    github_testhost_priv$get_repos_from_repos,
    "graphql_engine$get_repos_from_org",
    test_mocker$use("gh_repos_from_orgs")
  )
  mockery::stub(
    github_testhost_priv$get_repos_from_repos,
    "graphql_engine$prepare_repos_table",
    test_mocker$use("gh_repos_table")
  )
  github_testhost_priv$searching_scope <- c("org", "repo")
  github_testhost_priv$orgs_repos <- list("r-world-devs" = "GitStats")
  expect_snapshot(
    gh_repos_individual <- github_testhost_priv$get_repos_from_repos(
      add_languages = TRUE,
      verbose = TRUE,
      progress = FALSE
    )
  )
  expect_repos_table(
    gh_repos_individual
  )
  test_mocker$cache(gh_repos_individual)
})

test_that("`get_all_repos()` works as expected", {
  mockery::stub(
    github_testhost_priv$get_all_repos,
    "private$get_repos_from_orgs",
    test_mocker$use("gh_repos_from_orgs")
  )
  mockery::stub(
    github_testhost_priv$get_all_repos,
    "private$get_repos_from_repos",
    test_mocker$use("gh_repos_individual")
  )
  gh_repos_table <- github_testhost_priv$get_all_repos(
    add_languages = TRUE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_repos_table(
    gh_repos_table
  )
  test_mocker$cache(gh_repos_table)
})

test_that("`get_all_repos()` is set to scan whole git host", {
  github_testhost_all_priv <- create_github_testhost_all(
    orgs = "test_org",
    mode = "private"
  )
  mockery::stub(
    github_testhost_all_priv$get_all_repos,
    "private$get_orgs_from_host",
    "test_org"
  )
  mockery::stub(
    github_testhost_all_priv$get_all_repos,
    "private$get_repos_from_orgs",
    test_mocker$use("gh_repos_from_orgs")
  )
  mockery::stub(
    github_testhost_all_priv$get_all_repos,
    "private$get_repos_from_repos",
    test_mocker$use("gh_repos_individual")
  )
  expect_repos_table(
    github_testhost_all_priv$get_all_repos(
      add_languages = TRUE,
      verbose  = TRUE,
      progress = FALSE
    )
  )
})

test_that("GitHost adds `repo_api_url` column to GitHub repos table", {
  repos_table <- test_mocker$use("gh_repos_table")
  gh_repos_table_with_api_url <- github_testhost_priv$add_repo_api_url(repos_table)
  expect_true(all(grepl("api.github.com", gh_repos_table_with_api_url$api_url)))
  test_mocker$cache(gh_repos_table_with_api_url)
})

test_that("add_githost_info adds data on Git platform to repos table", {
  gh_repos_table_with_platform <- github_testhost_priv$add_githost_info(
    repos_table = test_mocker$use("gh_repos_table_with_api_url")
  )
  expect_repos_table(
    gh_repos_table_with_platform,
    with_cols = c("api_url", "githost")
  )
  test_mocker$cache(gh_repos_table_with_platform)
})

test_that("get_contributors_from_repo", {
  mockery::stub(
    test_rest_github_priv$get_contributors_from_repo,
    "private$paginate_results",
    test_fixtures$github_contributors_response
  )
  github_contributors <- test_rest_github_priv$get_contributors_from_repo(
    contributors_endpoint = "test_endpoint",
    user_name = "test_user"
  )
  expect_type(github_contributors, "character")
  expect_gt(length(github_contributors), 0)
  test_mocker$cache(github_contributors)
})

test_that("`get_repos_contributors()` adds contributors to repos table", {
  mockery::stub(
    test_rest_github$get_repos_contributors,
    "private$get_contributors_from_repo",
    test_mocker$use("github_contributors")
  )
  gh_repos_with_contributors <- test_rest_github$get_repos_contributors(
    repos_table = test_mocker$use("gh_repos_table_with_platform"),
    progress = FALSE
  )
  expect_repos_table(
    gh_repos_with_contributors,
    with_cols = c("api_url", "githost", "contributors")
  )
  expect_gt(
    length(gh_repos_with_contributors$contributors),
    0
  )
  test_mocker$cache(gh_repos_with_contributors)
})

test_that("`get_repos_contributors()` works on GitHost level", {
  mockery::stub(
    github_testhost_priv$get_repos_contributors,
    "rest_engine$get_repos_contributors",
    test_mocker$use("gh_repos_with_contributors")
  )
  expect_snapshot(
    gh_repos_with_contributors <- github_testhost_priv$get_repos_contributors(
      repos_table = test_mocker$use("gh_repos_table_with_platform"),
      verbose = TRUE,
      progress = FALSE
    )
  )
  expect_repos_table(
    gh_repos_with_contributors,
    with_cols = c("api_url", "githost", "contributors")
  )
  expect_gt(
    length(gh_repos_with_contributors$contributors),
    0
  )
  test_mocker$cache(gh_repos_with_contributors)
})

test_that("`get_repos()` works as expected", {
  mockery::stub(
    github_testhost$get_repos,
    "private$get_all_repos",
    test_mocker$use("gh_repos_table_with_platform")
  )
  gh_repos_table <- github_testhost$get_repos(
    add_contributors = FALSE,
    add_languages = TRUE,
    verbose = FALSE
  )
  expect_repos_table(
    gh_repos_table,
    with_cols = c("api_url", "githost")
  )
  test_mocker$cache(gh_repos_table)
})

test_that("`get_repos()` works as expected", {
  mockery::stub(
    github_testhost$get_repos,
    "private$get_all_repos",
    test_mocker$use("gh_repos_table_with_platform")
  )
  mockery::stub(
    github_testhost$get_repos,
    "private$get_repos_contributors",
    test_mocker$use("gh_repos_with_contributors")
  )
  gh_repos_table_full <- github_testhost$get_repos(
    add_contributors = TRUE,
    add_languages = TRUE,
    language = "R",
    verbose = FALSE
  )
  expect_repos_table(
    gh_repos_table_full,
    with_cols = c("api_url", "githost", "contributors")
  )
  test_mocker$cache(gh_repos_table_full)
})

test_that("filtering by language works", {
  repos_table <- test_mocker$use("gh_repos_table_full")
  repos_table$languages[1] <- "Python"
  filtered_repos <- github_testhost_priv$filter_repos_table_by_language(
    repos_table = repos_table,
    language_filter = "R"
  )
  expect_gt(
    nrow(filtered_repos),
    0
  )
  expect_gt(
    nrow(repos_table),
    nrow(filtered_repos)
  )
})

test_that("filtering by language does not fail when NULL is passed as input", {
  expect_no_error({
    filtered_repos <- github_testhost_priv$filter_repos_table_by_language(
      repos_table = NULL,
      language_filter = "R"
    )
  })
})

test_that("`get_repos_data` pulls repos data from org", {
  if (integration_tests_skipped) {
    mockery::stub(
      github_testhost_priv$get_repos_data,
      "graphql_engine$get_repos_from_org",
      test_mocker$use("gh_repos_from_org")
    )
  }
  github_testhost_priv$orgs_repos <- list("r-world-devs" = c("cohortBuilder", "GitStats", "GitAI"))
  github_testhost_priv$searching_scope <- "org"
  gh_repos_data <- github_testhost_priv$get_repos_data(
    org = "r-world-devs",
    verbose = FALSE
  )
  expect_type(gh_repos_data, "list")
  expect_type(gh_repos_data[["paths"]], "list")
  expect_gt(length(gh_repos_data[["paths"]]), 0)
  expect_type(gh_repos_data[["def_branches"]], "list")
  expect_gt(length(gh_repos_data[["def_branches"]]), 0)
  test_mocker$cache(gh_repos_data)
})

test_that("`get_repos_data` pulls repos data from cache for the second time", {
  mockery::stub(
    github_testhost_priv$get_repos_data,
    "graphql_engine$get_repos_from_org",
    test_mocker$use("gh_repos_from_org")
  )
  expect_gt(
    length(github_testhost_priv$cached_repos),
    0
  )
  github_testhost_priv$searching_scope <- "org"
  expect_snapshot(
    gh_repos_data <- github_testhost_priv$get_repos_data(
      org = "r-world-devs",
      verbose = TRUE
    )
  )
  expect_type(gh_repos_data, "list")
  expect_type(gh_repos_data[["paths"]], "list")
  expect_gt(length(gh_repos_data[["paths"]]), 0)
  expect_type(gh_repos_data[["def_branches"]], "list")
  expect_gt(length(gh_repos_data[["def_branches"]]), 0)
})

test_that("`get_repos_data` pulls repos data from repos", {
  if (integration_tests_skipped) {
    mockery::stub(
      github_testhost_priv$get_repos_data,
      "graphql_engine$get_repos_from_org",
      test_mocker$use("gh_repos_from_org")
    )
  }
  github_testhost_priv$searching_scope <- "repo"
  gh_repos_data <- github_testhost_priv$get_repos_data(
    org = "r-world-devs",
    repos = c("cohortBuilder", "GitStats", "GitAI"),
    verbose = FALSE
  )
  expect_type(gh_repos_data, "list")
  expect_type(gh_repos_data[["paths"]], "list")
  expect_gt(length(gh_repos_data[["paths"]]), 0)
  expect_type(gh_repos_data[["def_branches"]], "list")
  expect_gt(length(gh_repos_data[["def_branches"]]), 0)
})

test_that("`set_searching_scope` does not throw error when `orgs` or `repos` are defined", {
  expect_snapshot(
    gitlab_testhost_priv$set_searching_scope(orgs = "mbtests", repos = NULL, verbose = TRUE)
  )
  expect_snapshot(
    gitlab_testhost_priv$set_searching_scope(orgs = NULL, repos = "mbtests/GitStatsTesting", verbose = TRUE)
  )
})

test_that("`set_searching_scope` sets scope to whole host", {
  gitlab_testhost_priv$is_public <- FALSE
  expect_snapshot(
    gitlab_testhost_priv$set_searching_scope(orgs = NULL, repos = NULL, verbose = TRUE)
  )
  expect_true(
    gitlab_testhost_priv$scan_all
  )
})

test_that("`extract_repos_and_orgs` extracts fullnames vector into a list of GitLab organizations with assigned repositories", {
  repos_fullnames <- c(
    "mbtests/gitstatstesting", "mbtests/gitstats-testing-2", "mbtests/subgroup/test-project-in-subgroup"
  )
  expect_equal(
    gitlab_testhost_priv$extract_repos_and_orgs(repos_fullnames),
    list(
      "mbtests" = c("gitstatstesting", "gitstats-testing-2"),
      "mbtests/subgroup" = c("test-project-in-subgroup")
    )
  )
})

test_that("`extract_repos_and_orgs` extracts fullnames vector into a list of GitHub organizations with assigned repositories", {
  repos_fullnames <- c(
    "r-world-devs/GitStats", "r-world-devs/shinyCohortBuilder",
    "openpharma/DataFakeR", "openpharma/GithubMetrics"
  )
  expect_equal(
    github_testhost_priv$extract_repos_and_orgs(repos_fullnames),
    list(
      "r-world-devs" = c("GitStats", "shinyCohortBuilder"),
      "openpharma" = c("DataFakeR", "GithubMetrics")
    )
  )
})

test_that("When token is empty throw error", {
  expect_snapshot(
    error = TRUE,
    github_testhost_priv$check_token("")
  )
})

test_that("`check_token()` prints error when token exists but does not grant access", {
  token <- "does_not_grant_access"
  expect_snapshot_error(
    github_testhost_priv$check_token(token)
  )
})

test_that("when token is proper token is passed", {
  skip_on_cran()
  github_testhost_priv <- create_github_testhost(
    orgs = "test-org",
    token = Sys.getenv("GITHUB_PAT"),
    mode = "private"
  )
  mockery::stub(
    github_testhost_priv$check_token,
    "private$test_token",
    TRUE
  )
  expect_equal(
    github_testhost_priv$check_token(Sys.getenv("GITHUB_PAT")),
    Sys.getenv("GITHUB_PAT")
  )
})

test_that("check_endpoint returns TRUE if they are correct", {
  skip_on_cran()
  github_testhost_priv <- create_github_testhost(
    org = "test-org",
    token = Sys.getenv("GITHUB_PAT"),
    mode = "private"
  )
  expect_true(
    github_testhost_priv$check_endpoint(
      endpoint = "https://api.github.com/repos/r-world-devs/GitStats",
      type = "Repository"
    )
  )
  expect_true(
    github_testhost_priv$check_endpoint(
      endpoint = "https://api.github.com/orgs/openpharma",
    )
  )
})

test_that("check_endpoint returns error if they are not correct", {
  skip_on_cran()
  github_testhost_priv <- create_github_testhost(
    orgs = "test-org",
    token = Sys.getenv("GITHUB_PAT"),
    mode = "private"
  )
  expect_snapshot_error(
    check <- github_testhost_priv$check_endpoint(
      endpoint = "https://api.github.com/repos/r-worlddevs/GitStats",
      type = "Repository",
      verbose = TRUE,
      .error = TRUE
    )
  )
})

test_that("check_endpoint returns warning if they are not correct", {
  skip_on_cran()
  github_testhost_priv <- create_github_testhost(
    orgs = "test-org",
    token = Sys.getenv("GITHUB_PAT"),
    mode = "private"
  )
  expect_snapshot(
    check <- github_testhost_priv$check_endpoint(
      endpoint = "https://api.github.com/repos/r-worlddevs/GitStats",
      type = "Repository",
      verbose = TRUE,
      .error = FALSE
    )
  )
})

test_that("`check_if_public` works correctly", {
  expect_true(
    github_testhost_priv$check_if_public("api.github.com")
  )
  expect_false(
    github_testhost_priv$check_if_public("github.internal.com")
  )
})

test_that("`set_default_token` sets default token for public GitHub", {
  skip_on_cran()
  expect_snapshot(
    default_token <- github_testhost_priv$set_default_token(
      verbose = TRUE
    )
  )
  test_rest <- create_testrest(token = default_token)
  expect_equal(
    test_rest$perform_request(
      endpoint = "https://api.github.com",
      token = default_token
    )$status,
    200
  )
})

test_that("`set_default_token` returns error if none are found", {
  mockery::stub(
    github_testhost_priv$set_default_token,
    "private$test_token",
    FALSE
  )
  expect_error({
    github_testhost_priv$set_default_token(
      verbose = FALSE
    )
  })
})

test_that("`test_token` works properly", {
  skip_on_cran()
  expect_true(
    github_testhost_priv$test_token(Sys.getenv("GITHUB_PAT"))
  )
  expect_false(
    github_testhost_priv$test_token("false_token")
  )
})

test_that("`test_token` works properly", {
  skip_on_cran()
  expect_true(
    gitlab_testhost_priv$test_token(Sys.getenv("GITLAB_PAT_PUBLIC"))
  )
  expect_false(
    gitlab_testhost_priv$test_token("false_token")
  )
})

test_that("`check_token_scopes` works for GitHub", {
  skip_on_cran()
  github_pat <- Sys.getenv("GITHUB_PAT")
  github_response <- httr2::request("https://api.github.com") |>
    httr2::req_headers("Authorization" = paste0("Bearer ", github_pat)) |>
    httr2::req_perform()
  expect_true(
    github_testhost_priv$check_token_scopes(
      response = github_response,
      token = github_pat
    )
  )
})

test_that("`set_default_token` sets default token for GitLab", {
  skip_on_cran()
  expect_snapshot(
    withr::with_envvar(new = c("GITLAB_PAT" = Sys.getenv("GITLAB_PAT_PUBLIC")), {
      default_token <- gitlab_testhost_priv$set_default_token(verbose = TRUE)
    })
  )
  test_rest <- create_testrest(token = default_token)
  expect_equal(
    test_rest$perform_request(
      endpoint = "https://gitlab.com/api/v4/projects",
      token = default_token
    )$status,
    200
  )
})

test_that("pr_by_repo GitHub query is built properly", {
  gh_pr_from_repo_query <-
    test_gqlquery_gh$pull_requests_from_repo()
  expect_snapshot(
    gh_pr_from_repo_query
  )
})

test_that("`get_pr_page_from_repo()` pulls pr page from repository", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_github_priv$get_pr_page_from_repo,
      "self$gql_response",
      test_fixtures_pr$github_graphql_pr_response
    )
  }
  pr_page <- test_graphql_github_priv$get_pr_page_from_repo(
    org = "r-world-devs",
    repo = "GitStats"
  )
  expect_pr_gql_response(
    pr_page$data$repository$pullRequests$edges[[1]]
  )
  test_mocker$cache(pr_page)
})

test_that("`get_pr_from_one_repo()` prepares formatted list", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_github_priv$get_pr_from_one_repo,
      "private$get_pr_page_from_repo",
      test_mocker$use("pr_page")
    )
  }
  pr_from_repo <- test_graphql_github_priv$get_pr_from_one_repo(
    org = "r-world-devs",
    repo = "GitStats"
  )
  if (!integration_tests_skipped) {
    expect_gt(length(pr_from_repo), 100)
  }
  expect_pr_full_list(
    pr_from_repo
  )
  test_mocker$cache(pr_from_repo)
})

test_that("`get_pr_from_repos()` pulls pr from repos", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_github$get_pr_from_repos,
      "private$get_pr_from_one_repo",
      test_mocker$use("pr_from_repo")
    )
  }
  pr_from_repos <- test_graphql_github$get_pr_from_repos(
    org = "r-world-devs",
    repo = c("GitStats", "GitAI")
  )
  expect_pr_full_list(
    pr_from_repos[[1]]
  )
  test_mocker$cache(pr_from_repos)
})

test_that("`prepare_pr_table()` prepares pr table", {
  gh_pr_table <- test_graphql_github$prepare_pr_table(
    repos_list_with_pr = test_mocker$use("pr_from_repos"),
    org = "r-world-devs"
  )
  expect_pr_table(
    gh_pr_table
  )
  test_mocker$cache(gh_pr_table)
})

test_that("get_pr_from_orgs for GitHub works", {
  if (integration_tests_skipped) {
    mockery::stub(
      github_testhost_priv$get_pr_from_orgs,
      "graphql_engine$prepare_pr_table",
      test_mocker$use("gh_pr_table")
    )
    mockery::stub(
      github_testhost_priv$get_pr_from_orgs,
      "private$get_repos_names",
      test_mocker$use("gh_repos_names")
    )
  }
  github_testhost_priv$searching_scope <- "org"
  gh_pr_from_orgs <- github_testhost_priv$get_pr_from_orgs(
    verbose = FALSE,
    progress = FALSE
  )
  expect_pr_table(
    gh_pr_from_orgs
  )
  test_mocker$cache(gh_pr_from_orgs)
})

test_that("get_pr_from_orgs for GitHub prints messages", {
  mockery::stub(
    github_testhost_priv$get_pr_from_orgs,
    "graphql_engine$prepare_pr_table",
    test_mocker$use("gh_pr_table")
  )
  mockery::stub(
    github_testhost_priv$get_pr_from_orgs,
    "private$get_repos_data",
    test_mocker$use("gh_repos_data")
  )
  github_testhost_priv$searching_scope <- "org"
  expect_snapshot(
    gh_pr_from_orgs <- github_testhost_priv$get_pr_from_orgs(
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("get_pr_from_repos for GitHub works", {
  if (integration_tests_skipped) {
    mockery::stub(
      github_testhost_priv$get_pr_from_repos,
      "graphql_engine$prepare_pr_table",
      test_mocker$use("gh_pr_table")
    )
    github_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
    test_org <- "test_org"
    attr(test_org, "type") <- "organization"
    mockery::stub(
      github_testhost_priv$get_pr_from_repos,
      "graphql_engine$set_owner_type",
      test_org
    )
  } else {
    github_testhost_priv$orgs_repos <- list("r-world-devs" = "GitStats")
  }
  github_testhost_priv$searching_scope <- "repo"
  gh_pr_from_repos <- github_testhost_priv$get_pr_from_repos(
    verbose = FALSE,
    progress = FALSE
  )
  expect_pr_table(
    gh_pr_from_repos
  )
  test_mocker$cache(gh_pr_from_repos)
})

test_that("get_pr_from_repos for GitHub prints messages", {
  mockery::stub(
    github_testhost_priv$get_pr_from_repos,
    "graphql_engine$prepare_pr_table",
    test_mocker$use("gh_pr_table")
  )
  github_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    github_testhost_priv$get_pr_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  github_testhost_priv$searching_scope <- "repo"
  expect_snapshot(
    gh_pr_from_repos <- github_testhost_priv$get_pr_from_repos(
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("`get_pull_requests()` retrieves pr in the table format in a certain time span", {
  mockery::stub(
    github_testhost$get_pull_requests,
    "private$get_pr_from_orgs",
    test_mocker$use("gh_pr_from_orgs")
  )
  mockery::stub(
    github_testhost$get_pull_requests,
    "private$get_pr_from_repos",
    test_mocker$use("gh_pr_from_repos")
  )
  gh_pr_table <- github_testhost$get_pull_requests(
    since = "2024-01-01",
    until = "2026-03-06",
    verbose = FALSE,
    progress = FALSE
  )
  expect_pr_table(
    gh_pr_table
  )
  if (!integration_tests_skipped) {
    gh_pr_table_shorter <- github_testhost$get_pull_requests(
      since = "2024-02-01",
      until = "2025-01-01",
      verbose = FALSE,
      progress = FALSE
    )
    expect_true(
      nrow(gh_pr_table) > nrow(gh_pr_table_shorter)
    )
    expect_true(
      max(gh_pr_table_shorter$created_at) <= "2025-01-01"
    )
    expect_true(
      min(gh_pr_table_shorter$created_at) >= "2024-02-01"
    )
  }
  test_mocker$cache(gh_pr_table)
})

test_that("`get_pull_requests()` retrieves closed and open pr in the table format in a certain time span", {
  mockery::stub(
    github_testhost$get_pull_requests,
    "private$get_pr_from_orgs",
    test_mocker$use("gh_pr_from_orgs")
  )
  mockery::stub(
    github_testhost$get_pull_requests,
    "private$get_pr_from_repos",
    test_mocker$use("gh_pr_from_repos")
  )
  if (integration_tests_skipped) {
    gh_open_pr_table <- github_testhost$get_pull_requests(
      since = "2024-01-01",
      until = "2025-01-01",
      state = "open",
      verbose = FALSE,
      progress = FALSE
    )
    expect_pr_table(
      gh_open_pr_table
    )
    expect_true(
      all(gh_open_pr_table$state == "open")
    )
  }
  gh_closed_pr_table <- github_testhost$get_pull_requests(
    since = "2024-01-01",
    until = "2025-01-01",
    state = "closed",
    verbose = FALSE,
    progress = FALSE
  )
  expect_pr_table(
    gh_closed_pr_table
  )
  expect_true(
    all(gh_closed_pr_table$state == "closed")
  )
})

test_that("pr_by_repo GitLab query is built properly", {
  gl_pr_from_repo_query <-
    test_gqlquery_gl$pull_requests_from_repo()
  expect_snapshot(
    gl_pr_from_repo_query
  )
})

test_that("`get_pr_page_from_repo()` pulls pr page from repository", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_gitlab_priv$get_pr_page_from_repo,
      "self$gql_response",
      test_fixtures_pr$gitlab_graphql_pr_response
    )
  }
  gl_pr_page <- test_graphql_gitlab_priv$get_pr_page_from_repo(
    org = "mbtests",
    repo = "gitstatstesting"
  )

  expect_pr_gql_response(
    gl_pr_page$data$project$mergeRequests$edges[[1]]
  )
  test_mocker$cache(gl_pr_page)
})

test_that("`get_pr_from_one_repo()` prepares formatted list", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_gitlab_priv$get_pr_from_one_repo,
      "private$get_pr_page_from_repo",
      test_mocker$use("gl_pr_page")
    )
  }
  gl_pr_from_repo <- test_graphql_gitlab_priv$get_pr_from_one_repo(
    org = "mbtests",
    repo = "gitstatstesting"
  )
  expect_pr_full_list(
    gl_pr_from_repo
  )
  test_mocker$cache(gl_pr_from_repo)
})

test_that("`get_pr_from_repos()` pulls pr from repos", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_gitlab$get_pr_from_repos,
      "private$get_pr_from_one_repo",
      test_mocker$use("gl_pr_from_repo")
    )
  }
  pr_from_repos <- test_graphql_gitlab$get_pr_from_repos(
    org = "mbtests",
    repo = c("gitstatstesting", "graphql_tests")
  )
  expect_pr_full_list(
    pr_from_repos[[1]]
  )
  test_mocker$cache(pr_from_repos)
})

test_that("`prepare_pr_table()` prepares pr table", {
  gl_pr_table <- test_graphql_gitlab$prepare_pr_table(
    repos_list_with_pr = test_mocker$use("pr_from_repos"),
    org = "mbtests"
  )
  expect_pr_table(
    gl_pr_table
  )
  test_mocker$cache(gl_pr_table)
})

test_that("get_pr_from_orgs for GitLab works", {
  if (integration_tests_skipped) {
    mockery::stub(
      gitlab_testhost_priv$get_pr_from_orgs,
      "graphql_engine$prepare_pr_table",
      test_mocker$use("gl_pr_table")
    )
    mockery::stub(
      gitlab_testhost_priv$get_pr_from_orgs,
      "private$get_repos_names",
      test_mocker$use("gl_repos_names")
    )
  }
  gitlab_testhost_priv$searching_scope <- "org"
  gl_pr_from_orgs <- gitlab_testhost_priv$get_pr_from_orgs(
    verbose = FALSE,
    progress = FALSE
  )
  expect_pr_table(
    gl_pr_from_orgs
  )
  test_mocker$cache(gl_pr_from_orgs)
})

test_that("get_pr_from_orgs for GitLab prints message", {
  if (integration_tests_skipped) {
    mockery::stub(
      gitlab_testhost_priv$get_pr_from_orgs,
      "graphql_engine$prepare_pr_table",
      test_mocker$use("gl_pr_table")
    )
    mockery::stub(
      gitlab_testhost_priv$get_pr_from_orgs,
      "private$get_repos_names",
      test_mocker$use("gl_repos_names")
    )
  }
  gitlab_testhost_priv$searching_scope <- "org"
  expect_snapshot(
    gl_pr_from_orgs <- gitlab_testhost_priv$get_pr_from_orgs(
      verbose = TRUE,
      progress = FALSE
    )
  )  
})

test_that("get_pr_from_repos for GitLab works", {
  if (integration_tests_skipped) {
    mockery::stub(
      gitlab_testhost_priv$get_pr_from_repos,
      "graphql_engine$prepare_pr_table",
      test_mocker$use("gl_pr_table")
    )
    test_org <- "mbtests"
    attr(test_org, "type") <- "organization"
    mockery::stub(
      gitlab_testhost_priv$get_pr_from_repos,
      "graphql_engine$set_owner_type",
      test_org
    )
  }
  gitlab_testhost_priv$orgs_repos <- list("mbtests" = "gitstatstesting")
  gitlab_testhost_priv$searching_scope <- "repo"
  gl_pr_from_repos <- gitlab_testhost_priv$get_pr_from_repos(
    verbose = FALSE,
    progress = FALSE
  )
  expect_pr_table(
    gl_pr_from_repos
  )
  test_mocker$cache(gl_pr_from_repos)
})

test_that("get_pr_from_repos for GitLab prints message", {
  if (integration_tests_skipped) {
    mockery::stub(
      gitlab_testhost_priv$get_pr_from_repos,
      "graphql_engine$prepare_pr_table",
      test_mocker$use("gl_pr_table")
    )
    test_org <- "mbtests"
    attr(test_org, "type") <- "organization"
    mockery::stub(
      gitlab_testhost_priv$get_pr_from_repos,
      "graphql_engine$set_owner_type",
      test_org
    )
  }
  gitlab_testhost_priv$orgs_repos <- list("mbtests" = "gitstatstesting")
  gitlab_testhost_priv$searching_scope <- "repo"
  expect_snapshot(
    gl_pr_from_repos <- gitlab_testhost_priv$get_pr_from_repos(
      verbose = TRUE,
      progress = FALSE
    )
  )  
})

test_that("`get_pull_requests()` retrieves pr in the table format in a certain time span", {
  mockery::stub(
    gitlab_testhost$get_pull_requests,
    "private$get_pr_from_orgs",
    test_mocker$use("gl_pr_from_orgs")
  )
  mockery::stub(
    gitlab_testhost$get_pull_requests,
    "private$get_pr_from_repos",
    test_mocker$use("gl_pr_from_repos")
  )
  gl_pr_table <- gitlab_testhost$get_pull_requests(
    since = "2026-02-01",
    until = "2026-02-28",
    verbose = FALSE,
    progress = FALSE
  )
  expect_pr_table(
    gl_pr_table
  )
  if (!integration_tests_skipped) {
    gl_pr_table_shorter <- gitlab_testhost$get_pull_requests(
      since = "2026-02-01",
      until = "2026-02-24",
      verbose = FALSE,
      progress = FALSE
    )
    expect_true(
      nrow(gl_pr_table) > nrow(gl_pr_table_shorter)
    )
    expect_true(
      max(gl_pr_table_shorter$created_at) <= "2026-02-24"
    )
    expect_true(
      min(gl_pr_table_shorter$created_at) >= "2026-02-01"
    )
  }  
  test_mocker$cache(gl_pr_table)
})

test_that("`get_pull_requests()` retrieves open and closed pr in the table format in a certain time span", {
  skip_if(integration_tests_skipped)
  mockery::stub(
    gitlab_testhost$get_pull_requests,
    "private$get_pr_from_orgs",
    test_mocker$use("gl_pr_from_orgs")
  )
  mockery::stub(
    gitlab_testhost$get_pull_requests,
    "private$get_pr_from_repos",
    test_mocker$use("gl_pr_from_repos")
  )
  gl_open_pr_table <- gitlab_testhost$get_pull_requests(
    since = "2026-02-01",
    until = "2026-02-28",
    state = "open",
    verbose = FALSE,
    progress = FALSE
  )
  expect_pr_table(
    gl_open_pr_table
  )
  expect_true(
    all(gl_open_pr_table$state == "open")
  )
  gl_closed_pr_table <- gitlab_testhost$get_pull_requests(
    since = "2026-02-01",
    until = "2026-02-28",
    state = "closed",
    verbose = FALSE,
    progress = FALSE
  )
  expect_pr_table(
    gl_closed_pr_table
  )
  expect_true(
    all(gl_closed_pr_table$state == "closed")
  )
})

test_that("issues_by_repo GitLab query is built properly", {
  gl_issues_from_repo_query <-
    test_gqlquery_gl$issues_from_repo()
  expect_snapshot(
    gl_issues_from_repo_query
  )
})

test_that("issues page is pulled from repository", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_gitlab_priv$get_issues_page_from_repo,
      "self$gql_response",
      test_fixtures$gitlab_graphql_issues_response
    )
    org <- "test_org"
    repo <- "TestRepo"
  } else {
    org <- "mbtests"
    repo <- "gitstatstesting"
  }
  issues_page <- test_graphql_gitlab_priv$get_issues_page_from_repo(
    org = org,
    repo = repo
  )
  expect_issues_gitlab_page(issues_page)
  test_mocker$cache(issues_page)
})

test_that("`get_issues_from_one_repo()` prepares formatted list", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_gitlab_priv$get_issues_from_one_repo,
      "private$get_issues_page_from_repo",
      test_mocker$use("issues_page")
    )
  }
  issues_from_repo <- test_graphql_gitlab_priv$get_issues_from_one_repo(
    org = "mbtests",
    repo = "gitstatstesting"
  )
  expect_issues_full_list(
    issues_from_repo
  )
  test_mocker$cache(issues_from_repo)
})

test_that("`get_issues_from_repos()` pulls issues from repos", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_gitlab$get_issues_from_repos,
      "private$get_issues_from_one_repo",
      test_mocker$use("issues_from_repo")
    )
  }
  issues_from_repos <- test_graphql_gitlab$get_issues_from_repos(
    org = "mbtests",
    repo = c("gitstatstesting", "graphql_tests")
  )
  expect_issues_full_list(
    issues_from_repos[[1]]
  )
  test_mocker$cache(issues_from_repos)
})

test_that("`prepare_issues_table()` prepares issues table", {
  gl_issues_table <- test_graphql_gitlab$prepare_issues_table(
    repos_list_with_issues = test_mocker$use("issues_from_repos"),
    org = "r-world-devs"
  )
  expect_issues_table(
    gl_issues_table
  )
  test_mocker$cache(gl_issues_table)
})

test_that("get_issues_from_orgs for GitLab works", {
  if (integration_tests_skipped) {
    mockery::stub(
      gitlab_testhost_priv$get_issues_from_orgs,
      "graphql_engine$prepare_issues_table",
      test_mocker$use("gl_issues_table")
    )
    mockery::stub(
      gitlab_testhost_priv$get_issues_from_orgs,
      "private$get_repos_names",
      test_mocker$use("gl_repos_names")
    )
  }
  gitlab_testhost_priv$searching_scope <- "org"
  gl_issues_from_orgs <- gitlab_testhost_priv$get_issues_from_orgs(
    verbose = FALSE,
    progress = FALSE
  )
  expect_issues_table(
    gl_issues_from_orgs
  )
  test_mocker$cache(gl_issues_from_orgs)
})

test_that("get_issues_from_repos for GitLab works", {
  if (integration_tests_skipped) {
    mockery::stub(
      gitlab_testhost_priv$get_issues_from_repos,
      "graphql_engine$prepare_issues_table",
      test_mocker$use("gl_issues_table")
    )
    gitlab_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
    test_org <- "test_org"
    attr(test_org, "type") <- "organization"
    mockery::stub(
      gitlab_testhost_priv$get_issues_from_repos,
      "graphql_engine$set_owner_type",
      test_org
    )
  } else {
    gitlab_testhost_priv$orgs_repos <- list("mbtests" = "gitstatstesting")
  }
  gitlab_testhost_priv$searching_scope <- "repo"
  gl_issues_from_repos <- gitlab_testhost_priv$get_issues_from_repos(
    verbose = FALSE,
    progress = FALSE
  )
  expect_issues_table(
    gl_issues_from_repos
  )
  test_mocker$cache(gl_issues_from_repos)
})

test_that("`get_issues()` retrieves issues in the table format in a certain time span", {
  mockery::stub(
    gitlab_testhost$get_issues,
    "private$get_issues_from_orgs",
    test_mocker$use("gl_issues_from_orgs")
  )
  mockery::stub(
    gitlab_testhost$get_issues,
    "private$get_issues_from_repos",
    test_mocker$use("gl_issues_from_repos")
  )
  gl_issues_table <- gitlab_testhost$get_issues(
    since = "2023-01-01",
    until = "2025-03-06",
    verbose = FALSE,
    progress = FALSE
  )
  expect_issues_table(
    gl_issues_table
  )
  gl_issues_table_shorter <- gitlab_testhost$get_issues(
    since = "2023-02-01",
    until = "2024-01-01",
    verbose = FALSE,
    progress = FALSE
  )
  expect_true(
    nrow(gl_issues_table) > nrow(gl_issues_table_shorter)
  )
  expect_true(
    max(gl_issues_table_shorter$created_at) <= "2024-01-01"
  )
  expect_true(
    min(gl_issues_table_shorter$created_at) >= "2023-02-01"
  )
  test_mocker$cache(gl_issues_table)
})

test_that("`get_issues()` retrieves open issues in the table format in a certain time span", {
  mockery::stub(
    gitlab_testhost$get_issues,
    "private$get_issues_from_orgs",
    test_mocker$use("gl_issues_from_orgs")
  )
  mockery::stub(
    gitlab_testhost$get_issues,
    "private$get_issues_from_repos",
    test_mocker$use("gl_issues_from_repos")
  )
  gl_open_issues_table <- gitlab_testhost$get_issues(
    since = "2023-01-01",
    until = "2024-01-01",
    state = "open",
    verbose = FALSE,
    progress = FALSE
  )
  expect_issues_table(
    gl_open_issues_table
  )
  expect_true(
    all(gl_open_issues_table$state == "open")
  )
  gl_closed_issues_table <- gitlab_testhost$get_issues(
    since = "2023-01-01",
    until = "2024-01-01",
    state = "closed",
    verbose = FALSE,
    progress = FALSE
  )
  expect_issues_table(
    gl_closed_issues_table
  )
  expect_true(
    all(gl_closed_issues_table$state == "closed")
  )
})

test_that("get_repos_urls() works for whole orgs", {
  mockery::stub(
    test_rest_github$get_repos_urls,
    "private$paginate_results",
    test_fixtures$github_repositories_rest_response
  )
  gh_repos_urls <- test_rest_github$get_repos_urls(
    type = "web",
    org = "test_org",
    repos = NULL
  )
  expect_length(
    gh_repos_urls,
    3
  )
  test_mocker$cache(gh_repos_urls)
})

test_that("get_repos_urls() works for individual repos", {
  mockery::stub(
    test_rest_github$get_repos_urls,
    "private$paginate_results",
    test_fixtures$github_repositories_rest_response
  )
  gh_repos_urls <- test_rest_github$get_repos_urls(
    type = "web",
    org = "test_org",
    repos = c("testRepo", "testRepo2")
  )
  expect_length(
    gh_repos_urls,
    2
  )
  test_mocker$cache(gh_repos_urls)
})

test_that("get_repos_urls prepares api repo_urls vector", {
  mockery::stub(
    test_rest_github$get_repos_urls,
    "private$paginate_results",
    test_fixtures$github_repositories_rest_response
  )
  gh_api_repos_urls <- test_rest_github$get_repos_urls(
    org = "test_org",
    repos = NULL,
    type = "api"
  )
  expect_gt(length(gh_api_repos_urls), 0)
  expect_true(any(grepl("test-org", gh_api_repos_urls)))
  expect_true(all(grepl("api", gh_api_repos_urls)))
  test_mocker$cache(gh_api_repos_urls)
})

test_that("get_repos_urls_from_orgs prepares web repo_urls vector", {
  mockery::stub(
    github_testhost_priv$get_repos_urls_from_orgs,
    "rest_engine$get_repos_urls",
    test_mocker$use("gh_repos_urls")
  )
  github_testhost_priv$searching_scope <- "org"
  github_testhost_priv$orgs <- "test_org"
  expect_snapshot(
    gh_repos_urls_from_orgs <- github_testhost_priv$get_repos_urls_from_orgs(
      type = "web",
      verbose = TRUE,
      progress = FALSE
    )
  )
  expect_gt(length(gh_repos_urls_from_orgs), 0)
  expect_true(any(grepl("test-org", gh_repos_urls_from_orgs)))
  expect_true(all(grepl("https://testhost.com/", gh_repos_urls_from_orgs)))
  test_mocker$cache(gh_repos_urls_from_orgs)
})

test_that("get_repos_urls_from_repos prepares web repo_urls vector", {
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    github_testhost_priv$get_repos_urls_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  mockery::stub(
    github_testhost_priv$get_repos_urls_from_repos,
    "rest_engine$get_repos_urls",
    test_mocker$use("gh_repos_urls"),
    depth = 2L
  )
  github_testhost_priv$searching_scope <- c("repo")
  github_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
  expect_snapshot(
    gh_repos_urls <- github_testhost_priv$get_repos_urls_from_repos(
      type = "web",
      verbose = TRUE,
      progress = FALSE
    )
  )
  expect_gt(length(gh_repos_urls), 0)
  expect_true(any(grepl("test-org", gh_repos_urls)))
  expect_true(all(grepl("https://testhost.com/", gh_repos_urls)))
  test_mocker$cache(gh_repos_urls)
})

test_that("get_all_repos_urls prepares web repo_urls vector", {
  mockery::stub(
    github_testhost_priv$get_all_repos_urls,
    "private$get_repos_urls_from_orgs",
    test_mocker$use("gh_repos_urls_from_orgs")
  )
  mockery::stub(
    github_testhost_priv$get_all_repos_urls,
    "private$get_repos_urls_from_repos",
    test_mocker$use("gh_repos_urls")
  )
  expect_snapshot(
    gh_repos_urls <- github_testhost_priv$get_all_repos_urls(
      type = "web",
      verbose = TRUE
    )
  )
  expect_gt(length(gh_repos_urls), 0)
  expect_true(any(grepl("test-org", gh_repos_urls)))
  expect_true(all(grepl("https://testhost.com/", gh_repos_urls)))
  test_mocker$cache(gh_repos_urls)
})

test_that("get_all_repos_urls is set to scan whole host", {
  github_testhost_all_priv <- create_github_testhost_all(
    orgs = "test_org",
    mode = "private"
  )
  mockery::stub(
    github_testhost_all_priv$get_all_repos_urls,
    "private$get_orgs_from_host",
    "test_org"
  )
  mockery::stub(
    github_testhost_all_priv$get_all_repos_urls,
    "private$get_repos_urls_from_orgs",
    test_mocker$use("gh_repos_urls_from_orgs")
  )
  mockery::stub(
    github_testhost_all_priv$get_all_repos_urls,
    "private$get_repos_urls_from_repos",
    test_mocker$use("gh_repos_urls")
  )
  gh_repos_urls <- github_testhost_all_priv$get_all_repos_urls(
    type = "web",
    verbose = FALSE
  )
  expect_gt(length(gh_repos_urls), 0)
  expect_true(any(grepl("test-org", gh_repos_urls)))
  expect_true(all(grepl("https://testhost.com/", gh_repos_urls)))
})

test_that("get_repo_url_from_response retrieves repositories URLS", {
  gh_repo_api_urls <- github_testhost_priv$get_repo_url_from_response(
    search_response = test_mocker$use("gh_repos_raw_output"),
    type = "api"
  )
  expect_type(gh_repo_api_urls, "character")
  expect_gt(length(gh_repo_api_urls), 0)
  test_mocker$cache(gh_repo_api_urls)
  gh_repo_web_urls <- github_testhost_priv$get_repo_url_from_response(
    search_response = test_mocker$use("gh_repos_raw_output"),
    type = "web"
  )
  expect_type(gh_repo_web_urls, "character")
  expect_gt(length(gh_repo_web_urls), 0)
  test_mocker$cache(gh_repo_web_urls)
})

test_that("get_repos_urls_with_code_from_host returns repositories URLS", {
  mockery::stub(
    github_testhost_priv$get_repos_urls_with_code_from_host,
    "private$get_repo_url_from_response",
    test_mocker$use("gh_repo_web_urls")
  )
  gh_repos_urls_with_code_from_host <- github_testhost_priv$get_repos_urls_with_code_from_host(
    type = "api",
    code = "shiny",
    in_files = "DESCRIPTION",
    in_path = FALSE,
    verbose = FALSE,
    progress= FALSE
  )
  expect_type(gh_repos_urls_with_code_from_host, "character")
  expect_gt(length(gh_repos_urls_with_code_from_host), 0)
  test_mocker$cache(gh_repos_urls_with_code_from_host)
})

test_that("get_repos_urls_with_code_from_orgs returns repositories URLS", {
  mockery::stub(
    github_testhost_priv$get_repos_urls_with_code_from_orgs,
    "private$get_repo_url_from_response",
    test_mocker$use("gh_repo_web_urls")
  )
  gh_repos_urls_with_code_from_orgs <- github_testhost_priv$get_repos_urls_with_code_from_orgs(
    type = "web",
    code = "shiny",
    in_files = "DESCRIPTION",
    in_path = FALSE,
    verbose = FALSE,
    progress= FALSE
  )
  expect_type(gh_repos_urls_with_code_from_orgs, "character")
  expect_gt(length(gh_repos_urls_with_code_from_orgs), 0)
  test_mocker$cache(gh_repos_urls_with_code_from_orgs)
})

test_that("get_repos_urls_with_code_from_repos returns repositories URLS", {
  mockery::stub(
    github_testhost_priv$get_repos_urls_with_code_from_repos,
    "private$get_repo_url_from_response",
    test_mocker$use("gh_repo_web_urls")
  )
  gh_repos_urls_with_code_from_repos <- github_testhost_priv$get_repos_urls_with_code_from_repos(
    type = "web",
    code = "shiny",
    in_files = "DESCRIPTION",
    in_path = FALSE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_type(gh_repos_urls_with_code_from_repos, "character")
  expect_gt(length(gh_repos_urls_with_code_from_repos), 0)
  test_mocker$cache(gh_repos_urls_with_code_from_repos)
})

test_that("get_repos_urls_with_code returns repositories URLS", {
  mockery::stub(
    github_testhost_priv$get_repos_urls_with_code,
    "private$get_repos_urls_with_code_from_orgs",
    test_mocker$use("gh_repos_urls_with_code_from_orgs")
  )
  mockery::stub(
    github_testhost_priv$get_repos_urls_with_code,
    "private$get_repos_urls_with_code_from_repos",
    test_mocker$use("gh_repos_urls_with_code_from_repos")
  )
  gh_repos_urls_with_code_in_files <- github_testhost_priv$get_repos_urls_with_code(
    type = "web",
    code = "shiny",
    in_files = "DESCRIPTION",
    in_path = FALSE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_type(gh_repos_urls_with_code_in_files, "character")
  expect_gt(length(gh_repos_urls_with_code_in_files), 0)
  test_mocker$cache(gh_repos_urls_with_code_in_files)
})

test_that("get_repos_urls_with_code returns repositories URLS from whole host", {
  github_testhost_priv <- create_github_testhost(
    mode = "private"
  )
  mockery::stub(
    github_testhost_priv$get_repos_urls_with_code,
    "private$get_repos_urls_with_code_from_host",
    test_mocker$use("gh_repos_urls_with_code_from_host")
  )
  gh_repos_urls_with_code_in_files <- github_testhost_priv$get_repos_urls_with_code(
    type = "web",
    code = "shiny",
    in_files = "DESCRIPTION",
    in_path = FALSE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_type(gh_repos_urls_with_code_in_files, "character")
  expect_gt(length(gh_repos_urls_with_code_in_files), 0)
  test_mocker$cache(gh_repos_urls_with_code_in_files)
})

test_that("get_repos_urls returns repositories URLS", {
  mockery::stub(
    github_testhost$get_repos_urls,
    "private$get_repos_urls_with_code",
    test_mocker$use("gh_repos_urls_with_code_in_files")
  )
  gh_repos_urls_with_code_in_files <- github_testhost$get_repos_urls(
    type = "web",
    with_code = "shiny",
    in_files = "DESCRIPTION",
    with_file = NULL,
    verbose = FALSE,
    progress = FALSE
  )
  expect_type(gh_repos_urls_with_code_in_files, "character")
  expect_gt(length(gh_repos_urls_with_code_in_files), 0)
  test_mocker$cache(gh_repos_urls_with_code_in_files)
})

test_that("get_repos_urls returns repositories URLS", {
  mockery::stub(
    github_testhost$get_repos_urls,
    "private$get_repos_urls_with_code",
    test_mocker$use("gh_repos_urls_with_code_in_files")
  )
  gh_repos_urls_with_code_in_files <- github_testhost$get_repos_urls(
    type = "web",
    with_file = "DESCRIPTION",
    verbose = FALSE,
    progress = FALSE
  )
  expect_type(gh_repos_urls_with_code_in_files, "character")
  expect_gt(length(gh_repos_urls_with_code_in_files), 0)
  test_mocker$cache(gh_repos_urls_with_code_in_files)
})

test_that("get_repos_urls returns repositories URLS", {
  mockery::stub(
    github_testhost$get_repos_urls,
    "private$get_all_repos_urls",
    test_mocker$use("gh_repos_urls")
  )
  gh_repos_urls <- github_testhost$get_repos_urls(
    type = "web",
    verbose = FALSE,
    progress = FALSE
  )
  expect_type(gh_repos_urls, "character")
  expect_gt(length(gh_repos_urls), 0)
  test_mocker$cache(gh_repos_urls)
})

test_that("get_repos_urls() works for org", {
  mockery::stub(
    test_rest_gitlab$get_repos_urls,
    "self$get_repos_from_org",
    test_fixtures$gitlab_repositories_rest_response
  )
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  gl_api_repos_urls <- test_rest_gitlab$get_repos_urls(
    type = "api",
    org = test_org,
    repos = NULL
  )
  expect_length(
    gl_api_repos_urls,
    3
  )
  test_mocker$cache(gl_api_repos_urls)
  test_user <- "test_user"
  attr(test_user, "type") <- "user"
  gl_web_repos_urls <- test_rest_gitlab$get_repos_urls(
    type = "web",
    org = test_user,
    repos = NULL
  )
  expect_length(
    gl_web_repos_urls,
    3
  )
  test_mocker$cache(gl_web_repos_urls)
})

test_that("get_repos_urls() works for individual repos", {
  mockery::stub(
    test_rest_gitlab$get_repos_urls,
    "self$get_repos_from_org",
    test_mocker$use("gitlab_rest_repos_from_org_raw")
  )
  test_group <- "test_group"
  attr(test_group, "type") <- "organization"
  gl_api_repos_urls <- test_rest_gitlab$get_repos_urls(
    type = "api",
    org = test_group,
    repos = c("testRepo1", "testRepo2")
  )
  expect_length(
    gl_api_repos_urls,
    2
  )
  test_mocker$cache(gl_api_repos_urls)
  test_user <- "test_user"
  attr(test_user, "type") <- "user"
  gl_web_repos_urls <- test_rest_gitlab$get_repos_urls(
    type = "web",
    org = test_user,
    repos = c("testRepo1", "testRepo2")
  )
  expect_length(
    gl_web_repos_urls,
    2
  )
  test_mocker$cache(gl_web_repos_urls)
})

test_that("get_repos_urls_from_orgs prepares api repo_urls vector", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_urls_from_orgs,
    "rest_engine$get_repos_urls",
    test_mocker$use("gl_api_repos_urls")
  )
  gitlab_testhost_priv$searching_scope <- "org"
  gitlab_testhost_priv$orgs <- "test_org"
  gl_api_repos_urls <- gitlab_testhost_priv$get_repos_urls_from_orgs(
    type = "api",
    verbose = FALSE,
    progress = FALSE
  )
  expect_gt(length(gl_api_repos_urls), 0)
  expect_true(all(grepl("api", gl_api_repos_urls)))
  test_mocker$cache(gl_api_repos_urls)
})

test_that("get_repos_urls_from_repos prepares api repo_urls vector", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_urls_from_repos,
    "rest_engine$get_repos_urls",
    test_mocker$use("gl_web_repos_urls")
  )
  gitlab_testhost_priv$searching_scope <- c("repo")
  gitlab_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
  gl_web_repos_urls <- gitlab_testhost_priv$get_repos_urls_from_repos(
    type = "web",
    verbose = FALSE,
    progress = FALSE
  )
  expect_gt(length(gl_web_repos_urls), 0)
  test_mocker$cache(gl_web_repos_urls)
})


test_that("get_all_repos_urls prepares api repo_urls vector", {
  mockery::stub(
    gitlab_testhost_priv$get_all_repos_urls,
    "private$get_repos_urls_from_orgs",
    test_mocker$use("gl_api_repos_urls")
  )
  mockery::stub(
    gitlab_testhost_priv$get_all_repos_urls,
    "private$get_repos_urls_from_repos",
    NULL
  )
  gl_api_repos_urls <- gitlab_testhost_priv$get_all_repos_urls(
    type = "api",
    verbose = FALSE
  )
  expect_true(all(grepl("api", gl_api_repos_urls)))
  expect_gt(length(gl_api_repos_urls), 0)
  test_mocker$cache(gl_api_repos_urls)
})

test_that("`get_repo_url_from_response()` works", {
  gl_repo_web_urls <- gitlab_testhost_priv$get_repo_url_from_response(
    search_response = test_mocker$use("gl_repos_raw_output"),
    type = "web",
    progress = FALSE
  )
  expect_gt(length(gl_repo_web_urls), 0)
  expect_type(gl_repo_web_urls, "character")
  test_mocker$cache(gl_repo_web_urls)
})

test_that("get_repos_urls_with_code_from_orgs returns repositories URLS", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_urls_with_code_from_orgs,
    "private$get_repo_url_from_response",
    test_mocker$use("gl_repo_web_urls")
  )
  gl_repos_urls_with_code_from_orgs <- gitlab_testhost_priv$get_repos_urls_with_code_from_orgs(
    type = "web",
    code = "shiny",
    in_files = "DESCRIPTION",
    in_path = FALSE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_type(gl_repos_urls_with_code_from_orgs, "character")
  expect_gt(length(gl_repos_urls_with_code_from_orgs), 0)
  test_mocker$cache(gl_repos_urls_with_code_from_orgs)
})

test_that("get_repos_urls_with_code_from_repos returns repositories URLS", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_urls_with_code_from_repos,
    "private$get_repo_url_from_response",
    test_mocker$use("gl_repo_web_urls")
  )
  gl_repos_urls_with_code_from_repos <- gitlab_testhost_priv$get_repos_urls_with_code_from_repos(
    type = "web",
    code = "shiny",
    in_files = "DESCRIPTION",
    in_path = FALSE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_type(gl_repos_urls_with_code_from_repos, "character")
  expect_gt(length(gl_repos_urls_with_code_from_repos), 0)
  test_mocker$cache(gl_repos_urls_with_code_from_repos)
})

test_that("get_repos_urls_with_code_from_repos returns repositories URLS", {
  mockery::stub(
    gitlab_testhost_priv$get_repos_urls_with_code,
    "private$get_repos_urls_with_code_from_orgs",
    test_mocker$use("gl_repos_urls_with_code_from_orgs")
  )
  mockery::stub(
    gitlab_testhost_priv$get_repos_urls_with_code,
    "private$get_repos_urls_with_code_from_repos",
    test_mocker$use("gl_repos_urls_with_code_from_repos")
  )
  gl_repos_urls_with_code_in_files <- gitlab_testhost_priv$get_repos_urls_with_code(
    type = "web",
    code = "shiny",
    in_files = "DESCRIPTION",
    in_path = FALSE,
    verbose = FALSE,
    progress = FALSE
  )
  expect_type(gl_repos_urls_with_code_in_files, "character")
  expect_gt(length(gl_repos_urls_with_code_in_files), 0)
  test_mocker$cache(gl_repos_urls_with_code_in_files)
})

test_that("issues_by_repo GitHub query is built properly", {
  gh_issues_from_repo_query <-
    test_gqlquery_gh$issues_from_repo()
  expect_snapshot(
    gh_issues_from_repo_query
  )
})

test_that("issues page is pulled from repository", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_github_priv$get_issues_page_from_repo,
      "self$gql_response",
      test_fixtures$github_graphql_issues_response
    )
    org <- "test_org"
    repo <- "TestRepo"
  } else {
    org <- "r-world-devs"
    repo <- "GitStats"
  }
  issues_page <- test_graphql_github_priv$get_issues_page_from_repo(
    org = org,
    repo = repo
  )
  expect_issues_github_page(issues_page)
  test_mocker$cache(issues_page)
})

test_that("issues page works when encounters error", {
  response_error <- ""
  class(response_error) <- "graphql_error"
  mockery::stub(
    test_graphql_github_priv$get_issues_page_from_repo,
    "self$gql_response",
    response_error
  )
  org <- "test_org"
  repo <- "TestRepo"
  expect_no_error(
    issues_page <- test_graphql_github_priv$get_issues_page_from_repo(
      org = org,
      repo = repo
    )
  )
})


test_that("issues page with cursor is pulled from repository", {
  if (!integration_tests_skipped) {
    issues_page <- test_graphql_github_priv$get_issues_page_from_repo(
      org = "r-world-devs",
      repo = "GitStats",
      issues_cursor = "Y3Vyc29yOnYyOpK5MjAyNC0xMS0wN1QxMzowNTowOSswMTowMM6dZ-dm"
    )
    expect_issues_github_page(issues_page)
  }
})

test_that("`get_issues_from_one_repo()` prepares formatted list", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_github_priv$get_issues_from_one_repo,
      "private$get_issues_page_from_repo",
      test_mocker$use("issues_page")
    )
  }
  issues_from_repo <- test_graphql_github_priv$get_issues_from_one_repo(
    org = "r-world-devs",
    repo = "GitStats"
  )
  expect_issues_full_list(
    issues_from_repo
  )
  test_mocker$cache(issues_from_repo)
})

test_that("`get_issues_from_repos()` pulls issues from repos", {
  if (integration_tests_skipped) {
    mockery::stub(
      test_graphql_github$get_issues_from_repos,
      "private$get_issues_from_one_repo",
      test_mocker$use("issues_from_repo")
    )
  }
  issues_from_repos <- test_graphql_github$get_issues_from_repos(
    org = "r-world-devs",
    repo = c("GitStats", "GitAI")
  )
  expect_issues_full_list(
    issues_from_repos[[1]]
  )
  test_mocker$cache(issues_from_repos)
})

test_that("`prepare_issues_table()` prepares issues table", {
  gh_issues_table <- test_graphql_github$prepare_issues_table(
    repos_list_with_issues = test_mocker$use("issues_from_repos"),
    org = "r-world-devs"
  )
  expect_issues_table(
    gh_issues_table
  )
  test_mocker$cache(gh_issues_table)
})

test_that("get_issues_from_orgs for GitHub works", {
  if (integration_tests_skipped) {
    mockery::stub(
      github_testhost_priv$get_issues_from_orgs,
      "graphql_engine$prepare_issues_table",
      test_mocker$use("gh_issues_table")
    )
    mockery::stub(
      github_testhost_priv$get_issues_from_orgs,
      "private$get_repos_names",
      test_mocker$use("gh_repos_names")
    )
  }
  github_testhost_priv$searching_scope <- "org"
  gh_issues_from_orgs <- github_testhost_priv$get_issues_from_orgs(
    verbose = FALSE,
    progress = FALSE
  )
  expect_issues_table(
    gh_issues_from_orgs
  )
  test_mocker$cache(gh_issues_from_orgs)
})

test_that("get_issues_from_orgs for GitHub prints messages", {
  mockery::stub(
    github_testhost_priv$get_issues_from_orgs,
    "graphql_engine$prepare_issues_table",
    test_mocker$use("gh_issues_table")
  )
  mockery::stub(
    github_testhost_priv$get_issues_from_orgs,
    "private$get_repos_data",
    test_mocker$use("gh_repos_data")
  )
  github_testhost_priv$searching_scope <- "org"
  expect_snapshot(
    gh_issues_from_orgs <- github_testhost_priv$get_issues_from_orgs(
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("get_issues_from_repos for GitHub works", {
  if (integration_tests_skipped) {
    mockery::stub(
      github_testhost_priv$get_issues_from_repos,
      "graphql_engine$prepare_issues_table",
      test_mocker$use("gh_issues_table")
    )
    github_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
    test_org <- "test_org"
    attr(test_org, "type") <- "organization"
    mockery::stub(
      github_testhost_priv$get_issues_from_repos,
      "graphql_engine$set_owner_type",
      test_org
    )
  } else {
    github_testhost_priv$orgs_repos <- list("r-world-devs" = "GitStats")
  }
  github_testhost_priv$searching_scope <- "repo"
  gh_issues_from_repos <- github_testhost_priv$get_issues_from_repos(
    verbose = FALSE,
    progress = FALSE
  )
  expect_issues_table(
    gh_issues_from_repos
  )
  test_mocker$cache(gh_issues_from_repos)
})

test_that("get_issues_from_repos for GitHub prints messages", {
  mockery::stub(
    github_testhost_priv$get_issues_from_repos,
    "graphql_engine$prepare_issues_table",
    test_mocker$use("gh_issues_table")
  )
  github_testhost_priv$orgs_repos <- list("test_org" = "TestRepo")
  test_org <- "test_org"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    github_testhost_priv$get_issues_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  github_testhost_priv$searching_scope <- "repo"
  expect_snapshot(
    gh_issues_from_repos <- github_testhost_priv$get_issues_from_repos(
      verbose = TRUE,
      progress = FALSE
    )
  )
})

test_that("`get_issues()` retrieves issues in the table format in a certain time span", {
  mockery::stub(
    github_testhost$get_issues,
    "private$get_issues_from_orgs",
    test_mocker$use("gh_issues_from_orgs")
  )
  mockery::stub(
    github_testhost$get_issues,
    "private$get_issues_from_repos",
    test_mocker$use("gh_issues_from_repos")
  )
  gh_issues_table <- github_testhost$get_issues(
    since = "2023-01-01",
    until = "2025-03-06",
    verbose = FALSE,
    progress = FALSE
  )
  expect_issues_table(
    gh_issues_table
  )
  gh_issues_table_shorter <- github_testhost$get_issues(
    since = "2023-02-01",
    until = "2024-01-01",
    verbose = FALSE,
    progress = FALSE
  )
  expect_true(
   nrow(gh_issues_table) > nrow(gh_issues_table_shorter)
  )
  expect_true(
    max(gh_issues_table_shorter$created_at) <= "2024-01-01"
  )
  expect_true(
    min(gh_issues_table_shorter$created_at) >= "2023-02-01"
  )
  test_mocker$cache(gh_issues_table)
})

test_that("`get_issues()` retrieves open issues in the table format in a certain time span", {
  mockery::stub(
    github_testhost$get_issues,
    "private$get_issues_from_orgs",
    test_mocker$use("gh_issues_from_orgs")
  )
  mockery::stub(
    github_testhost$get_issues,
    "private$get_issues_from_repos",
    test_mocker$use("gh_issues_from_repos")
  )
  gh_open_issues_table <- github_testhost$get_issues(
    since = "2023-01-01",
    until = "2024-01-01",
    state = "open",
    verbose = FALSE,
    progress = FALSE
  )
  expect_issues_table(
    gh_open_issues_table
  )
  expect_true(
    all(gh_open_issues_table$state == "open")
  )
  gh_closed_issues_table <- github_testhost$get_issues(
    since = "2023-01-01",
    until = "2024-01-01",
    state = "closed",
    verbose = FALSE,
    progress = FALSE
  )
  expect_issues_table(
    gh_closed_issues_table
  )
  expect_true(
    all(gh_closed_issues_table$state == "closed")
  )
})

test_fixtures <- list()

test_fixtures$empty_gql_response <- list(
  "data" = list(
    "group" = list(
      "projects" = list(
        "count" = 0,
        "pageInfo" = list(
          "hasNextPage" = FALSE,
          "endCursor" = NULL
        ),
        "edges" = list()
      )
    )
  )
)

test_fixtures$half_empty_gql_response <- list(
  "data" = list(
    "group" = list(
      "projects" = list(
        "count" = 445,
        "pageInfo" = list(
          "hasNextPage" = TRUE,
          "endCursor" = NULL
        ),
        "edges" = list()
      )
    )
  )
)

GitHostGitHub <- R6::R6Class(
  classname = "GitHostGitHub",
  inherit = GitHost,
  public = list(
    initialize = function(orgs = NA,
                          repos = NA,
                          token = NA,
                          host = NA,
                          verbose = NA,
                          .error = TRUE) {
      super$initialize(orgs = orgs,
                       repos = repos,
                       token = token,
                       host = host,
                       verbose = verbose,
                       .error = .error)
      if (verbose) {
        cli::cli_alert_success("Set connection to GitHub.")
      }
    }
  ),
  private = list(

    host_name = "GitHub",
    api_version = 3,
    token_name = "GITHUB_PAT",
    min_access_scopes = c("public_repo", "read:org", "read:user"),

    access_scopes = list(
      org = c("read:org", "admin:org"),
      repo = c("public_repo", "repo"),
      user = c("read:user", "user")
    ),

    engine_methods = list(
      "graphql" = list(
        "repos",
        "commits",
        "release_logs"
      ),
      "rest" = list(
        "code",
        "contributors"
      )
    ),

    set_api_url = function(host) {
      if (is.null(host) ||
            host == "https://github.com" ||
            host == "http://github.com" ||
            host == "github.com") {
        private$api_url <- "https://api.github.com"
      } else {
        private$set_custom_api_url(host)
      }
    },

    set_web_url = function(host) {
      if (is.null(host)) {
        private$web_url <- "https://github.com"
      } else {
        private$set_custom_web_url(host)
      }
    },

    check_if_public = function(host) {
      private$is_public <- is.null(host) || grepl("github.com", host)
    },

    set_test_endpoint = function() {
      private$test_endpoint <- private$api_url
    },

    set_tokens_endpoint = function() {
      private$endpoints$tokens <- NULL
    },

    set_orgs_endpoint = function() {
      private$endpoints$orgs <- glue::glue("{private$api_url}/orgs")
      private$endpoints$users <- glue::glue("{private$api_url}/users")
    },

    set_repositories_endpoint = function() {
      private$endpoints$repositories <- glue::glue("{private$api_url}/repos")
    },

    setup_engines = function() {
      private$engines$rest <- EngineRestGitHub$new(
        rest_api_url = private$api_url,
        token = private$token,
        scan_all = private$scan_all
      )
      private$engines$graphql <- EngineGraphQLGitHub$new(
        gql_api_url = private$graphql_api_url,
        token = private$token,
        scan_all = private$scan_all
      )
    },

    check_token_scopes = function(response, token = NULL) {
      private$token_scopes <- response$headers$`x-oauth-scopes` |>
        stringr::str_split(", ") |>
        unlist()
      org_scopes <- any(private$access_scopes$org %in% private$token_scopes)
      repo_scopes <- any(private$access_scopes$repo %in% private$token_scopes)
      user_scopes <- any(private$access_scopes$user %in% private$token_scopes)
      all(c(org_scopes, repo_scopes, user_scopes))
    },

    add_repo_api_url = function(repos_table) {
      if (!is.null(repos_table) && nrow(repos_table) > 0) {
        repos_table <- dplyr::mutate(
          repos_table,
          api_url = paste0(private$endpoints$repositories, "/", organization, "/", repo_name),
        )
      }
      return(repos_table)
    },

    get_orgs_from_host = function(output, verbose) {
      if (private$is_public) {
        cli::cli_abort("This feature is not applicable for public hosts.",
                       call = NULL)
      }
      graphql_engine <- private$engines$graphql
      orgs <- graphql_engine$get_orgs(
        output = output,
        verbose = verbose
      )
      if (output == "full_table") {
        orgs <- orgs |>
          graphql_engine$prepare_orgs_table()
        private$orgs <- dplyr::pull(orgs, path)
      } else {
        private$orgs <- orgs
      }
      return(orgs)
    },

    get_repos_ids = function(search_response) {
      purrr::map_vec(search_response, ~.$repository$node_id) |> unique()
    },

    get_repo_url_from_response = function(search_response,
                                          repos_fullnames = NULL,
                                          type,
                                          verbose = TRUE,
                                          progress = TRUE) {
      if (!is.null(repos_fullnames)) {
        search_response <- search_response |>
          purrr::keep(~ paste0(.$organization$login, "/", .$repo_path) %in% repos_fullnames)
      }
      purrr::map_vec(search_response, function(project) {
        if (type == "api") {
          paste0(private$api_url, "/repos/", project$organization$login, "/", project$repo_name)
        } else {
          project$repo_url
        }
      })
    },

    get_commits_from_orgs = function(since,
                                     until,
                                     verbose,
                                     progress) {
      if ("org" %in% private$searching_scope) {
        graphql_engine <- private$engines$graphql
        commits_table <- gitstats_map(private$orgs, function(org) {
          commits_table_org <- NULL
          repos_data <- private$get_repos_data(
            org = org,
            verbose = verbose
          )
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = org,
              information = paste0("Pulling commits ", cli_icons$commit)
            )
          }
          commits_table_org <- graphql_engine$get_commits_from_repos(
            org = org,
            repos_names = repos_data[["paths"]],
            since = since,
            until = until
          ) |>
            graphql_engine$prepare_commits_table(
              org = org
            )
          return(commits_table_org)
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind()
        return(commits_table)
      }
    },

    get_commits_from_repos = function(since,
                                      until,
                                      verbose,
                                      progress) {
      if ("repo" %in% private$searching_scope) {
        graphql_engine <- private$engines$graphql
        orgs <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        )
        commits_table <- gitstats_map(orgs, function(org) {
          commits_table_org <- NULL
          if (!private$scan_all && verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = set_repo_scope(org, private),
              information = paste0("Pulling commits ", cli_icons$commit)
            )
          }
          commits_table_org <- graphql_engine$get_commits_from_repos(
            org = org,
            repos_names = private$orgs_repos[[org]],
            since = since,
            until = until
          ) |>
            graphql_engine$prepare_commits_table(
              org = org
            )
          return(commits_table_org)
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind()
        return(commits_table)
      }
    },

    get_files_content_from_repos = function(file_path,
                                            verbose = TRUE,
                                            progress = TRUE) {
      if ("repo" %in% private$searching_scope) {
        graphql_engine <- private$engines$graphql
        orgs <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        )
        files_table <- gitstats_map(orgs, function(org) {
          owner_type <- attr(org, "type") %||% "organization"
          repos_data <- private$get_repos_data(
            org = org,
            repos = private$orgs_repos[[org]],
            verbose = verbose
          )
          if (verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = set_repo_scope(org, private),
              information = glue::glue("Pulling files {cli_icons$file} content: [{paste0(file_path, collapse = ', ')}]")
            )
          }
          graphql_engine$get_files_from_org(
            org = org,
            owner_type = owner_type,
            repos_data = repos_data,
            file_paths = file_path,
            verbose = verbose
          ) |>
            graphql_engine$prepare_files_table(
              org = org
            )
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind() |>
          private$add_repo_api_url()
        return(files_table)
      }
    },

    get_files_content_from_files_structure = function(files_structure,
                                                      verbose,
                                                      progress) {
      if (length(files_structure) > 0) {
        graphql_engine <- private$engines$graphql
        orgs <- names(files_structure)
        files_table <- gitstats_map(orgs, function(org) {
          owner_type <- attr(org, "type") %||% "organization"
          repos <- files_structure[[org]] |> names()
          repos_data <- private$get_repos_data(
            org = org,
            repos = repos,
            verbose = verbose
          )
          if (verbose) {
            show_message(
              host = private$host_name,
              engine = "graphql",
              scope = org,
              information = paste0("Pulling files ", cli_icons$file, " from files structure")
            )
          }
          graphql_engine$get_files_from_org(
            org = org,
            owner_type = owner_type,
            repos_data = repos_data,
            host_files_structure = files_structure,
            verbose = verbose
          ) |>
            graphql_engine$prepare_files_table(
              org = org
            )
        }, .progress = set_progress_bar(progress, private)) |>
          purrr::list_rbind() |>
          private$add_repo_api_url()
        return(files_table)
      } else {
        cli::cli_alert_warning("[GitHub] No files {cli_icons$file} found. Skipping pulling files content.")
        return(NULL)
      }
    },

    get_files_structure_from_repos = function(pattern,
                                              depth,
                                              verbose  = TRUE,
                                              progress = TRUE) {
      if ("repo" %in% private$searching_scope) {
        rest_engine <- private$engines$rest
        graphql_engine <- private$engines$graphql
        orgs <- graphql_engine$set_owner_type(
          owners = names(private$orgs_repos),
          verbose = verbose
        )
        files_structure_list <- gitstats_map(orgs, function(org) {
          repos_data <- private$get_repos_data(
            org = org,
            repos = private$orgs_repos[[org]],
            verbose = verbose
          )
          if (verbose) {
            user_info <- if (!is.null(pattern)) {
              glue::glue(
                "Pulling repos {cli_icons$tree} [files matching pattern: '{paste0(pattern, collapse = '|')}']"
              )
            } else {
              glue::glue("Pulling repos {cli_icons$tree}")
            }
            show_message(
              host = private$host_name,
              engine = "rest",
              scope = set_repo_scope(org, private),
              information = user_info
            )
          }
          private$get_files_structure_from_repos_data(
            rest_engine = rest_engine,
            org = org,
            repos_data = repos_data,
            pattern = pattern,
            depth = depth,
            verbose = verbose
          )
        }, .progress = set_progress_bar(progress, private))
        names(files_structure_list) <- orgs
        files_structure_list <- files_structure_list |>
          purrr::discard(~ length(.) == 0)
        if (length(files_structure_list) == 0 && verbose) {
          cli::cli_alert_warning(
            cli::col_yellow(
              "For {private$host_name} no files {cli_icons$tree} structure found."
            )
          )
        }
        return(files_structure_list)
      }
    },

    get_repos_data = function(org, repos = NULL, verbose) {
      cached_repos <- private$get_cached_repos(org)
      if (is.null(cached_repos)) {
        if (verbose) cli::cli_alert("[{org}] Pulling repositories {cli_icons$repo} data...")
        owner_type <- attr(org, "type") %||% "organization"
        org <- url_decode(org)
        graphql_engine <- private$engines$graphql
        repos_from_org <- graphql_engine$get_repos_from_org(
          org = org,
          owner_type = owner_type,
          verbose = verbose
        )
        private$set_cached_repos(repos_from_org, org, verbose)
        tryCatch(
          private$save_repos_to_storage(
            repos_from_org, org, graphql_engine
          ),
          error = function(e) NULL
        )
      } else {
        if (verbose) cli::cli_alert("[{org}] Using cached repositories data...")
        repos_from_org <- cached_repos
      }
      if (!is.null(repos)) {
        repos_from_org <- repos_from_org |>
          purrr::keep(~ .$repo_path %in% repos)
      }
      repos_data <- list(
        "paths" = purrr::map(repos_from_org, ~ .$repo_path),
        "def_branches" = purrr::map(repos_from_org, ~ .$default_branch$name),
        "repo_ids" = purrr::map_chr(repos_from_org, ~ .$repo_id)
      )
      return(repos_data)
    },

    save_repos_to_storage = function(repos_from_org, org, engine) {
      storage <- private$storage_backend
      if (!is.null(storage) && storage$is_db()) {
        repos_table <- engine$prepare_repos_table(repos_from_org, org)
        if (!is.null(repos_table) && nrow(repos_table) > 0) {
          repos_table <- dplyr::as_tibble(repos_table)
          existing <- storage$load("repositories")
          if (!is.null(existing)) {
            repos_table <- dplyr::bind_rows(existing, repos_table) |>
              dplyr::distinct(repo_id, .keep_all = TRUE)
          }
          class(repos_table) <- c("gitstats_repos", class(repos_table))
          storage$save("repositories", repos_table)
        }
      }
    },

    set_repo_url = function(repo_fullname) {
      paste0(private$endpoints$repositories, "/", repo_fullname)
    }
  )
)

EngineGraphQLGitLab <- R6::R6Class(
  classname = "EngineGraphQLGitLab",
  inherit = EngineGraphQL,
  public = list(
    initialize = function(gql_api_url,
                          token,
                          scan_all = FALSE) {
      super$initialize(
        gql_api_url = gql_api_url,
        token = token,
        scan_all = scan_all
      )
      self$gql_query <- GQLQueryGitLab$new()
    },

    set_owner_type = function(owners, verbose = TRUE) {
      user_or_org_query <- self$gql_query$user_or_org_query
      login_types <- gitstats_map(owners, function(owner) {
        cached <- private[["owner_types_cache"]][[owner]]
        if (!is.null(cached)) {
          return(cached)
        }
        response <- self$gql_response(
          gql_query = user_or_org_query,
          vars = list(
            "username" = owner,
            "grouppath" = owner
          ),
          verbose = verbose
        )
        if (!all(purrr::map_lgl(response$data, is.null))) {
          type <- purrr::discard(response$data, is.null) |>
            names()
          if (type == "group") {
            type <- "organization"
          }
          attr(owner, "type") <- type
        } else {
          attr(owner, "type") <- "not found"
        }
        private[["owner_types_cache"]][[owner]] <- owner
        return(owner)
      })
      return(login_types)
    },

    get_orgs = function(orgs_count,
                        output = c("only_names", "full_table"),
                        verbose,
                        progress = verbose) {
      if (verbose) {
        cli::cli_alert("[Host:GitLab][Engine:{cli::col_yellow('GraphQL')}] Pulling organizations {cli_icons$org}...")
      }
      group_cursor <- ""
      iterations_number <- round(orgs_count / 100)
      full_orgs_list <- list()
      for (x in 1:iterations_number) {
        response <- self$gql_response(
          gql_query = self$gql_query$groups(),
          vars = list("groupCursor" = group_cursor),
          verbose = verbose
        )
        if (!inherits(response, "graphql_error")) {
          if (output == "only_names") {
            orgs_list <- purrr::map(response$data$groups$edges, ~ .$node$fullPath)
          } else {
            orgs_list <- purrr::map(response$data$groups$edges, ~ .$node)
          }
          group_cursor <- response$data$groups$pageInfo$endCursor
          full_orgs_list <- append(full_orgs_list, orgs_list)
        } else {
          full_orgs_list <- response
          break
        }
      }
      full_orgs_list <- private$handle_graphql_error(full_orgs_list, verbose)
      if (!inherits(full_orgs_list, "graphql_error")) {
        if (output == "only_names") {
          all_orgs <- unlist(full_orgs_list)
        } else if (output == "full_table") {
          all_orgs <- full_orgs_list
        }
      } else {
        all_orgs <- full_orgs_list
      }
      return(all_orgs)
    },

    get_org = function(org, verbose) {
      if (verbose) {
        cli::cli_alert("[Host:GitLab][Engine:{cli::col_yellow('GraphQL')}] Pulling {org} organization {cli_icons$org}...")
      }
      response <- self$gql_response(
        gql_query = self$gql_query$group(),
        vars = list("org" = org),
        verbose = verbose
      )
      if (length(response$data$group) == 0) {
        class(response) <- c(class(response), "graphql_error")
        return(response)
      } else {
        return(response$data$group)
      }
    },

    prepare_orgs_table = function(full_orgs_list) {
      orgs_table <- purrr::map(full_orgs_list, function(org_node) {
        org_node$avatarUrl <- org_node$avatarUrl %||% ""
        data.frame(org_node)
      }) |>
        purrr::list_rbind() |>
        dplyr::rename(path = fullPath,
                      url = webUrl,
                      repos_count = projectsCount,
                      members_count = groupMembersCount,
                      avatar_url = avatarUrl) |>
        dplyr::relocate(avatar_url, .before = repos_count) |>
        tibble::as_tibble()
      return(orgs_table)
    },

    get_repos = function(repos_ids, verbose) {
      full_repos_list <- list()
      next_page <- TRUE
      repo_cursor <- ""
      while (next_page) {
        repos_response <- private$get_repos_page(
          projects_ids = paste0("gid://gitlab/Project/", repos_ids),
          type = "projects",
          repo_cursor = repo_cursor,
          verbose = verbose
        )
        if (inherits(repos_response, "graphql_error")) {
          if (inherits(repos_response, "graphql_no_fields_error")) {
            full_repos_list <- repos_response
            break
          }
          if (inherits(repos_response, "graphql_limit_error")) {
            if (verbose) {
              cli::cli_alert_warning("GraphQL limit error when pulling repos by IDs ({length(repos_ids)} repos).")
              cli::cli_alert_info(
                cli::col_br_cyan("I will pull repos in batches.")
              )
            }
            full_repos_list <- private$get_repos_in_batches(
              repos_ids = repos_ids,
              verbose = verbose
            )
            return(full_repos_list)
          }
          repos_response <- private$get_repos_page(
            projects_ids = paste0("gid://gitlab/Project/", repos_ids),
            type = "projects",
            repo_cursor = repo_cursor,
            verbose = verbose
          )
        }
        core_response <- repos_response$data$projects
        repos_list <- core_response$edges
        next_page <- core_response$pageInfo$hasNextPage
        if (is.null(next_page)) next_page <- FALSE
        if (is.null(repos_list)) repos_list <- list()
        if (length(repos_list) == 0) next_page <- FALSE
        if (next_page) {
          repo_cursor <- core_response$pageInfo$endCursor
        } else {
          repo_cursor <- ""
        }
        full_repos_list <- append(full_repos_list, repos_list)
      }
      full_repos_list <- private$handle_graphql_error(full_repos_list, verbose)
      return(full_repos_list)
    },

    get_repos_by_fullpath = function(full_paths, verbose = TRUE) {
      repos_list <- purrr::map(full_paths, function(full_path) {
        response <- self$gql_response(
          gql_query = self$gql_query$repo_by_fullpath(),
          vars = list("fullPath" = full_path),
          verbose = verbose
        )
        if (inherits(response, "graphql_error") ||
              is.null(response$data$project)) {
          return(NULL)
        }
        # Wrap in node structure to match prepare_repos_table() expectations
        list(node = response$data$project)
      }) |>
        purrr::compact()
      return(repos_list)
    },

    get_repos_from_org = function(org  = NULL,
                                  owner_type = c("organization", "user"),
                                  verbose = TRUE) {
      full_repos_list <- list()
      next_page <- TRUE
      repo_cursor <- ""
      while (next_page) {
        repos_response <- private$get_repos_page(
          org = org,
          type = owner_type,
          repo_cursor = repo_cursor,
          verbose = verbose
        )
        if (inherits(repos_response, "graphql_error")) {
          full_repos_list <- repos_response
          break
        } else {
          core_response <- if (owner_type == "organization") {
            repos_response$data$group$projects
          } else {
            repos_response$data$projects
          }
          repos_list <- core_response$edges
          next_page <- core_response$pageInfo$hasNextPage
          if (is.null(next_page)) next_page <- FALSE
          if (is.null(repos_list)) repos_list <- list()
          if (length(repos_list) == 0) next_page <- FALSE
          if (next_page) {
            repo_cursor <- core_response$pageInfo$endCursor
          } else {
            repo_cursor <- ""
          }
          full_repos_list <- append(full_repos_list, repos_list)
        }
      }
      full_repos_list <- private$handle_graphql_error(full_repos_list, verbose)
      return(full_repos_list)
    },

    prepare_repos_table = function(repos_list, org) {
      if (length(repos_list) > 0) {
        repos_table <- purrr::map(repos_list, function(repo) {
          repo <- repo$node
          languages <- if (length(repo$languages) > 0) {
            purrr::map_chr(repo$languages, ~ .$name) |>
              paste0(collapse = ", ")
          } else {
            ""
          }
          if (!is.null(repo$namespace)) {
            org <- repo$namespace$path
          }
          if (is.null(org)) {
            org <- sub(paste0("/", repo$repo_path), "", repo$repo_url) |>
              stringr::str_replace_all("^https://[^/]+", "") |>
              stringr::str_replace_all("^/", "")
          }
          data.frame(
            repo_id = get_gitlab_repo_id(repo$repo_id),
            repo_name = repo$repo_path,
            repo_fullpath = repo$repo_fullpath,
            default_branch = repo$repository$rootRef %||% "",
            stars = repo$stars,
            forks = repo$forks,
            created_at = gts_to_posixt(repo$created_at),
            last_activity_at = as.POSIXct(repo$last_activity_at),
            languages = languages,
            issues_open = repo$issues$opened %||% 0,
            issues_closed = repo$issues$closed %||% 0,
            organization = org,
            repo_url = repo$repo_url,
            commit_sha = repo$repository$lastCommit$sha %||% NA_character_
          )
        }) |>
          purrr::list_rbind()
      } else {
        repos_table <- NULL
      }
      return(repos_table)
    },

    get_files_from_org = function(org,
                                  owner_type,
                                  repos_data,
                                  file_paths = NULL,
                                  host_files_structure = NULL,
                                  verbose = FALSE) {
      org <- url_decode(org)
      full_files_list <- list()
      next_page <- TRUE
      end_cursor <- ""
      if (!is.null(host_files_structure)) {
        file_paths <- private$get_path_from_files_structure(
          host_files_structure = host_files_structure,
          org = org
        )
      } else {
        file_paths <- file_paths[grepl(text_files_pattern, file_paths)]
      }
      if (owner_type == "organization") {
        while (next_page) {
          files_query <- self$gql_query$files_by_org(
            end_cursor = end_cursor
          )
          files_response <- self$gql_response(
            gql_query = files_query,
            vars = list(
              "org" = org,
              "file_paths" = file_paths
            ),
            verbose = verbose
          )
          if (inherits(files_response, "graphql_error")) {
            files_response <- list()
          }
          if (private$is_query_error(files_response)) {
            if (verbose) {
              purrr::walk(files_response$errors, ~ cli::cli_alert_warning(.))
            }
            if (private$is_complexity_error(files_response)) {
              if (verbose) {
                cli::cli_alert_info(
                  cli::col_br_cyan("I will switch to pulling files per repository.")
                )
              }
              full_files_list <- self$get_files_from_org_per_repo(
                org = org,
                owner_type = owner_type,
                repos_data = repos_data,
                file_paths = file_paths,
                host_files_structure = host_files_structure,
                verbose = verbose
              )
              return(full_files_list)
            }
          }
          if (length(files_response$data$group) == 0 && verbose) {
            cli::cli_alert_danger("Empty response.")
          }
          projects <- files_response$data$group$projects
          files_list <- purrr::map(projects$edges, function(edge) {
            edge$node
          }) |>
            purrr::discard(~ length(.$repository$blobs$nodes) == 0)
          if (is.null(files_list)) files_list <- list()
          if (length(files_list) > 0) {
            next_page <- projects$pageInfo$hasNextPage
          } else {
            next_page <- FALSE
          }
          if (is.null(next_page)) next_page <- FALSE
          if (next_page) {
            end_cursor <- projects$pageInfo$endCursor
          } else {
            end_cursor <- ""
          }
          full_files_list <- append(full_files_list, files_list)
        }
        if (!is.null(repos_data)) {
          full_files_list <- purrr::keep(full_files_list, function(project) {
            repo_name <- private$get_repo_name_from_url(project$webUrl)
            repo_name %in% repos_data$paths
          })
        }
      } else {
        full_files_list <- self$get_files_from_org_per_repo(
          org = org,
          owner_type = owner_type,
          repos_data = repos_data,
          file_paths = file_paths,
          host_files_structure = host_files_structure,
          verbose = verbose
        )
      }
      return(full_files_list)
    },

    get_files_from_org_per_repo = function(org,
                                           owner_type,
                                           repos_data,
                                           file_paths = NULL,
                                           host_files_structure = NULL,
                                           verbose = FALSE) {
      org_files_list <- gitstats_map(repos_data$paths, function(repo) {
        if (!is.null(host_files_structure)) {
          file_paths <- private$get_path_from_files_structure(
            host_files_structure = host_files_structure,
            org = org,
            repo = repo
          )
        }
        files_response <- private$get_file_blobs_response(
          org = org,
          repo = repo,
          file_paths = file_paths,
          verbose = verbose
        )
        if (private$is_complexity_error(files_response)) {
          if (verbose) {
            cli::cli_alert_warning("[{repo}] Encountered query complexity error. Too many files ({length(file_paths)})).")
            cli::cli_alert_info("I will run queries per 1 file.")
          }
          files_response <- private$get_file_blobs_response(
            org = org,
            repo = repo,
            file_paths = file_paths[1],
            verbose = verbose
          )
          nodes <- purrr::map(seq_along(file_paths), function(i) {
            files_part_response <- private$get_file_blobs_response(
              org = org,
              repo = repo,
              file_paths = file_paths[i],
              verbose = verbose
            )
            return(files_part_response$data$project$repository$blobs$nodes)
          }) |>
            purrr::list_flatten()
          files_response <- list(
            "data" = list(
              "project" = list(
                "name" = repo,
                "id" = files_response$data$project$id,
                "webUrl" = files_response$data$project$webUrl,
                "repository" = list(
                  "blobs" = list(
                    "nodes" = nodes
                  ),
                  "lastCommit" = list(
                    "sha" = files_response$data$project$repository$lastCommit$sha
                  )
                )
              )
            )
          )
        }
        return(files_response)
      })
      return(org_files_list)
    },

    prepare_files_table = function(files_response, org) {
      if (!is.null(files_response)) {
        if (private$response_prepared_by_iteration(files_response)) {
          files_table <- purrr::map(files_response, function(response_data) {
            private$prepare_files_table_row(
              project = response_data$data$project,
              org = org
            )
          }) |>
            purrr::list_rbind()
        } else {
          files_table <- purrr::map(files_response, function(project) {
            private$prepare_files_table_row(
              project = project,
              org = org
            )
          }) |>
            purrr::list_rbind()
        }
      } else {
        files_table <- NULL
      }
      return(files_table)
    },

    prepare_user_table = function(user_response) {
      if (!is.null(user_response$data$user)) {
        user_data <- user_response$data$user
        user_data[["name"]] <- user_data$name %||% ""
        user_data[["starred_repos"]] <- user_data$starred_repos$count
        user_data[["pull_requests"]] <- user_data$pull_requests$count
        user_data[["reviews"]] <- user_data$reviews$count
        user_data[["email"]] <- user_data$email %||% ""
        user_data[["location"]] <- user_data$location %||% ""
        user_data[["web_url"]] <- user_data$web_url %||% ""
        user_table <- tibble::as_tibble(user_data) |>
          dplyr::mutate(commits = NA,
                        issues = NA) |>
          dplyr::relocate(
            c(commits, issues),
            .after = starred_repos
          )
      } else {
        user_table <- NULL
      }
      return(user_table)
    },

    get_release_logs_from_org = function(repos_names, org, verbose = TRUE) {
      release_responses <- gitstats_map(repos_names, function(repository) {
        releases_from_repo_query <- self$gql_query$releases_from_repo()
        response <- self$gql_response(
          gql_query = releases_from_repo_query,
          vars = list(
            "project_path" = paste0(org, "/", url_decode(repository))
          ),
          verbose = verbose
        )
        return(response)
      }) |>
        purrr::discard(~ length(.$data$project$releases$nodes) == 0)
      return(release_responses)
    },

    prepare_releases_table = function(releases_response, org) {
      if (length(releases_response) > 0) {
        releases_table <-
          purrr::map(releases_response, function(release) {
            release_table <- purrr::map(release$data$project$releases$nodes, function(node) {
              data.frame(
                release_name = node$name,
                release_tag = node$tagName,
                published_at = gts_to_posixt(node$releasedAt),
                release_url = node$links$selfUrl,
                release_log = node$description
              )
            }) |>
              purrr::list_rbind() |>
              dplyr::mutate(
                repo_name = release$data$project$name,
                repo_url = release$data$project$webUrl
              ) |>
              dplyr::relocate(
                repo_name, repo_url,
                .before = release_name
              )
            return(release_table)
          }) |>
          purrr::list_rbind()
      } else {
        releases_table <- NULL
      }
      return(releases_table)
    }
  ),
  private = list(
    get_repos_page = function(org = NULL,
                              projects_ids = NULL,
                              type = "organization",
                              repo_cursor = "",
                              verbose = TRUE) {
      if (type == "organization") {
        response <- self$gql_response(
          gql_query = self$gql_query$repos_by_org(),
          vars = list(
            "org" = org,
            "repo_cursor" = repo_cursor
          ),
          verbose = verbose
        )
      } else if (type == "user") {
        response <- self$gql_response(
          gql_query = self$gql_query$repos_by_user(),
          vars = list(
            "username" = org,
            "repo_cursor" = repo_cursor
          ),
          verbose = verbose
        )
      } else if (type == "projects") {
        response <- self$gql_response(
          gql_query = self$gql_query$repos(repo_cursor),
          vars = list(
            "projects_ids" = as.character(projects_ids)
          ),
          verbose = verbose
        )
      }
      return(response)
    },

    get_repos_in_batches = function(repos_ids, batch_size = 50, verbose) {
      batches <- split(repos_ids, ceiling(seq_along(repos_ids) / batch_size))

      # Capture scalar values needed by workers — R6 envs don't serialize
      # to mirai workers. The actual request/error logic lives in the
      # standalone helpers from EngineGraphQL.R (graphql_response, etc.).
      gql_api_url <- self$gql_api_url
      token <- private$token

      # Pre-evaluate query templates so the closure doesn't need the
      # GQLQueryGitLab R6 object (which also has private fields).
      query_no_cursor <- self$gql_query$repos("")
      query_sample <- self$gql_query$repos("__CURSOR__")
      repos_query_fn <- function(repo_cursor) {
        if (nchar(repo_cursor) == 0) {
          query_no_cursor
        } else {
          gsub('after: "__CURSOR__"',
               paste0('after: "', repo_cursor, '"'),
               query_sample, fixed = TRUE)
        }
      }

      fetch_batch <- function(batch) {
        batch_repos <- list()
        next_page <- TRUE
        repo_cursor <- ""
        while (next_page) {
          repos_response <- graphql_response(
            gql_api_url = gql_api_url,
            token = token,
            gql_query = repos_query_fn(repo_cursor),
            vars = list("projects_ids" = paste0("gid://gitlab/Project/", batch))
          )
          if (inherits(repos_response, "graphql_error")) {
            if (inherits(repos_response, "graphql_no_fields_error")) {
              return(repos_response)
            }
            break
          }
          core_response <- repos_response$data$projects
          repos_list <- core_response$edges
          next_page <- core_response$pageInfo$hasNextPage
          if (is.null(next_page)) next_page <- FALSE
          if (is.null(repos_list)) repos_list <- list()
          if (length(repos_list) == 0) next_page <- FALSE
          if (next_page) {
            repo_cursor <- core_response$pageInfo$endCursor
          } else {
            repo_cursor <- ""
          }
          batch_repos <- append(batch_repos, repos_list)
        }
        batch_repos
      }

      batch_results <- gitstats_map(batches, fetch_batch)

      # Check if any batch returned a graphql error object
      for (result in batch_results) {
        if (inherits(result, "graphql_error")) {
          result <- private$handle_graphql_error(result, verbose)
          return(result)
        }
      }

      full_repos_list <- purrr::list_c(batch_results)
      full_repos_list <- private$handle_graphql_error(full_repos_list, verbose)
      return(full_repos_list)
    },

    get_repo_name_from_url = function(web_url) {
      url_split <- stringr::str_split(web_url, ":|/")[[1]]
      repo_name <- url_split[length(url_split)]
      return(repo_name)
    },

    get_file_blobs_response = function(org, repo, file_paths, verbose = TRUE) {
      file_blobs_response <- self$gql_response(
        gql_query = self$gql_query$file_blob_from_repo(),
        vars = list(
          "fullPath" = paste0(org, "/", repo),
          "file_paths" = file_paths
        ),
        verbose = verbose
      )
      return(file_blobs_response)
    },

    prepare_files_table_row = function(project, org) {
      purrr::map(project$repository$blobs$nodes, function(file) {
        if (!is.null(file)) {
          project_id <- project$id %||% NA_character_
          repo_url <- project$webUrl %||% NA_character_
          commit_sha <- project$repository$lastCommit$sha %||% NA_character_
          data.frame(
            "repo_id" = get_gitlab_repo_id(project_id),
            "repo_name" = project$path %||% project$name,
            "organization" = org,
            "file_path" = file$path,
            "file_content" = file$rawBlob,
            "file_size" = as.integer(file$size),
            "file_id" = file$oid,
            "repo_url" = repo_url,
            "commit_sha" = commit_sha
          )
        }
      }) |>
        purrr::list_rbind()
    },

    get_issues_from_one_repo = function(org,
                                        repo,
                                        verbose = TRUE) {
      next_page <- TRUE
      full_issues_list <- list()
      issues_cursor <- ""
      while (next_page) {
        issues_response <- private$get_issues_page_from_repo(
          org = org,
          repo = repo,
          issues_cursor = issues_cursor,
          verbose = verbose
        )
        issues_list <- issues_response$data$project$issues$edges
        next_page <- issues_response$data$project$issues$pageInfo$hasNextPage
        if (is.null(next_page)) next_page <- FALSE
        if (is.null(issues_list)) issues_list <- list()
        if (next_page) {
          issues_cursor <- issues_response$data$project$issues$pageInfo$endCursor
        } else {
          issues_cursor <- ""
        }
        full_issues_list <- append(full_issues_list, issues_list)
      }
      return(full_issues_list)
    },

    get_issues_page_from_repo = function(org,
                                         repo,
                                         issues_cursor = "",
                                         verbose = TRUE) {
      issues_from_repo_query <- self$gql_query$issues_from_repo(
        issues_cursor = issues_cursor
      )
      response <- self$gql_response(
        gql_query = issues_from_repo_query,
        vars = list(
          "fullPath" = paste0(org, "/", repo)
        ),
        verbose = verbose
      )
      return(response)
    },

    get_pr_from_one_repo = function(org,
                                    repo,
                                    verbose = TRUE) {
      next_page <- TRUE
      full_pr_list <- list()
      pr_cursor <- ""
      while (next_page) {
        pr_response <- private$get_pr_page_from_repo(
          org = org,
          repo = repo,
          pr_cursor = pr_cursor,
          verbose = verbose
        )
        pr_list <- pr_response$data$project$mergeRequests$edges
        next_page <- pr_response$data$project$mergeRequests$pageInfo$hasNextPage
        if (is.null(next_page)) next_page <- FALSE
        if (is.null(pr_list)) pr_list <- list()
        if (next_page) {
          pr_cursor <- pr_response$data$project$mergeRequests$pageInfo$endCursor
        } else {
          pr_cursor <- ""
        }
        full_pr_list <- append(full_pr_list, pr_list)
      }
      return(full_pr_list)
    },

    get_pr_page_from_repo = function(org,
                                     repo,
                                     pr_cursor = "",
                                     verbose = TRUE) {
      pr_from_repo_query <- self$gql_query$pull_requests_from_repo(
        pr_cursor = pr_cursor
      )
      response <- self$gql_response(
        gql_query = pr_from_repo_query,
        vars = list(
          "fullPath" = paste0(org, "/", repo)
        ),
        verbose = verbose
      )
      if (inherits(response, "graphql_error")) {
        response <- self$gql_response(
          gql_query = pr_from_repo_query,
          vars = list(
            "org" = org,
            "repo" = repo
          ),
          verbose = verbose
        )
      }
      return(response)
    },

    response_prepared_by_iteration = function(files_response) {
      !all(purrr::map_lgl(files_response, ~ all(c("name", "id", "webUrl", "repository") %in% names(.))))
    }
  )
)

test_that("file queries for GitLab are built properly", {
  gl_files_query <-
    test_gqlquery_gl$files_by_org()
  expect_snapshot(
    gl_files_query
  )
  gl_file_blobs_from_repo_query <-
    test_gqlquery_gl$file_blob_from_repo()
  expect_snapshot(
    gl_file_blobs_from_repo_query
  )
  test_mocker$cache(gl_file_blobs_from_repo_query)
})

test_that("get_file_blobs_response() works", {
  mockery::stub(
    test_graphql_gitlab_priv$get_file_blobs_response,
    "self$gql_response",
    test_fixtures$gitlab_file_repo_response
  )
  gl_file_blobs_response <- test_graphql_gitlab_priv$get_file_blobs_response(
    org = "mbtests",
    repo = "graphql_tests",
    file_path = c("project_metadata.yaml", "README.md")
  )
  expect_files_gitlab_blob_response(gl_file_blobs_response)
  test_mocker$cache(gl_file_blobs_response)
})

test_that("GitLab GraphQL Engine pulls files from a group", {
  mockery::stub(
    test_graphql_gitlab$get_files_from_org,
    "self$gql_response",
    test_fixtures$gitlab_file_org_response
  )
  gitlab_files_response <- test_graphql_gitlab$get_files_from_org(
    org = "mbtests",
    owner_type = "organization",
    repos = NULL,
    file_paths = "meta_data.yaml",
    host_files_structure = NULL
  )
  expect_files_gitlab_from_org_response(gitlab_files_response)
  test_mocker$cache(gitlab_files_response)
})

test_that("GitLab GraphQL Engine pulls files from org by iterating over repos", {
  mockery::stub(
    test_graphql_gitlab$get_files_from_org_per_repo,
    "private$get_file_blobs_response",
    test_mocker$use("gl_file_blobs_response")
  )
  gl_files_from_org_per_repo <- test_graphql_gitlab$get_files_from_org_per_repo(
    org = "test_org",
    owner_type = "organization",
    repos_data = list("paths" = "TestProject"),
    file_paths = c("project_metadata.yaml", "README.md")
  )
  expect_files_gitlab_from_org_by_repos_response(
    response = gl_files_from_org_per_repo,
    expected_files = c("project_metadata.yaml", "README.md")
  )
  test_mocker$cache(gl_files_from_org_per_repo)
})

test_that("is query error is FALSE when response is empty (non query error)", {
  expect_false(
    test_graphql_gitlab_priv$is_query_error(list())
  )
})

test_that("GitLab GraphQL Engine switches to pulliing files per repo when owner is not an organization", {
  mockery::stub(
    test_graphql_gitlab$get_files_from_org,
    "self$get_files_from_org_per_repo",
    test_mocker$use("gl_files_from_org_per_repo")
  )
  gitlab_files_response <- test_graphql_gitlab$get_files_from_org(
    org = "mbtests",
    owner_type = "user",
    repos = NULL,
    file_paths = "project_metadata.yaml",
    host_files_structure = NULL
  )
  expect_files_gitlab_from_org_by_repos_response(
    response = gitlab_files_response,
    expected_files = "project_metadata.yaml"
  )
})

test_that("Gitlab GraphQL switches to pulling files per repositories when query is too complex", {
  mockery::stub(
    test_graphql_gitlab$get_files_from_org,
    "private$is_query_error",
    TRUE
  )
  mockery::stub(
    test_graphql_gitlab$get_files_from_org,
    "private$is_complexity_error",
    TRUE
  )
  mockery::stub(
    test_graphql_gitlab$get_files_from_org,
    "self$get_files_from_org_per_repo",
    test_mocker$use("gl_files_from_org_per_repo")
  )
  gitlab_files_response_by_repos <- test_graphql_gitlab$get_files_from_org(
    org = "mbtests",
    owner_type = "organization",
    repos = NULL,
    file_paths = c("project_metadata.yaml", "README.md"),
    host_files_structure = NULL,
    verbose = FALSE
  )
  expect_files_gitlab_from_org_by_repos_response(
    response = gitlab_files_response_by_repos,
    expected_files = c("project_metadata.yaml", "README.md")
  )
  test_mocker$cache(gitlab_files_response_by_repos)
})

test_that("Gitlab GraphQL prints messages when switching to pulling files per repositories", {
  mockery::stub(
    test_graphql_gitlab$get_files_from_org,
    "private$is_query_error",
    TRUE
  )
  mockery::stub(
    test_graphql_gitlab$get_files_from_org,
    "private$is_complexity_error",
    TRUE
  )
  mockery::stub(
    test_graphql_gitlab$get_files_from_org,
    "self$get_files_from_org_per_repo",
    test_mocker$use("gl_files_from_org_per_repo")
  )
  expect_snapshot(
    gitlab_files_response_by_repos <- test_graphql_gitlab$get_files_from_org(
      org = "mbtests",
      owner_type = "organization",
      repos = NULL,
      file_paths = c("project_metadata.yaml", "README.md"),
      host_files_structure = NULL,
      verbose = TRUE
    )
  )
})

test_that("GitLab GraphQL switches to iteration when query is too complex", {
  mockery::stub(
    test_graphql_gitlab$get_files_from_org_per_repo,
    "private$get_file_blobs_response",
    test_mocker$use("gl_file_blobs_response")
  )
  mockery::stub(
    test_graphql_gitlab$get_files_from_org_per_repo,
    "private$is_complexity_error",
    TRUE
  )
  expect_snapshot(
    files_from_org_per_repo <- test_graphql_gitlab$get_files_from_org_per_repo(
      org = "mbtests",
      owner_type = "organization",
      repos_data = list("paths" = "gitstatstesting"),
      file_paths = c("project_metadata.yaml", "README.md"),
      host_files_structure = NULL,
      verbose = TRUE
    )
  )
})

test_that("checker properly identifies gitlab files responses", {
  expect_false(
    test_graphql_gitlab_priv$response_prepared_by_iteration(
      files_response = test_mocker$use("gitlab_files_response")
    )
  )
  expect_true(
    test_graphql_gitlab_priv$response_prepared_by_iteration(
      files_response = test_mocker$use("gitlab_files_response_by_repos")
    )
  )
})

test_that("GitLab prepares table from files response", {
  gl_files_table <- test_graphql_gitlab$prepare_files_table(
    files_response = test_mocker$use("gitlab_files_response"),
    org = "mbtests"
  )
  expect_files_table(gl_files_table)
  test_mocker$cache(gl_files_table)
})

test_that("GitLab prepares table from files response prepared in alternative way", {
  gl_files_table <- test_graphql_gitlab$prepare_files_table(
    files_response = test_mocker$use("gitlab_files_response_by_repos"),
    org = "mbtests"
  )
  expect_files_table(gl_files_table)
})

test_that("get_files_content_from_orgs for GitLab works", {
  mockery::stub(
    gitlab_testhost_priv$get_files_content_from_orgs,
    "graphql_engine$prepare_files_table",
    test_mocker$use("gl_files_table")
  )
  gitlab_testhost_priv$searching_scope <- "org"
  gl_files_table <- gitlab_testhost_priv$get_files_content_from_orgs(
    file_path = "meta_data.yaml",
    verbose = FALSE
  )
  expect_files_table(
    gl_files_table, with_cols = "api_url"
  )
  test_mocker$cache(gl_files_table)
})

test_that("get_files_content_from_repos for GitLab works", {
  mockery::stub(
    gitlab_testhost_priv$get_files_content_from_repos,
    "graphql_engine$prepare_files_table",
    test_mocker$use("gl_files_table")
  )
  test_org <- "test_group"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    gitlab_testhost_priv$get_files_content_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  gitlab_testhost_priv$searching_scope <- "repo"
  gl_files_table <- gitlab_testhost_priv$get_files_content_from_repos(
    file_path = "meta_data.yaml",
    verbose = FALSE
  )
  expect_files_table(
    gl_files_table, with_cols = "api_url"
  )
  gitlab_testhost_priv$searching_scope <- "org"
})

test_that("get_files_content_from_repos for GitLab prints message", {
  mockery::stub(
    gitlab_testhost_priv$get_files_content_from_repos,
    "graphql_engine$prepare_files_table",
    test_mocker$use("gl_files_table")
  )
  test_org <- "test_group"
  attr(test_org, "type") <- "organization"
  mockery::stub(
    gitlab_testhost_priv$get_files_content_from_repos,
    "graphql_engine$set_owner_type",
    test_org
  )
  gitlab_testhost_priv$searching_scope <- "repo"
  expect_snapshot(
    gl_files_table <- gitlab_testhost_priv$get_files_content_from_repos(
      file_path = "meta_data.yaml",
      verbose = TRUE
    )
  )
})

test_that("get_files_content makes use of files_structure", {
  mockery::stub(
    gitlab_testhost_priv$get_files_content_from_files_structure,
    "graphql_engine$prepare_files_table",
    test_mocker$use("gl_files_table")
  )
  expect_snapshot(
    files_content <- gitlab_testhost_priv$get_files_content_from_files_structure(
      files_structure = test_mocker$use("gl_files_structure_from_orgs"),
      verbose = TRUE
    )
  )
  expect_files_table(
    files_content,
    with_cols = "api_url"
  )
})

test_that("get_files_content skips if files structure is empty", {
  expect_snapshot(
    files_content <- gitlab_testhost_priv$get_files_content_from_files_structure(
      files_structure = test_mocker$use("gl_empty_files_structure"),
      verbose = TRUE
    )
  )
})

#' @title Get data on pull requests
#' @name get_pull_requests
#' @description List all pull requests from all repositories for an organization or a
#'   vector of repositories.
#' @inheritParams get_issues
#' @return A table of `tibble` and `gitstats_pull_requests` classes.
#' @examples
#' \dontrun{
#' my_gitstats <- create_gitstats() |>
#'   set_github_host(
#'     token = Sys.getenv("GITHUB_PAT"),
#'     repos = c("openpharma/DataFakeR", "openpharma/visR")
#'   ) |>
#'   set_gitlab_host(
#'     token = Sys.getenv("GITLAB_PAT_PUBLIC"),
#'     orgs = "mbtests"
#'   )
#'  get_pull_requests(my_gitstats, since = "2018-01-01", state = "open")
#' }
#' @export
get_pull_requests <- function(gitstats,
                              since = NULL,
                              until = Sys.Date(),
                              state = NULL,
                              cache = TRUE,
                              verbose = FALSE,
                              progress = TRUE) {
  start_time <- Sys.time()
  if (is.null(since)) {
    cli::cli_abort(cli::col_red("You need to pass date to `since` parameter."), call = NULL)
  }
  pr <- gitstats$get_pull_requests(
    since = since,
    until = until,
    state = state,
    cache = cache,
    verbose = verbose,
    progress = progress
  )
  end_time <- Sys.time()
  time_taken <- end_time - start_time
  cli::cli_alert_success("Data pulled in {round(time_taken, 1)} {attr(time_taken, 'units')}")
  return(pr)
}

#' @title Get pull requests statistics
#' @name get_pull_requests_stats
#' @description Prepare statistics from the pulled pull requests data.
#' @details To make function work, you need first to get pull requests data with
#'   `GitStats`. See examples section.
#' @param pull_requests A `gitstats_pr_stats` S3 class table object (output of `get_pull_requests()`).
#' @inheritParams get_issues_stats
#' @return A table of `pull_requests_stats` class.
#' @examples
#' \dontrun{
#'  my_gitstats <- create_gitstats() |>
#'    set_github_host(
#'      token = Sys.getenv("GITHUB_PAT"),
#'      repos = c("r-world-devs/GitStats", "openpharma/visR")
#'    ) |>
#'    get_pull_requests(my_gitstats, since = "2022-01-01") |>
#'    get_pull_requests_stats(
#'      time_aggregation = "month",
#'      group_var = author
#'    )
#' }
#' @export
get_pull_requests_stats <- function(pull_requests,
                                    time_aggregation = c("year", "month", "week", "day"),
                                    group_var) {
  if (!inherits(pull_requests, "gitstats_pull_requests")) {
    cli::cli_abort(c(
      "x" = "`pull_requests` must be a `gitstats_issues` object.",
      "i" = "Pull first your pull_requests with `get_pull_requests()` function."
    ))
  }
  pull_requests <- pull_requests |>
    dplyr::mutate(
      created_at = lubridate::floor_date(
        created_at,
        unit = time_aggregation
      ),
      merged_at = lubridate::floor_date(
        merged_at,
        unit = time_aggregation
      )
    )
  pull_requests_stats <- pull_requests |>
    dplyr::group_by(state, created_at, {{ group_var }}) |>
    dplyr::summarise(
      stats = dplyr::n()
    ) |>
    dplyr::arrange(
      created_at
    )
  pull_requests_stats <- set_pr_stats_class(
    object = pull_requests_stats,
    time_aggregation = time_aggregation
  )
  return(pull_requests_stats)
}

set_pr_stats_class <- function(object, time_aggregation) {
  stopifnot(inherits(object, "grouped_df"))
  object <- dplyr::ungroup(object)
  class(object) <- append(class(object), "gitstats_pr_stats")
  attr(object, "time_aggregation") <- time_aggregation
  object
}

# Standalone GraphQL helpers -----------------------------------------------
# These are plain functions (no R6 dependency) so they can be serialized
# to mirai workers and reused by the R6 methods below.

#' @noRd
graphql_perform_request <- function(gql_api_url, token, gql_query, vars) {
  response <- httr2::request(paste0(gql_api_url, "?")) |>
    httr2::req_headers("Authorization" = paste0("Bearer ", token)) |>
    httr2::req_body_json(list(query = gql_query, variables = vars)) |>
    httr2::req_error(is_error = function(resp) FALSE) |>
    httr2::req_perform()
  if (response$status_code %in% c(400, 502)) {
    response <- httr2::request(paste0(gql_api_url, "?")) |>
      httr2::req_headers("Authorization" = paste0("Bearer ", token)) |>
      httr2::req_body_json(list(query = gql_query, variables = vars)) |>
      httr2::req_retry(
        is_transient = ~ httr2::resp_status(.x) %in% c(400, 502),
        max_seconds = 60
      ) |>
      httr2::req_perform()
  }
  return(response)
}

#' @noRd
is_graphql_query_error <- function(response) {
  length(response) > 0 && any(names(response) == "errors")
}

#' @noRd
is_graphql_no_fields_error <- function(response) {
  any(purrr::map_lgl(response$errors, ~ grepl("doesn't exist on type", .$message)))
}

#' @noRd
is_graphql_complexity_error <- function(response) {
  any(purrr::map_lgl(response$errors, ~ grepl("Query has complexity", .$message)))
}

#' @noRd
is_graphql_server_error <- function(response) {
  any(purrr::map_lgl(response$errors, ~ grepl("Internal server error", .$message)))
}

#' @noRd
is_graphql_limit_error <- function(response) {
  any(purrr::map_lgl(response$errors, ~ grepl("cannot accept more than", .$message)))
}

#' @noRd
set_graphql_error_class <- function(response) {
  if (is_graphql_query_error(response)) {
    class(response) <- c(class(response), "graphql_error")
    if (is_graphql_no_fields_error(response)) {
      class(response) <- c(class(response), "graphql_no_fields_error")
    }
    if (is_graphql_complexity_error(response)) {
      class(response) <- c(class(response), "graphql_complexity_error")
    }
    if (is_graphql_limit_error(response)) {
      class(response) <- c(class(response), "graphql_limit_error")
    }
    if (is_graphql_server_error(response)) {
      class(response) <- c(class(response), "graphql_server_error")
    }
  }
  return(response)
}

#' @noRd
graphql_response <- function(gql_api_url, token, gql_query, vars) {
  response <- graphql_perform_request(gql_api_url, token, gql_query, vars)
  httr2::resp_body_json(response) |> set_graphql_error_class()
}

# EngineGraphQL R6 class ---------------------------------------------------

EngineGraphQL <- R6::R6Class(
  "EngineGraphQL",
  inherit = Engine,
  public = list(
    gql_api_url = NULL,

    gql_query = NULL,

    initialize = function(gql_api_url = NA,
                          token = NA,
                          scan_all = FALSE) {
      private$engine <- "graphql"
      self$gql_api_url <- gql_api_url
      private$token <- token
      private$scan_all <- scan_all
    },

    gql_response = function(gql_query, vars = "null", verbose) {
      graphql_response(
        gql_api_url = self$gql_api_url,
        token = private$token,
        gql_query = gql_query,
        vars = vars
      )
    },

    get_user = function(username) {
      response <- self$gql_response(
        gql_query = self$gql_query$user(),
        vars = list("user" = username)
      )
      return(response)
    },

    get_issues_from_repos = function(org,
                                     repos_names,
                                     verbose) {
      repos_list_with_issues <- gitstats_map(repos_names, function(repo) {
        private$get_issues_from_one_repo(
          org = org,
          repo = repo,
          verbose = verbose
        )
      })
      names(repos_list_with_issues) <- repos_names
      repos_list_with_issues <- repos_list_with_issues |>
        purrr::discard(~ length(.) == 0)
      return(repos_list_with_issues)
    },

    prepare_issues_table = function(repos_list_with_issues,
                                    org) {
      issues_table <- purrr::imap(repos_list_with_issues, function(repo, repo_name) {
        issues_row <- purrr::map_dfr(repo, function(issue_data) {
          state <- tolower(issue_data[["node"]][["state"]])
          if (state == "opened") {
            state <- "open"
          }
          get_node_data <- function(node_data) {
            issue_data[["node"]][[node_data]] %||% ""
          }
          author_login <- issue_data[["node"]][["author"]][["login"]] %||% NA_character_
          data.frame(
            "number" = as.character(get_node_data("number")),
            "title" = get_node_data("title"),
            "description" = get_node_data("description"),
            "created_at" = lubridate::as_datetime(get_node_data("created_at")),
            "closed_at" = lubridate::as_datetime(get_node_data("closed_at")),
            "state" = state,
            "url" = get_node_data("url"),
            "author" = author_login
          )
        })
        issues_row$repo_name <- repo_name
        issues_row
      }) |>
        purrr::discard(~ length(.) == 1) |>
        purrr::list_rbind()
      if (nrow(issues_table) > 0) {
        issues_table <- issues_table |>
          dplyr::mutate(
            organization = org,
            api_url = self$gql_api_url
          ) |>
          dplyr::relocate(
            repo_name,
            .before = number
          )
      }
      return(issues_table)
    },

    get_pr_from_repos = function(org,
                                 repos_names,
                                 verbose) {
      repos_list_with_pr <- gitstats_map(repos_names, function(repo) {
        private$get_pr_from_one_repo(
          org = org,
          repo = repo,
          verbose = verbose
        )
      })
      names(repos_list_with_pr) <- repos_names
      repos_list_with_pr <- repos_list_with_pr |>
        purrr::discard(~ length(.) == 0)
      return(repos_list_with_pr)
    },

    prepare_pr_table = function(repos_list_with_pr,
                                org) {
      pr_table <- purrr::imap(repos_list_with_pr, function(repo, repo_name) {
        pr_row <- purrr::map_dfr(repo, function(pr_data) {
          state <- tolower(pr_data[["node"]][["state"]])
          if (state == "opened") {
            state <- "open"
          }
          get_node_data <- function(node_data) {
            pr_data[["node"]][[node_data]] %||% ""
          }
          author_login <- pr_data[["node"]][["author"]][["login"]] %||% NA_character_
          data.frame(
            "number" = as.character(get_node_data("number")),
            "created_at" = lubridate::as_datetime(get_node_data("created_at")),
            "merged_at" = lubridate::as_datetime(get_node_data("merged_at")),
            "state" = state,
            "author" = author_login,
            "source_branch" = get_node_data("source_branch"),
            "target_branch" = get_node_data("target_branch")
          )
        })
        pr_row$repo_name <- repo_name
        pr_row
      }) |>
        purrr::discard(~ length(.) == 1) |>
        purrr::list_rbind()
      if (nrow(pr_table) > 0) {
        pr_table <- pr_table |>
          dplyr::mutate(
            organization = org,
            api_url = self$gql_api_url
          ) |>
          dplyr::relocate(
            repo_name,
            .before = number
          )
      }
      return(pr_table)
    }
  ),
  private = list(
    owner_types_cache = list(),

    perform_request = function(gql_query, vars, token = private$token, verbose = TRUE) {
      graphql_perform_request(
        gql_api_url = self$gql_api_url,
        token = token,
        gql_query = gql_query,
        vars = vars
      )
    },

    handle_graphql_error = function(responses_list, verbose) {
      if (verbose && inherits(responses_list, "graphql_error")) {
        cli::cli_alert_danger("GraphQL returned errors:")
        if (inherits(responses_list, "graphql_no_fields_error")) {
          error_fields <- purrr::map_vec(responses_list$errors, ~.$extensions$fieldName %||% "") |>
            purrr::discard(~ . == "")
          cli::cli_alert_info("Your GraphQL does not recognize [{error_fields}] field{?s}.")
          cli::cli_alert_warning("Check version of your GitLab.")
        } else {
          cli::cli_alert_warning(purrr::map_vec(responses_list$errors, ~.$message))
        }
      }
      return(responses_list)
    },

    is_query_error = function(response) {
      is_graphql_query_error(response)
    },

    is_no_fields_query_error = function(response) {
      is_graphql_no_fields_error(response)
    },

    is_complexity_error = function(response) {
      is_graphql_complexity_error(response)
    },

    is_server_error = function(response) {
      is_graphql_server_error(response)
    },

    is_limit_error = function(response) {
      is_graphql_limit_error(response)
    },

    set_graphql_error_class = function(response) {
      set_graphql_error_class(response)
    },

    get_path_from_files_structure = function(host_files_structure,
                                             org,
                                             repo = NULL) {
      if (is.null(repo)) {
        file_path <- host_files_structure[[org]] |>
          unlist(use.names = FALSE) |>
          unique()
      } else {
        file_path <- host_files_structure[[org]][[repo]]
      }
      file_path <- file_path[grepl(text_files_pattern, file_path)]
      return(file_path)
    }
  )
)

test_that("is_query_error works", {
  expect_true(test_graphql_github_priv$is_query_error(test_error_fixtures$graphql_error_no_groups))
  expect_true(test_graphql_github_priv$is_query_error(test_error_fixtures$graphql_error_no_count_languages))
})

test_that("is_no_fields_query_error works", {
  expect_true(test_graphql_github_priv$is_no_fields_query_error(test_error_fixtures$graphql_error_no_groups))
  expect_true(test_graphql_github_priv$is_no_fields_query_error(test_error_fixtures$graphql_error_no_count_languages))
  expect_false(test_graphql_github_priv$is_no_fields_query_error(test_error_fixtures$graphql_server_error))
})

test_that("is_server_error works", {
  expect_true(test_graphql_github_priv$is_server_error(test_error_fixtures$graphql_server_error))
  expect_false(test_graphql_github_priv$is_server_error(test_error_fixtures$graphql_error_no_count_languages))
})

test_that("is_complexity_error works", {
  expect_true(test_graphql_github_priv$is_complexity_error(test_error_fixtures$graphql_complexity_error))
  expect_false(test_graphql_github_priv$is_complexity_error(test_error_fixtures$graphql_error_no_groups))
})

test_that("is_limit_error works", {
  expect_true(test_graphql_github_priv$is_limit_error(test_error_fixtures$graphql_limit_error))
  expect_false(test_graphql_github_priv$is_limit_error(test_error_fixtures$graphql_error_no_groups))
})

test_that("set_graphql_error_class works", {
  error_class_object <- test_graphql_github_priv$set_graphql_error_class(
    test_error_fixtures$graphql_error_no_groups
  )
  expect_true(
    all(c("graphql_error", "graphql_no_fields_error") %in% class(error_class_object))
  )
  limit_error_object <- test_graphql_github_priv$set_graphql_error_class(
    test_error_fixtures$graphql_limit_error
  )
  expect_true(
    all(c("graphql_error", "graphql_limit_error") %in% class(limit_error_object))
  )
})

repo_gitstats_colnames <- c(
  "repo_id", "repo_name", "repo_fullpath", "default_branch", "stars", "forks",
  "created_at", "last_activity_at", "languages", "issues_open", "issues_closed",
  "organization", "repo_url", "commit_sha", "api_url", "githost", "last_activity"
)

repo_host_colnames <- c("repo_id", "repo_name", "repo_fullpath", "default_branch",
                        "stars", "forks", "created_at", "last_activity_at",
                        "languages", "issues_open", "issues_closed",
                        "organization", "repo_url", "commit_sha")

repo_min_colnames <- c("repo_id", "repo_name", "repo_fullpath", "default_branch",
                       "created_at", "organization", "repo_url", "commit_sha")

expect_orgs_table <- function(object, add_cols = NULL) {
  expect_s3_class(object, "data.frame")
  expect_named(object, c("name", "description", "path", "url", "avatar_url",
                         "repos_count", "members_count", add_cols))
  expect_gt(nrow(object), 0)
}

expect_repos_table_object <- function(repos_object, with_cols = NULL) {
  expect_repos_table(
    repos_object = repos_object,
    repo_cols = c(repo_gitstats_colnames, with_cols)
  )
  expect_s3_class(repos_object, "repos_table")
}

expect_repos_table <- function(repos_object, repo_cols = repo_host_colnames, with_cols = NULL) {
  repo_cols <- c(
    repo_cols, with_cols
  )
  expect_s3_class(repos_object, "data.frame")
  expect_named(repos_object, repo_cols)
  expect_gt(nrow(repos_object), 0)
}

expect_issues_table <- function(get_issues_object, with_stats = TRUE, exp_author = TRUE) {
  issue_cols <- c("repo_name", "number", "title", "description", "created_at",
                  "closed_at", "state", "url", "author", "organization", "api_url")

  expect_s3_class(get_issues_object, "data.frame")
  expect_named(get_issues_object, issue_cols)
  expect_gt(nrow(get_issues_object), 0)
  expect_s3_class(get_issues_object$created_at, "POSIXt")
  expect_s3_class(get_issues_object$closed_at, "POSIXt")
}

expect_pr_table <- function(get_pr_object, with_stats = TRUE, exp_author = TRUE) {
  pr_cols <- c("repo_name", "number", "created_at", "merged_at", "state", "author",
                  "source_branch", "target_branch", "organization","api_url")

  expect_s3_class(get_pr_object, "data.frame")
  expect_named(get_pr_object, pr_cols)
  expect_gt(nrow(get_pr_object), 0)
  expect_s3_class(get_pr_object$created_at, "POSIXt")
  expect_s3_class(get_pr_object$merged_at, "POSIXt")
}

expect_commits_table <- function(get_commits_object, with_stats = TRUE, exp_author = TRUE) {
  commit_cols <- if (exp_author) {
    c(
      "repo_name", "id", "committed_date", "author", "author_login", "author_name",
      "additions", "deletions", "organization", "repo_url", "api_url"
    )
  } else {
    c(
      "repo_name", "id", "committed_date", "author", "additions", "deletions",
      "organization", "repo_url", "api_url"
    )
  }
  expect_s3_class(get_commits_object, "data.frame")
  expect_named(get_commits_object, commit_cols)
  expect_gt(nrow(get_commits_object), 0)
  expect_s3_class(get_commits_object$committed_date, "POSIXt")
  if (with_stats) {
    expect_type(get_commits_object$additions, "integer")
    expect_type(get_commits_object$deletions, "integer")
  }
}

expect_users_table <- function(get_user_object, one_user = FALSE) {
  user_cols <- c(
    "id", "name", "login", "email", "location", "starred_repos",
    "commits", "issues", "pull_requests", "reviews",
    "avatar_url", "web_url"
  )
  expect_named(get_user_object, user_cols)
  if (one_user) {
    expect_equal(nrow(get_user_object), 1)
  } else {
    expect_gt(nrow(get_user_object), 1)
  }
}

expect_files_table <- function(files_object, with_cols = NULL) {
  expect_s3_class(files_object, "data.frame")
  expect_named(
    files_object,
    c("repo_id", "repo_name", "organization",
      "file_path", "file_content", "file_size", "file_id",
      "repo_url", "commit_sha", with_cols)
  )
  expect_type(files_object$file_size, "integer")
  expect_gt(nrow(files_object), 0)
}

expect_releases_table <- function(releases_object) {
  expect_s3_class(releases_object, "data.frame")
  expect_named(
    releases_object,
    c("repo_name", "repo_url",
      "release_name", "release_tag", "published_at", "release_url",
      "release_log")
  )
  expect_gt(nrow(releases_object), 0)
}

expect_empty_table <- function(object) {
  expect_s3_class(object, "data.frame")
  expect_equal(nrow(object), 0)
}

expect_list_contains <- function(object, elements) {
  act <- quasi_label(rlang::enquo(object), arg = "object")
  act$check <- any(elements %in% names(act$val))
  expect(
    act$check == TRUE,
    sprintf("%s does not contain specified elements", act$lab)
  )

  invisible(act$val)
}

expect_list_contains_only <- function(object, elements) {
  act <- quasi_label(rlang::enquo(object), arg = "object")
  act$check <- all(elements %in% names(act$val))
  expect(
    act$check == TRUE,
    sprintf("%s does not contain specified elements", act$lab)
  )

  invisible(act$val)
}

expect_search_github_response <- function(object) {
  expect_type(
    object,
    "list"
  )
  expect_list_contains(
    object[[1]],
    c("name", "path", "sha", "url", "git_url", "html_url", "repository", "score")
  )
}

expect_repos_gitlab_gql_response <- function(object, type = "organization") {
  expect_type(
    object,
    "list"
  )
  repo_node <- if (type == "organization") {
    object$data$group$projects$edges[[1]]$node
  } else {
    object$data$projects$edges[[1]]$node
  }
  expect_list_contains(
    repo_node,
    c(
      "id", "name", "repository", "stars", "forks", "created_at", "last_activity_at"
    )
  )
}

expect_repos_github_gql_response <- function(repo_node) {
  expect_in(
    names(repo_node),
    c("repo_id", "repo_name", "repo_path", "repo_fullpath",
      "default_branch", "stars", "forks", "created_at",
      "last_activity_at", "languages", "issues_open",
      "issues_closed", "organization", "repo_url",
      "defaultBranchRef", "commit_sha")
  )
}

expect_commit_gitlab_rest_response <- function(object) {
  expect_type(
    object,
    "list"
  )
  expect_list_contains(
    object,
    c(
      "id", "short_id", "created_at", "parent_ids", "title", "message",
      "author_name", "author_email", "authored_date", "committer_name",
      "committer_email"
    )
  )
}

expect_commit_github_gql_response <- function(object) {
  expect_type(
    object,
    "list"
  )
  expect_list_contains(
    object$node,
    c("id", "committed_date", "author", "additions", "deletions", "repository")
  )
}

expect_pr_gql_response <- function(object) {
  expect_type(
    object,
    "list"
  )
  expect_list_contains(
    object$node,
    c("number", "created_at", "merged_at", "state", "author", "source_branch", "target_branch")
  )
}

expect_pr_full_list <- function(object) {
  expect_named(
    object[[1]]$node,
    c("number", "created_at", "merged_at", "state", "author", "source_branch", "target_branch")
  )
}

expect_user_gql_response <- function(object) {
  expect_list_contains(
    object,
    "data"
  )
  expect_list_contains(
    object$data,
    "user"
  )
  expect_list_contains(
    object$data$user,
    c("id", "name", "email", "location", "starred_repos", "avatar_url", "web_url")
  )
}

expect_files_github_raw_response <- function(object) {
  purrr::walk(object$data$repository$object$entries, function(entry) {
    expect_equal(
      names(entry),
      c("name", "type")
    )
  })
}

expect_files_gitlab_blob_response <- function(object) {
  purrr::walk(object$data$project$repository$blobs$nodes, function(node) {
    expect_equal(
      names(node),
      c("path", "rawBlob", "size", "oid")
    )
  })
}

expect_files_gitlab_tree_response <- function(object) {
  purrr::walk(object$data$project$repository$tree$blobs$nodes, function(node) {
    expect_equal(
      names(node),
      c("name")
    )
  })
}

expect_files_github_response <- function(object) {
  expect_type(
    object,
    "list"
  )
  expect_gt(
    length(object[[1]]),
    0
  )
  purrr::walk(object, function(repository) {
    purrr::walk(repository, function(file_path) {
      expect_list_contains(
        file_path,
        c("repo_name", "repo_id", "repo_url", "object")
      )
      expect_list_contains(
        file_path$file,
        c("text", "byteSize")
      )
    })
  })
}

expect_files_gitlab_from_org_response <- function(object) {
  expect_type(
    object,
    "list"
  )
  expect_gt(
    length(object),
    0
  )
  purrr::walk(object, function(project) {
    expect_list_contains(
      project,
      c(
        "path", "id", "repository"
      )
    )
    expect_list_contains(
      project$repository$blobs$nodes[[1]],
      c("path", "rawBlob", "size", "oid")
    )
  })
}

expect_files_gitlab_from_org_by_repos_response <- function(response, expected_files) {
  expect_type(
    response,
    "list"
  )
  expect_gt(
    length(response),
    0
  )
  purrr::walk(response, function(repo) {
    purrr::walk(repo$data$project$repository$blobs$nodes, function(file) {
      expect_equal(
        names(file),
        c("path", "rawBlob", "size", "oid")
      )
    })
  })
  files_vec <- purrr::map(response, ~ purrr::map_vec(.$data$project$repository$blobs$nodes, ~ .$path)) |>
    unlist() |>
    unique()
  expect_in(expected_files, files_vec)
}

expect_releases_github_response <- function(object) {
  expect_type(
    object,
    "list"
  )
  purrr::walk(object, function(response) {
    expect_gt(length(response), 0)
    expect_list_contains(
      response$data$repository,
      c("releases")
    )
    purrr::walk(response$data$repository$releases$nodes, function(node) {
      expect_list_contains(
        node,
        c("name", "tagName", "publishedAt", "url", "description")
      )
    })
  })
}

expect_releases_gitlab_response <- function(object) {
  expect_type(
    object,
    "list"
  )
  purrr::walk(object, function(response) {
    expect_gt(length(response), 0)
    expect_list_contains(
      response$data$project,
      c("releases")
    )
    purrr::walk(response$data$project$releases$nodes, function(node) {
      expect_list_contains(
        node,
        c("name", "tagName", "releasedAt", "links", "description")
      )
    })
  })
}

issue_fields <- c("number", "title", "description", "created_at", "closed_at",
                  "state", "url", "author")

expect_issues_github_page <- function(object) {
  expect_type(
    object,
    "list"
  )
  if (names(object) == "errors") {
    cli::cli_alert_danger(object$errors)
    cli::cli_abort("Errors in GraphQL response.")
  }
  expect_named(
    object$data$repository$issues,
    c("pageInfo", "edges")
  )
  expect_named(
    object$data$repository$issues$edges[[1]]$node,
    issue_fields
  )
}

expect_issues_gitlab_page <- function(object) {
  expect_type(
    object,
    "list"
  )
  if (names(object) == "errors") {
    cli::cli_alert_danger(object$errors)
    cli::cli_abort("Errors in GraphQL response.")
  }
  expect_named(
    object$data$project$issues,
    c("pageInfo", "edges")
  )
  expect_named(
    object$data$project$issues$edges[[1]]$node,
    issue_fields
  )
}

expect_issues_full_list <- function(object) {
  expect_named(
    object[[1]]$node,
    issue_fields
  )
}

github_orgs_fields <- c("name", "description", "login", "url", "repositories",
                        "membersWithRole", "avatarUrl")

expect_orgs_github_full_list <- function(object) {
  expect_named(
    object[[1]],
    github_orgs_fields
  )
}

gitlab_orgs_fields <- c("name", "description", "fullPath", "webUrl", "projectsCount",
                        "groupMembersCount", "avatarUrl")

expect_orgs_gitlab_full_list <- function(object) {
  expect_named(
    object[[1]],
    gitlab_orgs_fields
  )
}

---
title: "Getting repositories with specific code or files"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting repositories with specific code or files}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4,
  message = FALSE
)
```

Apart from pulling all repositories from organizations, you can look for those that have a particular text in a `code blob`:

```{r, eval = FALSE}
library(GitStats)

github_stats <- create_gitstats() %>%
  set_github_host(
    orgs = c("r-world-devs", "openpharma"),
    token = Sys.getenv("GITHUB_PAT")
  ) %>%
  verbose_off()

repos_urls <- get_repos_urls(
  gitstats = github_stats,
  with_code = "shiny"
)
```

You can limit your search, as it is allowed with GitLab and GitHub API search endpoints, to certain files.

```{r, eval = FALSE}
repos_urls <- get_repos_urls(
  gitstats = github_stats,
  with_code =  c("purrr", "shiny"),
  in_files = c("DESCRIPTION", "NAMESPACE", "renv.lock")
)
```

You can also search for repositories with certain files (do not confuse `with_files` with `in_files`!).

```{r, eval = FALSE}
repos_urls <- get_repos_urls(
  gitstats = github_stats,
  with_files = c("renv.lock", "DESCRIPTION")
)
```

---
title: "Store your data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Store your data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
has_tokens <- nzchar(Sys.getenv("GITHUB_PAT")) && nzchar(Sys.getenv("GITLAB_PAT_PUBLIC"))
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4,
  eval = has_tokens
)
```

Set connections to hosts.

> Example workflow makes use of public GitHub and GitLab, but it is plausible, that you will use your internal git platforms, where you need to define `host` parameter. See `vignette("set_hosts")` article on that.

```{r}
library(GitStats)

git_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs",
    token = Sys.getenv("GITHUB_PAT")
  ) |>
  set_gitlab_host(
    orgs = c("mbtests"),
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  )
```

Optionally speed up processing.

```{r}
git_stats |>
  set_parallel(10L)
```

As scanning scope was set to `organizations` (`orgs` parameter in `set_*_host()`), `GitStats` will pull all repositories from these organizations.

```{r}
repos <- get_repos(git_stats, progress = FALSE)
dplyr::glimpse(repos)
```

You can always go for the lighter version of `get_repos`, i.e. `get_repos_urls()` which will print you a vector of URLs instead of whole table.

```{r}
repos_urls <- get_repos_urls(git_stats)
dplyr::glimpse(repos_urls)
```

## Local Storage

After pulling, the data is saved by default to `GitStats`.

```{r}
commits <- git_stats |>
  get_commits(
    since = "2025-06-01",
    until = "2025-06-14",
    progress = FALSE
  )
git_stats
dplyr::glimpse(commits)
```

## SQLite Storage

For local saving we recommend though using `SQLite` storage. You can set it up with `set_sqlite_storage()` function. Then, all data pulled with `get_*()` functions will be stored in the `SQLite` database and retrieved from there when you run the function again.

```{r}
commits <- git_stats |>
  set_sqlite_storage("my_local_db") |>
  get_commits(
    since = "2025-06-01",
    until = "2025-06-14",
    progress = FALSE
  )
dplyr::glimpse(commits)
git_stats
```

Therefore, it is now not be dependent on the `GitStats` object, but on the local database, so you can even create a new `GitStats` and connect it to the same database and data will be there.

```{r}
new_git_stats <- create_gitstats() |>
  set_github_host(
    orgs = "r-world-devs",
    token = Sys.getenv("GITHUB_PAT")
  ) |>
  set_gitlab_host(
    orgs = c("mbtests"),
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  ) |>
  set_sqlite_storage("my_local_db")

commits <- new_git_stats |>
  get_commits(
    since = "2025-06-01",
    until = "2025-06-14",
    verbose = TRUE
  )
dplyr::glimpse(commits)
```

Caching feature is by default turned on. You may switch it off:

```{r}
commits <- new_git_stats |>
  get_commits(
    since = "2025-06-01",
    until = "2025-06-14",
    verbose = TRUE,
    cache = FALSE,
    progress = FALSE
  )
dplyr::glimpse(commits)
```

## Incremental pulling

When you pull data with `get_*()` functions, it is stored in the local database. If you run the same function again, it will check if there is already data for the same parameters and pull only the missing data. This way, you can keep your database up to date without pulling all data again.


```{r}
commits <- new_git_stats |>
  get_commits(
    since = "2025-06-01",
    until = "2025-06-30",
    verbose = TRUE,
    progress = FALSE
  )
dplyr::glimpse(commits)
```

## Remove Storage

Remove storage if you wish.

```{r}
new_git_stats |>
  remove_sqlite_storage()
```

## Postgres Storage

For more permanent storage, you can set up a connection to your database with `set_postgres_storage()` function. Then, all data pulled with `get_*()` functions will be stored in the database and retrieved from there when you run the function again.

---
title: "Setting hosts"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Setting hosts}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4
)
```


To make `GitStats` work you need to set hosts after creating `gitstats`.

You can set GitLab host with `set_gitlab_host()` and GitHub host with `set_github_host()` or both.

When setting hosts you need to consider:

* Do you wish to connect to `private` or `public` hosts?

* What `scanning scope` do you wish to set? Do you want to scan specific `organizations` and/or `repositories` or maybe whole git platforms?

* Do you have `tokens` set up and stored in your environment variables that grant you access to APIs?

## Public and private hosts

If you connect to **public** hosts you simply call `set_github_host()` or `set_gitlab_host()` function without specifying `host` parameter.

```{r, eval = FALSE}
library(GitStats)
git_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs", "openpharma"),
    token = Sys.getenv("GITHUB_PAT")
  ) |>
  set_gitlab_host(
    orgs = c("mbtests"),
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  )
```

If you wish to connect to **internal** GitHub or GitLab, you need to pass names of the hosts to `host` parameter. Remember also to have tokens set up properly for these hosts (on tokens read below).

```r
git_stats <- create_gitstats() |>
  set_github_host(
    host = "github.internal.com",
    orgs = c("org_1", "org_2", "org_3"),
    token = Sys.getenv("YOUR_GITHUB_PAT")
  ) |>
  set_gitlab_host(
    host = "internal.host.com",
    orgs = c("internal_org"),
    token = Sys.getenv("YOUR_GITLAB_PAT")
  )
```

### API versions

`GitStats` is configured to connect to GitHub API (version 3) and GitLab API (version 4).

## Scanning scope

When setting hosts you choose what scanning scope of your `GitStats` will be:

* `organizations/groups` - in this case you need to pass character arguments (names of organizations (in case of GitHub) or groups (in case of GitLab)) to `orgs` parameter.

```{r, eval = FALSE}
git_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs", "openpharma"),
    token = Sys.getenv("GITHUB_PAT")
  ) |>
  set_gitlab_host(
    orgs = c("mbtests"),
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  )
```

* `repositories` - in this case you need to pass full names of repositories (`{org_name}/{repo_name}`) to the `repos` parameter.

```{r, eval = FALSE}
git_stats <- create_gitstats() |>
  set_github_host(
    repos = c("r-world-devs/GitStats", "r-world-devs/shinyCohortBuilder", "openpharma/DataFakeR"),
    token = Sys.getenv("GITHUB_PAT")
  ) |>
  set_gitlab_host(
    repos = "mbtests/gitstatstesting",
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  )
```

* `organizations/groups` and `repositories` - you can define both at the same time:

```{r, eval = FALSE}
git_stats <- create_gitstats() |>
  set_github_host(
    orgs = "openpharma",
    repos = c("r-world-devs/GitStats", "r-world-devs/shinyCohortBuilder"),
    token = Sys.getenv("GITHUB_PAT")
  )
```

* `whole hosts` - this is possible for the time being only in case of private hosts, as public ones are deemed to be too large. To set whole Git platform to be scanned just set hosts **without specifying** `orgs` or `repos`. On the other hand, remember that to connect with internal host, you need to pass argument to `host` parameter.

```r
git_stats <- create_gitstats() |>
  set_github_host(
    host = "github.internal.com",
    token = Sys.getenv("YOUR_GITHUB_PAT")
  ) |>
  set_gitlab_host(
    host = "internal.host.com",
    token = Sys.getenv("YOUR_GITLAB_PAT")
  )
```

## Authorize your access with tokens

Remember to pass to your `set_*_host()` functions `tokens` that authorize access.

You can store your tokens in environment variables, e.g. defined in `.Renviron` file as `GITHUB_PAT` for GitHub and `GITLAB_PAT` for GitLab.

On how to create your tokens refer to GitHub API and GitLab API documentation.

When creating tokens you will be asked to set access scopes of the tokens. For `GitStats` to work you need tokens with given scopes:

- [GitHub] `public_repo`, `read:org` and `read:user`,
- [GitLab] `read_api`.

If you have your access tokens stored in environment variables with such names as `GITHUB_PAT` or `GITHUB_PAT_*` and `GITLAB_PAT` or `GITLAB_PAT_*` you do not need to specify them in `set_*_host()` functions, `GitStats` will automatically find them.

```{r, eval = FALSE}
git_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs", "openpharma")
  ) |>
  set_gitlab_host(
    orgs = c("mbtests")
  )
```

---
title: "Setting hosts"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Setting hosts}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4
)
```


To make `GitStats` work you need to set hosts after creating `gitstats`.

You can set GitLab host with `set_gitlab_host()` and GitHub host with `set_github_host()` or both.

When setting hosts you need to consider:

* Do you wish to connect to `private` or `public` hosts?

* What `scanning scope` do you wish to set? Do you want to scan specific `organizations` and/or `repositories` or maybe whole git platforms?

* Do you have `tokens` set up and stored in your environment variables that grant you access to APIs?

## Public and private hosts

If you connect to **public** hosts you simply call `set_github_host()` or `set_gitlab_host()` function without specifying `host` parameter.

```{r, eval = FALSE}
library(GitStats)
git_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs", "openpharma"),
    token = Sys.getenv("GITHUB_PAT")
  ) |>
  set_gitlab_host(
    orgs = c("mbtests"),
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  )
```

If you wish to connect to **internal** GitHub or GitLab, you need to pass names of the hosts to `host` parameter. Remember also to have tokens set up properly for these hosts (on tokens read below).

```{r, eval = FALSE}
git_stats <- create_gitstats() |>
  set_github_host(
    host = "github.internal.com",
    orgs = c("org_1", "org_2", "org_3"),
    token = Sys.getenv("YOUR_GITHUB_PAT")
  ) |>
  set_gitlab_host(
    host = "internal.host.com",
    orgs = c("internal_org"),
    token = Sys.getenv("YOUR_GITLAB_PAT")
  )
```

### API versions

`GitStats` is configured to connect to GitHub API (version 3) and GitLab API (version 4).

## Scanning scope

When setting hosts you choose what scanning scope of your `GitStats` will be:

* `organizations/groups` - in this case you need to pass character arguments (names of organizations (in case of GitHub) or groups (in case of GitLab)) to `orgs` parameter.

```{r, eval = FALSE}
git_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs", "openpharma"),
    token = Sys.getenv("GITHUB_PAT")
  ) |>
  set_gitlab_host(
    orgs = c("mbtests"),
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  )
```

* `repositories` - in this case you need to pass full names of repositories (`{org_name}/{repo_name}`) to the `repos` parameter.

```{r, eval = FALSE}
git_stats <- create_gitstats() |>
  set_github_host(
    repos = c("r-world-devs/GitStats", "r-world-devs/shinyCohortBuilder", "openpharma/DataFakeR"),
    token = Sys.getenv("GITHUB_PAT")
  ) |>
  set_gitlab_host(
    repos = "mbtests/gitstatstesting",
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  )
```

* `organizations/groups` and `repositories` - you can define both at the same time:

```{r, eval = FALSE}
git_stats <- create_gitstats() |>
  set_github_host(
    orgs = "openpharma",
    repos = c("r-world-devs/GitStats", "r-world-devs/shinyCohortBuilder"),
    token = Sys.getenv("GITHUB_PAT")
  )
```

* `whole hosts` - this is possible for the time being only in case of private hosts, as public ones are deemed to be too large. To set whole Git platform to be scanned just set hosts **without specifying** `orgs` or `repos`. On the other hand, remember that to connect with internal host, you need to pass argument to `host` parameter.

```{r, eval = FALSE}
git_stats <- create_gitstats() |>
  set_github_host(
    host = "github.internal.com",
    token = Sys.getenv("YOUR_GITHUB_PAT")
  ) |>
  set_gitlab_host(
    host = "internal.host.com",
    token = Sys.getenv("YOUR_GITLAB_PAT")
  )
```

## Authorize your access with tokens

Remember to pass to your `set_*_host()` functions `tokens` that authorize access.

You can store your tokens in environment variables, e.g. defined in `.Renviron` file as `GITHUB_PAT` for GitHub and `GITLAB_PAT` for GitLab.

On how to create your tokens refer to GitHub API and GitLab API documentation.

When creating tokens you will be asked to set access scopes of the tokens. For `GitStats` to work you need tokens with given scopes:

- [GitHub] `public_repo`, `read:org` and `read:user`,
- [GitLab] `read_api`.

If you have your access tokens stored in environment variables with such names as `GITHUB_PAT` or `GITHUB_PAT_*` and `GITLAB_PAT` or `GITLAB_PAT_*` you do not need to specify them in `set_*_host()` functions, `GitStats` will automatically find them.

```{r, eval = FALSE}
git_stats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs", "openpharma")
  ) |>
  set_gitlab_host(
    orgs = c("mbtests")
  )
```

# get_repos_urls prints messages

    Code
      repo_urls <- test_gitstats$get_repos_urls(verbose = TRUE)
    Message
      > Pulling repositories 🌐 URLs...

# get_repos_urls prints time used to pull data

    Code
      get_repos_urls(gitstats = test_gitstats, with_code = "shiny", in_files = "DESCRIPTION",
        verbose = TRUE)
    Message
      v Data pulled in 0 secs
    Output
      # Repository URLs:
      - https://github.test.com/api/test_url
      - https://gitlab.com/mbtests/graphql_tests
      - https://gitlab.com/mbtests/gitstats-testing-2
      
      # Host Summary:
      - github.test.com: 1 URL(s)
      - gitlab.com: 2 URL(s)

# get_repos_urls prints when data is longer than 5

    Code
      get_repos_urls(gitstats = test_gitstats, with_code = "shiny", in_files = "DESCRIPTION",
        verbose = FALSE)
    Message
      v Data pulled in 0 secs
    Output
      # Repository URLs (showing first 5 of 15):
      - https://github.test.com/api/test_url
      - https://gitlab.com/mbtests/graphql_tests
      - https://gitlab.com/mbtests/gitstats-testing-2
      - https://github.test.com/api/test_url
      - https://gitlab.com/mbtests/graphql_tests
      
      # Host Summary:
      - github.test.com: 5 URL(s)
      - gitlab.com: 10 URL(s)

# user query is built properly

    Code
      gh_user_query
    Output
      [1] "\n        query GetUser($user: String!) {\n          user(login: $user) {\n            id\n            name\n            login\n            email\n            location\n            starred_repos: starredRepositories {\n              totalCount\n            }\n            contributions: contributionsCollection {\n              totalIssueContributions\n              totalCommitContributions\n              totalPullRequestContributions\n              totalPullRequestReviewContributions\n            }\n            avatar_url: avatarUrl\n            web_url: websiteUrl\n          }\n        }"

# groups GitLab query is built properly

    Code
      gl_orgs_query
    Output
      [1] "query GetGroups($groupCursor: String!) {\n            groups (after: $groupCursor) {\n              pageInfo {\n                endCursor\n                hasNextPage\n              }\n              edges {\n                node {\n      name\n      description\n      fullPath\n      webUrl\n      projectsCount\n      groupMembersCount\n      avatarUrl\n    }\n              }\n            }\n        }"

# group GitLab query is built properly

    Code
      gl_org_query
    Output
      [1] "\n      query GetGroup($org: ID!) {\n        group(fullPath: $org) {\n      name\n      description\n      fullPath\n      webUrl\n      projectsCount\n      groupMembersCount\n      avatarUrl\n    }\n      }\n    "

# get_orgs_count prints message

    Code
      orgs_count <- test_rest_gitlab$get_orgs_count(verbose = TRUE)
    Message
      > [Host:GitLab][Engine:REST] Pulling number of all organizations 🏛️...

# get_orgs prints message

    Code
      gl_orgs_full_response <- test_graphql_gitlab$get_orgs(orgs_count = 3L, output = "full_table",
        verbose = TRUE, progress = FALSE)
    Message
      > [Host:GitLab][Engine:GraphQL] Pulling organizations 🏛️...

# if get_orgs runs into GraphQL error, it prints warning

    Code
      gl_orgs_error_response <- test_graphql_gitlab$get_orgs(orgs_count = 3L, output = "full_table",
        verbose = TRUE, progress = FALSE)
    Message
      > [Host:GitLab][Engine:GraphQL] Pulling organizations 🏛️...
      x GraphQL returned errors:
      i Your GraphQL does not recognize [groups] field.
      ! Check version of your GitLab.

# get_org prints proper message

    Code
      gl_org_response <- test_graphql_gitlab$get_org(org = org, verbose = TRUE)
    Message
      > [Host:GitLab][Engine:GraphQL] Pulling mbtests organization 🏛️...

---

    Code
      gl_org_response <- test_rest_gitlab$get_org(org = "mbtests", verbose = TRUE)
    Message
      > [Host:GitLab][Engine:REST] Pulling mbtests organization 🏛️...

# REST method prints message

    Code
      gl_orgs_rest_list <- test_rest_gitlab$get_orgs(orgs_count = 300L, verbose = TRUE,
        progress = FALSE)
    Message
      > [Host:GitLab][Engine:REST] Pulling organizations 🏛️...

# if get_orgs_from_host runs into GraphQL error, it switches to REST API

    Code
      gitlab_orgs_vec <- gitlab_test_host_priv_2$get_orgs_from_host(output = "only_names",
        verbose = TRUE)
    Message
      > [Host:GitLab][Engine:REST] Pulling number of all organizations 🏛️...
      i  organizations found.
      i Switching to REST API

---

    Code
      gitlab_orgs_table <- gitlab_test_host_priv_2$get_orgs_from_host(output = "full_table",
        verbose = TRUE)
    Message
      > [Host:GitLab][Engine:REST] Pulling number of all organizations 🏛️...
      i  organizations found.
      i Switching to REST API

# get_orgs_from_host prints message on number of organizations

    Code
      gitlab_orgs_table <- gitlab_testhost_priv$get_orgs_from_host(output = "full_table",
        verbose = TRUE)
    Message
      i 3 organizations found.

# get_orgs_from_orgs_and_repos prints message when turning to REST engine

    Code
      gitlab_orgs_from_orgs_table <- gitlab_testhost_priv$
      get_orgs_from_orgs_and_repos(verbose = TRUE)
    Message
      i Switching to REST API
      i Switching to REST API
      i Switching to REST API
      i Switching to REST API
      i Switching to REST API
      i Switching to REST API

# get_orgs_from_host returns error when GitHost is public

    This feature is not applicable for public hosts.

# commits_by_repo GitHub query is built properly

    Code
      gh_commits_from_repo_query
    Output
      [1] "\n      query GetCommitsFromRepo($repo: String!\n                               $org: String!\n                               $since: GitTimestamp\n                               $until: GitTimestamp){\n          repository(name: $repo, owner: $org) {\n            defaultBranchRef {\n              target {\n                ... on Commit {\n                  history(since: $since\n                          until: $until\n                          ) {\n                    pageInfo {\n                      hasNextPage\n                      endCursor\n                    }\n                    edges {\n                      node {\n                        ... on Commit {\n                          id\n                          committed_date: committedDate\n                          author {\n                            name\n                            user {\n                              name\n                              login\n                            }\n                          }\n                          additions\n                          deletions\n                          repository {\n                            url\n                          }\n                        }\n                      }\n                    }\n                  }\n                }\n              }\n            }\n          }\n        }"

# get_commits_from_orgs for GitHub prints messages

    Code
      gh_commits_from_orgs <- github_testhost_priv$get_commits_from_orgs(since = "2023-03-01",
        until = "2023-04-01", verbose = TRUE, progress = FALSE)
    Message
      > [Host:GitHub][Engine:GraphQl][Scope:r-world-devs] Pulling commits 🕒...

# get_commits_from_repos for GitHub prints messages

    Code
      gh_commits_from_repos <- github_testhost_priv$get_commits_from_repos(since = "2023-03-01",
        until = "2023-04-01", verbose = TRUE, progress = FALSE)
    Message
      > [Host:GitHub][Engine:GraphQl][Scope:test_org: 1 repos] Pulling commits 🕒...

# get_pull_requests gets data from storage for the second time

    Code
      pr_table <- test_gitstats$get_pull_requests(since = "2023-01-01", until = "2025-03-06",
        state = NULL, verbose = FALSE)
    Message
      ! Getting cached pull_requests data.
      i If you wish to pull the data from API once more, set `cache` parameter to `FALSE`.

# get_pull_requests() returns warning if PR table is empty

    Code
      pr_table <- test_gitstats$get_pull_requests(since = "2023-01-01", until = "2025-03-06",
        cache = FALSE, state = NULL, verbose = TRUE)
    Message
      i Cache set to FALSE, I will pull data from API.
      > Getting pull requests 🔀...
      ! No pull requests 🔀 found.

# get_pull_requests() prints data on time used

    Code
      pr_data <- get_pull_requests(test_gitstats, since = "2023-01-01", until = "2025-03-06",
        verbose = TRUE)
    Message
      v Data pulled in 0 secs

# issues_by_repo GitHub query is built properly

    Code
      gh_issues_from_repo_query
    Output
      [1] "\n      query getIssuesFromRepo ($repo: String!\n                               $org: String!) {\n          repository(name: $repo, owner: $org) {\n            issues(first: 100\n                   ) {\n              pageInfo {\n                hasNextPage\n      \t\t\t\t  endCursor\n      \t\t\t\t}\n              edges {\n                node {\n                  number\n                  title\n                  description: body\n                  created_at: createdAt\n                  closed_at: closedAt\n                  state\n                  url\n                  author {\n                    login\n                  }\n                }\n              }\n            }\n          }\n        }\n      "

# get_issues_from_orgs for GitHub prints messages

    Code
      gh_issues_from_orgs <- github_testhost_priv$get_issues_from_orgs(verbose = TRUE,
        progress = FALSE)
    Message
      > [Host:GitHub][Engine:GraphQl][Scope:r-world-devs] Getting issues 🐛...

# get_issues_from_repos for GitHub prints messages

    Code
      gh_issues_from_repos <- github_testhost_priv$get_issues_from_repos(verbose = TRUE,
        progress = FALSE)
    Message
      > [Host:GitHub][Engine:GraphQl][Scope:test_org: 1 repos] Getting issues 🐛...

# pr_by_repo GitLab query is built properly

    Code
      gl_pr_from_repo_query
    Output
      [1] "\n      query getPullRequestsFromRepo ($fullPath: ID!) {\n          project(fullPath: $fullPath) {\n            mergeRequests(first: 100, ) {\n              pageInfo {\n                endCursor\n                hasNextPage\n              }\n              edges {\n                node {\n                  number: iid\n                  created_at: createdAt\n                  merged_at: mergedAt\n                  state\n                  author {\n                    username\n                  }\n                  source_branch: sourceBranch\n                  target_branch: targetBranch\n                }\n              }\n            }\n          }\n        }\n      "

# get_pr_from_orgs for GitLab prints message

    Code
      gl_pr_from_orgs <- gitlab_testhost_priv$get_pr_from_orgs(verbose = TRUE,
        progress = FALSE)
    Message
      > Using cached repositories data...
      > [Host:GitLab][Engine:GraphQl][Scope:mbtests] Pulling pull requests 🔀...

# get_pr_from_repos for GitLab prints message

    Code
      gl_pr_from_repos <- gitlab_testhost_priv$get_pr_from_repos(verbose = TRUE,
        progress = FALSE)
    Message
      > [Host:GitLab][Engine:GraphQl][Scope:mbtests: 1 repos] Getting pull requests 🔀...

# get_issues gets data from storage for the second time

    Code
      release_logs_table <- test_gitstats$get_release_logs(since = "2023-08-01",
        until = "2023-09-30", verbose = FALSE)
    Message
      ! Getting cached release_logs data.
      i If you wish to pull the data from API once more, set `cache` parameter to `FALSE`.

# get_release_logs returns warning when releases table is empty

    Code
      release_logs_data <- test_gitstats$get_release_logs(since = "2023-08-01",
        until = "2023-09-30", cache = FALSE, verbose = TRUE)
    Message
      i Cache set to FALSE, I will pull data from API.
      > Pulling release logs 🏷️..
      ! No release logs 🏷️ found.

# get_release_logs prints time used to pull data

    Code
      release_logs_table <- get_release_logs(gitstats = test_gitstats, since = "2023-08-01",
        until = "2023-09-30", verbose = TRUE)
    Message
      v Data pulled in 0 secs

# get_orgs() pulls data from storage for the second time

    Code
      orgs_table <- test_gitstats$get_orgs()
    Message
      ! Getting cached organizations data.
      i If you wish to pull the data from API once more, set `cache` parameter to `FALSE`.

# get_orgs prints info on time used to pull data

    Code
      orgs_table <- get_orgs(test_gitstats, verbose = TRUE)
    Message
      v Data pulled in 0 secs

# orgs GitHub query is built properly

    Code
      gh_orgs_query
    Output
      [1] "query {\n          search(first: 100, type: USER, query: \"type:org\" ) {\n            pageInfo {\n               hasNextPage\n               endCursor\n            }\n            edges {\n              node{\n                ... on Organization {\n      name\n      description\n      login\n      url\n      repositories (first: 100) {\n        totalCount\n      }\n      membersWithRole(first: 100) {\n        totalCount\n      }\n      avatarUrl\n    }\n              }\n            }\n          }\n        }"

# org GitHub query is built properly

    Code
      gh_org_query
    Output
      [1] "\n        query GetOrg($org: String!) {\n          organization(login: $org) {\n      name\n      description\n      login\n      url\n      repositories (first: 100) {\n        totalCount\n      }\n      membersWithRole(first: 100) {\n        totalCount\n      }\n      avatarUrl\n    }\n        }"

# get_orgs prints message

    Code
      gh_orgs_full_response <- test_graphql_github$get_orgs(output = "full_table",
        verbose = TRUE)
    Message
      > [Host:GitHub][Engine:GraphQL] Pulling organizations 🏛️...

# get_orgs_from_host returns error when GitHost is public

    This feature is not applicable for public hosts.

# issues_by_repo GitLab query is built properly

    Code
      gl_issues_from_repo_query
    Output
      [1] "\n      query getIssuesFromRepo ($fullPath: ID!) {\n          project(fullPath: $fullPath) {\n            issues(first: 100\n                   ) {\n              pageInfo {\n                hasNextPage\n      \t\t\t\t  endCursor\n      \t\t\t\t}\n              edges {\n                node {\n                  number: iid\n                  title\n                  description\n                  created_at: createdAt\n                  closed_at: closedAt\n                  state\n                  url: webUrl\n                  author {\n                    login: username\n                  }\n                }\n              }\n            }\n          }\n        }\n      "

# file queries for GitHub are built properly

    Code
      gh_file_blobs_from_repo_query
    Output
      [1] "query GetFileBlobFromRepo($org: String!, $repo: String!, $expression: String!) {\n          repository(owner: $org, name: $repo) {\n            repo_id: id\n            repo_name: name\n            repo_url: url\n            defaultBranchRef {\n              target {\n                ... on Commit {\n                  oid\n                }\n              }\n            }\n            file: object(expression: $expression) {\n              ... on Blob {\n                text\n                byteSize\n                oid\n              }\n            }\n          }\n      }"

# get_files_content_from_orgs for GitHub prints messages

    Code
      gh_files_from_orgs <- github_testhost_priv$get_files_content_from_orgs(
        file_path = "DESCRIPTION", verbose = TRUE)
    Message
      > [Host:GitHub][Engine:GraphQl][Scope:r-world-devs] Pulling files 📄 content: [DESCRIPTION]...

# get_files_content_from_repos for GitHub prints messages

    Code
      gh_files_from_repos <- github_testhost_priv$get_files_content_from_repos(
        file_path = "DESCRIPTION", verbose = TRUE)
    Message
      > [Host:GitHub][Engine:GraphQl][Scope:test_org: 1 repos] Pulling files 📄 content: [DESCRIPTION]...

# get_files_content makes use of files_structure

    Code
      files_content <- github_testhost_priv$get_files_content_from_files_structure(
        files_structure = test_mocker$use("gh_files_structure_from_orgs"), verbose = TRUE,
        progress = FALSE)
    Message
      > [Host:GitHub][Engine:GraphQl][Scope:r-world-devs] Pulling files 📄 from files structure...

# get_files_content skips if no files found in files_structure

    Code
      files_content <- github_testhost_priv$get_files_content_from_files_structure(
        files_structure = test_mocker$use("gh_empty_files_structure"), verbose = TRUE,
        progress = FALSE)
    Message
      ! [GitHub] No files 📄 found. Skipping pulling files content.

# get_files_structure_from_orgs prints messages without pattern

    Code
      gl_files_structure <- gitlab_testhost_priv$get_files_structure_from_orgs(
        pattern = NULL, depth = Inf, verbose = TRUE, progress = FALSE)
    Message
      > Using cached repositories data...
      > [Host:GitLab][Engine:REST][Scope:mbtests] Pulling repos 🌳...

# get_files_structure_from_orgs warns when no structure found

    Code
      gl_files_structure <- gitlab_testhost_priv$get_files_structure_from_orgs(
        pattern = NULL, depth = Inf, verbose = TRUE, progress = FALSE)
    Message
      > Using cached repositories data...
      > [Host:GitLab][Engine:REST][Scope:mbtests] Pulling repos 🌳...
      ! For GitLab no files structure found.

# get_files_structure_from_repos prints message

    Code
      gl_files_structure_from_repos <- gitlab_testhost_priv$
        get_files_structure_from_repos(pattern = "\\.md", depth = 1L, verbose = TRUE,
        progress = FALSE)
    Message
      > [Host:GitLab][Engine:REST][Scope:test_group: 0 repos] Pulling repos 🌳 [files matching pattern: '\.md']...

# GitStats makes use of stored data

    Code
      repos_trees <- test_gitstats$get_repos_trees(pattern = NULL, depth = Inf,
        cache = TRUE, verbose = TRUE, progress = FALSE)
    Message
      ! Getting cached repos_trees data.
      i If you wish to pull the data from API once more, set `cache` parameter to `FALSE`.

# GitStats prints warning when no repos trees found

    Code
      repos_trees <- test_gitstats$get_repos_trees(pattern = NULL, depth = Inf,
        cache = FALSE, verbose = TRUE, progress = FALSE)
    Message
      i Cache set to FALSE, I will pull data from API.
      > Pulling repositories 🌳 structure...
      ! No repos 🌳 trees found.

# get_repos_urls_from_orgs prepares web repo_urls vector

    Code
      gh_repos_urls_from_orgs <- github_testhost_priv$get_repos_urls_from_orgs(type = "web",
        verbose = TRUE, progress = FALSE)
    Message
      > [Host:GitHub][Engine:REST][Scope:test_org] Pulling repositories 🌐 (URLs)...

# get_repos_urls_from_repos prepares web repo_urls vector

    Code
      gh_repos_urls <- github_testhost_priv$get_repos_urls_from_repos(type = "web",
        verbose = TRUE, progress = FALSE)
    Message
      > [Host:GitHub][Engine:REST][Scope:test_org/TestRepo] Pulling repositories 🌐 (URLs)...

# get_all_repos_urls prepares web repo_urls vector

    Code
      gh_repos_urls <- github_testhost_priv$get_all_repos_urls(type = "web", verbose = TRUE)

# get_commits() pulls data from storage for the second time

    Code
      commits_table <- test_gitstats$get_commits(since = "2023-06-15", until = "2023-06-30",
        verbose = FALSE)
    Message
      ! Getting cached commits data.
      i If you wish to pull the data from API once more, set `cache` parameter to `FALSE`.

# get_commits prints warning when commits table is empty

    Code
      commits_table <- test_gitstats$get_commits(since = "2023-06-15", until = "2023-06-30",
        cache = FALSE, verbose = TRUE)
    Message
      i Cache set to FALSE, I will pull data from API.
      > Pulling commits 🕒...
      ! No commits 🕒 found.

# get_commits() returns error when since is not defined

    You need to pass date to `since` parameter.

# get_commits() prints time used

    Code
      commits <- get_commits(test_gitstats, since = "2023-06-15", until = "2023-06-30",
        verbose = TRUE)
    Message
      v Data pulled in 0 secs

# `get_commits_authors_handles_and_names()` prints messages

    Code
      gl_commits_table <- test_rest_gitlab$get_commits_authors_handles_and_names(
        commits_table = test_mocker$use("gl_commits_table"), verbose = TRUE)
    Message
      > [Host:GitLab][Engine:REST] Looking up for authors' 🙍 names and logins...

# get_commits_from_orgs works

    Code
      gl_commits_table <- gitlab_testhost_priv$get_commits_from_orgs(since = "2023-03-01",
        until = "2023-04-01", verbose = TRUE, progress = FALSE)
    Message
      > [mbtests] Pulling repositories 🌐 data...
      > Caching repositories for [mbtests]...
      > [Host:GitLab][Engine:REST][Scope:mbtests] Pulling commits 🕒...

# get_commits_from_repos works

    Code
      gl_commits_table <- gitlab_testhost_priv$get_commits_from_repos(since = "2023-03-01",
        until = "2023-04-01", verbose = TRUE, progress = FALSE)
    Message
      > [Host:GitLab][Engine:REST][Scope:test_org: 1 repos] Pulling commits 🕒...

# GitStats get_users prints messages

    Code
      users_result <- test_gitstats$get_users(c("test_user1", "test_user2"), verbose = TRUE)
    Message
      > Pulling users 🙍 data...

# get_users prints info on data used to pull data

    Code
      users_result <- get_users(gitstats = test_gitstats, c("test_user1",
        "test_user2"), verbose = TRUE)
    Message
      v Data pulled in 0 secs

# Set connection returns appropriate messages

    Code
      set_github_host(gitstats = test_gitstats, token = Sys.getenv("GITHUB_PAT"),
      orgs = c("openpharma", "r-world-devs"))
    Message
      > Checking owners...
      v Set connection to GitHub.

---

    Code
      set_gitlab_host(test_gitstats, token = Sys.getenv("GITLAB_PAT_PUBLIC"), orgs = c(
        "mbtests"))
    Message
      > Checking owners...
      v Set connection to GitLab.

# When empty token for GitHub, GitStats pulls default token

    Code
      test_gitstats <- set_github_host(create_gitstats(), orgs = c("openpharma",
        "r-world-devs"))
    Message
      i Using PAT from GITHUB_PAT envar.
      > Checking owners...
      v Set connection to GitHub.

# When empty token for GitLab, GitStats pulls default token

    Code
      withr::with_envvar(new = c(GITLAB_PAT = Sys.getenv("GITLAB_PAT_PUBLIC")), {
        test_gitstats <- set_gitlab_host(create_gitstats(), orgs = "mbtests")
      })
    Message
      i Using PAT from GITLAB_PAT envar.
      > Checking owners...
      v Set connection to GitLab.

# Set GitHub host with particular repos vector instead of orgs

    Code
      set_github_host(test_gitstats, token = Sys.getenv("GITHUB_PAT"), repos = c(
        "r-world-devs/GitStats", "r-world-devs/shinyCohortBuilder",
        "openpharma/GithubMetrics", "openpharma/DataFakeR"))
    Message
      > Checking repositories...
      v Set connection to GitHub.

# Set GitLab host with particular repos vector instead of orgs

    Code
      set_gitlab_host(test_gitstats, token = Sys.getenv("GITLAB_PAT_PUBLIC"), repos = c(
        "mbtests/gitstatstesting", "mbtests/gitstats-testing-2"))
    Message
      > Checking repositories...
      v Set connection to GitLab.

# When no organizations or repositories are set information is printed

    Code
      set_github_host(test_gitstats, token = Sys.getenv("GITHUB_PAT"), verbose = TRUE)
    Message
      i No `orgs` nor `repos` specified.
      i Searching scope set to [all].
      v Set connection to GitHub.

# Error shows, when wrong input is passed when setting connection and host is not passed

    x Token exists but does not grant access.
    i Check if you use correct token.
    ! Scope that is needed: [read_api].

# Error pops out, when two clients of the same url api are passed as input

    Code
      set_github_host(set_github_host(test_gitstats, token = Sys.getenv("GITHUB_PAT"),
      orgs = "pharmaverse"), token = Sys.getenv("GITHUB_PAT"), orgs = "openpharma")
    Message
      > Checking owners...
      v Set connection to GitHub.
      > Checking owners...
      v Set connection to GitHub.
    Condition
      Error:
      x You can not provide two hosts of the same API urls.

# Error pops out when `org` does not exist

    Code
      test_gitstats <- set_github_host(create_gitstats(), token = Sys.getenv(
        "GITHUB_PAT"), orgs = c("openparma"))
    Message
      > Checking owners...
    Condition
      Error in `purrr::map()`:
      i In index: 1.
      Caused by error:
      x Org/user you provided does not exist or its name was passed in a wrong way: openparma
      ! Please type your org/user name the way you see it in web URL.

---

    Code
      test_gitstats <- set_gitlab_host(create_gitstats(), token = Sys.getenv(
        "GITLAB_PAT_PUBLIC"), orgs = c("openparma", "mbtests"))
    Message
      > Checking owners...
    Condition
      Error in `purrr::map()`:
      i In index: 1.
      Caused by error:
      x Org/user you provided does not exist or its name was passed in a wrong way: openparma
      ! Please type your org/user name the way you see it in web URL.

---

    Code
      test_gitstats <- set_github_host(create_gitstats(), token = Sys.getenv(
        "GITHUB_PAT"), orgs = c("openpharma", "r_world_devs"))
    Message
      > Checking owners...
    Condition
      Error in `purrr::map()`:
      i In index: 2.
      Caused by error:
      x Org/user you provided does not exist or its name was passed in a wrong way: r_world_devs
      ! Please type your org/user name the way you see it in web URL.

# When wrong orgs and repos are passed they are excluded but host is created

    Code
      test_gitstats <- set_github_host(create_gitstats(), orgs = c("openpharma",
        "r_world_devs"), repos = c("r-world-devs/GitStats", "r-world-devs/GitMetrics"),
      verbose = TRUE, .error = FALSE)
    Message
      i Using PAT from GITHUB_PAT envar.
      > Checking owners...
      ! Org/user you provided does not exist: r_world_devs
      > Checking repositories...
      x HTTP 404 Not Found.
      ! Repository you provided does not exist: https://api.github.com/repos/r-world-devs/GitMetrics
      v Set connection to GitHub.

# get_files gets data from storage for the second time

    Code
      files_table <- test_gitstats$get_files(pattern = NULL, depth = Inf, file_path = "meta_data.yaml",
        verbose = FALSE)
    Message
      ! Getting cached files data.
      i If you wish to pull the data from API once more, set `cache` parameter to `FALSE`.

# get_files prints message when files structure is empty

    Code
      files_table <- test_gitstats$get_files(pattern = "\\.test", depth = 0L,
        verbose = FALSE, progress = FALSE)
    Message
      > Pulling files 📄 content...
      ! [GitHub] No files 📄 found. Skipping pulling files content.
      ! [GitLab] No files 📄 found. Skipping pulling files content.

# error shows when file_path and pattern are defined at the same time

    Please choose either `pattern` or `file_path`.

# get_files prints time used to pull data

    Code
      files_table <- get_files(test_gitstats, file_path = "meta_data.yaml", verbose = TRUE,
        progress = FALSE)
    Message
      v Data pulled in 0 secs

# releases query is built properly

    Code
      gh_releases_query
    Output
      [1] "query GetReleasesFromRepo ($org: String!, $repo: String!) {\n          repository(owner:$org, name:$repo){\n            name\n            url\n            releases (last: 100) {\n              nodes {\n                name\n                tagName\n                publishedAt\n                url\n                description\n              }\n            }\n          }\n        }"

# `set_searching_scope` does not throw error when `orgs` or `repos` are defined

    Code
      gitlab_testhost_priv$set_searching_scope(orgs = "mbtests", repos = NULL,
        verbose = TRUE)

---

    Code
      gitlab_testhost_priv$set_searching_scope(orgs = NULL, repos = "mbtests/GitStatsTesting",
        verbose = TRUE)

# `set_searching_scope` sets scope to whole host

    Code
      gitlab_testhost_priv$set_searching_scope(orgs = NULL, repos = NULL, verbose = TRUE)
    Message
      i No `orgs` nor `repos` specified.
      i Searching scope set to [all].

# When token is empty throw error

    Code
      github_testhost_priv$check_token("")
    Condition
      Error in `github_testhost_priv$check_token()`:
      i No token provided.

# `check_token()` prints error when token exists but does not grant access

    x Token exists but does not grant access.
    i Check if you use correct token.
    ! Scope that is needed: [public_repo, read:org, read:user].

# check_endpoint returns error if they are not correct

    x Repository you provided does not exist or its name was passed in a wrong way: https://api.github.com/repos/r-worlddevs/GitStats
    ! Please type your repository name as you see it in web URL.
    i E.g. do not use spaces. Repository names as you see on the page may differ from their web 'address'.
    i Repository names should be provided as full paths: "org/repo_name".

# check_endpoint returns warning if they are not correct

    Code
      check <- github_testhost_priv$check_endpoint(endpoint = "https://api.github.com/repos/r-worlddevs/GitStats",
        type = "Repository", verbose = TRUE, .error = FALSE)
    Message
      x HTTP 404 Not Found.
      ! Repository you provided does not exist: https://api.github.com/repos/r-worlddevs/GitStats

# `set_default_token` sets default token for public GitHub

    Code
      default_token <- github_testhost_priv$set_default_token(verbose = TRUE)
    Message
      i Using PAT from GITHUB_PAT envar.

# `set_default_token` sets default token for GitLab

    Code
      withr::with_envvar(new = c(GITLAB_PAT = Sys.getenv("GITLAB_PAT_PUBLIC")), {
        default_token <- gitlab_testhost_priv$set_default_token(verbose = TRUE)
      })
    Message
      i Using PAT from GITLAB_PAT envar.

# user query is built properly

    Code
      gl_user_query
    Output
      [1] "\n        query GetUser($user: String!) {\n          user(username: $user) {\n            id\n            name\n            login: username\n            email: publicEmail\n            location\n            starred_repos: starredProjects {\n              count\n            }\n            pull_requests: authoredMergeRequests {\n              count\n            }\n            reviews: reviewRequestedMergeRequests {\n              count\n            }\n            avatar_url: avatarUrl\n            web_url: webUrl\n          }\n        }\n      "

# get_pull_requests_stats returns error when issues is not pr_data object

    x `pull_requests` must be a `gitstats_issues` object.
    i Pull first your pull_requests with `get_pull_requests()` function.

# releases query is built properly

    Code
      gl_releases_query
    Output
      [1] "query GetReleasesFromRepo($project_path: ID!) {\n              project(fullPath: $project_path) {\n                name\n                webUrl\n    \t\t\t\t\t\treleases {\n                  nodes{\n                    name\n                    tagName\n                    releasedAt\n                    links {\n                      selfUrl\n                    }\n                    description\n                  }\n                }\n              }\n          }"

# `get_release_logs_from_orgs()` prints proper message

    Code
      releases_from_orgs <- gitlab_testhost_priv$get_release_logs_from_orgs(verbose = TRUE,
        progress = FALSE)
    Message
      > [Host:GitLab][Engine:GraphQl][Scope:mbtests] Pulling release logs 🏷️...

# `get_release_logs_from_repos()` works

    Code
      releases_from_repos <- gitlab_testhost_priv$get_release_logs_from_repos(
        verbose = TRUE, progress = FALSE)
    Message
      > [Host:GitLab][Engine:GraphQl][Scope:test_org: 1 repos] Pulling release logs 🏷️...

# pr_by_repo GitHub query is built properly

    Code
      gh_pr_from_repo_query
    Output
      [1] "query GetPullRequestsFromRepo($org: String!, $repo: String!) {\n          repository(owner: $org, name: $repo) {\n            pullRequests(first: 100, ) {\n              pageInfo {\n                 hasNextPage\n                endCursor\n              }\n              edges {\n                node {\n                  number\n                  created_at: createdAt\n                  merged_at: mergedAt\n                  state\n                  author {\n                    login\n                  }\n                  source_branch: headRefName\n                  target_branch: baseRefName\n                }\n              }\n            }\n          }\n        }"

# get_pr_from_orgs for GitHub prints messages

    Code
      gh_pr_from_orgs <- github_testhost_priv$get_pr_from_orgs(verbose = TRUE,
        progress = FALSE)
    Message
      > [Host:GitHub][Engine:GraphQl][Scope:r-world-devs] Pulling pull requests 🔀...

# get_pr_from_repos for GitHub prints messages

    Code
      gh_pr_from_repos <- github_testhost_priv$get_pr_from_repos(verbose = TRUE,
        progress = FALSE)
    Message
      > [Host:GitHub][Engine:GraphQl][Scope:test_org: 1 repos] Getting pull requests 🔀...

# get_files_structure_from_orgs() prints message

    Code
      gh_files_structure_from_orgs <- github_testhost_priv$
        get_files_structure_from_orgs(pattern = "\\.md|\\.qmd|\\.Rmd", depth = Inf,
        verbose = TRUE)
    Message
      > [r-world-devs] Pulling repositories 🌐 data...
      > Caching repositories for [r-world-devs]...
      > [Host:GitHub][Engine:REST][Scope:r-world-devs] Pulling repos 🌳 [files matching pattern: '\.md|\.qmd|\.Rmd']...

---

    Code
      gh_files_structure_from_orgs <- github_testhost_priv$
        get_files_structure_from_orgs(pattern = NULL, depth = Inf, verbose = TRUE)
    Message
      > [r-world-devs] Pulling repositories 🌐 data...
      > Caching repositories for [r-world-devs]...
      > [Host:GitHub][Engine:REST][Scope:r-world-devs] Pulling repos 🌳...

# get_files_structure_from_repos() prints message

    Code
      gh_files_structure_from_repos <- github_testhost_priv$
        get_files_structure_from_repos(pattern = "\\.md|\\.qmd|\\.Rmd", depth = Inf,
        verbose = TRUE)
    Message
      > [test_org] Pulling repositories 🌐 data...
      > Caching repositories for [test_org]...
      > [Host:GitHub][Engine:REST][Scope:test_org: 1 repos] Pulling repos 🌳 [files matching pattern: '\.md|\.qmd|\.Rmd']...

---

    Code
      gh_files_structure_from_repos <- github_testhost_priv$
        get_files_structure_from_repos(pattern = NULL, depth = Inf, verbose = TRUE)
    Message
      > [test_org] Pulling repositories 🌐 data...
      > Caching repositories for [test_org]...
      > [Host:GitHub][Engine:REST][Scope:test_org: 1 repos] Pulling repos 🌳...

# when files_structure is empty, appropriate message is returned

    Code
      github_testhost_priv$get_files_structure_from_repos(pattern = "\\.png", depth = 1L,
        verbose = TRUE)
    Message
      ! For GitHub no files 🌳 structure found.
    Output
      named list()

# GitStats prints empty fields.

    Code
      test_gitstats
    Output
      A GitStats object for 0 hosts: 
      Hosts: 
      Scanning scope: 
       Organizations: [0] 
       Repositories: [0] 
      Storage [local]: <no data in storage>

# GitStats prints the proper info when connections are added.

    Code
      test_gitstats
    Output
      A GitStats object for 2 hosts: 
      Hosts: https://api.github.com, https://gitlab.com/api/v4
      Scanning scope: 
       Organizations: [2] github_test_org, gitlab_test_group
       Repositories: [0] 
      Storage [local]: <no data in storage>

# GitStats prints the proper info when repos are passed instead of orgs.

    Code
      test_gitstats
    Output
      A GitStats object for 2 hosts: 
      Hosts: https://api.github.com, https://gitlab.com/api/v4
      Scanning scope: 
       Organizations: [0] 
       Repositories: [4] r-world-devs/GitStats, openpharma/GithubMetrics, mbtests/gitstatstesting, mbtests/gitstats-testing-2
      Storage [local]: <no data in storage>

# check_for_host returns error when no hosts are passed

    Add first your hosts with `set_github_host()` or `set_gitlab_host()`.

# check_params_conflict returns error

    ! Passing files to `in_files` parameter works only when you search code with `with_code` parameter.
    i If you want to search for repositories with [DESCRIPTION] files you should instead use `with_files` parameter.

---

    x Both `with_code` and `with_files` parameters are defined.
    ! Use either `with_code` of `with_files` parameter.
    i If you want to search for [shiny] code in given files - use `in_files` parameter together with `with_code` instead.

# print_storage_attribute

    Code
      test_gitstats_priv$print_storage_attribute(storage_data = test_mocker$use(
        "commits_table"), storage_name = "commits")
    Output
      <cli_ansi_string>
      [1] [date range: 2023-06-15 - 2023-06-30]

---

    Code
      test_gitstats_priv$print_storage_attribute(storage_data = test_mocker$use(
        "release_logs_table"), storage_name = "release_logs")
    Output
      <cli_ansi_string>
      [1] [date range: 2023-08-01 - 2023-09-30]

---

    Code
      test_gitstats_priv$print_storage_attribute(storage_data = test_mocker$use(
        "repos_trees"), storage_name = "repos_trees")
    Output
      <cli_ansi_string>
      [1] [file pattern: ]

# subgroups are cleanly printed in GitStats

    Code
      test_gitstats
    Output
      A GitStats object for 1 hosts: 
      Hosts: https://gitlab.com/api/v4
      Scanning scope: 
       Organizations: [1] mbtests/subgroup
       Repositories: [0] 
      Storage [local]: <no data in storage>

# get_issues gets data from storage for the second time

    Code
      issues_table <- test_gitstats$get_issues(since = "2023-01-01", until = "2025-03-06",
        state = NULL, verbose = FALSE)
    Message
      ! Getting cached issues data.
      i If you wish to pull the data from API once more, set `cache` parameter to `FALSE`.

# get_issues() returns warning if issues table is empty

    Code
      issues_table <- test_gitstats$get_issues(since = "2023-01-01", until = "2025-03-06",
        cache = FALSE, state = NULL, verbose = TRUE)
    Message
      i Cache set to FALSE, I will pull data from API.
      > Getting issues 🐛...
      ! No issues 🐛 found.

# get_issues() prints data on time used

    Code
      issues_data <- get_issues(test_gitstats, since = "2023-01-01", until = "2025-03-06",
        verbose = TRUE)
    Message
      v Data pulled in 0 secs

# get_repos returns warning when repositories table is empty

    Code
      repos_data <- test_gitstats$get_repos(cache = FALSE, verbose = TRUE)
    Message
      i Cache set to FALSE, I will pull data from API.
      > Pulling repositories 🌐 data...
      ! No repositories 🌐 found.

# get_repos prints time used to pull data

    Code
      repos_data <- get_repos(gitstats = test_gitstats, verbose = TRUE)
    Message
      v Data pulled in 0 secs

# repos_by_org query is built properly

    Code
      gh_repos_by_org_query
    Output
      [1] "\n        query GetReposByOrg($login: String!) {\n          repositoryOwner(login: $login) {\n            ... on Organization {\n              \n      repositories(first: 100) {\n        totalCount\n        pageInfo {\n          endCursor\n          hasNextPage\n        }\n        nodes {\n      repo_id: id\n      repo_name: name\n      repo_path: name\n      repo_fullpath: nameWithOwner\n      default_branch: defaultBranchRef {\n        name\n      }\n      stars: stargazerCount\n      forks: forkCount\n      created_at: createdAt\n      last_activity_at: pushedAt\n      languages(first: 5) {\n        nodes {\n          name\n        }\n      }\n      issues_open: issues(first: 100, states: [OPEN]) {\n        totalCount\n      }\n      issues_closed: issues(first: 100, states: [CLOSED]) {\n        totalCount\n      }\n      organization: owner {\n        login\n      }\n      repo_url: url\n      defaultBranchRef {\n        target {\n          ... on Commit {\n            oid\n          }\n        }\n      }\n    }\n      }\n            }\n          }\n        }"

# repos_by_user query is built properly

    Code
      gh_repos_by_user_query
    Output
      [1] "\n        query GetUsersRepos($login: String!){\n          user(login: $login) {\n            \n      repositories(first: 100) {\n        totalCount\n        pageInfo {\n          endCursor\n          hasNextPage\n        }\n        nodes {\n      repo_id: id\n      repo_name: name\n      repo_path: name\n      repo_fullpath: nameWithOwner\n      default_branch: defaultBranchRef {\n        name\n      }\n      stars: stargazerCount\n      forks: forkCount\n      created_at: createdAt\n      last_activity_at: pushedAt\n      languages(first: 5) {\n        nodes {\n          name\n        }\n      }\n      issues_open: issues(first: 100, states: [OPEN]) {\n        totalCount\n      }\n      issues_closed: issues(first: 100, states: [CLOSED]) {\n        totalCount\n      }\n      organization: owner {\n        login\n      }\n      repo_url: url\n      defaultBranchRef {\n        target {\n          ... on Commit {\n            oid\n          }\n        }\n      }\n    }\n      }\n          }\n        }"

# repos_by_ids query is built properly

    Code
      gh_repos_by_ids_query
    Output
      [1] "\n        query GetReposByIds($ids: [ID!]!) {\n          nodes(ids: $ids) {\n            ... on Repository {\n      repo_id: id\n      repo_name: name\n      repo_path: name\n      repo_fullpath: nameWithOwner\n      default_branch: defaultBranchRef {\n        name\n      }\n      stars: stargazerCount\n      forks: forkCount\n      created_at: createdAt\n      last_activity_at: pushedAt\n      languages(first: 5) {\n        nodes {\n          name\n        }\n      }\n      issues_open: issues(first: 100, states: [OPEN]) {\n        totalCount\n      }\n      issues_closed: issues(first: 100, states: [CLOSED]) {\n        totalCount\n      }\n      organization: owner {\n        login\n      }\n      repo_url: url\n      defaultBranchRef {\n        target {\n          ... on Commit {\n            oid\n          }\n        }\n      }\n    \n            }\n          }\n        }\n        "

# parse_search_response prints message

    Code
      gh_repos_raw_output <- github_testhost_priv$parse_search_response(
        search_response = test_mocker$use("gh_search_repos_for_code"), org = gh_org,
        output = "raw", verbose = TRUE)
    Message
      > Parsing search response with GraphQL...

# `get_repos_with_code_from_orgs()` pulls raw response

    Code
      repos_with_code_from_orgs_raw <- github_testhost_priv$
        get_repos_with_code_from_orgs(code = "shiny", in_files = c("DESCRIPTION",
        "NAMESPACE"), language = "R", output = "raw", verbose = TRUE)
    Message
      > [Host:GitHub][Engine:REST][Scope:r-world-devs] Pulling repositories 🌐...

# `get_repos_with_code_from_host()` pulls and parses output into table

    Code
      repos_with_code_from_host_table <- github_testhost_priv$
        get_repos_with_code_from_host(code = "DESCRIPTION", in_path = TRUE, output = "table_full",
        verbose = TRUE)
    Message
      > [Host:GitHub][Engine:REST] Pulling repositories 🌐...

# `get_repos_with_code_from_repos()` works

    Code
      repos_with_code_from_repos_full <- github_testhost_priv$
        get_repos_with_code_from_repos(code = "tests", output = "table", verbose = TRUE)
    Message
      > [Host:GitHub][Engine:REST][Scope:] Pulling repositories 🌐...

# get_repos_from_repos works

    Code
      gh_repos_individual <- github_testhost_priv$get_repos_from_repos(add_languages = TRUE,
        verbose = TRUE, progress = FALSE)
    Message
      > [Host:GitHub][Engine:GraphQl][Scope:r-world-devs: 1 repos] Pulling repositories 🌐...

# `get_repos_contributors()` works on GitHost level

    Code
      gh_repos_with_contributors <- github_testhost_priv$get_repos_contributors(
        repos_table = test_mocker$use("gh_repos_table_with_platform"), verbose = TRUE,
        progress = FALSE)
    Message
      > [Host:GitHub][Engine:REST] Pulling contributors 🙍...

# `get_repos_data` pulls repos data from cache for the second time

    Code
      gh_repos_data <- github_testhost_priv$get_repos_data(org = "r-world-devs",
        verbose = TRUE)
    Message
      > [r-world-devs] Using cached repositories data...

# repos queries are built properly

    Code
      gl_repos_by_org_query
    Output
      [1] "\n        query GetReposByOrg($org: ID! $repo_cursor: String!) {\n          group(fullPath: $org) {\n            projects(first: 100 after: $repo_cursor) {\n            \n      count\n      pageInfo {\n        hasNextPage\n        endCursor\n      }\n      edges {\n        node {\n          repo_id: id\n          repo_name: name\n          repo_path: path\n          repo_fullpath: fullPath\n          ... on Project {\n            repository {\n              rootRef\n              lastCommit {\n                sha\n              }\n            }\n          }\n          stars: starCount\n          forks: forksCount\n          created_at: createdAt\n          last_activity_at: lastActivityAt\n          languages {\n            name\n          }\n          issues: issueStatusCounts {\n            all\n            closed\n            opened\n          }\n          namespace {\n            path: fullPath\n          }\n          repo_url: webUrl\n        }\n      }\n            }\n          }\n        }"

---

    Code
      gl_repos_query
    Output
      [1] "\n        query GetRepo($projects_ids: [ID!]!) {\n          projects(ids: $projects_ids first:100) {\n          \n      count\n      pageInfo {\n        hasNextPage\n        endCursor\n      }\n      edges {\n        node {\n          repo_id: id\n          repo_name: name\n          repo_path: path\n          repo_fullpath: fullPath\n          ... on Project {\n            repository {\n              rootRef\n              lastCommit {\n                sha\n              }\n            }\n          }\n          stars: starCount\n          forks: forksCount\n          created_at: createdAt\n          last_activity_at: lastActivityAt\n          languages {\n            name\n          }\n          issues: issueStatusCounts {\n            all\n            closed\n            opened\n          }\n          namespace {\n            path: fullPath\n          }\n          repo_url: webUrl\n        }\n      }\n          }\n        }"

# repo_by_fullpath query is built properly

    Code
      gl_repo_by_fullpath_query
    Output
      [1] "\n        query GetRepoByFullPath($fullPath: ID!) {\n          project(fullPath: $fullPath) {\n            \n          repo_id: id\n          repo_name: name\n          repo_path: path\n          repo_fullpath: fullPath\n          ... on Project {\n            repository {\n              rootRef\n              lastCommit {\n                sha\n              }\n            }\n          }\n          stars: starCount\n          forks: forksCount\n          created_at: createdAt\n          last_activity_at: lastActivityAt\n          languages {\n            name\n          }\n          issues: issueStatusCounts {\n            all\n            closed\n            opened\n          }\n          namespace {\n            path: fullPath\n          }\n          repo_url: webUrl\n          }\n        }"

# error handler prints proper messages

    Code
      output <- test_graphql_gitlab_priv$handle_graphql_error(responses_list = test_mocker$
        use("repos_graphql_error"), verbose = TRUE)
    Message
      x GraphQL returned errors:
      i Your GraphQL does not recognize [count and languages] fields.
      ! Check version of your GitLab.

# get_repos prints messages when falling back to batching

    Code
      gitlab_repos <- test_graphql_gitlab$get_repos(repos_ids = c("test_id_1",
        "test_id_2"), verbose = TRUE)
    Message
      ! GraphQL limit error when pulling repos by IDs (2 repos).
      i I will pull repos in batches.

# `search_for_code()` works

    Code
      gl_search_repos_by_code <- test_rest_gitlab$search_for_code(code = "test",
        filename = "TESTFILE", verbose = TRUE, page_max = 2)
    Message
      > Searching for code [test]...

# `search_for_code()` handles the 1000 response limit

    Code
      gl_search_repos_by_code <- test_rest_gitlab$search_for_code(code = "test",
        filename = "TESTFILE", verbose = TRUE, page_max = 2)
    Message
      > Searching for code [test]...
      ! Reached 10 thousand response limit.

# `search_repos_for_code()` works

    Code
      gl_search_repos_by_code <- test_rest_gitlab$search_repos_for_code(code = "test",
        repos = "TestRepo", filename = "TESTFILE", verbose = TRUE, page_max = 2)
    Message
      > Searching for code [test]...

# get_repos_from_org prints proper message

    Code
      gl_repos_from_orgs <- gitlab_testhost_priv$get_repos_from_orgs(add_languages = TRUE,
        verbose = TRUE, progress = FALSE)
    Message
      > [Host:GitLab][Engine:GraphQl][Scope:mbtests] Pulling repositories 🌐...

# GitLab Host prints message when turning to REST engine (from orgs)

    Code
      gl_repos_from_orgs <- gitlab_testhost_priv$get_repos_from_orgs(add_languages = TRUE,
        verbose = TRUE, progress = FALSE)
    Message
      > [Host:GitLab][Engine:GraphQl][Scope:mbtests] Pulling repositories 🌐...
      i Switching to REST API
      > [Host:GitLab][Engine:REST][Scope:mbtests] Pulling repositories 🌐...

# `get_repos_from_repos()` prints proper message

    Code
      gl_repos_from_repos <- gitlab_testhost_priv$get_repos_from_repos(add_languages = TRUE,
        verbose = TRUE, progress = FALSE)
    Message
      > [Host:GitLab][Engine:GraphQl][Scope:mbtests/gitstatstesting] Pulling repositories 🌐...

# `get_repos_data` uses cached data from org

    Code
      gl_repos_data <- gitlab_testhost_priv$get_repos_data(org = "mbtests", verbose = TRUE)
    Message
      > Using cached repositories data...

# get_repos_data prints message when turns to REST engine

    Code
      gl_repos_data <- gitlab_testhost_priv$get_repos_data(org = "test_org", verbose = TRUE)
    Message
      > [test_org] Pulling repositories 🌐 data...
      i Switching to REST API...
      > Caching repositories for [test_org]...

# `fill_repos_commit_sha()` fills missing commit_sha via REST

    Code
      repos_commit_sha <- gitlab_testhost_fill$fill_repos_commit_sha(repos_table,
        verbose = TRUE)
    Message
      > [Host:GitLab][Engine:REST] Fetching missing commit SHAs for 1 repo...

# file queries for GitLab are built properly

    Code
      gl_files_query
    Output
      [1] "query GetFilesByOrg($org: ID!, $file_paths: [String!]!) {\n            group(fullPath: $org) {\n              projects(first: 100) {\n          count\n          pageInfo {\n            hasNextPage\n            endCursor\n          }\n          edges {\n            node {\n              name\n              path\n              id\n              webUrl\n              repository {\n                blobs(paths: $file_paths) {\n                  nodes {\n                    path\n                    rawBlob\n                    size\n                    oid\n                  }\n                }\n                lastCommit {\n                  sha\n                }\n              }\n            }\n          }\n        }\n      }\n    }"

---

    Code
      gl_file_blobs_from_repo_query
    Output
      [1] "\n      query GetFilesByRepo($fullPath: ID!, $file_paths: [String!]!) {\n        project(fullPath: $fullPath) {\n          name\n          path\n          id\n          webUrl\n          repository {\n            blobs(paths: $file_paths) {\n              nodes {\n                path\n                rawBlob\n                size\n                oid\n              }\n            }\n            lastCommit {\n              sha\n            }\n          }\n        }\n      }\n      "

# Gitlab GraphQL prints messages when switching to pulling files per repositories

    Code
      gitlab_files_response_by_repos <- test_graphql_gitlab$get_files_from_org(org = "mbtests",
        owner_type = "organization", repos = NULL, file_paths = c(
          "project_metadata.yaml", "README.md"), host_files_structure = NULL,
        verbose = TRUE)
    Message
      i I will switch to pulling files per repository.

# GitLab GraphQL switches to iteration when query is too complex

    Code
      files_from_org_per_repo <- test_graphql_gitlab$get_files_from_org_per_repo(org = "mbtests",
        owner_type = "organization", repos_data = list(paths = "gitstatstesting"),
        file_paths = c("project_metadata.yaml", "README.md"), host_files_structure = NULL,
        verbose = TRUE)
    Message
      ! [gitstatstesting] Encountered query complexity error. Too many files (2)).
      i I will run queries per 1 file.

# get_files_content_from_repos for GitLab prints message

    Code
      gl_files_table <- gitlab_testhost_priv$get_files_content_from_repos(file_path = "meta_data.yaml",
        verbose = TRUE)
    Message
      > [Host:GitLab][Engine:GraphQl][Scope:test_group: 0 repos] Pulling files 📄 content: [meta_data.yaml]...

# get_files_content makes use of files_structure

    Code
      files_content <- gitlab_testhost_priv$get_files_content_from_files_structure(
        files_structure = test_mocker$use("gl_files_structure_from_orgs"), verbose = TRUE)
    Message
      > [Host:GitLab][Engine:GraphQl][Scope:mbtests] Pulling files 📄 from files structure...

# get_files_content skips if files structure is empty

    Code
      files_content <- gitlab_testhost_priv$get_files_content_from_files_structure(
        files_structure = test_mocker$use("gl_empty_files_structure"), verbose = TRUE)
    Message
      ! [GitLab] No files 📄 found. Skipping pulling files content.

=encoding utf8

=head1 NAME

gitstats - git history statistics generator

=head1 SYNOPSIS

B<gitstats> [options] <repository dir> <output dir>

=head1 DESCRIPTION

B<gitstats> is a statistics generator for L<git(1)> repositories. It examines the repository and produces some interesting statistics from the history of it. Currently HTML is the only output format.

=head1 OPTIONS

-c option=value

Override a default configuration value. Defaults can be seen by running B<gitstats> without parameters.

=head2 Values:

=over

=item authors_top

How many top authors to show.

=item commit_begin, commit_end

Specify a commit range to generate statistics from. You can specify only commit_end limit statistics to a certain commit or another branch.

=item linear_linestats

When enabled, the lines of code statistics are collected from linear history.
The downside is that commits of long feature branches appear only at the point
where a merge commit is made.

If disabled (old behaviour), the problem is that if two branches contain the
same changes (for example, removal of same lines), the statistics get skewed.

Defaults to on.

=item max_authors

How many authors to show in the list of authors.

=item max_domains

How many domains to show in domains by commits.

=item max_ext_length

Maximum file extension length.

=item processes

Number of concurrent processes to use when extracting git repository data.

=item project_name

Project name to show on the generated pages. Default is to use basename of the repository directory.

=item start_date

Specify a starting date to pass with --since to git.

=item style

CSS stylesheet to use.

=back

=head1 FAQ

Q: How do I generate statistics of a non-master branch?

A: Use C<-c commit_end=web> parameter.

Q: I have files in my git repository that I would like to exclude from the statistics, how do I do that?

A: At the moment the only way is to use L<git-filter-branch(1)> to create a temporary repository and generate the statistics from that.

Q: How do I merge author information when the same author has made commits using different names or emails ?

A: Use git .mailmap feature described in B<MAPPING AUTHORS> of L<git-shortlog(1)>.

=head1 EXAMPLES

=over

=item Generates statistics from a git repository in C<foo> and outputs the result in a directory C<foo_stats>:

  gitstats foo foo_stats

=item As above, but only analyzes the last 10 commits:

  gitstats -c commit_begin='HEAD~10' foo foo_stats

=back

=head1 AUTHORS

B<gitstats> was written by Heikki Hokkanen and others.

See the git repository at https://github.com/hoxu/gitstats for an up-to-date full list of contributors.

=head1 WWW

http://gitstats.sourceforge.net/

=head1 SEE ALSO

L<git(1)>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Storage.R
\name{remove_postgres_storage}
\alias{remove_postgres_storage}
\title{Remove PostgreSQL storage}
\usage{
remove_postgres_storage(gitstats)
}
\arguments{
\item{gitstats}{A GitStats object.}
}
\value{
A \code{GitStats} object (invisibly).
}
\description{
Drops the GitStats schema (with all tables and metadata) from
the PostgreSQL database, closes the connection, and reverts storage to the
default in-memory local backend. Errors if no PostgreSQL backend is
currently set.
}
\examples{
\dontrun{
  my_gitstats <- create_gitstats() |>
    set_postgres_storage(
      dbname = "my_database",
      host = "localhost",
      user = "postgres",
      password = "secret"
    )
  remove_postgres_storage(my_gitstats)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_orgs.R
\name{get_orgs}
\alias{get_orgs}
\title{Get data on organizations}
\usage{
get_orgs(gitstats, cache = TRUE, verbose = FALSE)
}
\arguments{
\item{gitstats}{A GitStats object.}

\item{cache}{A logical, if set to \code{TRUE} GitStats will retrieve the last
result from its storage.}

\item{verbose}{A logical, \code{TRUE} by default. If \code{FALSE} messages and printing
output is switched off.}
}
\value{
A data.frame.
}
\description{
Pulls data on all organizations from a Git host and parses it
into table format.
}
\examples{
\dontrun{
my_gitstats <- create_gitstats() |>
  set_github_host(
    orgs = c("r-world-devs", "openpharma"),
    token = Sys.getenv("GITHUB_PAT")
  ) |>
  set_gitlab_host(
    orgs = "mbtests",
    token = Sys.getenv("GITLAB_PAT_PUBLIC")
  )
get_orgs(my_gitstats)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_files.R
\name{get_files}
\alias{get_files}
\title{Get files}
\usage{
get_files(
  gitstats,
  pattern = NULL,
  depth = Inf,
  file_path = NULL,
  cache = TRUE,
  verbose = FALSE,
  progress = TRUE
)
}
\arguments{
\item{gitstats}{A \code{GitStats} object.}

\item{pattern}{A regular expression. If defined, it pulls content of all
files in a repository matching this pattern reaching to the level of
directories defined by \code{depth} parameter. Works only if \code{file_path} stays
\code{NULL}.}

\item{depth}{Defines level of directories to retrieve files from. E.g. if set
to \code{0}, it will pull files only from \code{root}, if \code{1L}, will take data from
\code{root} directory and directories visible in \code{root} directory. If left with
no argument, will pull files from all directories.}

\item{file_path}{A specific path to file(s) in repositories. May be a
character vector if multiple files are to be pulled. If defined, the
function pulls content of this specific \code{file_path}. Can be defined only if
\code{pattern} stays \code{NULL}.}

\item{cache}{A logical, if set to \code{TRUE} \code{GitStats} will retrieve the last
result from its storage.}

\item{verbose}{A logical, \code{TRUE} by default. If \code{FALSE} messages and printing
output is switched off.}

\item{progress}{A logical, by default set to \code{verbose} value. If \code{FALSE} no
\code{cli} progress bar will be displayed.}
}
\value{
A data.frame.
}
\description{
Pulls text files and their content.
}
\details{
\code{get_files()} may be used in two ways: either with \code{pattern} (with
optional \code{depth}) or \code{file_path} argument defined.

In the first scenario, \code{GitStats} will pull first a files structure
responding to the passed \code{pattern} and \code{depth} arguments and afterwards
files content for all of these files. In the second scenario, \code{GitStats}
will pull only the content of files for the specific \code{file_path} of the
repository.

If user wants to pull a particular file or files, a \code{file_path} approach
seems more reasonable, as it is a faster way since it omits pulling the
whole file structure from the repo.

For example, if user wants to pull content of \code{README.md} and/or \code{NEWS.md}
files placed in the \code{root} directories of the repositories, he should take
the \code{file_path} approach as he already knows precisely paths of the files.

On the other hand, if user wants to pull specific type of files (e.g. all
\code{.md} or \code{.Rmd} files in the repository), without knowing their path, it is
recommended to use a \code{pattern} approach, which will trigger \code{GitStats} to
find all the files in the repository on the given level of directories
(\code{pattern} argument) and afterwards pull their content.

The latter approach is slower than the former but may be more useful
depending on users' goals. Both approaches return data in the same format:
\code{tibble} with data on \code{files}, namely their \code{path} and their \code{content}.
}
\examples{
\dontrun{
 git_stats <- create_gitstats() |>
  set_github_host(
    token = Sys.getenv("GITHUB_PAT"),
    orgs = c("r-world-devs")
  ) |>
  set_gitlab_host(
    token = Sys.getenv("GITLAB_PAT_PUBLIC"),
    orgs = "mbtests"
  )

 rmd_files <- get_files(
   gitstats = git_stats,
   pattern = "\\\\.Rmd",
   depth = 2L
 )

 app_files <- get_files(
   gitstats = git_stats,
   file_path = c("R/app.R", "R/ui.R", "R/server.R")
 )

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GitStats.R
\name{is_parallel}
\alias{is_parallel}
\title{Check if parallel processing is active}
\usage{
is_parallel(gitstats)
}
\arguments{
\item{gitstats}{A GitStats object.}
}
\value{
A logical scalar.
}
\description{
Returns \code{TRUE} when mirai daemons are running on the given
\code{GitStats} object (i.e. \code{set_parallel()} has been called), \code{FALSE}
otherwise.
}
\examples{
\dontrun{
  my_gitstats <- create_gitstats()
  is_parallel(my_gitstats)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Storage.R
\name{set_local_storage}
\alias{set_local_storage}
\title{Set local (in-memory) storage}
\usage{
set_local_storage(gitstats)
}
\arguments{
\item{gitstats}{A GitStats object.}
}
\value{
A \code{GitStats} object (invisibly).
}
\description{
Reset storage to the default in-memory backend.
}
\examples{
\dontrun{
  my_gitstats <- create_gitstats() |>
    set_local_storage()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Storage.R
\name{get_storage}
\alias{get_storage}
\title{Get data from \code{GitStats} storage}
\usage{
get_storage(gitstats, storage = NULL)
}
\arguments{
\item{gitstats}{A GitStats object.}

\item{storage}{A character, type of the data you want to get from storage:
\code{commits}, \code{repositories}, \code{release_logs}, \code{users}, \code{files},
\code{files_structure}, \code{R_package_usage} or \code{release_logs}.}
}
\value{
A list of tibbles (if \code{storage} set to \code{NULL}) or a tibble (if
\code{storage} defined).
}
\description{
Retrieves whole or particular data (see \code{storage} parameter)
pulled earlier with \code{GitStats}.
}
\examples{
\dontrun{
 my_gitstats <- create_gitstats() |>
  set_github_host(
    token = Sys.getenv("GITHUB_PAT"),
    orgs = c("r-world-devs", "openpharma")
  )
  get_release_logs(my_gistats, since = "2024-01-01")
  get_repos(my_gitstats)

  release_logs <- get_storage(
    gitstats = my_gitstats,
    storage = "release_logs"
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GitStats.R
\name{verbose_off}
\alias{verbose_off}
\title{Switch off verbose mode}
\usage{
verbose_off(gitstats)
}
\arguments{
\item{gitstats}{A GitStats object.}
}
\value{
A GitStats object.
}
\description{
Stop printing messages and output.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Storage.R
\name{set_sqlite_storage}
\alias{set_sqlite_storage}
\title{Set SQLite storage}
\usage{
set_sqlite_storage(gitstats, dbname = ":memory:")
}
\arguments{
\item{gitstats}{A GitStats object.}

\item{dbname}{A character, path to SQLite file. Defaults to \code{":memory:"}
for an in-memory database.}
}
\value{
A \code{GitStats} object (invisibly).
}
\description{
Persist GitStats data in a SQLite database. R classes,
custom attributes, and column types are preserved via a \verb{_metadata} table.
Requires \code{DBI}, \code{RSQLite}, and \code{jsonlite} packages.
}
\examples{
\dontrun{
  # File-based
  my_gitstats <- create_gitstats() |>
    set_sqlite_storage(dbname = "gitstats.sqlite")

  # In-memory
  my_gitstats <- create_gitstats() |>
    set_sqlite_storage()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_commits.R
\name{get_commits_stats}
\alias{get_commits_stats}
\title{Get commits statistics}
\usage{
get_commits_stats(
  commits,
  time_aggregation = c("year", "month", "week", "day"),
  group_var
)
}
\arguments{
\item{commits}{A \code{gitstats_commits} S3 class table object (output of \code{get_commits()}).}

\item{time_aggregation}{A character, specifying time aggregation of
statistics.}

\item{group_var}{Other grouping variable to be passed to \code{dplyr::group_by()}
function apart from \code{stats_date} and \code{githost}. Could be: \code{author},
\code{author_login}, \code{author_name} or \code{organization}. Should be passed without
quotation marks.}
}
\value{
A table of \code{commits_stats} class.
}
\description{
Prepare statistics from the pulled commits data.
}
\details{
To make function work, you need first to get commits data with
\code{GitStats}. See examples section.
}
\examples{
\dontrun{
 my_gitstats <- create_gitstats() |>
   set_github_host(
     token = Sys.getenv("GITHUB_PAT"),
     repos = c("r-world-devs/GitStats", "openpharma/visR")
   ) |>
   get_commits(my_gitstats, since = "2022-01-01") |>
   get_commits_stats(
     time_aggregation = "year",
     group_var = author
   )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GitStats.R
\name{create_gitstats}
\alias{create_gitstats}
\title{Create a \code{GitStats} object}
\usage{
create_gitstats()
}
\value{
A \code{GitStats} object.
}
\description{
Create a \code{GitStats} object
}
\examples{
my_gitstats <- create_gitstats()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Storage.R
\name{remove_sqlite_storage}
\alias{remove_sqlite_storage}
\title{Remove SQLite storage}
\usage{
remove_sqlite_storage(gitstats)
}
\arguments{
\item{gitstats}{A GitStats object.}
}
\value{
A \code{GitStats} object (invisibly).
}
\description{
Closes the SQLite connection and, for file-based databases,
deletes the database file. For in-memory databases the data is simply
discarded. Storage is reverted to the default in-memory local backend.
Errors if no SQLite backend is currently set.
}
\examples{
\dontrun{
  my_gitstats <- create_gitstats() |>
    set_sqlite_storage(dbname = "gitstats.sqlite")
  remove_sqlite_storage(my_gitstats)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GitStats.R
\name{show_orgs}
\alias{show_orgs}
\title{Show organizations set in \code{GitStats}}
\usage{
show_orgs(gitstats)
}
\arguments{
\item{gitstats}{A GitStats object.}
}
\value{
A vector of organizations.
}
\description{
Retrieves organizations set or pulled by \code{GitStats}. Especially
helpful when user is scanning whole git platform and wants to have a
glimpse at organizations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GitStats.R
\name{set_parallel}
\alias{set_parallel}
\title{Enable parallel processing}
\usage{
set_parallel(gitstats, workers = 10L)
}
\arguments{
\item{gitstats}{A GitStats object.}

\item{workers}{Number of parallel workers. Set to \code{TRUE} for automatic
detection, a positive integer for a specific count, or \code{FALSE}/\code{0} to
disable parallelism.}
}
\value{
A \code{GitStats} object.
}
\description{
Set up parallel processing for API calls using mirai daemons.
When enabled, GitStats fetches data from multiple repositories
concurrently. Call \code{set_parallel(FALSE)} or \code{set_parallel(0)} to revert
to sequential execution.
}
\examples{
\dontrun{
  my_gitstats <- create_gitstats() |>
    set_github_host(
      token = Sys.getenv("GITHUB_PAT"),
      orgs = c("r-world-devs", "openpharma")
    ) |>
    set_parallel(4)
  get_commits(my_gitstats, since = "2024-01-01")
  my_gitstats |> set_parallel(FALSE) # revert to sequential
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GitStats.R
\name{show_hosts}
\alias{show_hosts}
\title{Show hosts set in \code{GitStats}}
\usage{
show_hosts(gitstats)
}
\arguments{
\item{gitstats}{A GitStats object.}
}
\value{
A list of hosts.
}
\description{
Retrieves hosts set by \code{GitStats} with \verb{set_*_host()} functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GitStats.R
\name{is_verbose}
\alias{is_verbose}
\title{Is verbose mode switched on}
\usage{
is_verbose(gitstats)
}
\arguments{
\item{gitstats}{A GitStats object.}
}
\description{
Is verbose mode switched on
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Storage.R
\name{set_postgres_storage}
\alias{set_postgres_storage}
\title{Set PostgreSQL storage}
\usage{
set_postgres_storage(
  gitstats,
  host = NULL,
  port = NULL,
  dbname = NULL,
  user = NULL,
  password = NULL,
  schema = "git_stats",
  ...
)
}
\arguments{
\item{gitstats}{A GitStats object.}

\item{host}{A character, database host.}

\item{port}{An integer, database port.}

\item{dbname}{A character, database name.}

\item{user}{A character, database user.}

\item{password}{A character, database password.}

\item{schema}{A character, database schema (default \code{"git_stats"}).}

\item{...}{Additional arguments passed to
\code{DBI::dbConnect(RPostgres::Postgres(), ...)}.}
}
\value{
A \code{GitStats} object (invisibly).
}
\description{
Persist GitStats data in a PostgreSQL database. R classes,
custom attributes, and column types are preserved via a \verb{_metadata} table.
Requires \code{DBI}, \code{RPostgres}, and \code{jsonlite} packages.
}
\examples{
\dontrun{
  my_gitstats <- create_gitstats() |>
    set_postgres_storage(
      dbname = "my_database",
      host = "localhost",
      user = "postgres",
      password = "secret"
    )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_repos.R
\name{get_repos}
\alias{get_repos}
\title{Get data on repositories}
\usage{
get_repos(
  gitstats,
  add_contributors = TRUE,
  add_languages = TRUE,
  with_code = NULL,
  in_files = NULL,
  with_files = NULL,
  language = NULL,
  cache = TRUE,
  verbose = FALSE,
  progress = TRUE,
  fill_empty_sha = FALSE
)
}
\arguments{
\item{gitstats}{A GitStats object.}

\item{add_contributors}{A logical parameter to decide whether to add
information about repositories' contributors to the repositories output
(table). If set to \code{FALSE} it makes function run faster as, in the case of
\code{org} search mode, it reaches only \code{GraphQL} endpoint with a query on
repositories, and in the case of \code{code} search mode it reaches only
\verb{repositories REST API} endpoint. However, the pitfall is that the result
does not convey information on contributors. \cr\cr When set to \code{TRUE} (by
default), \code{GitStats} iterates additionally over pulled repositories and
reaches to the \verb{contributors APIs}, which makes it slower, but gives
additional information.}

\item{add_languages}{A logical, \code{TRUE} by default. If set to \code{FALSE},
languages data will not be included in the repositories output which
may speed up the process for GitLab REST engine.}

\item{with_code}{A character vector, if defined, GitStats will pull
repositories with specified code phrases in code blobs.}

\item{in_files}{A character vector of file names. Works when \code{with_code} is
set - then it searches code blobs only in files passed to \code{in_files}
parameter.}

\item{with_files}{A character vector, if defined, GitStats will pull
repositories with specified files.}

\item{language}{A character. If defined, GitStats will return only
repositories with given language.}

\item{cache}{A logical, if set to \code{TRUE} GitStats will retrieve the last
result from its storage.}

\item{verbose}{A logical, \code{TRUE} by default. If \code{FALSE} messages and printing
output is switched off.}

\item{progress}{A logical, by default set to \code{verbose} value. If \code{FALSE} no
\code{cli} progress bar will be displayed.}

\item{fill_empty_sha}{A logical, \code{FALSE} by default. If \code{TRUE}, GitStats
will try to fetch missing \code{commit_sha} values (e.g. for archived GitLab
projects) via the REST Branches API. This may slow down the call for
large numbers of repositories.}
}
\value{
A data.frame.
}
\description{
Pulls data on all repositories for an organization, individual
user or those with a given text in code blobs (\code{with_code} parameter) or a
file (\code{with_files} parameter) and parse it into table format.
}
\examples{
\dontrun{
my_gitstats <- create_gitstats() |>
  set_github_host(
    token = Sys.getenv("GITHUB_PAT"),
    orgs = c("r-world-devs", "openpharma")
  ) |>
  set_gitlab_host(
    token = Sys.getenv("GITLAB_PAT_PUBLIC"),
    orgs = "mbtests"
  )
get_repos(my_gitstats)
get_repos(my_gitstats, add_contributors = FALSE)
get_repos(my_gitstats, add_languages = FALSE)
get_repos(my_gitstats, with_code = "Shiny", in_files = "renv.lock")
get_repos(my_gitstats, with_files = "DESCRIPTION")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Storage.R
\name{remove_from_storage}
\alias{remove_from_storage}
\title{Remove a table from \code{GitStats} storage}
\usage{
remove_from_storage(gitstats, storage)
}
\arguments{
\item{gitstats}{A GitStats object.}

\item{storage}{A character, name of the table to remove (e.g. \code{"commits"},
\code{"repositories"}).}
}
\value{
A \code{GitStats} object (invisibly).
}
\description{
Removes a named table from the active storage backend.
}
\examples{
\dontrun{
  my_gitstats <- create_gitstats() |>
    set_github_host(
      token = Sys.getenv("GITHUB_PAT"),
      orgs = "r-world-devs"
    )
  get_commits(my_gitstats, since = "2024-01-01")
  remove_from_storage(my_gitstats, storage = "commits")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_host.R
\name{set_gitlab_host}
\alias{set_gitlab_host}
\title{Set GitLab host}
\usage{
set_gitlab_host(
  gitstats,
  host = NULL,
  token = NULL,
  orgs = NULL,
  repos = NULL,
  verbose = is_verbose(gitstats),
  .error = TRUE
)
}
\arguments{
\item{gitstats}{A GitStats object.}

\item{host}{A character, optional, URL name of the host. If not passed, a
public host will be used.}

\item{token}{A token.}

\item{orgs}{An optional character vector of organisations}

\item{repos}{An optional character vector of repositories full names
(organization and repository name, e.g. "r-world-devs/GitStats").}

\item{verbose}{A logical, \code{TRUE} by default. If \code{FALSE} messages and printing
output is switched off.}

\item{.error}{A logical to control if passing wrong input
(\code{repositories} and \code{organizations}) should end with an error or not.}
}
\value{
A \code{GitStats} object with added information on host.
}
\description{
Set GitLab host
}
\details{
If you do not define \code{orgs} and \code{repos}, \code{GitStats} will be set to
scan whole Git platform (such as enterprise version of GitHub or GitLab),
unless it is a public platform. In case of a public one (like GitHub) you
need to define \code{orgs} or \code{repos} as scanning through all organizations may
take large amount of time.
}
\examples{
\dontrun{
my_gitstats <- create_gitstats() |>
  set_gitlab_host(
    token = Sys.getenv("GITLAB_PAT_PUBLIC"),
    orgs = "mbtests"
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_repos.R
\name{get_repos_urls}
\alias{get_repos_urls}
\title{Get repository URLS}
\usage{
get_repos_urls(
  gitstats,
  type = "api",
  with_code = NULL,
  in_files = NULL,
  with_files = NULL,
  cache = TRUE,
  verbose = FALSE,
  progress = TRUE
)
}
\arguments{
\item{gitstats}{A GitStats object.}

\item{type}{A character, choose if \code{api} or \code{web} (\code{html}) URLs should be
returned. \code{api} type is set by default as setting \code{web} results in parsing
which may be time consuming in case of large number of repositories.}

\item{with_code}{A character vector, if defined, \code{GitStats} will pull
repositories with specified code phrases in code blobs.}

\item{in_files}{A character vector of file names. Works when \code{with_code} is
set - then it searches code blobs only in files passed to \code{in_files}
parameter.}

\item{with_files}{A character vector, if defined, GitStats will pull
repositories with specified files.}

\item{cache}{A logical, if set to \code{TRUE} GitStats will retrieve the last
result from its storage.}

\item{verbose}{A logical, \code{TRUE} by default. If \code{FALSE} messages and printing
output is switched off.}

\item{progress}{A logical, by default set to \code{verbose} value. If \code{FALSE} no
\code{cli} progress bar will be displayed.}
}
\value{
A character vector.
}
\description{
Pulls a vector of repositories URLs (web or API): either all for
an organization or those with a given text in code blobs (\code{with_code}
parameter) or a file (\code{with_files} parameter).
}
\examples{
\dontrun{
my_gitstats <- create_gitstats() |>
  set_github_host(
    token = Sys.getenv("GITHUB_PAT"),
    orgs = c("r-world-devs", "openpharma")
  ) |>
  set_gitlab_host(
    token = Sys.getenv("GITLAB_PAT_PUBLIC"),
    orgs = "mbtests"
  )
get_repos_urls(my_gitstats, with_files = c("DESCRIPTION", "LICENSE"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_commits.R
\name{get_commits}
\alias{get_commits}
\title{Get data on commits}
\usage{
get_commits(
  gitstats,
  since = NULL,
  until = Sys.Date(),
  cache = TRUE,
  verbose = FALSE,
  progress = TRUE
)
}
\arguments{
\item{gitstats}{A \code{GitStats} object.}

\item{since}{A starting date.}

\item{until}{An end date.}

\item{cache}{A logical, if set to \code{TRUE} GitStats will retrieve the last
result from its storage.}

\item{verbose}{A logical, \code{TRUE} by default. If \code{FALSE} messages and printing
output is switched off.}

\item{progress}{A logical, by default set to \code{verbose} value. If \code{FALSE} no
\code{cli} progress bar will be displayed.}
}
\value{
A table of \code{tibble} and \code{gitstats_commits} classes.
}
\description{
List all commits from all repositories for an organization or a
vector of repositories.
}
\examples{
\dontrun{
my_gitstats <- create_gitstats() |>
  set_github_host(
    token = Sys.getenv("GITHUB_PAT"),
    repos = c("openpharma/DataFakeR", "openpharma/visR")
  ) |>
  set_gitlab_host(
    token = Sys.getenv("GITLAB_PAT_PUBLIC"),
    orgs = "mbtests"
  )
 get_commits(my_gitstats, since = "2018-01-01")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_users.R
\name{get_users}
\alias{get_users}
\title{Get users data}
\usage{
get_users(gitstats, logins, cache = TRUE, verbose = FALSE)
}
\arguments{
\item{gitstats}{A GitStats object.}

\item{logins}{A character vector of logins.}

\item{cache}{A logical, if set to \code{TRUE} GitStats will retrieve the last
result from its storage.}

\item{verbose}{A logical, \code{TRUE} by default. If \code{FALSE} messages and
printing output is switched off.}
}
\value{
A data.frame.
}
\description{
Get users data
}
\examples{
\dontrun{
 my_gitstats <- create_gitstats() |>
  set_github_host(
    token = Sys.getenv("GITHUB_PAT"),
    orgs = c("r-world-devs")
  ) |>
  set_gitlab_host(
    token = Sys.getenv("GITLAB_PAT_PUBLIC"),
    orgs = "mbtests"
  )
 get_users(my_gitstats, c("maciekabanas", "marcinkowskak"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_repos.R
\name{get_repos_trees}
\alias{get_repos_trees}
\title{Get data on files trees across repositories}
\usage{
get_repos_trees(
  gitstats,
  pattern = NULL,
  depth = Inf,
  cache = TRUE,
  verbose = FALSE,
  progress = TRUE
)
}
\arguments{
\item{gitstats}{A GitStats object.}

\item{pattern}{A regular expression. If defined, it pulls structure of files
in a repository matching this pattern reaching to the level of directories
defined by \code{depth} parameter.}

\item{depth}{Defines level of directories to reach for files structure from.
E.g. if set to \code{0}, it will pull files tree only from \code{root}, if \code{1L}, will
take data from \code{root} directory and directories visible in \code{root}
directory. If left with no argument, will pull files tree down to every
directory in a repo.}

\item{cache}{A logical, if set to \code{TRUE} GitStats will retrieve the last
result from its storage.}

\item{verbose}{A logical, \code{TRUE} by default. If \code{FALSE} messages and printing
output is switched off.}

\item{progress}{A logical, by default set to \code{verbose} value. If \code{FALSE} no
\code{cli} progress bar will be displayed.}
}
\value{
A \code{tibble}.
}
\description{
Pulls files tree (structure) per repository. Files trees are
then stored as character vectors in \code{files_tree} column of output table.
}
\examples{
\dontrun{
 my_gitstats <- create_gitstats() |>
  set_github_host(
    token = Sys.getenv("GITHUB_PAT"),
    orgs = c("r-world-devs", "openpharma")
  )

 get_repos_trees(
   gitstats = my_gitstats,
   pattern = "\\\\.md"
 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Storage.R
\name{get_storage_metadata}
\alias{get_storage_metadata}
\title{Get metadata for a storage table}
\usage{
get_storage_metadata(gitstats, storage = NULL)
}
\arguments{
\item{gitstats}{A GitStats object.}

\item{storage}{A character, name of the table (e.g. \code{"commits"},
\code{"repositories"}). If \code{NULL} (default), metadata for all tables in storage
will be returned.}
}
\value{
A list with metadata fields: \code{class}, \code{attributes}, and (for
database backends) \code{column_types}.
}
\description{
Retrieves metadata (R classes, custom attributes, column types)
for a table stored in the active storage backend.
}
\examples{
\dontrun{
  my_gitstats <- create_gitstats() |>
    set_github_host(
      token = Sys.getenv("GITHUB_PAT"),
      orgs = "r-world-devs"
    )
  get_commits(my_gitstats, since = "2024-01-01")
  get_storage_metadata(my_gitstats, storage = "commits")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_issues.R
\name{get_issues_stats}
\alias{get_issues_stats}
\title{Get issues statistics}
\usage{
get_issues_stats(
  issues,
  time_aggregation = c("year", "month", "week", "day"),
  group_var
)
}
\arguments{
\item{issues}{A \code{gitstats_issue} S3 class table object (output of \code{get_issues()}).}

\item{time_aggregation}{A character, specifying time aggregation of
statistics.}

\item{group_var}{Other grouping variable to be passed to \code{dplyr::group_by()}
function apart from \code{stats_date} and \code{githost}. Could be: \code{author},
\code{state} or \code{organization}. Should be passed without
quotation marks.}
}
\value{
A table of \code{issues_stats} class.
}
\description{
Prepare statistics from the pulled issues data.
}
\details{
To make function work, you need first to get issues data with
\code{GitStats}. See examples section.
}
\examples{
\dontrun{
 my_gitstats <- create_gitstats() |>
   set_github_host(
     token = Sys.getenv("GITHUB_PAT"),
     repos = c("r-world-devs/GitStats", "openpharma/visR")
   ) |>
   get_issues(my_gitstats, since = "2022-01-01") |>
   get_issues_stats(
     time_aggregation = "month",
     group_var = state
   )
}
}
